package com.avaje.ebeaninternal.server.autofetch;

import com.avaje.ebean.bean.CallStack;
import com.avaje.ebean.bean.NodeUsageCollector;
import com.avaje.ebean.bean.ObjectGraphNode;
import com.avaje.ebean.bean.ObjectGraphOrigin;
import com.avaje.ebean.config.AutofetchConfig;
import com.avaje.ebean.config.AutofetchMode;
import com.avaje.ebean.config.ServerConfig;
import com.avaje.ebeaninternal.api.ClassUtil;
import com.avaje.ebeaninternal.api.SpiEbeanServer;
import com.avaje.ebeaninternal.api.SpiQuery;
import com.avaje.ebeaninternal.server.deploy.BeanDescriptor;
import com.avaje.ebeaninternal.server.querydefn.OrmQueryDetail;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import javax.persistence.PersistenceException;





public class DefaultAutoFetchManager
  implements AutoFetchManager, Serializable
{
  private static final long serialVersionUID = -6826119882781771722L;
  private final String statisticsMonitor = new String();
  


  private final String fileName;
  

  private Map<String, Statistics> statisticsMap = new ConcurrentHashMap();
  



  private Map<String, TunedQueryInfo> tunedQueryInfoMap = new ConcurrentHashMap();
  
  private transient long defaultGarbageCollectionWait = 100L;
  




  private transient int tunedQueryCount;
  



  private transient double profilingRate = 0.1D;
  
  private transient int profilingBase = 10;
  
  private transient int profilingMin = 1;
  

  private transient boolean profiling;
  

  private transient boolean queryTuning;
  

  private transient boolean queryTuningAddVersion;
  

  private transient AutofetchMode mode;
  
  private transient boolean useFileLogging;
  
  private transient SpiEbeanServer server;
  
  private transient DefaultAutoFetchManagerLogging logging;
  

  public DefaultAutoFetchManager(String fileName)
  {
    this.fileName = fileName;
  }
  


  public void setOwner(SpiEbeanServer server, ServerConfig serverConfig)
  {
    this.server = server;
    logging = new DefaultAutoFetchManagerLogging(serverConfig, this);
    
    AutofetchConfig autofetchConfig = serverConfig.getAutofetchConfig();
    
    useFileLogging = autofetchConfig.isUseFileLogging();
    queryTuning = autofetchConfig.isQueryTuning();
    queryTuningAddVersion = autofetchConfig.isQueryTuningAddVersion();
    profiling = autofetchConfig.isProfiling();
    profilingMin = autofetchConfig.getProfilingMin();
    profilingBase = autofetchConfig.getProfilingBase();
    
    setProfilingRate(autofetchConfig.getProfilingRate());
    

    defaultGarbageCollectionWait = autofetchConfig.getGarbageCollectionWait();
    


    mode = autofetchConfig.getMode();
    
    if ((profiling) || (queryTuning))
    {
      String msg = "AutoFetch queryTuning[" + queryTuning + "] profiling[" + profiling + "] mode[" + mode + "]  profiling rate[" + profilingRate + "] min[" + profilingMin + "] base[" + profilingBase + "]";
      

      logging.logToJavaLogger(msg);
    }
  }
  

  public void clearQueryStatistics()
  {
    server.clearQueryStatistics();
  }
  


  public int getTotalTunedQueryCount()
  {
    return tunedQueryCount;
  }
  


  public int getTotalTunedQuerySize()
  {
    return tunedQueryInfoMap.size();
  }
  


  public int getTotalProfileSize()
  {
    return statisticsMap.size();
  }
  

  public int clearTunedQueryInfo()
  {
    tunedQueryCount = 0;
    

    int size = tunedQueryInfoMap.size();
    tunedQueryInfoMap.clear();
    return size;
  }
  
  public int clearProfilingInfo() {
    int size = statisticsMap.size();
    statisticsMap.clear();
    return size;
  }
  

  public void serialize()
  {
    File autoFetchFile = new File(fileName);
    try
    {
      FileOutputStream fout = new FileOutputStream(autoFetchFile);
      
      ObjectOutputStream oout = new ObjectOutputStream(fout);
      oout.writeObject(this);
      oout.flush();
      oout.close();
    }
    catch (Exception e) {
      String msg = "Error serializing autofetch file";
      logging.logError(Level.SEVERE, msg, e);
    }
  }
  


  public TunedQueryInfo getTunedQueryInfo(String originKey)
  {
    return (TunedQueryInfo)tunedQueryInfoMap.get(originKey);
  }
  


  public Statistics getStatistics(String originKey)
  {
    return (Statistics)statisticsMap.get(originKey);
  }
  
  public Iterator<TunedQueryInfo> iterateTunedQueryInfo() {
    return tunedQueryInfoMap.values().iterator();
  }
  
  public Iterator<Statistics> iterateStatistics() {
    return statisticsMap.values().iterator();
  }
  
  public boolean isProfiling() {
    return profiling;
  }
  




  public void setProfiling(boolean profiling)
  {
    this.profiling = profiling;
  }
  
  public boolean isQueryTuning() {
    return queryTuning;
  }
  
  public void setQueryTuning(boolean queryTuning) {
    this.queryTuning = queryTuning;
  }
  
  public double getProfilingRate() {
    return profilingRate;
  }
  
  public AutofetchMode getMode() {
    return mode;
  }
  
  public void setMode(AutofetchMode mode) {
    this.mode = mode;
  }
  
  public void setProfilingRate(double rate) {
    if (rate < 0.0D) {
      rate = 0.0D;
    } else if (rate > 1.0D) {
      rate = 1.0D;
    }
    profilingRate = rate;
  }
  
  public int getProfilingBase() {
    return profilingBase;
  }
  
  public void setProfilingBase(int profilingBase) {
    this.profilingBase = profilingBase;
  }
  
  public int getProfilingMin() {
    return profilingMin;
  }
  
  public void setProfilingMin(int profilingMin) {
    this.profilingMin = profilingMin;
  }
  







  public void shutdown()
  {
    collectUsageViaGC(-1L);
    if (useFileLogging) {
      serialize();
    }
  }
  







  public String collectUsageViaGC(long waitMillis)
  {
    
    






    try
    {
      if (waitMillis < 0L) {
        waitMillis = defaultGarbageCollectionWait;
      }
      Thread.sleep(waitMillis);
    } catch (InterruptedException e) {
      String msg = "Error while sleeping after System.gc() request.";
      logging.logError(Level.SEVERE, msg, e);
      return msg;
    }
    return updateTunedQueryInfo();
  }
  



  public String updateTunedQueryInfo()
  {
    if (!profiling)
    {

      return "Not profiling";
    }
    
    synchronized (statisticsMonitor)
    {
      Counters counters = new Counters(null);
      
      Iterator<Statistics> it = statisticsMap.values().iterator();
      while (it.hasNext()) {
        Statistics queryPointStatistics = (Statistics)it.next();
        if (!queryPointStatistics.hasUsage())
        {
          counters.incrementNoUsage();
        } else {
          updateTunedQueryFromUsage(counters, queryPointStatistics);
        }
      }
      
      String summaryInfo = counters.toString();
      
      if (counters.isInteresting())
      {
        logging.logSummary(summaryInfo);
      }
      
      return summaryInfo;
    }
  }
  
  private static class Counters
  {
    int newPlan;
    int modified;
    int unchanged;
    int noUsage;
    
    void incrementNoUsage() {
      noUsage += 1;
    }
    
    void incrementNew() { newPlan += 1; }
    
    void incrementModified() {
      modified += 1;
    }
    
    void incrementUnchanged() { unchanged += 1; }
    
    boolean isInteresting() {
      return (newPlan > 0) || (modified > 0);
    }
    
    public String toString() { return "new[" + newPlan + "] modified[" + modified + "] unchanged[" + unchanged + "] nousage[" + noUsage + "]"; }
  }
  

  private void updateTunedQueryFromUsage(Counters counters, Statistics statistics)
  {
    ObjectGraphOrigin queryPoint = statistics.getOrigin();
    String beanType = queryPoint.getBeanType();
    try
    {
      Class<?> beanClass = ClassUtil.forName(beanType, getClass());
      BeanDescriptor<?> beanDescriptor = server.getBeanDescriptor(beanClass);
      if (beanDescriptor != null)
      {




        OrmQueryDetail newFetchDetail = statistics.buildTunedFetch(beanDescriptor);
        

        TunedQueryInfo currentFetch = (TunedQueryInfo)tunedQueryInfoMap.get(queryPoint.getKey());
        
        if (currentFetch == null)
        {
          counters.incrementNew();
          
          currentFetch = statistics.createTunedFetch(newFetchDetail);
          logging.logNew(currentFetch);
          tunedQueryInfoMap.put(queryPoint.getKey(), currentFetch);
        }
        else if (!currentFetch.isSame(newFetchDetail))
        {
          counters.incrementModified();
          
          logging.logChanged(currentFetch, newFetchDetail);
          currentFetch.setTunedDetail(newFetchDetail);
        }
        else
        {
          counters.incrementUnchanged();
        }
        
        currentFetch.setProfileCount(statistics.getCounter());
      }
    }
    catch (ClassNotFoundException e)
    {
      String msg = e.toString() + " updating autoFetch tuned query for " + beanType + ". It isLikely this bean has been renamed or moved";
      
      logging.logError(Level.INFO, msg, null);
      statisticsMap.remove(statistics.getOrigin().getKey());
    }
  }
  



  private boolean useAutoFetch(SpiQuery<?> query)
  {
    if ((query.isLoadBeanCache()) || (query.isSharedInstance()))
    {

      return false;
    }
    
    Boolean autoFetch = query.isAutofetch();
    if (autoFetch != null)
    {
      return autoFetch.booleanValue();
    }
    

    switch (1.$SwitchMap$com$avaje$ebean$config$AutofetchMode[mode.ordinal()]) {
    case 1: 
      return true;
    
    case 2: 
      return false;
    
    case 3: 
      return query.isDetailEmpty();
    }
    
    throw new PersistenceException("Invalid autoFetchMode " + mode);
  }
  





  public boolean tuneQuery(SpiQuery<?> query)
  {
    if ((!queryTuning) && (!profiling)) {
      return false;
    }
    
    if (!useAutoFetch(query))
    {
      return false;
    }
    
    ObjectGraphNode parentAutoFetchNode = query.getParentNode();
    if (parentAutoFetchNode != null)
    {

      query.setAutoFetchManager(this);
      return true;
    }
    

    CallStack stack = server.createCallStack();
    ObjectGraphNode origin = query.setOrigin(stack);
    

    TunedQueryInfo tunedFetch = (TunedQueryInfo)tunedQueryInfoMap.get(origin.getOriginQueryPoint().getKey());
    

    int profileCount = tunedFetch == null ? 0 : tunedFetch.getProfileCount();
    
    if (profiling)
    {
      if (tunedFetch == null) {
        query.setAutoFetchManager(this);
      }
      else if (profileCount < profilingBase) {
        query.setAutoFetchManager(this);
      }
      else if (tunedFetch.isPercentageProfile(profilingRate)) {
        query.setAutoFetchManager(this);
      }
    }
    
    if ((queryTuning) && 
      (tunedFetch != null) && (profileCount >= profilingMin))
    {

      if (tunedFetch.autoFetchTune(query))
      {


        tunedQueryCount += 1;
      }
      return true;
    }
    

    return false;
  }
  





  public void collectQueryInfo(ObjectGraphNode node, int beans, int micros)
  {
    if (node != null) {
      ObjectGraphOrigin origin = node.getOriginQueryPoint();
      if (origin != null) {
        Statistics stats = getQueryPointStats(origin);
        stats.collectQueryInfo(node, beans, micros);
      }
    }
  }
  







  public void collectNodeUsage(NodeUsageCollector usageCollector)
  {
    ObjectGraphOrigin origin = usageCollector.getNode().getOriginQueryPoint();
    
    Statistics stats = getQueryPointStats(origin);
    
    if (logging.isTraceUsageCollection()) {
      System.out.println("... NodeUsageCollector " + usageCollector);
    }
    
    stats.collectUsageInfo(usageCollector);
    
    if (logging.isTraceUsageCollection()) {
      System.out.println("stats\n" + stats);
    }
  }
  
  private Statistics getQueryPointStats(ObjectGraphOrigin originQueryPoint) {
    synchronized (statisticsMonitor) {
      Statistics stats = (Statistics)statisticsMap.get(originQueryPoint.getKey());
      if (stats == null) {
        stats = new Statistics(originQueryPoint, queryTuningAddVersion);
        statisticsMap.put(originQueryPoint.getKey(), stats);
      }
      return stats;
    }
  }
  
  public String toString() {
    synchronized (statisticsMonitor) {
      return statisticsMap.values().toString();
    }
  }
}
