package com.avaje.ebeaninternal.server.cache;

import com.avaje.ebean.BackgroundExecutor;
import com.avaje.ebean.EbeanServer;
import com.avaje.ebean.cache.ServerCache;
import com.avaje.ebean.cache.ServerCacheOptions;
import com.avaje.ebean.cache.ServerCacheStatistics;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;








public class DefaultServerCache
  implements ServerCache
{
  private static final Logger logger = Logger.getLogger(DefaultServerCache.class.getName());
  
  private static final CacheEntryComparator comparator = new CacheEntryComparator(null);
  
  private final ConcurrentHashMap<Object, CacheEntry> map = new ConcurrentHashMap();
  
  private final AtomicInteger missCount = new AtomicInteger();
  
  private final AtomicInteger removedHitCount = new AtomicInteger();
  
  private final Object monitor = new Object();
  
  private final String name;
  
  private int maxSize;
  
  private long trimFrequency;
  
  private int maxIdleSecs;
  private int maxSecsToLive;
  
  public DefaultServerCache(String name, ServerCacheOptions options)
  {
    this(name, options.getMaxSize(), options.getMaxIdleSecs(), options.getMaxSecsToLive());
  }
  
  public DefaultServerCache(String name, int maxSize, int maxIdleSecs, int maxSecsToLive) {
    this.name = name;
    this.maxSize = maxSize;
    this.maxIdleSecs = maxIdleSecs;
    this.maxSecsToLive = maxSecsToLive;
    trimFrequency = 60L;
  }
  

  public void init(EbeanServer server)
  {
    TrimTask trim = new TrimTask(null);
    
    BackgroundExecutor executor = server.getBackgroundExecutor();
    executor.executePeriodically(trim, trimFrequency, TimeUnit.SECONDS);
  }
  



  public ServerCacheStatistics getStatistics(boolean reset)
  {
    ServerCacheStatistics s = new ServerCacheStatistics();
    s.setCacheName(name);
    s.setMaxSize(maxSize);
    



    int mc = reset ? missCount.getAndSet(0) : missCount.get();
    int hc = getHitCount(reset);
    int size = size();
    
    s.setSize(size);
    s.setHitCount(hc);
    s.setMissCount(mc);
    
    return s;
  }
  
  public int getHitRatio()
  {
    int mc = missCount.get();
    int hc = getHitCount(false);
    
    int totalCount = hc + mc;
    if (totalCount == 0) {
      return 0;
    }
    return hc * 100 / totalCount;
  }
  


  private int getHitCount(boolean reset)
  {
    int hc = reset ? removedHitCount.getAndSet(0) : removedHitCount.get();
    
    Iterator<CacheEntry> it = map.values().iterator();
    while (it.hasNext()) {
      CacheEntry cacheEntry = (CacheEntry)it.next();
      hc += cacheEntry.getHitCount(reset);
    }
    
    return hc;
  }
  
  public ServerCacheOptions getOptions()
  {
    synchronized (monitor) {
      ServerCacheOptions o = new ServerCacheOptions();
      o.setMaxIdleSecs(maxIdleSecs);
      o.setMaxSize(maxSize);
      o.setMaxSecsToLive(maxSecsToLive);
      return o;
    }
  }
  
  public void setOptions(ServerCacheOptions o) {
    synchronized (monitor) {
      maxIdleSecs = o.getMaxIdleSecs();
      maxSize = o.getMaxSize();
      maxSecsToLive = o.getMaxSecsToLive();
    }
  }
  



  public int getMaxSize()
  {
    return maxSize;
  }
  


  public void setMaxSize(int maxSize)
  {
    synchronized (monitor) {
      this.maxSize = maxSize;
    }
  }
  


  public long getMaxIdleSecs()
  {
    return maxIdleSecs;
  }
  


  public void setMaxIdleSecs(int maxIdleSecs)
  {
    synchronized (monitor) {
      this.maxIdleSecs = maxIdleSecs;
    }
  }
  


  public long getMaxSecsToLive()
  {
    return maxSecsToLive;
  }
  


  public void setMaxSecsToLive(int maxSecsToLive)
  {
    synchronized (monitor) {
      this.maxSecsToLive = maxSecsToLive;
    }
  }
  


  public String getName()
  {
    return name;
  }
  


  public void clear()
  {
    map.clear();
  }
  



  public Object get(Object key)
  {
    CacheEntry entry = (CacheEntry)map.get(key);
    
    if (entry == null) {
      missCount.incrementAndGet();
      return null;
    }
    


    return entry.getValue();
  }
  




  public Object put(Object key, Object value)
  {
    CacheEntry entry = (CacheEntry)map.put(key, new CacheEntry(key, value));
    if (entry == null) {
      return null;
    }
    int removedHits = entry.getHitCount(true);
    removedHitCount.addAndGet(removedHits);
    return entry.getValue();
  }
  



  public Object putIfAbsent(Object key, Object value)
  {
    CacheEntry entry = (CacheEntry)map.putIfAbsent(key, new CacheEntry(key, value));
    if (entry == null) {
      return null;
    }
    return entry.getValue();
  }
  



  public Object remove(Object key)
  {
    CacheEntry entry = (CacheEntry)map.remove(key);
    if (entry == null) {
      return null;
    }
    int removedHits = entry.getHitCount(true);
    removedHitCount.addAndGet(removedHits);
    return entry.getValue();
  }
  



  public int size()
  {
    return map.size();
  }
  
  private Iterator<CacheEntry> cacheEntries() {
    return map.values().iterator();
  }
  
  private class TrimTask
    implements Runnable
  {
    private TrimTask() {}
    
    public void run()
    {
      long startTime = System.currentTimeMillis();
      
      if (DefaultServerCache.logger.isLoggable(Level.FINER)) {
        DefaultServerCache.logger.finer("trimming cache " + name);
      }
      
      int trimmedByIdle = 0;
      int trimmedByTTL = 0;
      int trimmedByLRU = 0;
      
      boolean trimMaxSize = (maxSize > 0) && (maxSize < size());
      
      ArrayList<DefaultServerCache.CacheEntry> activeList = new ArrayList();
      
      long idleExpire = System.currentTimeMillis() - maxIdleSecs * 1000;
      long ttlExpire = System.currentTimeMillis() - maxSecsToLive * 1000;
      
      Iterator<DefaultServerCache.CacheEntry> it = DefaultServerCache.this.cacheEntries();
      while (it.hasNext()) {
        DefaultServerCache.CacheEntry cacheEntry = (DefaultServerCache.CacheEntry)it.next();
        if ((maxIdleSecs > 0) && (idleExpire > cacheEntry.getLastAccessTime())) {
          it.remove();
          trimmedByIdle++;
        }
        else if ((maxSecsToLive > 0) && (ttlExpire > cacheEntry.getCreateTime())) {
          it.remove();
          trimmedByTTL++;
        }
        else if (trimMaxSize) {
          activeList.add(cacheEntry);
        }
      }
      
      if (trimMaxSize) {
        trimmedByLRU = activeList.size() - maxSize;
        
        if (trimmedByLRU > 0)
        {
          Collections.sort(activeList, DefaultServerCache.comparator);
          for (int i = maxSize; i < activeList.size(); i++)
          {
            map.remove(((DefaultServerCache.CacheEntry)activeList.get(i)).getKey());
          }
        }
      }
      
      long exeTime = System.currentTimeMillis() - startTime;
      
      if (DefaultServerCache.logger.isLoggable(Level.FINE)) {
        DefaultServerCache.logger.fine("Executed trim of cache " + name + " in [" + exeTime + "]millis  idle[" + trimmedByIdle + "] timeToLive[" + trimmedByTTL + "] accessTime[" + trimmedByLRU + "]");
      }
    }
  }
  




  private static class CacheEntryComparator
    implements Comparator<DefaultServerCache.CacheEntry>, Serializable
  {
    private static final long serialVersionUID = 1L;
    



    public int compare(DefaultServerCache.CacheEntry o1, DefaultServerCache.CacheEntry o2)
    {
      return o1.getLastAccessLong().compareTo(o2.getLastAccessLong());
    }
  }
  

  public static class CacheEntry
  {
    private final Object key;
    
    private final Object value;
    
    private final long createTime;
    private final AtomicInteger hitCount = new AtomicInteger();
    private Long lastAccessTime;
    
    public CacheEntry(Object key, Object value) {
      this.key = key;
      this.value = value;
      createTime = System.currentTimeMillis();
      lastAccessTime = Long.valueOf(createTime);
    }
    
    public Object getKey() {
      return key;
    }
    
    public Object getValue()
    {
      hitCount.incrementAndGet();
      lastAccessTime = Long.valueOf(System.currentTimeMillis());
      return value;
    }
    
    public long getCreateTime() {
      return createTime;
    }
    
    public long getLastAccessTime() {
      return lastAccessTime.longValue();
    }
    
    public Long getLastAccessLong() {
      return lastAccessTime;
    }
    
    public int getHitCount(boolean reset) {
      if (reset) {
        return hitCount.getAndSet(0);
      }
      
      return hitCount.get();
    }
    
    public int getHitCount() {
      return hitCount.get();
    }
  }
}
