package com.avaje.ebeaninternal.server.cluster;

import com.avaje.ebean.EbeanServer;
import com.avaje.ebean.config.GlobalProperties;
import com.avaje.ebeaninternal.api.ClassUtil;
import com.avaje.ebeaninternal.server.cluster.mcast.McastClusterManager;
import com.avaje.ebeaninternal.server.cluster.socket.SocketClusterBroadcast;
import com.avaje.ebeaninternal.server.lucene.cluster.SLuceneClusterFactory;
import com.avaje.ebeaninternal.server.transaction.RemoteTransactionEvent;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;























public class ClusterManager
{
  private static final Logger logger = Logger.getLogger(ClusterManager.class.getName());
  
  private final ClusterBroadcast broadcast;
  
  private final ConcurrentHashMap<String, EbeanServer> serverMap = new ConcurrentHashMap();
  
  private LuceneClusterListener luceneListener;
  
  private LuceneClusterIndexSync luceneIndexSync;
  
  private boolean started;
  
  public ClusterManager()
  {
    String clusterType = GlobalProperties.get("ebean.cluster.type", null);
    if ((clusterType == null) || (clusterType.trim().length() == 0))
    {
      broadcast = null;

    }
    else
    {
      LuceneClusterFactory luceneFactory = new SLuceneClusterFactory();
      
      int lucenePort = GlobalProperties.getInt("ebean.cluster.lucene.port", 9991);
      luceneListener = luceneFactory.createListener(this, lucenePort);
      
      String masterHostPort = GlobalProperties.get("ebean.cluster.lucene.masterHostPort", null);
      luceneIndexSync = luceneFactory.createIndexSync();
      luceneIndexSync.setMasterHost(masterHostPort);
      luceneIndexSync.setMode(masterHostPort == null ? LuceneClusterIndexSync.Mode.MASTER_MODE : LuceneClusterIndexSync.Mode.SLAVE_MODE);
      

      logger.info("... luceneListener using [" + lucenePort + "]");
      try {
        if ("mcast".equalsIgnoreCase(clusterType)) {
          broadcast = new McastClusterManager();
        }
        else if ("socket".equalsIgnoreCase(clusterType)) {
          broadcast = new SocketClusterBroadcast();
        }
        else {
          logger.info("Clustering using [" + clusterType + "]");
          broadcast = ((ClusterBroadcast)ClassUtil.newInstance(clusterType));
        }
      }
      catch (Exception e) {
        String msg = "Error initialising ClusterManager type [" + clusterType + "]";
        logger.log(Level.SEVERE, msg, e);
        throw new RuntimeException(e);
      }
    }
  }
  
  public void registerServer(EbeanServer server) {
    synchronized (serverMap) {
      if (!started) {
        startup();
      }
      serverMap.put(server.getName(), server);
    }
  }
  
  public LuceneClusterIndexSync getLuceneClusterIndexSync() {
    return luceneIndexSync;
  }
  
  public EbeanServer getServer(String name) {
    synchronized (serverMap) {
      return (EbeanServer)serverMap.get(name);
    }
  }
  
  private void startup() {
    started = true;
    if (broadcast != null) {
      broadcast.startup(this);
    }
    if (luceneListener != null) {
      luceneListener.startup();
    }
  }
  


  public boolean isClustering()
  {
    return broadcast != null;
  }
  


  public void broadcast(RemoteTransactionEvent remoteTransEvent)
  {
    if (broadcast != null) {
      broadcast.broadcast(remoteTransEvent);
    }
  }
  


  public void shutdown()
  {
    if (luceneListener != null) {
      luceneListener.shutdown();
    }
    if (broadcast != null) {
      logger.info("ClusterManager shutdown ");
      broadcast.shutdown();
    }
  }
}
