package com.avaje.ebeaninternal.server.cluster.mcast;

import com.avaje.ebean.config.GlobalProperties;
import com.avaje.ebeaninternal.api.SpiEbeanServer;
import com.avaje.ebeaninternal.server.cluster.ClusterBroadcast;
import com.avaje.ebeaninternal.server.cluster.ClusterManager;
import com.avaje.ebeaninternal.server.cluster.Packet;
import com.avaje.ebeaninternal.server.cluster.PacketWriter;
import com.avaje.ebeaninternal.server.transaction.RemoteTransactionEvent;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;






































public class McastClusterManager
  implements ClusterBroadcast, Runnable
{
  private static final Logger logger = Logger.getLogger(McastClusterManager.class.getName());
  



  private ClusterManager clusterManager;
  



  private final Thread managerThread;
  


  private final McastPacketControl packageControl;
  


  private final McastListener listener;
  


  private final McastSender localSender;
  


  private final String localSenderHostPort;
  


  private final PacketWriter packetWriter;
  


  private final ArrayList<MessageResend> resendMessages = new ArrayList();
  



  private final ArrayList<MessageControl> controlMessages = new ArrayList();
  



  private final OutgoingPacketsCache outgoingPacketsCache = new OutgoingPacketsCache();
  



  private final IncomingPacketsLastAck incomingPacketsLastAck = new IncomingPacketsLastAck();
  






  private final int maxResendOutgoing;
  






  private long managerSleepMillis;
  






  private boolean sendWithNoMembers;
  






  private long minAcked;
  






  private long minAckedFromListener;
  






  private int currentGroupSize = -1;
  



  private long lastSendTime;
  



  private int lastSendTimeFreqMillis;
  



  private long lastStatusTime = System.currentTimeMillis();
  
  private int lastStatusTimeFreqMillis;
  
  private long totalTxnEventsSent;
  
  private long totalTxnEventsReceived;
  
  private long totalPacketsSent;
  
  private long totalBytesSent;
  
  private long totalPacketsResent;
  
  private long totalBytesResent;
  
  private long totalPacketsReceived;
  
  private long totalBytesReceived;
  

  public McastClusterManager()
  {
    managerSleepMillis = GlobalProperties.getInt("ebean.cluster.mcast.managerSleepMillis", 80);
    lastSendTimeFreqMillis = (1000 * GlobalProperties.getInt("ebean.cluster.mcast.pingFrequencySecs", 300));
    lastStatusTimeFreqMillis = (1000 * GlobalProperties.getInt("ebean.cluster.mcast.statusFrequencySecs", 600));
    

    maxResendOutgoing = GlobalProperties.getInt("ebean.cluster.mcast.maxResendOutgoing", 200);
    
    int maxResendIncoming = GlobalProperties.getInt("ebean.cluster.mcast.maxResendIncoming", 50);
    

    int port = GlobalProperties.getInt("ebean.cluster.mcast.listen.port", 0);
    String addr = GlobalProperties.get("ebean.cluster.mcast.listen.address", null);
    
    int sendPort = GlobalProperties.getInt("ebean.cluster.mcast.send.port", 0);
    String sendAddr = GlobalProperties.get("ebean.cluster.mcast.send.address", null);
    


    int maxSendPacketSize = GlobalProperties.getInt("ebean.cluster.mcast.send.maxPacketSize", 1500);
    
    sendWithNoMembers = GlobalProperties.getBoolean("ebean.cluster.mcast.send.sendWithNoMembers", true);
    


    boolean disableLoopback = GlobalProperties.getBoolean("ebean.cluster.mcast.listen.disableLoopback", false);
    int ttl = GlobalProperties.getInt("ebean.cluster.mcast.listen.ttl", -1);
    int timeout = GlobalProperties.getInt("ebean.cluster.mcast.listen.timeout", 1000);
    int bufferSize = GlobalProperties.getInt("ebean.cluster.mcast.listen.bufferSize", 65500);
    
    String mcastAddr = GlobalProperties.get("ebean.cluster.mcast.listen.mcastAddress", null);
    
    InetAddress mcastAddress = null;
    if (mcastAddr != null) {
      try {
        mcastAddress = InetAddress.getByName(mcastAddr);
      } catch (UnknownHostException e) {
        String msg = "Error getting Multicast InetAddress for " + mcastAddr;
        throw new RuntimeException(msg, e);
      }
    }
    
    if ((port == 0) || (addr == null)) {
      String msg = "One of these Multicast settings has not been set. ebean.cluster.mcast.listen.port=" + port + ", ebean.cluster.mcast.listen.address=" + addr;
      

      throw new IllegalArgumentException(msg);
    }
    
    managerThread = new Thread(this, "EbeanClusterMcastManager");
    
    packetWriter = new PacketWriter(maxSendPacketSize);
    localSender = new McastSender(port, addr, sendPort, sendAddr);
    localSenderHostPort = localSender.getSenderHostPort();
    
    packageControl = new McastPacketControl(this, localSenderHostPort, maxResendIncoming);
    
    listener = new McastListener(this, packageControl, port, addr, bufferSize, timeout, localSenderHostPort, disableLoopback, ttl, mcastAddress);
  }
  





  protected void fromListenerTimeoutNoMembers()
  {
    synchronized (managerThread) {
      currentGroupSize = 0;
    }
  }
  













  protected void fromListener(long newMinAcked, MessageControl msgControl, MessageResend msgResend, int groupSize, long totalPacketsReceived, long totalBytesReceived, long totalTxnEventsReceived)
  {
    synchronized (managerThread) {
      if (newMinAcked > minAckedFromListener) {
        minAckedFromListener = newMinAcked;
      }
      if (msgControl != null) {
        controlMessages.add(msgControl);
      }
      if (msgResend != null) {
        resendMessages.add(msgResend);
      }
      
      currentGroupSize = groupSize;
      

      this.totalPacketsReceived = totalPacketsReceived;
      this.totalBytesReceived = totalBytesReceived;
      this.totalTxnEventsReceived = totalTxnEventsReceived;
    }
  }
  


  public McastStatus getStatus(boolean reset)
  {
    synchronized (managerThread)
    {
      long currentPacketId = packetWriter.currentPacketId();
      String lastAcks = incomingPacketsLastAck.toString();
      
      return new McastStatus(currentGroupSize, outgoingPacketsCache.size(), currentPacketId, minAcked, lastAcks, totalTxnEventsSent, totalTxnEventsReceived, totalPacketsSent, totalPacketsResent, totalPacketsReceived, totalBytesSent, totalBytesResent, totalBytesReceived);
    }
  }
  



  public void run()
  {
    try
    {
      String msg;
      

      for (;;)
      {
        Thread.sleep(managerSleepMillis);
        
        synchronized (managerThread)
        {
          handleControlMessages();
          
          handleResendMessages();
          
          if (currentGroupSize == 0)
          {
            int trimmedCount = outgoingPacketsCache.trimAll();
            if (trimmedCount > 0) {
              logger.fine("Cluster has no other members. Trimmed " + trimmedCount);
            }
          }
          else if (minAckedFromListener > minAcked)
          {
            outgoingPacketsCache.trimAcknowledgedMessages(minAckedFromListener);
            minAcked = minAckedFromListener;
          }
          






          AckResendMessages ackResendMessages = packageControl.getAckResendMessages(incomingPacketsLastAck);
          
          if (ackResendMessages.size() > 0)
          {

            if (sendMessages(false, ackResendMessages.getMessages()))
            {
              incomingPacketsLastAck.updateLastAck(ackResendMessages);
            }
          }
          
          if (lastSendTime < System.currentTimeMillis() - lastSendTimeFreqMillis)
          {
            sendPing();
          }
          
          if ((lastStatusTimeFreqMillis > 0) && 
            (lastStatusTime < System.currentTimeMillis() - lastStatusTimeFreqMillis)) {
            McastStatus status = getStatus(false);
            logger.info("Cluster Status: " + status.getSummary());
            lastStatusTime = System.currentTimeMillis();
          }
        }
      }
    }
    catch (Exception e) {
      msg = "Error with Cluster Mcast Manager thread";
      logger.log(Level.SEVERE, msg, e);
    }
  }
  




  private void handleResendMessages()
  {
    if (resendMessages.size() > 0)
    {
      TreeSet<Long> s = new TreeSet();
      for (int i = 0; i < resendMessages.size(); i++) {
        MessageResend resendMsg = (MessageResend)resendMessages.get(i);
        s.addAll(resendMsg.getResendPacketIds());
      }
      
      totalPacketsResent += s.size();
      
      Iterator<Long> it = s.iterator();
      while (it.hasNext()) {
        Long resendPacketId = (Long)it.next();
        Packet packet = outgoingPacketsCache.getPacket(resendPacketId);
        if (packet == null) {
          String msg = "Cluster unable to resend packet[" + resendPacketId + "] as it is no longer in the outgoingPacketsCache";
          logger.log(Level.SEVERE, msg);
        } else {
          int resendCount = packet.incrementResendCount();
          if (resendCount <= maxResendOutgoing) {
            resendPacket(packet);
          } else {
            String msg = "Cluster maxResendOutgoing [" + maxResendOutgoing + "] hit for packet " + resendPacketId + ". We will not try to send it anymore, removing it from the outgoingPacketsCache.";
            
            logger.log(Level.SEVERE, msg);
            outgoingPacketsCache.remove(packet);
          }
        }
      }
    }
  }
  

  private void resendPacket(Packet packet)
  {
    try
    {
      totalPacketsResent += 1L;
      totalBytesResent += localSender.sendPacket(packet);
    } catch (IOException e) {
      String msg = "Error trying to resend packet " + packet.getPacketId();
      logger.log(Level.SEVERE, msg, e);
    }
  }
  



  private void handleControlMessages()
  {
    boolean pingReponse = false;
    boolean joinReponse = false;
    
    for (int i = 0; i < controlMessages.size(); i++) {
      MessageControl message = (MessageControl)controlMessages.get(i);
      
      short type = message.getControlType();
      switch (type)
      {
      case 1: 
        logger.info("Cluster member Joined [" + message.getFromHostPort() + "]");
        joinReponse = true;
        break;
      
      case 7: 
        logger.info("Cluster member Online [" + message.getFromHostPort() + "]");
        
        break;
      
      case 3: 
        pingReponse = true;
        break;
      

      case 8: 
        break;
      


      case 2: 
        incomingPacketsLastAck.remove(message.getFromHostPort());
      }
      
    }
    


    controlMessages.clear();
    
    if (joinReponse) {
      sendJoinResponse();
    }
    if (pingReponse) {
      sendPingResponse();
    }
  }
  


  public void shutdown()
  {
    sendLeave();
    listener.shutdown();
  }
  


  public void startup(ClusterManager clusterManager)
  {
    this.clusterManager = clusterManager;
    listener.startListening();
    
    managerThread.setDaemon(true);
    managerThread.start();
    
    sendJoin();
  }
  
  protected SpiEbeanServer getEbeanServer(String serverName) {
    return (SpiEbeanServer)clusterManager.getServer(serverName);
  }
  
  private void sendJoin() {
    sendControlMessage(true, (short)1);
  }
  
  private void sendLeave() {
    sendControlMessage(false, (short)2);
  }
  
  private void sendJoinResponse() {
    sendControlMessage(true, (short)7);
  }
  
  private void sendPingResponse() {
    sendControlMessage(true, (short)8);
  }
  
  private void sendPing() {
    sendControlMessage(true, (short)3);
  }
  
  private void sendControlMessage(boolean requiresAck, short controlType) {
    sendMessage(requiresAck, new MessageControl(controlType, localSenderHostPort));
  }
  
  private void sendMessage(boolean requiresAck, Message msg) {
    ArrayList<Message> messages = new ArrayList(1);
    messages.add(msg);
    sendMessages(requiresAck, messages);
  }
  
  private boolean sendMessages(boolean requiresAck, List<? extends Message> messages)
  {
    synchronized (managerThread)
    {
      try {
        List<Packet> packets = packetWriter.write(requiresAck, messages);
        sendPackets(requiresAck, packets);
        return true;
      }
      catch (IOException e) {
        String msg = "Error sending Messages " + messages;
        logger.log(Level.SEVERE, msg, e);
        return false;
      }
    }
  }
  
  private boolean sendPackets(boolean requiresAck, List<Packet> packets) throws IOException {
    if ((currentGroupSize == 0) && (!sendWithNoMembers))
    {
      return false;
    }
    
    if (requiresAck)
    {
      outgoingPacketsCache.registerPackets(packets);
    }
    totalPacketsSent += packets.size();
    totalBytesSent += localSender.sendPackets(packets);
    
    lastSendTime = System.currentTimeMillis();
    
    return true;
  }
  




  public void broadcast(RemoteTransactionEvent remoteTransEvent)
  {
    synchronized (managerThread) {
      try {
        List<Packet> packets = packetWriter.write(remoteTransEvent);
        if (sendPackets(true, packets)) {
          totalTxnEventsSent += 1L;
        }
      } catch (IOException e) {
        String msg = "Error sending RemoteTransactionEvent " + remoteTransEvent;
        logger.log(Level.SEVERE, msg, e);
      }
    }
  }
  


  public void setManagerSleepMillis(long managerSleepMillis)
  {
    synchronized (managerThread) {
      this.managerSleepMillis = managerSleepMillis;
    }
  }
  


  public long getManagerSleepMillis()
  {
    synchronized (managerThread) {
      return managerSleepMillis;
    }
  }
}
