package com.avaje.ebeaninternal.server.lib.sql;

import com.avaje.ebean.config.DataSourceConfig;
import com.avaje.ebean.config.GlobalProperties;
import com.avaje.ebeaninternal.api.ClassUtil;
import com.avaje.ebeaninternal.server.lib.BackgroundRunnable;
import com.avaje.ebeaninternal.server.lib.BackgroundThread;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;





















public class DataSourceManager
  implements DataSourceNotify
{
  private static final Logger logger = Logger.getLogger(DataSourceManager.class.getName());
  



  private final DataSourceAlertListener alertlistener;
  



  private final Hashtable<String, DataSourcePool> dsMap = new Hashtable();
  



  private final Object monitor = new Object();
  



  private final BackgroundRunnable dbChecker;
  



  private final int dbUpFreqInSecs;
  



  private final int dbDownFreqInSecs;
  



  private boolean shuttingDown;
  



  public DataSourceManager()
  {
    alertlistener = createAlertListener();
    

    dbUpFreqInSecs = GlobalProperties.getInt("datasource.heartbeatfreq", 30);
    dbDownFreqInSecs = GlobalProperties.getInt("datasource.deadbeatfreq", 10);
    dbChecker = new BackgroundRunnable(new Checker(null), dbUpFreqInSecs);
    try
    {
      BackgroundThread.add(dbChecker);
    }
    catch (Exception e) {
      logger.log(Level.SEVERE, null, e);
    }
  }
  
  private DataSourceAlertListener createAlertListener() throws DataSourceException
  {
    String alertCN = GlobalProperties.get("datasource.alert.class", null);
    if (alertCN == null) {
      return new SimpleAlerter();
    }
    try
    {
      return (DataSourceAlertListener)ClassUtil.newInstance(alertCN, getClass());
    }
    catch (Exception ex) {
      throw new DataSourceException(ex);
    }
  }
  




  public void notifyDataSourceUp(String dataSourceName)
  {
    dbChecker.setFreqInSecs(dbUpFreqInSecs);
    
    if (alertlistener != null) {
      alertlistener.dataSourceUp(dataSourceName);
    }
  }
  



  public void notifyDataSourceDown(String dataSourceName)
  {
    dbChecker.setFreqInSecs(dbDownFreqInSecs);
    
    if (alertlistener != null) {
      alertlistener.dataSourceDown(dataSourceName);
    }
  }
  


  public void notifyWarning(String subject, String msg)
  {
    if (alertlistener != null) {
      alertlistener.warning(subject, msg);
    }
  }
  


  public boolean isShuttingDown()
  {
    synchronized (monitor) {
      return shuttingDown;
    }
  }
  



  public void shutdown()
  {
    synchronized (monitor)
    {
      shuttingDown = true;
      
      Iterator<DataSourcePool> it = dsMap.values().iterator();
      while (it.hasNext()) {
        try {
          DataSourcePool ds = (DataSourcePool)it.next();
          ds.shutdown();
        }
        catch (DataSourceException e)
        {
          logger.log(Level.SEVERE, null, e);
        }
      }
    }
  }
  


  public List<DataSourcePool> getPools()
  {
    synchronized (monitor)
    {
      ArrayList<DataSourcePool> list = new ArrayList();
      list.addAll(dsMap.values());
      return list;
    }
  }
  


  public DataSourcePool getDataSource(String name)
  {
    return getDataSource(name, null);
  }
  

  public DataSourcePool getDataSource(String name, DataSourceConfig dsConfig)
  {
    if (name == null) {
      throw new IllegalArgumentException("name not defined");
    }
    
    synchronized (monitor) {
      DataSourcePool pool = (DataSourcePool)dsMap.get(name);
      if (pool == null) {
        if (dsConfig == null) {
          dsConfig = new DataSourceConfig();
          dsConfig.loadSettings(name);
        }
        pool = new DataSourcePool(this, name, dsConfig);
        dsMap.put(name, pool);
      }
      return pool;
    }
  }
  




  private void checkDataSource()
  {
    synchronized (monitor) {
      if (!isShuttingDown()) {
        Iterator<DataSourcePool> it = dsMap.values().iterator();
        while (it.hasNext()) {
          DataSourcePool ds = (DataSourcePool)it.next();
          ds.checkDataSource();
        }
      }
    }
  }
  
  private final class Checker implements Runnable
  {
    private Checker() {}
    
    public void run()
    {
      DataSourceManager.this.checkDataSource();
    }
  }
}
