package com.avaje.ebeaninternal.server.lib.sql;

import com.avaje.ebean.config.DataSourceConfig;
import com.avaje.ebeaninternal.api.ClassUtil;
import com.avaje.ebeaninternal.server.lib.cron.CronManager;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.PersistenceException;
import javax.sql.DataSource;

































public class DataSourcePool
  implements DataSource
{
  private static final Logger logger = Logger.getLogger(DataSourcePool.class.getName());
  




  private final String name;
  




  private final DataSourceNotify notify;
  



  private final DataSourcePoolListener poolListener;
  



  private final Properties connectionProps;
  



  private final String databaseUrl;
  



  private final String databaseDriver;
  



  private final String heartbeatsql;
  



  private final int transactionIsolation;
  



  private final boolean autoCommit;
  



  private boolean captureStackTrace;
  



  private int maxStackTraceSize;
  



  private boolean dataSourceDownAlertSent;
  



  private long lastTrimTime;
  



  private boolean dataSourceUp = true;
  



  private boolean inWarningMode;
  



  private int minConnections;
  



  private int maxConnections;
  



  private int warningSize;
  



  private int waitTimeoutMillis;
  


  private int pstmtCacheSize;
  


  private int maxInactiveTimeSecs;
  


  private final PooledConnectionQueue queue;
  


  private long leakTimeMinutes;
  



  public DataSourcePool(DataSourceNotify notify, String name, DataSourceConfig params)
  {
    this.notify = notify;
    this.name = name;
    poolListener = createPoolListener(params.getPoolListener());
    
    autoCommit = false;
    transactionIsolation = params.getIsolationLevel();
    
    maxInactiveTimeSecs = params.getMaxInactiveTimeSecs();
    leakTimeMinutes = params.getLeakTimeMinutes();
    captureStackTrace = params.isCaptureStackTrace();
    maxStackTraceSize = params.getMaxStackTraceSize();
    databaseDriver = params.getDriver();
    databaseUrl = params.getUrl();
    pstmtCacheSize = params.getPstmtCacheSize();
    
    minConnections = params.getMinConnections();
    maxConnections = params.getMaxConnections();
    waitTimeoutMillis = params.getWaitTimeoutMillis();
    heartbeatsql = params.getHeartbeatSql();
    
    queue = new PooledConnectionQueue(this);
    
    String un = params.getUsername();
    String pw = params.getPassword();
    if (un == null) {
      throw new RuntimeException("DataSource user is null?");
    }
    if (pw == null) {
      throw new RuntimeException("DataSource password is null?");
    }
    connectionProps = new Properties();
    connectionProps.setProperty("user", un);
    connectionProps.setProperty("password", pw);
    try
    {
      initialise();
    } catch (SQLException ex) {
      throw new DataSourceException(ex);
    }
  }
  


  private DataSourcePoolListener createPoolListener(String cn)
  {
    if (cn == null) {
      return null;
    }
    try {
      return (DataSourcePoolListener)ClassUtil.newInstance(cn, getClass());
    } catch (Exception e) {
      throw new DataSourceException(e);
    }
  }
  
  private void initialise() throws SQLException
  {
    try
    {
      ClassUtil.forName(databaseDriver, getClass());
    } catch (Throwable e) {
      throw new PersistenceException("Problem loading Database Driver [" + databaseDriver + "]: " + e.getMessage(), e);
    }
    

    String transIsolation = TransactionIsolation.getLevelDescription(transactionIsolation);
    StringBuffer sb = new StringBuffer();
    sb.append("DataSourcePool [").append(name);
    sb.append("] autoCommit[").append(autoCommit);
    sb.append("] transIsolation[").append(transIsolation);
    sb.append("] min[").append(minConnections);
    sb.append("] max[").append(maxConnections).append("]");
    
    logger.info(sb.toString());
    
    queue.ensureMinimumConnections();
  }
  

  public boolean isWrapperFor(Class<?> arg0)
    throws SQLException
  {
    return false;
  }
  

  public <T> T unwrap(Class<T> arg0)
    throws SQLException
  {
    throw new SQLException("Not Implemented");
  }
  


  public String getName()
  {
    return name;
  }
  





  public int getMaxStackTraceSize()
  {
    return maxStackTraceSize;
  }
  


  public boolean isDataSourceUp()
  {
    return dataSourceUp;
  }
  



  protected void notifyWarning(String msg)
  {
    if (!inWarningMode)
    {
      inWarningMode = true;
      logger.warning(msg);
      if (notify != null) {
        String subject = "DataSourcePool [" + name + "] warning";
        notify.notifyWarning(subject, msg);
      }
    }
  }
  
  private void notifyDataSourceIsDown(SQLException ex)
  {
    if (isExpectedToBeDownNow()) {
      if (dataSourceUp) {
        String msg = "DataSourcePool [" + name + "] is down but in downtime!";
        logger.log(Level.WARNING, msg, ex);
      }
    }
    else if (!dataSourceDownAlertSent)
    {
      String msg = "FATAL: DataSourcePool [" + name + "] is down!!!";
      logger.log(Level.SEVERE, msg, ex);
      if (notify != null) {
        notify.notifyDataSourceDown(name);
      }
      dataSourceDownAlertSent = true;
    }
    
    if (dataSourceUp) {
      reset();
    }
    dataSourceUp = false;
  }
  
  private void notifyDataSourceIsUp() {
    if (dataSourceDownAlertSent) {
      String msg = "RESOLVED FATAL: DataSourcePool [" + name + "] is back up!";
      logger.log(Level.SEVERE, msg);
      if (notify != null) {
        notify.notifyDataSourceUp(name);
      }
      dataSourceDownAlertSent = false;
    }
    else if (!dataSourceUp) {
      logger.log(Level.WARNING, "DataSourcePool [" + name + "] is back up!");
    }
    
    if (!dataSourceUp) {
      dataSourceUp = true;
      reset();
    }
  }
  


  protected void checkDataSource()
  {
    Connection conn = null;
    try
    {
      conn = getConnection();
      testConnection(conn);
      
      notifyDataSourceIsUp();
      
      if (System.currentTimeMillis() > lastTrimTime + maxInactiveTimeSecs * 1000) {
        queue.trim(maxInactiveTimeSecs);
        lastTrimTime = System.currentTimeMillis();
      }
      return;
    } catch (SQLException ex) {
      notifyDataSourceIsDown(ex);
    } finally {
      try {
        if (conn != null) {
          conn.close();
        }
      } catch (SQLException ex) {
        logger.log(Level.WARNING, "Can't close connection in checkDataSource!");
      }
    }
  }
  



  private boolean isExpectedToBeDownNow()
  {
    return CronManager.isDowntime();
  }
  










  public Connection createUnpooledConnection()
    throws SQLException
  {
    try
    {
      Connection conn = DriverManager.getConnection(databaseUrl, connectionProps);
      conn.setAutoCommit(autoCommit);
      conn.setTransactionIsolation(transactionIsolation);
      return conn;
    }
    catch (SQLException ex) {
      notifyDataSourceIsDown(null);
      throw ex;
    }
  }
  




  public void setMaxSize(int max)
  {
    queue.setMaxSize(max);
    maxConnections = max;
  }
  


  public int getMaxSize()
  {
    return maxConnections;
  }
  


  public void setMinSize(int min)
  {
    queue.setMinSize(min);
    minConnections = min;
  }
  


  public int getMinSize()
  {
    return minConnections;
  }
  





  public void setWarningSize(int warningSize)
  {
    queue.setWarningSize(warningSize);
    this.warningSize = warningSize;
  }
  



  public int getWarningSize()
  {
    return warningSize;
  }
  




  public int getWaitTimeoutMillis()
  {
    return waitTimeoutMillis;
  }
  


  public void setMaxInactiveTimeSecs(int maxInactiveTimeSecs)
  {
    this.maxInactiveTimeSecs = maxInactiveTimeSecs;
  }
  


  public int getMaxInactiveTimeSecs()
  {
    return maxInactiveTimeSecs;
  }
  
  private void testConnection(Connection conn) throws SQLException
  {
    if (heartbeatsql == null) {
      return;
    }
    Statement stmt = null;
    ResultSet rset = null;
    
    try
    {
      stmt = conn.createStatement();
      rset = stmt.executeQuery(heartbeatsql);
      conn.commit(); return;
    }
    finally {
      try {
        if (rset != null) {
          rset.close();
        }
      } catch (SQLException e) {
        logger.log(Level.SEVERE, null, e);
      }
      try {
        if (stmt != null) {
          stmt.close();
        }
      } catch (SQLException e) {
        logger.log(Level.SEVERE, null, e);
      }
    }
  }
  


  protected boolean validateConnection(PooledConnection conn)
  {
    try
    {
      if (heartbeatsql == null) {
        logger.info("Can not test connection as heartbeatsql is not set");
        return false;
      }
      
      testConnection(conn);
      return true;
    }
    catch (Exception e) {
      String desc = "heartbeatsql test failed on connection[" + conn.getName() + "]";
      logger.warning(desc); }
    return false;
  }
  














  protected void returnConnection(PooledConnection pooledConnection)
  {
    if (poolListener != null) {
      poolListener.onBeforeReturnConnection(pooledConnection);
    }
    queue.returnPooledConnection(pooledConnection);
  }
  



  public String getBusyConnectionInformation()
  {
    return queue.getBusyConnectionInformation();
  }
  







  public void dumpBusyConnectionInformation()
  {
    queue.dumpBusyConnectionInformation();
  }
  












  public void closeBusyConnections(long leakTimeMinutes)
  {
    queue.closeBusyConnections(leakTimeMinutes);
  }
  





  protected PooledConnection createConnectionForQueue(int connId)
    throws SQLException
  {
    try
    {
      Connection c = createUnpooledConnection();
      
      PooledConnection pc = new PooledConnection(this, connId, c);
      pc.resetForUse();
      
      if (!dataSourceUp) {
        notifyDataSourceIsUp();
      }
      return pc;
    }
    catch (SQLException ex) {
      notifyDataSourceIsDown(ex);
      throw ex;
    }
  }
  













  public void reset()
  {
    queue.reset(leakTimeMinutes);
    inWarningMode = false;
  }
  

  public Connection getConnection()
    throws SQLException
  {
    return getPooledConnection();
  }
  






  public PooledConnection getPooledConnection()
    throws SQLException
  {
    PooledConnection c = queue.getPooledConnection();
    
    if (captureStackTrace) {
      c.setStackTrace(Thread.currentThread().getStackTrace());
    }
    
    if (poolListener != null) {
      poolListener.onAfterBorrowConnection(c);
    }
    return c;
  }
  




  public void testAlert()
  {
    String subject = "Test DataSourcePool [" + name + "]";
    String msg = "Just testing if alert message is sent successfully.";
    
    if (notify != null) {
      notify.notifyWarning(subject, msg);
    }
  }
  








  public void shutdown()
  {
    queue.shutdown();
  }
  




  public boolean getAutoCommit()
  {
    return autoCommit;
  }
  





  public int getTransactionIsolation()
  {
    return transactionIsolation;
  }
  






  public boolean isCaptureStackTrace()
  {
    return captureStackTrace;
  }
  




  public void setCaptureStackTrace(boolean captureStackTrace)
  {
    this.captureStackTrace = captureStackTrace;
  }
  

  public Connection getConnection(String username, String password)
    throws SQLException
  {
    throw new SQLException("Method not supported");
  }
  

  public int getLoginTimeout()
    throws SQLException
  {
    throw new SQLException("Method not supported");
  }
  

  public void setLoginTimeout(int seconds)
    throws SQLException
  {
    throw new SQLException("Method not supported");
  }
  


  public PrintWriter getLogWriter()
  {
    return null;
  }
  

  public void setLogWriter(PrintWriter writer)
    throws SQLException
  {
    throw new SQLException("Method not supported");
  }
  









  public void setLeakTimeMinutes(long leakTimeMinutes)
  {
    this.leakTimeMinutes = leakTimeMinutes;
  }
  



  public long getLeakTimeMinutes()
  {
    return leakTimeMinutes;
  }
  


  public int getPstmtCacheSize()
  {
    return pstmtCacheSize;
  }
  


  public void setPstmtCacheSize(int pstmtCacheSize)
  {
    this.pstmtCacheSize = pstmtCacheSize;
  }
  






  public Status getStatus(boolean reset)
  {
    return queue.getStatus(reset);
  }
  
  public static class Status
  {
    private final String name;
    private final int minSize;
    private final int maxSize;
    private final int free;
    private final int busy;
    private final int waiting;
    private final int highWaterMark;
    private final int waitCount;
    private final int hitCount;
    
    protected Status(String name, int minSize, int maxSize, int free, int busy, int waiting, int highWaterMark, int waitCount, int hitCount)
    {
      this.name = name;
      this.minSize = minSize;
      this.maxSize = maxSize;
      this.free = free;
      this.busy = busy;
      this.waiting = waiting;
      this.highWaterMark = highWaterMark;
      this.waitCount = waitCount;
      this.hitCount = hitCount;
    }
    
    public String toString() {
      return "min:" + minSize + " max:" + maxSize + " free:" + free + " busy:" + busy + " waiting:" + waiting + " highWaterMark:" + highWaterMark + " waitCount:" + waitCount + " hitCount:" + hitCount;
    }
    



    public String getName()
    {
      return name;
    }
    


    public int getMinSize()
    {
      return minSize;
    }
    


    public int getMaxSize()
    {
      return maxSize;
    }
    


    public int getFree()
    {
      return free;
    }
    


    public int getBusy()
    {
      return busy;
    }
    


    public int getWaiting()
    {
      return waiting;
    }
    


    public int getHighWaterMark()
    {
      return highWaterMark;
    }
    


    public int getWaitCount()
    {
      return waitCount;
    }
    







    public int getHitCount()
    {
      return hitCount;
    }
  }
}
