package com.avaje.ebeaninternal.server.lib.thread;

import java.util.logging.Level;
import java.util.logging.Logger;





















public class PooledThread
  implements Runnable
{
  private static final Logger logger = Logger.getLogger(PooledThread.class.getName());
  




  protected PooledThread(ThreadPool threadPool, String name, boolean isDaemon, Integer threadPriority)
  {
    this.name = name;
    this.threadPool = threadPool;
    lastUsedTime = System.currentTimeMillis();
    
    thread = new Thread(this, name);
    thread.setDaemon(isDaemon);
    
    if (threadPriority != null) {
      thread.setPriority(threadPriority.intValue());
    }
  }
  
  protected void start()
  {
    thread.start();
  }
  



  public boolean assignWork(Work work)
  {
    synchronized (workMonitor) {
      this.work = work;
      workMonitor.notifyAll();
    }
    return true;
  }
  



  public void run()
  {
    synchronized (workMonitor) {
      while (!isStopping) {
        try {
          if (work == null) {
            workMonitor.wait();
          }
        }
        catch (InterruptedException e) {}
        doTheWork();
      }
    }
    

    synchronized (threadMonitor) {
      threadMonitor.notifyAll();
    }
    
    isStopped = true;
  }
  


  private void doTheWork()
  {
    if (isStopping) {
      return;
    }
    
    long startTime = System.currentTimeMillis();
    if (work != null)
    {
      try
      {

        work.setStartTime(startTime);
        work.getRunnable().run();
      }
      catch (Throwable ex) {
        logger.log(Level.SEVERE, null, ex);
        
        if (wasInterrupted) {
          isStopping = true;
          threadPool.removeThread(this);
          logger.info("PooledThread [" + name + "] removed due to interrupt");
          try {
            thread.interrupt();
          } catch (Exception e) {
            String msg = "Error interrupting PooledThead[" + name + "]";
            logger.log(Level.SEVERE, msg, e);
          }
          return;
        }
      }
    }
    lastUsedTime = System.currentTimeMillis();
    totalWorkCount += 1;
    totalWorkExecutionTime = (totalWorkExecutionTime + lastUsedTime - startTime);
    work = null;
    threadPool.returnThread(this);
  }
  









  public void interrupt()
  {
    wasInterrupted = true;
    try {
      thread.interrupt();
    }
    catch (SecurityException ex) {
      wasInterrupted = false;
      throw ex;
    }
  }
  


  public boolean isStopped()
  {
    return isStopped;
  }
  



  protected void stop()
  {
    isStopping = true;
    
    synchronized (threadMonitor)
    {
      assignWork(null);
      try
      {
        threadMonitor.wait(10000L);
      }
      catch (InterruptedException e) {}
    }
    


    thread = null;
    threadPool.removeThread(this);
  }
  


  public String getName()
  {
    return name;
  }
  

  public Work getWork()
  {
    return work;
  }
  


  public int getTotalWorkCount()
  {
    return totalWorkCount;
  }
  


  public long getTotalWorkExecutionTime()
  {
    return totalWorkExecutionTime;
  }
  


  public long getLastUsedTime()
  {
    return lastUsedTime;
  }
  



  private boolean wasInterrupted = false;
  



  private long lastUsedTime;
  



  private Work work = null;
  



  private boolean isStopping = false;
  



  private boolean isStopped = false;
  


  private Thread thread = null;
  



  private ThreadPool threadPool;
  



  private String name = null;
  



  private Object threadMonitor = new Object();
  



  private Object workMonitor = new Object();
  



  private int totalWorkCount = 0;
  



  private long totalWorkExecutionTime = 0L;
}
