package com.avaje.ebeaninternal.server.lib.thread;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;


























public class ThreadPool
{
  private static final Logger logger = Logger.getLogger(ThreadPool.class.getName());
  



  private long maxIdleTime;
  



  private String poolName;
  



  private int minSize;
  



  private boolean isDaemon;
  



  private boolean isStopping = false;
  




  private Integer threadPriority;
  




  private int uniqueThreadID;
  



  private Vector<PooledThread> freeList = new Vector();
  



  private Vector<PooledThread> busyList = new Vector();
  



  private Vector<Work> workOverflowQueue = new Vector();
  




  private int maxSize = 100;
  



  private boolean stopThePool;
  




  public ThreadPool(String poolName, boolean isDaemon, Integer threadPriority)
  {
    this.poolName = poolName;
    stopThePool = false;
    this.isDaemon = isDaemon;
    this.threadPriority = threadPriority;
  }
  


  public boolean isStopping()
  {
    return isStopping;
  }
  


  public String getName()
  {
    return poolName;
  }
  


  public void setMinSize(int minSize)
  {
    if (minSize > 0) {
      if (minSize > maxSize) {
        maxSize = minSize;
      }
      this.minSize = minSize;
      maintainPoolSize();
    }
  }
  


  public int getMinSize()
  {
    return minSize;
  }
  


  public void setMaxSize(int maxSize)
  {
    if (maxSize > 0) {
      if (minSize > maxSize) {
        minSize = maxSize;
      }
      this.maxSize = maxSize;
      maintainPoolSize();
    }
  }
  


  public int getMaxSize()
  {
    return maxSize;
  }
  


  public int size()
  {
    return busyList.size() + freeList.size();
  }
  


  public int getBusyCount()
  {
    return busyList.size();
  }
  















  public boolean assign(Runnable work, boolean addToQueueIfFull)
  {
    if (stopThePool) {
      throw new RuntimeException("Pool is stopping... no more work please.");
    }
    
    Work runWork = new Work(work);
    

    PooledThread thread = getNextAvailableThread();
    if (thread != null)
    {
      busyList.add(thread);
      thread.assignWork(runWork);
      return true;
    }
    
    if (addToQueueIfFull) {
      runWork.setEnterQueueTime(System.currentTimeMillis());
      workOverflowQueue.add(runWork);
    }
    return false;
  }
  




  protected void removeThread(PooledThread thread)
  {
    synchronized (freeList) {
      busyList.remove(thread);
      freeList.remove(thread);
      freeList.notify();
    }
  }
  








  protected void returnThread(PooledThread thread)
  {
    synchronized (freeList)
    {

      busyList.remove(thread);
      
      if (!workOverflowQueue.isEmpty())
      {
        Work queuedWork = (Work)workOverflowQueue.remove(0);
        

        queuedWork.setExitQueueTime(System.currentTimeMillis());
        busyList.add(thread);
        thread.assignWork(queuedWork);
      }
      else
      {
        freeList.add(thread);
        
        freeList.notify();
      }
    }
  }
  






  private PooledThread getNextAvailableThread()
  {
    synchronized (freeList) {
      if (!freeList.isEmpty()) {
        return (PooledThread)freeList.remove(0);
      }
      if (size() < maxSize) {
        return growPool(true);
      }
      return null;
    }
  }
  






  public Iterator<PooledThread> getBusyThreads()
  {
    synchronized (freeList) {
      return busyList.iterator();
    }
  }
  




  protected void shutdown()
  {
    synchronized (freeList) {
      isStopping = true;
      
      int size = size();
      
      if (size > 0) {
        String msg = null;
        msg = "ThreadPool [" + poolName + "] Shutting down; threadCount[" + size() + "] busyCount[" + getBusyCount() + "]";
        

        logger.info(msg);
      }
      
      stopThePool = true;
      
      while (!freeList.isEmpty()) {
        PooledThread thread = (PooledThread)freeList.remove(0);
        thread.stop();
      }
      try
      {
        while (getBusyCount() > 0)
        {
          String msg = "ThreadPool [" + poolName + "] has [" + getBusyCount() + "] busy threads, waiting for those to finish.";
          
          logger.info(msg);
          
          Iterator<PooledThread> busyThreads = getBusyThreads();
          while (busyThreads.hasNext()) {
            PooledThread busyThread = (PooledThread)busyThreads.next();
            
            String threadName = busyThread.getName();
            Work work = busyThread.getWork();
            
            String busymsg = "Busy thread [" + threadName + "] work[" + work + "]";
            logger.info(busymsg);
          }
          

          freeList.wait();
          PooledThread thread = (PooledThread)freeList.remove(0);
          

          if (thread != null) {
            thread.stop();
          }
        }
      }
      catch (InterruptedException e)
      {
        logger.log(Level.SEVERE, null, e);
      }
    }
  }
  


  protected void maintainPoolSize()
  {
    synchronized (freeList) {
      if (isStopping)
      {
        return;
      }
      
      int numToStop = size() - minSize;
      if (numToStop > 0)
      {
        long usedAfter = System.currentTimeMillis() - maxIdleTime;
        ArrayList<PooledThread> stopList = new ArrayList();
        Iterator<PooledThread> it = freeList.iterator();
        while ((it.hasNext()) && (numToStop > 0)) {
          PooledThread thread = (PooledThread)it.next();
          if (thread.getLastUsedTime() < usedAfter) {
            stopList.add(thread);
            numToStop--;
          }
        }
        Iterator<PooledThread> stopIt = stopList.iterator();
        while (stopIt.hasNext()) {
          PooledThread thread = (PooledThread)stopIt.next();
          thread.stop();
        }
      }
      int numToAdd = minSize - size();
      if (numToAdd > 0)
      {
        for (int i = 0; i < numToAdd; i++) {
          growPool(false);
        }
      }
    }
  }
  
























  public PooledThread interrupt(String threadName)
  {
    PooledThread thread = getBusyThread(threadName);
    if (thread != null) {
      thread.interrupt();
      return thread;
    }
    return null;
  }
  



  public PooledThread getBusyThread(String threadName)
  {
    synchronized (freeList) {
      Iterator<PooledThread> it = getBusyThreads();
      while (it.hasNext()) {
        PooledThread pt = (PooledThread)it.next();
        if (pt.getName().equals(threadName)) {
          return pt;
        }
      }
      return null;
    }
  }
  




  private PooledThread growPool(boolean andReturn)
  {
    synchronized (freeList)
    {
      String threadName = poolName + "." + uniqueThreadID++;
      PooledThread bgw = new PooledThread(this, threadName, isDaemon, threadPriority);
      bgw.start();
      
      if (logger.isLoggable(Level.FINE)) {
        logger.fine("ThreadPool grow created [" + threadName + "] size[" + size() + "]");
      }
      if (andReturn) {
        return bgw;
      }
      freeList.add(bgw);
      return null;
    }
  }
  




  public long getMaxIdleTime()
  {
    return maxIdleTime;
  }
  



  public void setMaxIdleTime(long maxIdleTime)
  {
    this.maxIdleTime = maxIdleTime;
  }
}
