package com.avaje.ebeaninternal.server.lucene;

import com.avaje.ebean.BackgroundExecutor;
import com.avaje.ebean.Query.UseIndex;
import com.avaje.ebean.config.lucene.IndexDefn;
import com.avaje.ebeaninternal.api.SpiEbeanServer;
import com.avaje.ebeaninternal.api.SpiTransaction;
import com.avaje.ebeaninternal.server.cluster.ClusterManager;
import com.avaje.ebeaninternal.server.cluster.LuceneClusterIndexSync;
import com.avaje.ebeaninternal.server.deploy.BeanDescriptor;
import com.avaje.ebeaninternal.server.transaction.IndexEvent;
import com.avaje.ebeaninternal.server.transaction.RemoteTransactionEvent;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.lucene.analysis.Analyzer;




















public class DefaultLuceneIndexManager
  implements LuceneIndexManager, Runnable
{
  private static final Logger logger = Logger.getLogger(DefaultLuceneIndexManager.class.getName());
  
  private final ConcurrentHashMap<String, LIndex> indexMap;
  
  private final ClusterManager clusterManager;
  
  private final LuceneClusterIndexSync clusterIndexSync;
  
  private final BackgroundExecutor backgroundExecutor;
  
  private final Analyzer defaultAnalyzer;
  
  private final String baseDir;
  
  private final LIndexFactory indexFactory;
  
  private final boolean luceneAvailable;
  
  private final Query.UseIndex defaultUseIndex;
  
  private final String serverName;
  
  private SpiEbeanServer server;
  
  private Thread thread;
  
  private volatile boolean shutdown;
  
  private volatile boolean shutdownComplete;
  
  private long manageFreqMillis = 100L;
  




  public DefaultLuceneIndexManager(ClusterManager clusterManager, BackgroundExecutor backgroundExecutor, Analyzer defaultAnalyzer, String baseDir, String serverName, Query.UseIndex defaultUseIndex)
  {
    luceneAvailable = true;
    this.serverName = serverName;
    this.clusterManager = clusterManager;
    clusterIndexSync = clusterManager.getLuceneClusterIndexSync();
    this.backgroundExecutor = backgroundExecutor;
    this.defaultUseIndex = defaultUseIndex;
    this.defaultAnalyzer = defaultAnalyzer;
    this.baseDir = (baseDir + File.separator + serverName + File.separator);
    indexMap = new ConcurrentHashMap();
    indexFactory = new LIndexFactory(this);
    
    thread = new Thread(this, "Ebean-" + serverName + "-LuceneManager");
  }
  
  public void notifyCluster(IndexEvent event)
  {
    if ((clusterIndexSync != null) && (clusterIndexSync.isMaster()))
    {
      System.out.println("-- notifyCluster commit ... ");
      
      RemoteTransactionEvent e = new RemoteTransactionEvent(serverName);
      e.addIndexEvent(event);
      clusterManager.broadcast(e);
    }
  }
  
  protected void execute(LIndexSync indexSync) {
    if (clusterIndexSync != null) {
      IndexSynchRun r = new IndexSynchRun(clusterIndexSync, indexSync, null);
      backgroundExecutor.execute(r);
    }
  }
  
  public void processEvent(IndexEvent indexEvent)
  {
    if (clusterIndexSync == null) {
      return;
    }
    
    String masterHost = clusterIndexSync.getMasterHost();
    if (masterHost == null) {
      logger.warning("Master got IndexEvent " + indexEvent + " ?");
    }
    else {
      String idxName = indexEvent.getIndexName();
      if (idxName != null) {
        LIndex index = getIndex(idxName);
        if (index == null) {
          logger.warning("Can't find Lucene Index [" + idxName + "]");
        } else {
          index.queueSync(masterHost);
        }
      }
    }
  }
  
  public void processEvent(RemoteTransactionEvent txnEvent, SpiTransaction localTransaction)
  {
    Collection<IndexUpdates> events = IndexUpdatesBuilder.create(server, txnEvent);
    for (IndexUpdates e : events) {
      BeanDescriptor<?> beanDescriptor = e.getBeanDescriptor();
      LIndex luceneIndex = beanDescriptor.getLuceneIndex();
      if (luceneIndex != null) {
        LIndexUpdateFuture future = luceneIndex.process(e);
        if (localTransaction != null) {
          localTransaction.addIndexUpdateFuture(future);
        }
      }
    }
  }
  
  public LuceneClusterIndexSync getClusterIndexSync()
  {
    return clusterIndexSync;
  }
  
  public boolean isLuceneAvailable() {
    return luceneAvailable;
  }
  
  public Query.UseIndex getDefaultUseIndex() {
    return defaultUseIndex;
  }
  
  public LIndex create(IndexDefn<?> indexDefn, BeanDescriptor<?> descriptor) throws IOException {
    return indexFactory.create(indexDefn, descriptor);
  }
  
  public SpiEbeanServer getServer() {
    return server;
  }
  
  public void setServer(SpiEbeanServer server) {
    this.server = server;
  }
  
  public Analyzer getDefaultAnalyzer() {
    return defaultAnalyzer;
  }
  
  public void addIndex(LIndex index) throws IOException {
    synchronized (indexMap) {
      indexMap.put(index.getName(), index);
    }
  }
  
  public LIndex getIndex(String name) {
    return (LIndex)indexMap.get(name);
  }
  
  public String getIndexDirectory(String indexName) {
    return baseDir + indexName;
  }
  
  public void start() {
    thread.setDaemon(true);
    thread.start();
    logger.info("Lucene Manager started");
  }
  
  public void shutdown()
  {
    shutdown = true;
    synchronized (thread)
    {
      try {
        thread.wait(20000L);
      } catch (InterruptedException e) {
        logger.info("InterruptedException:" + e);
      }
    }
    
    if (!shutdownComplete) {
      String msg = "WARNING: Shutdown of Lucene Manager did not complete?";
      System.err.println(msg);
      logger.warning(msg);
    }
  }
  
  private void fireOnStartup() { String masterHost;
    if ((clusterIndexSync != null) && (!clusterIndexSync.isMaster())) {
      masterHost = clusterIndexSync.getMasterHost();
      if (masterHost != null) {
        for (LIndex index : indexMap.values()) {
          index.queueSync(masterHost);
        }
      }
    }
  }
  
  public void run()
  {
    fireOnStartup();
    
    while (!shutdown) {
      synchronized (indexMap) {
        long start = System.currentTimeMillis();
        for (LIndex idx : indexMap.values()) {
          idx.manage(this);
        }
        long exeTime = System.currentTimeMillis() - start;
        long sleepMillis = manageFreqMillis - exeTime;
        if (sleepMillis > 0L) {
          try {
            Thread.sleep(sleepMillis);
          } catch (InterruptedException e) {
            logger.log(Level.INFO, "Interrupted", e);
          }
        }
      }
    }
    shutdownComplete = true;
    synchronized (thread) {
      thread.notifyAll();
    }
  }
  
  private static class IndexSynchRun implements Runnable
  {
    private final LuceneClusterIndexSync clusterIndexSync;
    private final LIndex index;
    private final String masterHost;
    
    private IndexSynchRun(LuceneClusterIndexSync clusterIndexSync, LIndexSync indexSync)
    {
      this.clusterIndexSync = clusterIndexSync;
      index = indexSync.getIndex();
      masterHost = indexSync.getMasterHost();
    }
    
    public void run() {
      boolean success = false;
      try {
        clusterIndexSync.sync(index, masterHost);
        success = true;
      } catch (IOException e) {
        String msg = "Failed to sync Lucene index " + index;
        DefaultLuceneIndexManager.logger.log(Level.SEVERE, msg, e);
      } finally {
        index.syncFinished(success);
      }
    }
  }
}
