package com.avaje.ebeaninternal.server.lucene;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.lucene.index.IndexCommit;
import org.apache.lucene.index.IndexDeletionPolicy;




















public class HoldAwareIndexDeletionPolicy
  implements IndexDeletionPolicy
{
  private static final Logger logger = Logger.getLogger(HoldAwareIndexDeletionPolicy.class.getName());
  
  private final Map<Long, CommitRefCount> commitRefCounts = new HashMap();
  
  private IndexCommit lastCommit;
  private final String indexDir;
  
  public HoldAwareIndexDeletionPolicy(String indexDir)
  {
    this.indexDir = indexDir;
  }
  


  public void onInit(List<? extends IndexCommit> commits)
  {
    onCommit(commits);
  }
  


  public void onCommit(List<? extends IndexCommit> commits)
  {
    synchronized (commitRefCounts) {
      int size = commits.size();
      
      lastCommit = ((IndexCommit)commits.get(size - 1));
      

      for (int i = 0; i < size - 1; i++) {
        IndexCommit indexCommit = (IndexCommit)commits.get(i);
        if (!commitRefCounts.containsKey(Long.valueOf(indexCommit.getVersion())))
        {


          potentialIndexCommitDelete(indexCommit);
        }
      }
    }
  }
  
  private void potentialIndexCommitDelete(IndexCommit indexCommit)
  {
    indexCommit.delete();
  }
  
  public long getLastVersion() {
    synchronized (commitRefCounts) {
      if (lastCommit == null) {
        return 0L;
      }
      return lastCommit.getVersion();
    }
  }
  









  public LIndexCommitInfo obtainLastIndexCommitIfNewer(long remoteIndexVersion)
  {
    synchronized (commitRefCounts)
    {
      if ((remoteIndexVersion != 0L) && (remoteIndexVersion == lastCommit.getVersion()))
      {
        return null;
      }
      
      incRefIndexCommit(lastCommit);
      return new LIndexCommitInfo(indexDir, lastCommit);
    }
  }
  


  private void incRefIndexCommit(IndexCommit indexCommit)
  {
    Long commitVersion = Long.valueOf(indexCommit.getVersion());
    CommitRefCount refCount = (CommitRefCount)commitRefCounts.get(commitVersion);
    if (refCount == null) {
      refCount = new CommitRefCount(null);
      commitRefCounts.put(commitVersion, refCount);
    }
    refCount.inc();
  }
  



  public void releaseIndexCommit(long indexCommitVersion)
  {
    synchronized (commitRefCounts) {
      Long commitVersion = Long.valueOf(indexCommitVersion);
      CommitRefCount refCount = (CommitRefCount)commitRefCounts.get(commitVersion);
      if (refCount == null) {
        logger.log(Level.WARNING, "No Reference counter for indexCommitVersion: " + commitVersion);
      }
      else if (refCount.dec() <= 0)
      {

        commitRefCounts.remove(commitVersion);
      }
    }
  }
  
  public void touch(long indexCommitVersion)
  {
    synchronized (commitRefCounts) {
      Long commitVersion = Long.valueOf(indexCommitVersion);
      CommitRefCount refCount = (CommitRefCount)commitRefCounts.get(commitVersion);
      if (refCount == null) {
        logger.warning("No Reference counter for indexCommitVersion: " + commitVersion);
      } else {
        refCount.touch();
      }
    }
  }
  
  public long getLastTouched(long indexCommitVersion) {
    synchronized (commitRefCounts) {
      Long commitVersion = Long.valueOf(indexCommitVersion);
      CommitRefCount refCount = (CommitRefCount)commitRefCounts.get(commitVersion);
      if (refCount == null) {
        return 0L;
      }
      return refCount.getLastTouched();
    }
  }
  
  private static class CommitRefCount
  {
    private int refCount;
    private long lastTouched = System.currentTimeMillis();
    
    public void inc() {
      refCount += 1;
    }
    
    public int dec() {
      return --refCount;
    }
    
    public void touch() {
      lastTouched = System.currentTimeMillis();
    }
    
    public long getLastTouched() {
      return lastTouched;
    }
  }
}
