package com.avaje.ebeaninternal.server.lucene;

import com.avaje.ebean.config.lucene.LuceneIndex;
import com.avaje.ebeaninternal.api.SpiQuery;
import com.avaje.ebeaninternal.api.TransactionEventTable.TableIUD;
import com.avaje.ebeaninternal.server.deploy.BeanDescriptor;
import com.avaje.ebeaninternal.server.querydefn.OrmQueryDetail;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Set;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexWriter.MaxFieldLength;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.queryParser.QueryParser.Operator;
































public class LIndex
  implements LuceneIndex
{
  private final DefaultLuceneIndexManager manager;
  private final String name;
  private final Analyzer analyzer;
  private final IndexWriter.MaxFieldLength maxFieldLength;
  private final LIndexFields fieldDefn;
  private final BeanDescriptor<?> desc;
  private final OrmQueryDetail ormQueryDetail;
  private final LIndexIo indexIo;
  private final LIndexFieldId idField;
  private final Object syncMonitor = new Object();
  
  private boolean runningSync;
  
  private LIndexSync queuedSync;
  
  public LIndex(DefaultLuceneIndexManager manager, String indexName, String indexDir, Analyzer analyzer, IndexWriter.MaxFieldLength maxFieldLength, BeanDescriptor<?> desc, LIndexFields fieldDefn, String[] updateProps)
    throws IOException
  {
    this.manager = manager;
    name = desc.getFullName();
    this.analyzer = analyzer;
    this.maxFieldLength = maxFieldLength;
    this.desc = desc;
    this.fieldDefn = fieldDefn;
    idField = fieldDefn.getIdField();
    ormQueryDetail = fieldDefn.getOrmQueryDetail();
    
    indexIo = new LIndexIo(manager, indexDir, this, updateProps);
    manager.addIndex(this);
    fieldDefn.registerIndexWithProperties(this);
  }
  




  protected void syncFinished(boolean success)
  {
    synchronized (syncMonitor) {
      runningSync = false;
    }
  }
  







  public void queueSync(String masterHost)
  {
    synchronized (syncMonitor) {
      LIndexSync sync = new LIndexSync(this, masterHost);
      if (!runningSync)
      {
        runningSync = true;
        manager.execute(sync);

      }
      else
      {

        queuedSync = sync;
      }
    }
  }
  



  public void manage(LuceneIndexManager indexManager)
  {
    synchronized (syncMonitor) {
      indexIo.manage(indexManager);
      if ((!runningSync) && (queuedSync != null))
      {
        LIndexSync sync = queuedSync;
        runningSync = true;
        queuedSync = null;
        manager.execute(sync);
      }
    }
  }
  
  public void addNotifyCommitRunnable(Runnable r) {
    indexIo.addNotifyCommitRunnable(r);
  }
  
  public LIndexVersion getLastestVersion() {
    return indexIo.getLastestVersion();
  }
  
  public File getIndexDir() {
    return indexIo.getIndexDir();
  }
  
  public void refresh(boolean nearRealTime) {
    indexIo.refresh(nearRealTime);
  }
  
  public LIndexFileInfo getLocalFile(String fileName) {
    return indexIo.getLocalFile(fileName);
  }
  
  public LIndexCommitInfo obtainLastIndexCommitIfNewer(long remoteIndexVersion) {
    return indexIo.obtainLastIndexCommitIfNewer(remoteIndexVersion);
  }
  
  public void releaseIndexCommit(long remoteIndexVersion) {
    indexIo.releaseIndexCommit(remoteIndexVersion);
  }
  
  public LIndexFileInfo getFile(long remoteIndexVersion, String fileName) {
    return indexIo.getFile(remoteIndexVersion, fileName);
  }
  
  public Term createIdTerm(Object id) {
    return idField.createTerm(id);
  }
  
  public void shutdown() {
    indexIo.shutdown();
  }
  
  public LIndexUpdateFuture rebuild() {
    LIndexUpdateFuture future = new LIndexUpdateFuture(desc.getBeanType());
    indexIo.addWorkToQueue(LIndexWork.newRebuild(future));
    return future;
  }
  
  public LIndexUpdateFuture update() {
    LIndexUpdateFuture future = new LIndexUpdateFuture(desc.getBeanType());
    indexIo.addWorkToQueue(LIndexWork.newQueryUpdate(future));
    return future;
  }
  
  public String toString() {
    return name;
  }
  
  public String getName() {
    return name;
  }
  
  public Class<?> getBeanType() {
    return desc.getBeanType();
  }
  
  public BeanDescriptor<?> getBeanDescriptor() {
    return desc;
  }
  
  public LIndexSearch getIndexSearch() {
    return indexIo.getIndexSearch();
  }
  
  public Analyzer getAnalyzer() {
    return analyzer;
  }
  
  public IndexWriter.MaxFieldLength getMaxFieldLength() {
    return maxFieldLength;
  }
  
  public QueryParser createQueryParser(String fieldName) {
    QueryParser p = fieldDefn.createQueryParser(fieldName);
    p.setDefaultOperator(QueryParser.Operator.AND);
    return p;
  }
  
  public LIndexFields getIndexFieldDefn() {
    return fieldDefn;
  }
  
  public Set<String> getResolvePropertyNames() {
    return fieldDefn.getResolvePropertyNames();
  }
  
  public OrmQueryDetail getOrmQueryDetail() {
    return ormQueryDetail;
  }
  
  public Object readDocument(Document doc)
  {
    Object bean = desc.createEntityBean();
    fieldDefn.readDocument(doc, bean);
    return bean;
  }
  
  public DocFieldWriter createDocFieldWriter() {
    return fieldDefn.createDocFieldWriter();
  }
  
  public SpiQuery<?> createQuery() {
    return indexIo.createQuery();
  }
  
  public LIndexUpdateFuture process(IndexUpdates indexUpdates)
  {
    List<TransactionEventTable.TableIUD> tableList = indexUpdates.getTableList();
    if ((tableList != null) && (tableList.size() > 0)) {
      boolean bulkDelete = false;
      for (int i = 0; i < tableList.size(); i++) {
        TransactionEventTable.TableIUD bulkTableEvent = (TransactionEventTable.TableIUD)tableList.get(i);
        if (bulkTableEvent.isDelete()) {
          bulkDelete = true;
        }
      }
      if (bulkDelete) {
        return rebuild();
      }
      return update();
    }
    

    if (indexUpdates.isInvalidate()) {
      return update();
    }
    
    LIndexUpdateFuture f = new LIndexUpdateFuture(desc.getBeanType());
    indexIo.addWorkToQueue(LIndexWork.newTxnUpdate(f, indexUpdates));
    return f;
  }
}
