package com.avaje.ebeaninternal.server.lucene;

import com.avaje.ebean.BackgroundExecutor;
import com.avaje.ebean.ExpressionList;
import com.avaje.ebean.Junction;
import com.avaje.ebean.Query.UseIndex;
import com.avaje.ebean.QueryListener;
import com.avaje.ebeaninternal.api.SpiEbeanServer;
import com.avaje.ebeaninternal.api.SpiQuery;
import com.avaje.ebeaninternal.server.deploy.BeanDescriptor;
import com.avaje.ebeaninternal.server.querydefn.OrmQueryDetail;
import com.avaje.ebeaninternal.server.transaction.IndexEvent;
import java.io.File;
import java.io.IOException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.PersistenceException;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriter.MaxFieldLength;
import org.apache.lucene.index.Term;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;






















public class LIndexIo
{
  private static final Logger logger = Logger.getLogger(LIndexIo.class.getName());
  
  private final LuceneIndexManager manager;
  
  private final String indexDir;
  
  private final LIndex index;
  
  private final Analyzer analyzer;
  
  private final IndexWriter.MaxFieldLength maxFieldLength;
  
  private final Class<?> beanType;
  
  private final OrmQueryDetail ormQueryDetail;
  
  private final Directory directory;
  
  private final BeanDescriptor<?> beanDescriptor;
  
  private final IndexWriter indexWriter;
  
  private final LIndexIoSearcher ioSearcher;
  
  private final HoldAwareIndexDeletionPolicy commitDeletionPolicy;
  
  private final String[] updateProps;
  
  private final Object writeMonitor = new Object();
  
  private final Object workQueueMonitor = new Object();
  
  private final ArrayList<LIndexWork> workQueue = new ArrayList();
  
  private final ArrayList<Runnable> notifyCommitRunnables = new ArrayList();
  
  private long lastUpdateTime;
  
  private long queueCommitStart;
  
  private int queueCommitCount;
  
  private int totalCommitCount;
  private long totalCommitNanos;
  private long totalPostCommitNanos;
  
  public LIndexIo(LuceneIndexManager manager, String indexDir, LIndex index, String[] updateProps)
    throws IOException
  {
    this.manager = manager;
    this.indexDir = indexDir;
    this.index = index;
    this.updateProps = updateProps;
    analyzer = index.getAnalyzer();
    maxFieldLength = index.getMaxFieldLength();
    beanType = index.getBeanType();
    ormQueryDetail = index.getOrmQueryDetail();
    directory = createDirectory();
    beanDescriptor = index.getBeanDescriptor();
    
    commitDeletionPolicy = new HoldAwareIndexDeletionPolicy(indexDir);
    indexWriter = createIndexWriter();
    
    ioSearcher = createIoSearcher();
  }
  
  public LIndexVersion getLastestVersion() {
    return ioSearcher.getLastestVersion();
  }
  
  public long getLastVersion() {
    return commitDeletionPolicy.getLastVersion();
  }
  
  public LIndexCommitInfo obtainLastIndexCommitIfNewer(long remoteIndexVersion) {
    return commitDeletionPolicy.obtainLastIndexCommitIfNewer(remoteIndexVersion);
  }
  
  public File getIndexDir() {
    return new File(indexDir);
  }
  
  public LIndexFileInfo getLocalFile(String fileName)
  {
    File f = new File(indexDir, fileName);
    return new LIndexFileInfo(f);
  }
  
  public void refresh(boolean nearRealTime) {
    ioSearcher.refresh(nearRealTime);
  }
  

  public LIndexFileInfo getFile(long remoteIndexVersion, String fileName)
  {
    commitDeletionPolicy.touch(remoteIndexVersion);
    
    File f = new File(indexDir, fileName);
    return new LIndexFileInfo(f);
  }
  
  public void releaseIndexCommit(long remoteIndexVersion) {
    commitDeletionPolicy.releaseIndexCommit(remoteIndexVersion);
  }
  
  public void shutdown() {
    synchronized (writeMonitor) {
      try {
        if (queueCommitStart > 0L) {
          indexWriter.commit();
        }
        




        try
        {
          indexWriter.close();
        } catch (Exception e) {
          String msg = "Error closing IndexWriter for " + indexDir;
          logger.log(Level.SEVERE, msg, e);
          
          e.printStackTrace();
        }
      }
      catch (Exception e)
      {
        String msg = "Error committing queued changes for IndexWriter for " + indexDir;
        logger.log(Level.SEVERE, msg, e);
        
        e.printStackTrace();
      } finally {
        try { String msg;
          indexWriter.close();
        } catch (Exception e) {
          String msg = "Error closing IndexWriter for " + indexDir;
          logger.log(Level.SEVERE, msg, e);
          
          e.printStackTrace();
        }
      }
    }
  }
  
  protected void manage(LuceneIndexManager indexManager) {
    processWorkQueue();
    commit(false);
  }
  
  protected void addWorkToQueue(LIndexWork work) {
    synchronized (workQueueMonitor) {
      workQueue.add(work);
    }
  }
  
  protected void processWorkQueue()
  {
    synchronized (workQueueMonitor) {
      if (!workQueue.isEmpty())
      {
        LIndexWork.WorkType maxWorkType = null;
        
        for (int i = 0; i < workQueue.size(); i++) {
          LIndexWork work = (LIndexWork)workQueue.get(i);
          if ((maxWorkType == null) || (maxWorkType.ordinal() < work.getWorkType().ordinal())) {
            maxWorkType = work.getWorkType();
          }
        }
        List<LIndexWork> workQueueClone = (List)workQueue.clone();
        workQueue.clear();
        
        Callable<Integer> workCallable = getWorkCallable(maxWorkType, workQueueClone);
        FutureTask<Integer> ft = new FutureTask(workCallable);
        
        for (int i = 0; i < workQueueClone.size(); i++) {
          ((LIndexWork)workQueueClone.get(i)).getFuture().setTask(ft);
        }
        
        manager.getServer().getBackgroundExecutor().execute(ft);
      }
    }
  }
  
  private Callable<Integer> getWorkCallable(LIndexWork.WorkType maxWorkType, List<LIndexWork> workQueueClone)
  {
    switch (2.$SwitchMap$com$avaje$ebeaninternal$server$lucene$LIndexWork$WorkType[maxWorkType.ordinal()]) {
    case 1: 
      return newRebuildCallable(workQueueClone);
    case 2: 
      return newQueryUpdateCallable(workQueueClone);
    case 3: 
      return newTxnUpdateCallable(workQueueClone);
    }
    
    throw new IllegalStateException("Unknown workType " + maxWorkType);
  }
  

  private Callable<Integer> newTxnUpdateCallable(List<LIndexWork> workQueueClone)
  {
    final List<LIndexWork> updates = workQueueClone;
    
    new Callable() {
      public String toString() {
        return "TxnUpdate";
      }
      
      public Integer call() throws IOException
      {
        int totalDocs = 0;
        
        for (int i = 0; i < updates.size(); i++) {
          LIndexWork lIndexWork = (LIndexWork)updates.get(i);
          IndexUpdates indexUpdates = lIndexWork.getIndexUpdates();
          LIndexDeltaHandler h = LIndexIo.this.createDeltaHandler(indexUpdates);
          totalDocs += h.process();
        }
        
        LIndexIo.this.queueCommit(updates);
        
        return Integer.valueOf(totalDocs);
      }
    };
  }
  
  private Callable<Integer> newRebuildCallable(List<LIndexWork> workQueueClone) {
    return new QueryUpdater(true, workQueueClone, null);
  }
  
  private Callable<Integer> newQueryUpdateCallable(List<LIndexWork> workQueueClone) {
    return new QueryUpdater(false, workQueueClone, null);
  }
  
  class QueryUpdater implements Callable<Integer>
  {
    private final boolean rebuild;
    private final List<LIndexWork> workQueueClone;
    
    private QueryUpdater(List<LIndexWork> rebuild) {
      this.rebuild = rebuild;
      this.workQueueClone = workQueueClone;
    }
    
    public String toString() {
      return rebuild ? "Rebuild" : "QueryUpdate";
    }
    
    public Integer call() throws Exception {
      if (rebuild) {
        return Integer.valueOf(LIndexIo.this.rebuildIndex(workQueueClone));
      }
      return Integer.valueOf(LIndexIo.this.updateIndex(workQueueClone));
    }
  }
  

  private LIndexDeltaHandler createDeltaHandler(IndexUpdates indexUpdates)
  {
    LIndexSearch search = getIndexSearch();
    IndexWriter indexWriter = this.indexWriter;
    DocFieldWriter docFieldWriter = index.createDocFieldWriter();
    return new LIndexDeltaHandler(index, search, indexWriter, analyzer, beanDescriptor, docFieldWriter, indexUpdates);
  }
  
  public LIndexSearch getIndexSearch() {
    return ioSearcher.getIndexSearch();
  }
  












  private void queueCommit(List<LIndexWork> workQueueClone)
  {
    synchronized (workQueueMonitor) {
      if (queueCommitStart == 0L) {
        queueCommitStart = System.currentTimeMillis();
      }
      queueCommitCount += 1;
      
      for (LIndexWork w : workQueueClone)
      {
        notifyCommitRunnables.add(w.getFuture().getCommitRunnable());
      }
    }
  }
  
  protected void addNotifyCommitRunnable(Runnable r) {
    synchronized (workQueueMonitor) {
      notifyCommitRunnables.add(r);
    }
  }
  
  protected long getQueueCommitStart(boolean reset) {
    synchronized (workQueueMonitor) {
      long start = queueCommitStart;
      if (reset) {
        queueCommitStart = 0L;
        queueCommitCount = 0;
      }
      return start;
    }
  }
  




  private boolean commit(boolean force)
  {
    synchronized (writeMonitor)
    {
      long start = 0L;
      long count = 0L;
      ArrayList<Runnable> notifyRunnables = new ArrayList();
      
      synchronized (workQueueMonitor) {
        start = queueCommitStart;
        count = queueCommitCount;
        queueCommitStart = 0L;
        queueCommitCount = 0;
        notifyRunnables.addAll(notifyCommitRunnables);
        notifyCommitRunnables.clear();
      }
      try
      {
        if ((!force) && (start == 0L))
        {


          if (!notifyRunnables.isEmpty())
          {
            for (int i = 0; i < notifyRunnables.size(); i++) {
              addNotifyCommitRunnable((Runnable)notifyRunnables.get(i));
            }
          }
          return false;
        }
        if (logger.isLoggable(Level.INFO)) { String delayMsg;
          String delayMsg;
          if (queueCommitStart > 0L) {
            long delay = System.currentTimeMillis() - start;
            delayMsg = " queueDelayMillis:" + delay + " queueCount:" + count;
          } else {
            delayMsg = "";
          }
          String m = "Lucene commit " + indexDir + delayMsg;
          logger.info(m);
        }
        long nanoStart = System.nanoTime();
        

        indexWriter.commit();
        

        long nanoCommit = System.nanoTime();
        long nanoCommitExe = nanoCommit - nanoStart;
        

        ioSearcher.postCommit();
        
        for (int i = 0; i < notifyRunnables.size(); i++) {
          ((Runnable)notifyRunnables.get(i)).run();
        }
        
        long nanoPostCommitExe = System.nanoTime() - nanoCommitExe;
        
        totalCommitCount += 1;
        totalCommitNanos += nanoCommitExe;
        totalPostCommitNanos += nanoPostCommitExe;
        
        IndexEvent indexEvent = new IndexEvent(1, index.getName());
        manager.notifyCluster(indexEvent);
        
        return true;
      }
      catch (IOException e) {
        String msg = "Error committing changes on index " + indexDir;
        throw new PersistenceLuceneException(msg, e);
      }
    }
  }
  
  private int rebuildIndex(List<LIndexWork> workQueueClone) throws IOException
  {
    synchronized (writeMonitor)
    {
      logger.info("Lucene rebuild " + indexDir);
      
      try
      {
        indexWriter.deleteAll();
        lastUpdateTime = System.currentTimeMillis();
        SpiQuery<?> query = createQuery();
        
        WriteListener writeListener = new WriteListener(index, indexWriter, false, null);
        query.setListener(writeListener);
        
        manager.getServer().findList(query, null);
        
        int i = writeListener.getCount();
        

        queueCommit(workQueueClone);
        commit(false);
        return i;
      }
      finally
      {
        queueCommit(workQueueClone);
        commit(false);
      }
    }
  }
  
  private int updateIndex(List<LIndexWork> workQueueClone) throws IOException
  {
    synchronized (writeMonitor)
    {
      logger.info("Lucene update " + indexDir);
      try
      {
        long updateTime = System.currentTimeMillis();
        SpiQuery<?> query = createUpdateQuery();
        lastUpdateTime = updateTime;
        
        WriteListener writeListener = new WriteListener(index, indexWriter, true, null);
        query.setListener(writeListener);
        
        manager.getServer().findList(query, null);
        
        int i = writeListener.getCount();
        

        queueCommit(workQueueClone);
        return i;
      }
      finally
      {
        queueCommit(workQueueClone);
      }
    }
  }
  
  private SpiQuery<?> createUpdateQuery()
  {
    SpiQuery<?> q = createQuery();
    Junction<?> disjunction = q.where().disjunction();
    
    Timestamp lastUpdate = new Timestamp(lastUpdateTime);
    
    for (int i = 0; i < updateProps.length; i++) {
      disjunction.ge(updateProps[i], lastUpdate);
    }
    
    return q;
  }
  
  protected SpiQuery<?> createQuery()
  {
    SpiEbeanServer server = manager.getServer();
    SpiQuery<?> query = (SpiQuery)server.createQuery(beanType);
    query.setUseIndex(Query.UseIndex.NO);
    query.getDetail().tuneFetchProperties(ormQueryDetail);
    
    return query;
  }
  
  private Directory createDirectory() throws IOException {
    File dir = new File(indexDir);
    return FSDirectory.open(dir);
  }
  
  private IndexWriter createIndexWriter() {
    try {
      boolean create = true;
      return new IndexWriter(directory, analyzer, create, commitDeletionPolicy, maxFieldLength);
    } catch (IOException e) {
      String msg = "Error getting Lucene IndexWriter for " + indexDir;
      throw new PersistenceLuceneException(msg, e);
    }
  }
  
  private LIndexIoSearcher createIoSearcher()
  {
    return new LIndexIoSearcherDefault(indexWriter, index.getName());
  }
  
  private static class WriteListener
    implements QueryListener
  {
    private final boolean updateMode;
    private final LIndex index;
    private final BeanDescriptor beanDescriptor;
    private final IndexWriter indexWriter;
    private final DocFieldWriter docFieldWriter;
    private final Document document = new Document();
    private int count;
    
    private WriteListener(LIndex index, IndexWriter indexWriter, boolean updateMode) {
      this.updateMode = updateMode;
      this.index = index;
      beanDescriptor = index.getBeanDescriptor();
      this.indexWriter = indexWriter;
      docFieldWriter = index.createDocFieldWriter();
    }
    
    public void process(Object bean) {
      try {
        if (updateMode) {
          Object id = beanDescriptor.getId(bean);
          Term term = index.createIdTerm(id);
          indexWriter.deleteDocuments(term);
        }
        docFieldWriter.writeValue(bean, document);
        indexWriter.addDocument(document);
        count += 1;
      } catch (Exception e) {
        throw new PersistenceException(e);
      }
    }
    
    public int getCount() {
      return count;
    }
  }
}
