package com.avaje.ebeaninternal.server.persist.dml;

import com.avaje.ebeaninternal.api.SpiTransaction;
import com.avaje.ebeaninternal.api.TransactionEvent;
import com.avaje.ebeaninternal.server.core.PersistRequestBean;
import com.avaje.ebeaninternal.server.core.PstmtBatch;
import com.avaje.ebeaninternal.server.deploy.BeanProperty;
import com.avaje.ebeaninternal.server.persist.BatchControl;
import com.avaje.ebeaninternal.server.persist.BatchedPstmt;
import com.avaje.ebeaninternal.server.persist.BatchedPstmtHolder;
import com.avaje.ebeaninternal.server.persist.dmlbind.BindableRequest;
import com.avaje.ebeaninternal.server.transaction.BeanDelta;
import com.avaje.ebeaninternal.server.type.DataBind;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.OptimisticLockException;























public abstract class DmlHandler
  implements PersistHandler, BindableRequest
{
  protected static final Logger logger = Logger.getLogger(DmlHandler.class.getName());
  

  protected final PersistRequestBean<?> persistRequest;
  

  protected final StringBuilder bindLog;
  

  protected final Set<String> loadedProps;
  

  protected final SpiTransaction transaction;
  
  protected final boolean emptyStringToNull;
  
  protected final boolean logLevelSql;
  
  protected DataBind dataBind;
  
  protected String sql;
  
  protected ArrayList<UpdateGenValue> updateGenValues;
  
  private Set<String> additionalProps;
  
  private boolean checkDelta;
  
  private BeanDelta deltaBean;
  

  protected DmlHandler(PersistRequestBean<?> persistRequest, boolean emptyStringToNull)
  {
    this.persistRequest = persistRequest;
    this.emptyStringToNull = emptyStringToNull;
    loadedProps = persistRequest.getLoadedProperties();
    transaction = persistRequest.getTransaction();
    logLevelSql = transaction.isLogSql();
    if (logLevelSql) {
      bindLog = new StringBuilder();
    } else {
      bindLog = null;
    }
  }
  
  protected void setCheckDelta(boolean checkDelta) {
    this.checkDelta = checkDelta;
  }
  
  public PersistRequestBean<?> getPersistRequest() {
    return persistRequest;
  }
  


  public abstract void bind()
    throws SQLException;
  

  public abstract void execute()
    throws SQLException;
  

  protected void checkRowCount(int rowCount)
    throws SQLException, OptimisticLockException
  {
    try
    {
      persistRequest.checkRowCount(rowCount);
      persistRequest.postExecute();
    }
    catch (OptimisticLockException e) {
      String m = e.getMessage() + " sql[" + sql + "] bind[" + bindLog + "]";
      persistRequest.getTransaction().log("OptimisticLockException:" + m);
      throw new OptimisticLockException(m, null, e.getEntity());
    }
  }
  

  public void addBatch()
    throws SQLException
  {
    PstmtBatch pstmtBatch = persistRequest.getPstmtBatch();
    if (pstmtBatch != null) {
      pstmtBatch.addBatch(dataBind.getPstmt());
    } else {
      dataBind.getPstmt().addBatch();
    }
  }
  

  public void close()
  {
    try
    {
      if (dataBind != null) {
        dataBind.close();
      }
    } catch (SQLException ex) {
      logger.log(Level.SEVERE, null, ex);
    }
  }
  


  public String getBindLog()
  {
    return bindLog == null ? "" : bindLog.toString();
  }
  



  public void setIdValue(Object idValue)
  {
    persistRequest.setBoundId(idValue);
  }
  


  protected void logBinding()
  {
    if (logLevelSql) {
      transaction.logInternal(bindLog.toString());
    }
  }
  


  protected void logSql(String sql)
  {
    if (logLevelSql) {
      transaction.logInternal(sql);
    }
  }
  
  public boolean isIncluded(BeanProperty prop)
  {
    return (loadedProps == null) || (loadedProps.contains(prop.getName()));
  }
  
  public boolean isIncludedWhere(BeanProperty prop) {
    if (prop.isDbEncrypted())
    {


      return isIncluded(prop);
    }
    return (prop.isDbUpdatable()) && ((loadedProps == null) || (loadedProps.contains(prop.getName())));
  }
  

  public Object bind(String propName, Object value, int sqlType)
    throws SQLException
  {
    if (logLevelSql) {
      bindLog.append(propName).append("=");
      bindLog.append(value).append(", ");
    }
    dataBind.setObject(value, sqlType);
    return value;
  }
  
  public Object bindNoLog(Object value, int sqlType, String logPlaceHolder) throws SQLException {
    if (logLevelSql) {
      bindLog.append(logPlaceHolder).append(" ");
    }
    dataBind.setObject(value, sqlType);
    return value;
  }
  

  public Object bind(Object value, BeanProperty prop, String propName, boolean bindNull)
    throws SQLException
  {
    return bindInternal(logLevelSql, value, prop, propName, bindNull);
  }
  

  public Object bindNoLog(Object value, BeanProperty prop, String propName, boolean bindNull)
    throws SQLException
  {
    return bindInternal(false, value, prop, propName, bindNull);
  }
  
  private Object bindInternal(boolean log, Object value, BeanProperty prop, String propName, boolean bindNull) throws SQLException
  {
    if ((!bindNull) && 
      (emptyStringToNull) && ((value instanceof String)) && (((String)value).length() == 0))
    {

      value = null;
    }
    

    if ((!bindNull) && (value == null))
    {
      if (log) {
        bindLog.append(propName).append("=");
        bindLog.append("null, ");
      }
    } else {
      if (log) {
        bindLog.append(propName).append("=");
        if (prop.isLob()) {
          bindLog.append("[LOB]");
        } else {
          String sv = String.valueOf(value);
          if (sv.length() > 50) {
            sv = sv.substring(0, 47) + "...";
          }
          bindLog.append(sv);
        }
        bindLog.append(", ");
      }
      
      prop.bind(dataBind, value);
      if ((checkDelta) && 
        (!prop.isId()) && (prop.isDeltaRequired())) {
        if (deltaBean == null) {
          deltaBean = persistRequest.createDeltaBean();
          transaction.getEvent().addBeanDelta(deltaBean);
        }
        deltaBean.add(prop, value);
      }
    }
    
    return value;
  }
  


  protected void bindLogAppend(String comment)
  {
    if (logLevelSql) {
      bindLog.append(comment);
    }
  }
  



  public final void registerAdditionalProperty(String propertyName)
  {
    if ((loadedProps != null) && (!loadedProps.contains(propertyName))) {
      if (additionalProps == null) {
        additionalProps = new HashSet();
      }
      additionalProps.add(propertyName);
    }
  }
  



  protected void setAdditionalProperties()
  {
    if (additionalProps != null)
    {

      additionalProps.addAll(loadedProps);
      persistRequest.setLoadedProps(additionalProps);
    }
  }
  








  public void registerUpdateGenValue(BeanProperty prop, Object bean, Object value)
  {
    if (updateGenValues == null) {
      updateGenValues = new ArrayList();
    }
    updateGenValues.add(new UpdateGenValue(prop, bean, value, null));
    registerAdditionalProperty(prop.getName());
  }
  





  public void setUpdateGenValues()
  {
    if (updateGenValues != null) {
      for (int i = 0; i < updateGenValues.size(); i++) {
        UpdateGenValue updGenVal = (UpdateGenValue)updateGenValues.get(i);
        updGenVal.setValue();
      }
    }
  }
  


  protected PreparedStatement getPstmt(SpiTransaction t, String sql, boolean genKeys)
    throws SQLException
  {
    Connection conn = t.getInternalConnection();
    if (genKeys)
    {


      int[] columns = { 1 };
      return conn.prepareStatement(sql, columns);
    }
    
    return conn.prepareStatement(sql);
  }
  




  protected PreparedStatement getPstmt(SpiTransaction t, String sql, PersistRequestBean<?> request, boolean genKeys)
    throws SQLException
  {
    BatchedPstmtHolder batch = t.getBatchControl().getPstmtHolder();
    PreparedStatement stmt = batch.getStmt(sql, request);
    
    if (stmt != null) {
      return stmt;
    }
    
    if (logLevelSql) {
      t.logInternal(sql);
    }
    
    stmt = getPstmt(t, sql, genKeys);
    
    PstmtBatch pstmtBatch = request.getPstmtBatch();
    if (pstmtBatch != null) {
      pstmtBatch.setBatchSize(stmt, t.getBatchControl().getBatchSize());
    }
    
    BatchedPstmt bs = new BatchedPstmt(stmt, genKeys, sql, request.getPstmtBatch(), true);
    batch.addStmt(bs, request);
    return stmt;
  }
  


  private static final class UpdateGenValue
  {
    private final BeanProperty property;
    
    private final Object bean;
    
    private final Object value;
    

    private UpdateGenValue(BeanProperty property, Object bean, Object value)
    {
      this.property = property;
      this.bean = bean;
      this.value = value;
    }
    



    private void setValue()
    {
      property.setValueIntercept(bean, value);
    }
  }
}
