package com.avaje.ebeaninternal.server.query;

import com.avaje.ebean.BackgroundExecutor;
import com.avaje.ebean.bean.BeanCollection;
import com.avaje.ebean.bean.EntityBeanIntercept;
import com.avaje.ebean.bean.PersistenceContext;
import com.avaje.ebeaninternal.api.BeanIdList;
import com.avaje.ebeaninternal.api.SpiQuery;
import com.avaje.ebeaninternal.api.SpiQuery.Mode;
import com.avaje.ebeaninternal.api.SpiTransaction;
import com.avaje.ebeaninternal.server.core.OrmQueryRequest;
import com.avaje.ebeaninternal.server.core.ReferenceOptions;
import com.avaje.ebeaninternal.server.core.SpiOrmQueryRequest;
import com.avaje.ebeaninternal.server.deploy.BeanDescriptor;
import com.avaje.ebeaninternal.server.deploy.BeanPropertyAssocMany;
import com.avaje.ebeaninternal.server.deploy.BeanPropertyAssocOne;
import com.avaje.ebeaninternal.server.deploy.DbReadContext;
import com.avaje.ebeaninternal.server.deploy.id.IdBinder;
import com.avaje.ebeaninternal.server.type.DataBind;
import com.avaje.ebeaninternal.server.type.DataReader;
import com.avaje.ebeaninternal.server.type.RsetDataReader;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.FutureTask;
import java.util.logging.Level;
import java.util.logging.Logger;























public class CQueryFetchIds
{
  private static final Logger logger = Logger.getLogger(CQueryFetchIds.class.getName());
  

  private final OrmQueryRequest<?> request;
  

  private final BeanDescriptor<?> desc;
  

  private final SpiQuery<?> query;
  

  private final BackgroundExecutor backgroundExecutor;
  

  private final CQueryPredicates predicates;
  

  private final String sql;
  

  private RsetDataReader dataReader;
  

  private PreparedStatement pstmt;
  

  private String bindLog;
  

  private long startNano;
  

  private int executionTimeMicros;
  

  private int rowCount;
  

  private final int maxRows;
  

  private final int bgFetchAfter;
  

  public CQueryFetchIds(OrmQueryRequest<?> request, CQueryPredicates predicates, String sql, BackgroundExecutor backgroundExecutor)
  {
    this.backgroundExecutor = backgroundExecutor;
    this.request = request;
    query = request.getQuery();
    this.sql = sql;
    maxRows = query.getMaxRows();
    bgFetchAfter = query.getBackgroundFetchAfter();
    
    query.setGeneratedSql(sql);
    
    desc = request.getBeanDescriptor();
    this.predicates = predicates;
  }
  



  public String getSummary()
  {
    StringBuilder sb = new StringBuilder();
    sb.append("FindIds exeMicros[").append(executionTimeMicros).append("] rows[").append(rowCount).append("] type[").append(desc.getName()).append("] predicates[").append(predicates.getLogWhereSql()).append("] bind[").append(bindLog).append("]");
    




    return sb.toString();
  }
  


  public String getBindLog()
  {
    return bindLog;
  }
  


  public String getGeneratedSql()
  {
    return sql;
  }
  
  public SpiOrmQueryRequest<?> getQueryRequest() {
    return request;
  }
  


  public BeanIdList findIds()
    throws SQLException
  {
    boolean useBackgroundToContinueFetch = false;
    
    startNano = System.nanoTime();
    


    try
    {
      List<Object> idList = query.getIdList();
      if (idList == null)
      {
        idList = Collections.synchronizedList(new ArrayList());
        query.setIdList(idList);
      }
      
      BeanIdList result = new BeanIdList(idList);
      
      SpiTransaction t = request.getTransaction();
      Connection conn = t.getInternalConnection();
      pstmt = conn.prepareStatement(sql);
      
      if (query.getBufferFetchSizeHint() > 0) {
        pstmt.setFetchSize(query.getBufferFetchSizeHint());
      }
      
      if (query.getTimeout() > 0) {
        pstmt.setQueryTimeout(query.getTimeout());
      }
      
      bindLog = predicates.bind(new DataBind(pstmt));
      
      ResultSet rset = pstmt.executeQuery();
      dataReader = new RsetDataReader(rset);
      
      boolean hitMaxRows = false;
      boolean hasMoreRows = false;
      rowCount = 0;
      
      DbReadContext ctx = new DbContext();
      
      while (rset.next()) {
        Object idValue = desc.getIdBinder().read(ctx);
        idList.add(idValue);
        
        dataReader.resetColumnPosition();
        rowCount += 1;
        
        if ((maxRows > 0) && (rowCount == maxRows)) {
          hitMaxRows = true;
          hasMoreRows = rset.next();
          break;
        }
        if ((bgFetchAfter > 0) && (rowCount >= bgFetchAfter)) {
          useBackgroundToContinueFetch = true;
          break;
        }
      }
      
      if (hitMaxRows) {
        result.setHasMore(hasMoreRows);
      }
      
      if (useBackgroundToContinueFetch)
      {

        request.setBackgroundFetching();
        

        BackgroundIdFetch bgFetch = new BackgroundIdFetch(t, rset, pstmt, ctx, desc, result);
        FutureTask<Integer> future = new FutureTask(bgFetch);
        backgroundExecutor.execute(future);
        

        result.setBackgroundFetch(future);
      }
      
      long exeNano = System.nanoTime() - startNano;
      executionTimeMicros = ((int)exeNano / 1000);
      
      return result;
    }
    finally {
      if (!useBackgroundToContinueFetch)
      {

        close();
      }
    }
  }
  





  private void close()
  {
    try
    {
      if (dataReader != null) {
        dataReader.close();
        dataReader = null;
      }
    } catch (SQLException e) {
      logger.log(Level.SEVERE, null, e);
    }
    try {
      if (pstmt != null) {
        pstmt.close();
        pstmt = null;
      }
    } catch (SQLException e) {
      logger.log(Level.SEVERE, null, e);
    }
  }
  
  class DbContext implements DbReadContext {
    DbContext() {}
    
    public int getParentState() {
      return 0;
    }
    
    public void propagateState(Object e) {
      throw new RuntimeException("Not Called");
    }
    
    public SpiQuery.Mode getQueryMode() {
      return SpiQuery.Mode.NORMAL;
    }
    
    public DataReader getDataReader() {
      return dataReader;
    }
    
    public boolean isVanillaMode() {
      return false;
    }
    
    public boolean isSharedInstance() {
      return false;
    }
    
    public boolean isReadOnly() {
      return false;
    }
    
    public boolean isRawSql() {
      return false;
    }
    

    public void register(String path, EntityBeanIntercept ebi) {}
    

    public void register(String path, BeanCollection<?> bc) {}
    
    public ReferenceOptions getReferenceOptionsFor(BeanPropertyAssocOne<?> beanProp)
    {
      return null;
    }
    
    public BeanPropertyAssocMany<?> getManyProperty()
    {
      return null;
    }
    
    public PersistenceContext getPersistenceContext()
    {
      return null;
    }
    
    public boolean isAutoFetchProfiling() {
      return false;
    }
    
    public void profileBean(EntityBeanIntercept ebi, String prefix) {}
    
    public void setCurrentPrefix(String currentPrefix, Map<String, String> pathMap) {}
    
    public void setLoadedBean(Object loadedBean, Object id) {}
    
    public void setLoadedManyBean(Object loadedBean) {}
  }
}
