package com.avaje.ebeaninternal.server.transaction;

import com.avaje.ebean.LogLevel;
import com.avaje.ebean.bean.PersistenceContext;
import com.avaje.ebean.config.lucene.IndexUpdateFuture;
import com.avaje.ebeaninternal.api.SpiTransaction;
import com.avaje.ebeaninternal.api.TransactionEvent;
import com.avaje.ebeaninternal.server.lucene.LIndexUpdateFuture;
import com.avaje.ebeaninternal.server.lucene.PersistenceLuceneException;
import com.avaje.ebeaninternal.server.persist.BatchControl;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.PersistenceException;
import javax.persistence.RollbackException;

























public class JdbcTransaction
  implements SpiTransaction
{
  private static final Logger logger = Logger.getLogger(JdbcTransaction.class.getName());
  



  private static final String illegalStateMessage = "Transaction is Inactive";
  



  protected final TransactionManager manager;
  



  final String id;
  



  final boolean explicit;
  



  final TransactionManager.OnQueryOnly onQueryOnly;
  



  boolean active;
  



  Connection connection;
  



  BatchControl batchControl;
  



  TransactionEvent event;
  


  PersistenceContext persistenceContext;
  


  boolean persistCascade = true;
  




  boolean queryOnly = true;
  

  boolean localReadOnly;
  

  LogLevel logLevel;
  

  boolean batchMode;
  
  int batchSize = -1;
  
  boolean batchFlushOnQuery = true;
  


  Boolean batchGetGeneratedKeys;
  

  Boolean batchFlushOnMixed;
  

  int depth = 0;
  
  HashSet<Integer> persistingBeans;
  
  TransactionLogBuffer logBuffer;
  
  List<LIndexUpdateFuture> indexUpdateFutures;
  

  public JdbcTransaction(String id, boolean explicit, LogLevel logLevel, Connection connection, TransactionManager manager)
  {
    try
    {
      active = true;
      this.id = id;
      this.explicit = explicit;
      this.logLevel = logLevel;
      this.manager = manager;
      this.connection = connection;
      onQueryOnly = (manager == null ? TransactionManager.OnQueryOnly.ROLLBACK : manager.getOnQueryOnly());
      persistenceContext = new DefaultPersistenceContext();
      
      logBuffer = new TransactionLogBuffer(50, id);
    }
    catch (Exception e) {
      throw new PersistenceException(e);
    }
  }
  
  public String toString() {
    return "Trans[" + id + "]";
  }
  
  public void addIndexUpdateFuture(LIndexUpdateFuture future)
  {
    if (indexUpdateFutures == null) {
      indexUpdateFutures = new ArrayList();
    }
    indexUpdateFutures.add(future);
  }
  
  public void waitForIndexUpdates() {
    if (indexUpdateFutures != null) {
      try {
        for (IndexUpdateFuture f : indexUpdateFutures) {
          f.get();
        }
      } catch (InterruptedException e) {
        throw new PersistenceLuceneException(e);
      } catch (ExecutionException e) {
        throw new PersistenceLuceneException(e);
      }
    }
  }
  






  public void registerBean(Integer persistingBean)
  {
    if (persistingBeans == null) {
      persistingBeans = new HashSet();
    }
    persistingBeans.add(persistingBean);
  }
  


  public void unregisterBean(Integer persistedBean)
  {
    if (persistingBeans != null) {
      persistingBeans.remove(persistedBean);
    }
  }
  


  public boolean isRegisteredBean(Integer persistingBean)
  {
    if (persistingBeans == null) {
      return false;
    }
    return persistingBeans.contains(persistingBean);
  }
  
















  public int depth(int diff)
  {
    depth += diff;
    return depth;
  }
  
  public boolean isReadOnly() {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    try {
      return connection.isReadOnly();
    } catch (SQLException e) {
      throw new PersistenceException(e);
    }
  }
  
  public void setReadOnly(boolean readOnly) {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    try {
      localReadOnly = readOnly;
      connection.setReadOnly(readOnly);
    } catch (SQLException e) {
      throw new PersistenceException(e);
    }
  }
  
  public void setBatchMode(boolean batchMode) {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    this.batchMode = batchMode;
  }
  
  public void setBatchGetGeneratedKeys(boolean getGeneratedKeys) {
    batchGetGeneratedKeys = Boolean.valueOf(getGeneratedKeys);
    if (batchControl != null) {
      batchControl.setGetGeneratedKeys(Boolean.valueOf(getGeneratedKeys));
    }
  }
  
  public void setBatchFlushOnMixed(boolean batchFlushOnMixed) {
    this.batchFlushOnMixed = Boolean.valueOf(batchFlushOnMixed);
    if (batchControl != null) {
      batchControl.setBatchFlushOnMixed(batchFlushOnMixed);
    }
  }
  





  public int getBatchSize()
  {
    return batchSize;
  }
  
  public void setBatchSize(int batchSize) {
    this.batchSize = batchSize;
    if (batchControl != null) {
      batchControl.setBatchSize(batchSize);
    }
  }
  
  public boolean isBatchFlushOnQuery() {
    return batchFlushOnQuery;
  }
  
  public void setBatchFlushOnQuery(boolean batchFlushOnQuery) {
    this.batchFlushOnQuery = batchFlushOnQuery;
  }
  



  public boolean isBatchThisRequest()
  {
    if ((!explicit) && (depth <= 0))
    {

      return false;
    }
    return batchMode;
  }
  
  public BatchControl getBatchControl()
  {
    return batchControl;
  }
  



  public void setBatchControl(BatchControl batchControl)
  {
    queryOnly = false;
    this.batchControl = batchControl;
    
    if (batchGetGeneratedKeys != null) {
      batchControl.setGetGeneratedKeys(batchGetGeneratedKeys);
    }
    if (batchSize != -1) {
      batchControl.setBatchSize(batchSize);
    }
    if (batchFlushOnMixed != null) {
      batchControl.setBatchFlushOnMixed(batchFlushOnMixed.booleanValue());
    }
  }
  






  public void flushBatch()
  {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    if (batchControl != null) {
      batchControl.flush();
    }
  }
  
  public void batchFlush() {
    flushBatch();
  }
  



  public PersistenceContext getPersistenceContext()
  {
    return persistenceContext;
  }
  







  public void setPersistenceContext(PersistenceContext context)
  {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    persistenceContext = context;
  }
  


  public TransactionEvent getEvent()
  {
    queryOnly = false;
    if (event == null) {
      event = new TransactionEvent();
    }
    return event;
  }
  


  public void setLoggingOn(boolean loggingOn)
  {
    if (loggingOn) {
      logLevel = LogLevel.SQL;
    } else {
      logLevel = LogLevel.NONE;
    }
  }
  


  public boolean isExplicit()
  {
    return explicit;
  }
  
  public boolean isLogSql()
  {
    return logLevel.ordinal() >= LogLevel.SQL.ordinal();
  }
  
  public boolean isLogSummary() {
    return logLevel.ordinal() >= LogLevel.SUMMARY.ordinal();
  }
  
  public LogLevel getLogLevel() {
    return logLevel;
  }
  
  public void setLogLevel(LogLevel logLevel) {
    this.logLevel = logLevel;
  }
  


  public void log(String msg)
  {
    if (isLogSummary()) {
      logInternal(msg);
    }
  }
  



  public void logInternal(String msg)
  {
    if ((manager != null) && 
      (logBuffer.add(msg)))
    {
      manager.log(logBuffer);
      logBuffer = logBuffer.newBuffer();
    }
  }
  



  public String getId()
  {
    return id;
  }
  


  public Connection getInternalConnection()
  {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    return connection;
  }
  


  public Connection getConnection()
  {
    queryOnly = false;
    return getInternalConnection();
  }
  
  protected void deactivate() {
    try {
      if (localReadOnly) {
        connection.setReadOnly(false);
      }
    } catch (SQLException e) {
      logger.log(Level.SEVERE, "Error setting to readOnly?", e);
    }
    try {
      connection.close();
    }
    catch (Exception ex)
    {
      logger.log(Level.SEVERE, "Error closing connection", ex);
    }
    connection = null;
    active = false;
  }
  
  public TransactionLogBuffer getLogBuffer() {
    return logBuffer;
  }
  


  protected void notifyCommit()
  {
    if (manager == null) {
      return;
    }
    if (queryOnly) {
      manager.notifyOfQueryOnly(true, this, null);
    } else {
      manager.notifyOfCommit(this);
    }
  }
  





  private void commitQueryOnly()
  {
    try
    {
      switch (1.$SwitchMap$com$avaje$ebeaninternal$server$transaction$TransactionManager$OnQueryOnly[onQueryOnly.ordinal()]) {
      case 1: 
        connection.rollback();
        break;
      case 2: 
        connection.commit();
        break;
      
      case 3: 
        break;
      
      default: 
        connection.rollback();
      }
    } catch (SQLException e) {
      String m = "Error when ending a query only transaction via " + onQueryOnly;
      logger.log(Level.SEVERE, m, e);
    }
  }
  

  public void commit()
    throws RollbackException
  {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    try {
      if (queryOnly)
      {
        commitQueryOnly();
      }
      else {
        if ((batchControl != null) && (!batchControl.isEmpty())) {
          batchControl.flush();
        }
        connection.commit();
      }
      
      deactivate();
      notifyCommit();
    }
    catch (Exception e) {
      throw new RollbackException(e);
    }
  }
  


  protected void notifyRollback(Throwable cause)
  {
    if (manager == null) {
      return;
    }
    if (queryOnly) {
      manager.notifyOfQueryOnly(false, this, cause);
    } else {
      manager.notifyOfRollback(this, cause);
    }
  }
  

  public void rollback()
    throws PersistenceException
  {
    rollback(null);
  }
  


  public void rollback(Throwable cause)
    throws PersistenceException
  {
    if (!isActive()) {
      throw new IllegalStateException("Transaction is Inactive");
    }
    try {
      connection.rollback();
      

      deactivate();
      notifyRollback(cause);
    }
    catch (Exception ex) {
      throw new PersistenceException(ex);
    }
  }
  

  public void end()
    throws PersistenceException
  {
    if (isActive()) {
      rollback();
    }
  }
  


  public boolean isActive()
  {
    return active;
  }
  
  public boolean isPersistCascade() {
    return persistCascade;
  }
  
  public void setPersistCascade(boolean persistCascade) {
    this.persistCascade = persistCascade;
  }
  
  public void addModification(String tableName, boolean inserts, boolean updates, boolean deletes) {
    getEvent().add(tableName, inserts, updates, deletes);
  }
  
  public final TransactionManager getTransactionManger() {
    return manager;
  }
}
