package com.google.common.base;

import com.google.common.annotations.Beta;
import com.google.common.annotations.GwtCompatible;
import com.google.common.annotations.GwtIncompatible;
import java.util.concurrent.TimeUnit;


































































@Beta
@GwtCompatible(emulated=true)
public final class Stopwatch
{
  private final Ticker ticker;
  private boolean isRunning;
  private long elapsedNanos;
  private long startTick;
  
  public Stopwatch()
  {
    this(Ticker.systemTicker());
  }
  



  public Stopwatch(Ticker ticker)
  {
    this.ticker = ((Ticker)Preconditions.checkNotNull(ticker));
  }
  




  public boolean isRunning()
  {
    return isRunning;
  }
  




  public Stopwatch start()
  {
    Preconditions.checkState(!isRunning);
    isRunning = true;
    startTick = ticker.read();
    return this;
  }
  





  public Stopwatch stop()
  {
    long tick = ticker.read();
    Preconditions.checkState(isRunning);
    isRunning = false;
    elapsedNanos += tick - startTick;
    return this;
  }
  



  public Stopwatch reset()
  {
    elapsedNanos = 0L;
    isRunning = false;
    return this;
  }
  
  private long elapsedNanos() {
    return isRunning ? ticker.read() - startTick + elapsedNanos : elapsedNanos;
  }
  







  public long elapsedTime(TimeUnit desiredUnit)
  {
    return desiredUnit.convert(elapsedNanos(), TimeUnit.NANOSECONDS);
  }
  




  public long elapsedMillis()
  {
    return elapsedTime(TimeUnit.MILLISECONDS);
  }
  



  @GwtIncompatible("String.format()")
  public String toString()
  {
    return toString(4);
  }
  





  @GwtIncompatible("String.format()")
  public String toString(int significantDigits)
  {
    long nanos = elapsedNanos();
    
    TimeUnit unit = chooseUnit(nanos);
    double value = nanos / TimeUnit.NANOSECONDS.convert(1L, unit);
    

    return String.format("%." + significantDigits + "g %s", new Object[] { Double.valueOf(value), abbreviate(unit) });
  }
  
  private static TimeUnit chooseUnit(long nanos)
  {
    if (TimeUnit.SECONDS.convert(nanos, TimeUnit.NANOSECONDS) > 0L) {
      return TimeUnit.SECONDS;
    }
    if (TimeUnit.MILLISECONDS.convert(nanos, TimeUnit.NANOSECONDS) > 0L) {
      return TimeUnit.MILLISECONDS;
    }
    if (TimeUnit.MICROSECONDS.convert(nanos, TimeUnit.NANOSECONDS) > 0L) {
      return TimeUnit.MICROSECONDS;
    }
    return TimeUnit.NANOSECONDS;
  }
  
  private static String abbreviate(TimeUnit unit) {
    switch (1.$SwitchMap$java$util$concurrent$TimeUnit[unit.ordinal()]) {
    case 1: 
      return "ns";
    case 2: 
      return "μs";
    case 3: 
      return "ms";
    case 4: 
      return "s";
    }
    throw new AssertionError();
  }
}
