package com.google.common.cache;

import com.google.common.annotations.Beta;
import com.google.common.base.Ascii;
import com.google.common.base.Equivalence;
import com.google.common.base.Objects;
import com.google.common.base.Objects.ToStringHelper;
import com.google.common.base.Preconditions;
import com.google.common.base.Supplier;
import com.google.common.base.Suppliers;
import com.google.common.base.Ticker;
import com.google.common.collect.ForwardingConcurrentMap;
import com.google.common.util.concurrent.ExecutionError;
import com.google.common.util.concurrent.UncheckedExecutionException;
import java.io.Serializable;
import java.util.AbstractMap;
import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;






































































































@Beta
public final class CacheBuilder<K, V>
{
  private static final int DEFAULT_INITIAL_CAPACITY = 16;
  private static final int DEFAULT_CONCURRENCY_LEVEL = 4;
  private static final int DEFAULT_EXPIRATION_NANOS = 0;
  static final Supplier<? extends AbstractCache.StatsCounter> DEFAULT_STATS_COUNTER = Suppliers.ofInstance(new AbstractCache.StatsCounter()
  {
    public void recordHit() {}
    

    public void recordLoadSuccess(long loadTime) {}
    

    public void recordLoadException(long loadTime) {}
    

    public void recordConcurrentMiss() {}
    

    public void recordEviction() {}
    

    public CacheStats snapshot()
    {
      return CacheBuilder.EMPTY_STATS;
    }
  });
  




















  static final CacheStats EMPTY_STATS = new CacheStats(0L, 0L, 0L, 0L, 0L, 0L);
  
  static final Supplier<AbstractCache.SimpleStatsCounter> CACHE_STATS_COUNTER = new Supplier()
  {

    public AbstractCache.SimpleStatsCounter get() {
      return new AbstractCache.SimpleStatsCounter(); }
  };
  static final int UNSET_INT = -1;
  
  static enum NullListener implements RemovalListener<Object, Object> {
    INSTANCE;
    

    private NullListener() {}
    
    public void onRemoval(RemovalNotification<Object, Object> notification) {}
  }
  
  int initialCapacity = -1;
  int concurrencyLevel = -1;
  int maximumSize = -1;
  
  CustomConcurrentHashMap.Strength keyStrength;
  
  CustomConcurrentHashMap.Strength valueStrength;
  long expireAfterWriteNanos = -1L;
  long expireAfterAccessNanos = -1L;
  

  RemovalCause nullRemovalCause;
  

  Equivalence<Object> keyEquivalence;
  

  Equivalence<Object> valueEquivalence;
  

  RemovalListener<? super K, ? super V> removalListener;
  
  Ticker ticker;
  

  public static CacheBuilder<Object, Object> newBuilder()
  {
    return new CacheBuilder();
  }
  
  private boolean useNullCache() {
    return nullRemovalCause == null;
  }
  





  CacheBuilder<K, V> keyEquivalence(Equivalence<Object> equivalence)
  {
    Preconditions.checkState(keyEquivalence == null, "key equivalence was already set to %s", new Object[] { keyEquivalence });
    keyEquivalence = ((Equivalence)Preconditions.checkNotNull(equivalence));
    return this;
  }
  
  Equivalence<Object> getKeyEquivalence() {
    return (Equivalence)Objects.firstNonNull(keyEquivalence, getKeyStrength().defaultEquivalence());
  }
  






  CacheBuilder<K, V> valueEquivalence(Equivalence<Object> equivalence)
  {
    Preconditions.checkState(valueEquivalence == null, "value equivalence was already set to %s", new Object[] { valueEquivalence });
    
    valueEquivalence = ((Equivalence)Preconditions.checkNotNull(equivalence));
    return this;
  }
  
  Equivalence<Object> getValueEquivalence() {
    return (Equivalence)Objects.firstNonNull(valueEquivalence, getValueStrength().defaultEquivalence());
  }
  









  public CacheBuilder<K, V> initialCapacity(int initialCapacity)
  {
    Preconditions.checkState(this.initialCapacity == -1, "initial capacity was already set to %s", new Object[] { Integer.valueOf(this.initialCapacity) });
    
    Preconditions.checkArgument(initialCapacity >= 0);
    this.initialCapacity = initialCapacity;
    return this;
  }
  
  int getInitialCapacity() {
    return initialCapacity == -1 ? 16 : initialCapacity;
  }
  

















  public CacheBuilder<K, V> concurrencyLevel(int concurrencyLevel)
  {
    Preconditions.checkState(this.concurrencyLevel == -1, "concurrency level was already set to %s", new Object[] { Integer.valueOf(this.concurrencyLevel) });
    
    Preconditions.checkArgument(concurrencyLevel > 0);
    this.concurrencyLevel = concurrencyLevel;
    return this;
  }
  
  int getConcurrencyLevel() {
    return concurrencyLevel == -1 ? 4 : concurrencyLevel;
  }
  














  public CacheBuilder<K, V> maximumSize(int size)
  {
    Preconditions.checkState(maximumSize == -1, "maximum size was already set to %s", new Object[] { Integer.valueOf(maximumSize) });
    
    Preconditions.checkArgument(size >= 0, "maximum size must not be negative");
    maximumSize = size;
    if (maximumSize == 0)
    {
      nullRemovalCause = RemovalCause.SIZE;
    }
    return this;
  }
  




  CacheBuilder<K, V> strongKeys()
  {
    return setKeyStrength(CustomConcurrentHashMap.Strength.STRONG);
  }
  












  public CacheBuilder<K, V> weakKeys()
  {
    return setKeyStrength(CustomConcurrentHashMap.Strength.WEAK);
  }
  
  CacheBuilder<K, V> setKeyStrength(CustomConcurrentHashMap.Strength strength) {
    Preconditions.checkState(keyStrength == null, "Key strength was already set to %s", new Object[] { keyStrength });
    keyStrength = ((CustomConcurrentHashMap.Strength)Preconditions.checkNotNull(strength));
    return this;
  }
  
  CustomConcurrentHashMap.Strength getKeyStrength() {
    return (CustomConcurrentHashMap.Strength)Objects.firstNonNull(keyStrength, CustomConcurrentHashMap.Strength.STRONG);
  }
  




  CacheBuilder<K, V> strongValues()
  {
    return setValueStrength(CustomConcurrentHashMap.Strength.STRONG);
  }
  















  public CacheBuilder<K, V> weakValues()
  {
    return setValueStrength(CustomConcurrentHashMap.Strength.WEAK);
  }
  


















  public CacheBuilder<K, V> softValues()
  {
    return setValueStrength(CustomConcurrentHashMap.Strength.SOFT);
  }
  
  CacheBuilder<K, V> setValueStrength(CustomConcurrentHashMap.Strength strength) {
    Preconditions.checkState(valueStrength == null, "Value strength was already set to %s", new Object[] { valueStrength });
    valueStrength = ((CustomConcurrentHashMap.Strength)Preconditions.checkNotNull(strength));
    return this;
  }
  
  CustomConcurrentHashMap.Strength getValueStrength() {
    return (CustomConcurrentHashMap.Strength)Objects.firstNonNull(valueStrength, CustomConcurrentHashMap.Strength.STRONG);
  }
  

















  public CacheBuilder<K, V> expireAfterWrite(long duration, TimeUnit unit)
  {
    checkExpiration(duration, unit);
    expireAfterWriteNanos = unit.toNanos(duration);
    if ((duration == 0L) && (nullRemovalCause == null))
    {
      nullRemovalCause = RemovalCause.EXPIRED;
    }
    return this;
  }
  
  private void checkExpiration(long duration, TimeUnit unit) {
    Preconditions.checkState(expireAfterWriteNanos == -1L, "expireAfterWrite was already set to %s ns", new Object[] { Long.valueOf(expireAfterWriteNanos) });
    
    Preconditions.checkState(expireAfterAccessNanos == -1L, "expireAfterAccess was already set to %s ns", new Object[] { Long.valueOf(expireAfterAccessNanos) });
    
    Preconditions.checkArgument(duration >= 0L, "duration cannot be negative: %s %s", new Object[] { Long.valueOf(duration), unit });
  }
  
  long getExpireAfterWriteNanos() {
    return expireAfterWriteNanos == -1L ? 0L : expireAfterWriteNanos;
  }
  



















  public CacheBuilder<K, V> expireAfterAccess(long duration, TimeUnit unit)
  {
    checkExpiration(duration, unit);
    expireAfterAccessNanos = unit.toNanos(duration);
    if ((duration == 0L) && (nullRemovalCause == null))
    {
      nullRemovalCause = RemovalCause.EXPIRED;
    }
    return this;
  }
  
  long getExpireAfterAccessNanos() {
    return expireAfterAccessNanos == -1L ? 0L : expireAfterAccessNanos;
  }
  









  public CacheBuilder<K, V> ticker(Ticker ticker)
  {
    Preconditions.checkState(this.ticker == null);
    this.ticker = ((Ticker)Preconditions.checkNotNull(ticker));
    return this;
  }
  
  Ticker getTicker() {
    return (Ticker)Objects.firstNonNull(ticker, Ticker.systemTicker());
  }
  
























  @CheckReturnValue
  public <K1 extends K, V1 extends V> CacheBuilder<K1, V1> removalListener(RemovalListener<? super K1, ? super V1> listener)
  {
    Preconditions.checkState(removalListener == null);
    


    CacheBuilder<K1, V1> me = this;
    removalListener = ((RemovalListener)Preconditions.checkNotNull(listener));
    return me;
  }
  

  <K1 extends K, V1 extends V> RemovalListener<K1, V1> getRemovalListener()
  {
    return (RemovalListener)Objects.firstNonNull(removalListener, NullListener.INSTANCE);
  }
  











  public <K1 extends K, V1 extends V> Cache<K1, V1> build(CacheLoader<? super K1, V1> loader)
  {
    return useNullCache() ? new ComputingCache(this, CACHE_STATS_COUNTER, loader) : new NullCache(this, CACHE_STATS_COUNTER, loader);
  }
  






  public String toString()
  {
    Objects.ToStringHelper s = Objects.toStringHelper(this);
    if (initialCapacity != -1) {
      s.add("initialCapacity", Integer.valueOf(initialCapacity));
    }
    if (concurrencyLevel != -1) {
      s.add("concurrencyLevel", Integer.valueOf(concurrencyLevel));
    }
    if (maximumSize != -1) {
      s.add("maximumSize", Integer.valueOf(maximumSize));
    }
    if (expireAfterWriteNanos != -1L) {
      s.add("expireAfterWrite", expireAfterWriteNanos + "ns");
    }
    if (expireAfterAccessNanos != -1L) {
      s.add("expireAfterAccess", expireAfterAccessNanos + "ns");
    }
    if (keyStrength != null) {
      s.add("keyStrength", Ascii.toLowerCase(keyStrength.toString()));
    }
    if (valueStrength != null) {
      s.add("valueStrength", Ascii.toLowerCase(valueStrength.toString()));
    }
    if (keyEquivalence != null) {
      s.addValue("keyEquivalence");
    }
    if (valueEquivalence != null) {
      s.addValue("valueEquivalence");
    }
    if (removalListener != null) {
      s.addValue("removalListener");
    }
    return s.toString();
  }
  
  static class NullConcurrentMap<K, V>
    extends AbstractMap<K, V> implements ConcurrentMap<K, V>, Serializable
  {
    private static final long serialVersionUID = 0L;
    private final RemovalListener<K, V> removalListener;
    private final RemovalCause removalCause;
    
    NullConcurrentMap(CacheBuilder<? super K, ? super V> builder)
    {
      removalListener = builder.getRemovalListener();
      removalCause = nullRemovalCause;
    }
    


    public boolean containsKey(@Nullable Object key)
    {
      return false;
    }
    
    public boolean containsValue(@Nullable Object value)
    {
      return false;
    }
    
    public V get(@Nullable Object key)
    {
      return null;
    }
    
    void notifyRemoval(K key, V value) {
      RemovalNotification<K, V> notification = new RemovalNotification(key, value, removalCause);
      
      removalListener.onRemoval(notification);
    }
    
    public V put(K key, V value)
    {
      Preconditions.checkNotNull(key);
      Preconditions.checkNotNull(value);
      notifyRemoval(key, value);
      return null;
    }
    
    public V putIfAbsent(K key, V value)
    {
      return put(key, value);
    }
    
    public V remove(@Nullable Object key)
    {
      return null;
    }
    
    public boolean remove(@Nullable Object key, @Nullable Object value)
    {
      return false;
    }
    
    public V replace(K key, V value)
    {
      Preconditions.checkNotNull(key);
      Preconditions.checkNotNull(value);
      return null;
    }
    
    public boolean replace(K key, @Nullable V oldValue, V newValue)
    {
      Preconditions.checkNotNull(key);
      Preconditions.checkNotNull(newValue);
      return false;
    }
    
    public Set<Map.Entry<K, V>> entrySet()
    {
      return Collections.emptySet();
    }
  }
  

  static final class NullComputingConcurrentMap<K, V>
    extends CacheBuilder.NullConcurrentMap<K, V>
  {
    private static final long serialVersionUID = 0L;
    final CacheLoader<? super K, ? extends V> loader;
    
    NullComputingConcurrentMap(CacheBuilder<? super K, ? super V> builder, CacheLoader<? super K, ? extends V> loader)
    {
      super();
      this.loader = ((CacheLoader)Preconditions.checkNotNull(loader));
    }
    

    public V get(Object k)
    {
      K key = k;
      V value = compute(key);
      Preconditions.checkNotNull(value, loader + " returned null for key " + key + ".");
      notifyRemoval(key, value);
      return value;
    }
    
    private V compute(K key) {
      Preconditions.checkNotNull(key);
      try {
        return loader.load(key);
      } catch (Exception e) {
        throw new UncheckedExecutionException(e);
      } catch (Error e) {
        throw new ExecutionError(e);
      }
    }
  }
  
  static final class NullCache<K, V>
    extends AbstractCache<K, V>
  {
    final CacheBuilder.NullConcurrentMap<K, V> map;
    final CacheLoader<? super K, V> loader;
    final AbstractCache.StatsCounter statsCounter;
    ConcurrentMap<K, V> asMap;
    
    NullCache(CacheBuilder<? super K, ? super V> builder, Supplier<? extends AbstractCache.StatsCounter> statsCounterSupplier, CacheLoader<? super K, V> loader)
    {
      map = new CacheBuilder.NullConcurrentMap(builder);
      statsCounter = ((AbstractCache.StatsCounter)statsCounterSupplier.get());
      this.loader = ((CacheLoader)Preconditions.checkNotNull(loader));
    }
    
    public V get(K key) throws ExecutionException
    {
      V value = compute(key);
      map.notifyRemoval(key, value);
      return value;
    }
    
    private V compute(K key) throws ExecutionException {
      Preconditions.checkNotNull(key);
      long start = System.nanoTime();
      V value = null;
      try {
        value = loader.load(key);
      } catch (RuntimeException e) { long elapsed;
        throw new UncheckedExecutionException(e);
      } catch (Exception e) {
        throw new ExecutionException(e);
      } catch (Error e) {
        throw new ExecutionError(e);
      } finally {
        long elapsed = System.nanoTime() - start;
        if (value == null) {
          statsCounter.recordLoadException(elapsed);
        } else {
          statsCounter.recordLoadSuccess(elapsed);
        }
        statsCounter.recordEviction();
      }
      if (value == null) {
        throw new NullPointerException();
      }
      return value;
    }
    

    public long size()
    {
      return 0L;
    }
    


    public void invalidate(Object key) {}
    


    public void invalidateAll() {}
    

    public CacheStats stats()
    {
      return statsCounter.snapshot();
    }
    


    public ConcurrentMap<K, V> asMap()
    {
      ConcurrentMap<K, V> am = asMap;
      return this.asMap = new CacheBuilder.CacheAsMap(map);
    }
  }
  
  static final class CacheAsMap<K, V> extends ForwardingConcurrentMap<K, V> {
    private final ConcurrentMap<K, V> delegate;
    
    CacheAsMap(ConcurrentMap<K, V> delegate) {
      this.delegate = delegate;
    }
    
    protected ConcurrentMap<K, V> delegate()
    {
      return delegate;
    }
    
    public V put(K key, V value)
    {
      throw new UnsupportedOperationException();
    }
    
    public void putAll(Map<? extends K, ? extends V> map)
    {
      throw new UnsupportedOperationException();
    }
    
    public V putIfAbsent(K key, V value)
    {
      throw new UnsupportedOperationException();
    }
    
    public V replace(K key, V value)
    {
      throw new UnsupportedOperationException();
    }
    
    public boolean replace(K key, V oldValue, V newValue)
    {
      throw new UnsupportedOperationException();
    }
  }
}
