package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import java.io.Serializable;
import java.util.AbstractCollection;
import java.util.AbstractMap;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.RandomAccess;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import javax.annotation.Nullable;


















































































@GwtCompatible
abstract class AbstractMultimap<K, V>
  implements Multimap<K, V>, Serializable
{
  private transient Map<K, Collection<V>> map;
  private transient int totalSize;
  private transient Set<K> keySet;
  private transient Multiset<K> multiset;
  private transient Collection<V> valuesCollection;
  private transient Collection<Map.Entry<K, V>> entries;
  private transient Map<K, Collection<V>> asMap;
  private static final long serialVersionUID = 2447537837011683357L;
  
  protected AbstractMultimap(Map<K, Collection<V>> map)
  {
    Preconditions.checkArgument(map.isEmpty());
    this.map = map;
  }
  
  final void setMap(Map<K, Collection<V>> map)
  {
    this.map = map;
    totalSize = 0;
    for (Collection<V> values : map.values()) {
      Preconditions.checkArgument(!values.isEmpty());
      totalSize += values.size();
    }
  }
  










  abstract Collection<V> createCollection();
  









  Collection<V> createCollection(@Nullable K key)
  {
    return createCollection();
  }
  
  Map<K, Collection<V>> backingMap() {
    return map;
  }
  


  public int size()
  {
    return totalSize;
  }
  
  public boolean isEmpty()
  {
    return totalSize == 0;
  }
  
  public boolean containsKey(@Nullable Object key)
  {
    return map.containsKey(key);
  }
  
  public boolean containsValue(@Nullable Object value)
  {
    for (Collection<V> collection : map.values()) {
      if (collection.contains(value)) {
        return true;
      }
    }
    
    return false;
  }
  
  public boolean containsEntry(@Nullable Object key, @Nullable Object value)
  {
    Collection<V> collection = (Collection)map.get(key);
    return (collection != null) && (collection.contains(value));
  }
  


  public boolean put(@Nullable K key, @Nullable V value)
  {
    Collection<V> collection = getOrCreateCollection(key);
    
    if (collection.add(value)) {
      totalSize += 1;
      return true;
    }
    return false;
  }
  
  private Collection<V> getOrCreateCollection(@Nullable K key)
  {
    Collection<V> collection = (Collection)map.get(key);
    if (collection == null) {
      collection = createCollection(key);
      map.put(key, collection);
    }
    return collection;
  }
  
  public boolean remove(@Nullable Object key, @Nullable Object value)
  {
    Collection<V> collection = (Collection)map.get(key);
    if (collection == null) {
      return false;
    }
    
    boolean changed = collection.remove(value);
    if (changed) {
      totalSize -= 1;
      if (collection.isEmpty()) {
        map.remove(key);
      }
    }
    return changed;
  }
  


  public boolean putAll(@Nullable K key, Iterable<? extends V> values)
  {
    if (!values.iterator().hasNext()) {
      return false;
    }
    Collection<V> collection = getOrCreateCollection(key);
    int oldSize = collection.size();
    
    boolean changed = false;
    if ((values instanceof Collection)) {
      Collection<? extends V> c = Collections2.cast(values);
      changed = collection.addAll(c);
    } else {
      for (V value : values) {
        changed |= collection.add(value);
      }
    }
    
    totalSize += collection.size() - oldSize;
    return changed;
  }
  
  public boolean putAll(Multimap<? extends K, ? extends V> multimap)
  {
    boolean changed = false;
    for (Map.Entry<? extends K, ? extends V> entry : multimap.entries()) {
      changed |= put(entry.getKey(), entry.getValue());
    }
    return changed;
  }
  






  public Collection<V> replaceValues(@Nullable K key, Iterable<? extends V> values)
  {
    Iterator<? extends V> iterator = values.iterator();
    if (!iterator.hasNext()) {
      return removeAll(key);
    }
    
    Collection<V> collection = getOrCreateCollection(key);
    Collection<V> oldValues = createCollection();
    oldValues.addAll(collection);
    
    totalSize -= collection.size();
    collection.clear();
    
    while (iterator.hasNext()) {
      if (collection.add(iterator.next())) {
        totalSize += 1;
      }
    }
    
    return unmodifiableCollectionSubclass(oldValues);
  }
  





  public Collection<V> removeAll(@Nullable Object key)
  {
    Collection<V> collection = (Collection)map.remove(key);
    Collection<V> output = createCollection();
    
    if (collection != null) {
      output.addAll(collection);
      totalSize -= collection.size();
      collection.clear();
    }
    
    return unmodifiableCollectionSubclass(output);
  }
  
  private Collection<V> unmodifiableCollectionSubclass(Collection<V> collection)
  {
    if ((collection instanceof SortedSet))
      return Collections.unmodifiableSortedSet((SortedSet)collection);
    if ((collection instanceof Set))
      return Collections.unmodifiableSet((Set)collection);
    if ((collection instanceof List)) {
      return Collections.unmodifiableList((List)collection);
    }
    return Collections.unmodifiableCollection(collection);
  }
  


  public void clear()
  {
    for (Collection<V> collection : map.values()) {
      collection.clear();
    }
    map.clear();
    totalSize = 0;
  }
  







  public Collection<V> get(@Nullable K key)
  {
    Collection<V> collection = (Collection)map.get(key);
    if (collection == null) {
      collection = createCollection(key);
    }
    return wrapCollection(key, collection);
  }
  





  private Collection<V> wrapCollection(@Nullable K key, Collection<V> collection)
  {
    if ((collection instanceof SortedSet))
      return new WrappedSortedSet(key, (SortedSet)collection, null);
    if ((collection instanceof Set))
      return new WrappedSet(key, (Set)collection);
    if ((collection instanceof List)) {
      return wrapList(key, (List)collection, null);
    }
    return new WrappedCollection(key, collection, null);
  }
  

  private List<V> wrapList(@Nullable K key, List<V> list, @Nullable AbstractMultimap<K, V>.WrappedCollection ancestor)
  {
    return (list instanceof RandomAccess) ? new RandomAccessWrappedList(key, list, ancestor) : new WrappedList(key, list, ancestor);
  }
  




  private class WrappedCollection
    extends AbstractCollection<V>
  {
    final K key;
    



    Collection<V> delegate;
    


    final AbstractMultimap<K, V>.WrappedCollection ancestor;
    


    final Collection<V> ancestorDelegate;
    



    WrappedCollection(Collection<V> key, @Nullable AbstractMultimap<K, V>.WrappedCollection delegate)
    {
      this.key = key;
      this.delegate = delegate;
      this.ancestor = ancestor;
      ancestorDelegate = (ancestor == null ? null : ancestor.getDelegate());
    }
    







    void refreshIfEmpty()
    {
      if (ancestor != null) {
        ancestor.refreshIfEmpty();
        if (ancestor.getDelegate() != ancestorDelegate) {
          throw new ConcurrentModificationException();
        }
      } else if (delegate.isEmpty()) {
        Collection<V> newDelegate = (Collection)map.get(key);
        if (newDelegate != null) {
          delegate = newDelegate;
        }
      }
    }
    



    void removeIfEmpty()
    {
      if (ancestor != null) {
        ancestor.removeIfEmpty();
      } else if (delegate.isEmpty()) {
        map.remove(key);
      }
    }
    
    K getKey() {
      return key;
    }
    






    void addToMap()
    {
      if (ancestor != null) {
        ancestor.addToMap();
      } else {
        map.put(key, delegate);
      }
    }
    
    public int size() {
      refreshIfEmpty();
      return delegate.size();
    }
    
    public boolean equals(@Nullable Object object) {
      if (object == this) {
        return true;
      }
      refreshIfEmpty();
      return delegate.equals(object);
    }
    
    public int hashCode() {
      refreshIfEmpty();
      return delegate.hashCode();
    }
    
    public String toString() {
      refreshIfEmpty();
      return delegate.toString();
    }
    
    Collection<V> getDelegate() {
      return delegate;
    }
    
    public Iterator<V> iterator() {
      refreshIfEmpty();
      return new WrappedIterator();
    }
    
    class WrappedIterator implements Iterator<V>
    {
      final Iterator<V> delegateIterator;
      final Collection<V> originalDelegate = delegate;
      
      WrappedIterator() {
        delegateIterator = AbstractMultimap.this.iteratorOrListIterator(delegate);
      }
      
      WrappedIterator() {
        this.delegateIterator = delegateIterator;
      }
      



      void validateIterator()
      {
        refreshIfEmpty();
        if (delegate != originalDelegate) {
          throw new ConcurrentModificationException();
        }
      }
      
      public boolean hasNext()
      {
        validateIterator();
        return delegateIterator.hasNext();
      }
      
      public V next()
      {
        validateIterator();
        return delegateIterator.next();
      }
      
      public void remove()
      {
        delegateIterator.remove();
        AbstractMultimap.access$210(AbstractMultimap.this);
        removeIfEmpty();
      }
      
      Iterator<V> getDelegateIterator() {
        validateIterator();
        return delegateIterator;
      }
    }
    
    public boolean add(V value) {
      refreshIfEmpty();
      boolean wasEmpty = delegate.isEmpty();
      boolean changed = delegate.add(value);
      if (changed) {
        AbstractMultimap.access$208(AbstractMultimap.this);
        if (wasEmpty) {
          addToMap();
        }
      }
      return changed;
    }
    
    AbstractMultimap<K, V>.WrappedCollection getAncestor() {
      return ancestor;
    }
    

    public boolean addAll(Collection<? extends V> collection)
    {
      if (collection.isEmpty()) {
        return false;
      }
      int oldSize = size();
      boolean changed = delegate.addAll(collection);
      if (changed) {
        int newSize = delegate.size();
        AbstractMultimap.access$212(AbstractMultimap.this, newSize - oldSize);
        if (oldSize == 0) {
          addToMap();
        }
      }
      return changed;
    }
    
    public boolean contains(Object o) {
      refreshIfEmpty();
      return delegate.contains(o);
    }
    
    public boolean containsAll(Collection<?> c) {
      refreshIfEmpty();
      return delegate.containsAll(c);
    }
    
    public void clear() {
      int oldSize = size();
      if (oldSize == 0) {
        return;
      }
      delegate.clear();
      AbstractMultimap.access$220(AbstractMultimap.this, oldSize);
      removeIfEmpty();
    }
    
    public boolean remove(Object o) {
      refreshIfEmpty();
      boolean changed = delegate.remove(o);
      if (changed) {
        AbstractMultimap.access$210(AbstractMultimap.this);
        removeIfEmpty();
      }
      return changed;
    }
    
    public boolean removeAll(Collection<?> c) {
      if (c.isEmpty()) {
        return false;
      }
      int oldSize = size();
      boolean changed = delegate.removeAll(c);
      if (changed) {
        int newSize = delegate.size();
        AbstractMultimap.access$212(AbstractMultimap.this, newSize - oldSize);
        removeIfEmpty();
      }
      return changed;
    }
    
    public boolean retainAll(Collection<?> c) {
      Preconditions.checkNotNull(c);
      int oldSize = size();
      boolean changed = delegate.retainAll(c);
      if (changed) {
        int newSize = delegate.size();
        AbstractMultimap.access$212(AbstractMultimap.this, newSize - oldSize);
        removeIfEmpty();
      }
      return changed;
    }
  }
  
  private Iterator<V> iteratorOrListIterator(Collection<V> collection) {
    return (collection instanceof List) ? ((List)collection).listIterator() : collection.iterator();
  }
  
  private class WrappedSet
    extends AbstractMultimap<K, V>.WrappedCollection implements Set<V>
  {
    WrappedSet(Set<V> key)
    {
      super(key, delegate, null);
    }
  }
  

  private class WrappedSortedSet
    extends AbstractMultimap<K, V>.WrappedCollection
    implements SortedSet<V>
  {
    WrappedSortedSet(SortedSet<V> key, @Nullable AbstractMultimap<K, V>.WrappedCollection delegate)
    {
      super(key, delegate, ancestor);
    }
    
    SortedSet<V> getSortedSetDelegate() {
      return (SortedSet)getDelegate();
    }
    
    public Comparator<? super V> comparator()
    {
      return getSortedSetDelegate().comparator();
    }
    
    public V first()
    {
      refreshIfEmpty();
      return getSortedSetDelegate().first();
    }
    
    public V last()
    {
      refreshIfEmpty();
      return getSortedSetDelegate().last();
    }
    
    public SortedSet<V> headSet(V toElement)
    {
      refreshIfEmpty();
      return new WrappedSortedSet(AbstractMultimap.this, getKey(), getSortedSetDelegate().headSet(toElement), getAncestor() == null ? this : getAncestor());
    }
    


    public SortedSet<V> subSet(V fromElement, V toElement)
    {
      refreshIfEmpty();
      return new WrappedSortedSet(AbstractMultimap.this, getKey(), getSortedSetDelegate().subSet(fromElement, toElement), getAncestor() == null ? this : getAncestor());
    }
    


    public SortedSet<V> tailSet(V fromElement)
    {
      refreshIfEmpty();
      return new WrappedSortedSet(AbstractMultimap.this, getKey(), getSortedSetDelegate().tailSet(fromElement), getAncestor() == null ? this : getAncestor());
    }
  }
  
  private class WrappedList
    extends AbstractMultimap<K, V>.WrappedCollection
    implements List<V>
  {
    WrappedList(List<V> key, @Nullable AbstractMultimap<K, V>.WrappedCollection delegate)
    {
      super(key, delegate, ancestor);
    }
    
    List<V> getListDelegate() {
      return (List)getDelegate();
    }
    
    public boolean addAll(int index, Collection<? extends V> c)
    {
      if (c.isEmpty()) {
        return false;
      }
      int oldSize = size();
      boolean changed = getListDelegate().addAll(index, c);
      if (changed) {
        int newSize = getDelegate().size();
        AbstractMultimap.access$212(AbstractMultimap.this, newSize - oldSize);
        if (oldSize == 0) {
          addToMap();
        }
      }
      return changed;
    }
    
    public V get(int index)
    {
      refreshIfEmpty();
      return getListDelegate().get(index);
    }
    
    public V set(int index, V element)
    {
      refreshIfEmpty();
      return getListDelegate().set(index, element);
    }
    
    public void add(int index, V element)
    {
      refreshIfEmpty();
      boolean wasEmpty = getDelegate().isEmpty();
      getListDelegate().add(index, element);
      AbstractMultimap.access$208(AbstractMultimap.this);
      if (wasEmpty) {
        addToMap();
      }
    }
    
    public V remove(int index)
    {
      refreshIfEmpty();
      V value = getListDelegate().remove(index);
      AbstractMultimap.access$210(AbstractMultimap.this);
      removeIfEmpty();
      return value;
    }
    
    public int indexOf(Object o)
    {
      refreshIfEmpty();
      return getListDelegate().indexOf(o);
    }
    
    public int lastIndexOf(Object o)
    {
      refreshIfEmpty();
      return getListDelegate().lastIndexOf(o);
    }
    
    public ListIterator<V> listIterator()
    {
      refreshIfEmpty();
      return new WrappedListIterator();
    }
    
    public ListIterator<V> listIterator(int index)
    {
      refreshIfEmpty();
      return new WrappedListIterator(index);
    }
    
    public List<V> subList(int fromIndex, int toIndex)
    {
      refreshIfEmpty();
      return AbstractMultimap.this.wrapList(getKey(), getListDelegate().subList(fromIndex, toIndex), getAncestor() == null ? this : getAncestor());
    }
    
    private class WrappedListIterator
      extends AbstractMultimap<K, V>.WrappedCollection.WrappedIterator implements ListIterator<V>
    {
      WrappedListIterator()
      {
        super();
      }
      
      public WrappedListIterator(int index) { super(getListDelegate().listIterator(index)); }
      
      private ListIterator<V> getDelegateListIterator()
      {
        return (ListIterator)getDelegateIterator();
      }
      
      public boolean hasPrevious()
      {
        return getDelegateListIterator().hasPrevious();
      }
      
      public V previous()
      {
        return getDelegateListIterator().previous();
      }
      
      public int nextIndex()
      {
        return getDelegateListIterator().nextIndex();
      }
      
      public int previousIndex()
      {
        return getDelegateListIterator().previousIndex();
      }
      
      public void set(V value)
      {
        getDelegateListIterator().set(value);
      }
      
      public void add(V value)
      {
        boolean wasEmpty = isEmpty();
        getDelegateListIterator().add(value);
        AbstractMultimap.access$208(AbstractMultimap.this);
        if (wasEmpty) {
          addToMap();
        }
      }
    }
  }
  


  private class RandomAccessWrappedList
    extends AbstractMultimap.WrappedList
    implements RandomAccess
  {
    RandomAccessWrappedList(List<V> key, @Nullable AbstractMultimap<K, V>.WrappedCollection delegate)
    {
      super(key, delegate, ancestor);
    }
  }
  


  public Set<K> keySet()
  {
    Set<K> result = keySet;
    return result == null ? (this.keySet = createKeySet()) : result;
  }
  
  private Set<K> createKeySet() {
    return (map instanceof SortedMap) ? new SortedKeySet((SortedMap)map) : new KeySet(map);
  }
  


  private class KeySet
    extends Maps.KeySet<K, Collection<V>>
  {
    final Map<K, Collection<V>> subMap;
    

    KeySet()
    {
      this.subMap = subMap;
    }
    
    Map<K, Collection<V>> map()
    {
      return subMap;
    }
    
    public Iterator<K> iterator() {
      new Iterator() {
        final Iterator<Map.Entry<K, Collection<V>>> entryIterator = subMap.entrySet().iterator();
        
        Map.Entry<K, Collection<V>> entry;
        
        public boolean hasNext()
        {
          return entryIterator.hasNext();
        }
        
        public K next() {
          entry = ((Map.Entry)entryIterator.next());
          return entry.getKey();
        }
        
        public void remove() {
          Preconditions.checkState(entry != null);
          Collection<V> collection = (Collection)entry.getValue();
          entryIterator.remove();
          AbstractMultimap.access$220(AbstractMultimap.this, collection.size());
          collection.clear();
        }
      };
    }
    

    public boolean remove(Object key)
    {
      int count = 0;
      Collection<V> collection = (Collection)subMap.remove(key);
      if (collection != null) {
        count = collection.size();
        collection.clear();
        AbstractMultimap.access$220(AbstractMultimap.this, count);
      }
      return count > 0;
    }
    
    public void clear()
    {
      Iterators.clear(iterator());
    }
    
    public boolean containsAll(Collection<?> c) {
      return subMap.keySet().containsAll(c);
    }
    
    public boolean equals(@Nullable Object object) {
      return (this == object) || (subMap.keySet().equals(object));
    }
    
    public int hashCode() {
      return subMap.keySet().hashCode();
    }
  }
  
  private class SortedKeySet extends AbstractMultimap<K, V>.KeySet implements SortedSet<K>
  {
    SortedKeySet() {
      super(subMap);
    }
    
    SortedMap<K, Collection<V>> sortedMap() {
      return (SortedMap)subMap;
    }
    
    public Comparator<? super K> comparator()
    {
      return sortedMap().comparator();
    }
    
    public K first()
    {
      return sortedMap().firstKey();
    }
    
    public SortedSet<K> headSet(K toElement)
    {
      return new SortedKeySet(AbstractMultimap.this, sortedMap().headMap(toElement));
    }
    
    public K last()
    {
      return sortedMap().lastKey();
    }
    
    public SortedSet<K> subSet(K fromElement, K toElement)
    {
      return new SortedKeySet(AbstractMultimap.this, sortedMap().subMap(fromElement, toElement));
    }
    
    public SortedSet<K> tailSet(K fromElement)
    {
      return new SortedKeySet(AbstractMultimap.this, sortedMap().tailMap(fromElement));
    }
  }
  


  public Multiset<K> keys()
  {
    Multiset<K> result = multiset;
    if (result == null) {
      this.multiset = new Multimaps.Keys() {
        Multimap<K, V> multimap() {
          return AbstractMultimap.this;
        }
      };
    }
    return result;
  }
  

  private int removeValuesForKey(Object key)
  {
    Collection<V> collection;
    
    try
    {
      collection = (Collection)map.remove(key);
    } catch (NullPointerException e) {
      return 0;
    } catch (ClassCastException e) {
      return 0;
    }
    
    int count = 0;
    if (collection != null) {
      count = collection.size();
      collection.clear();
      totalSize -= count;
    }
    return count;
  }
  







  public Collection<V> values()
  {
    Collection<V> result = valuesCollection;
    if (result == null) {
      this.valuesCollection = new Multimaps.Values() {
        Multimap<K, V> multimap() {
          return AbstractMultimap.this;
        }
      };
    }
    return result;
  }
  


















  public Collection<Map.Entry<K, V>> entries()
  {
    Collection<Map.Entry<K, V>> result = entries;
    return result == null ? (this.entries = createEntries()) : result;
  }
  
  Collection<Map.Entry<K, V>> createEntries() {
    if ((this instanceof SetMultimap)) {
      new Multimaps.EntrySet() {
        Multimap<K, V> multimap() {
          return AbstractMultimap.this;
        }
        
        public Iterator<Map.Entry<K, V>> iterator() {
          return createEntryIterator();
        }
      };
    }
    new Multimaps.Entries() {
      Multimap<K, V> multimap() {
        return AbstractMultimap.this;
      }
      
      public Iterator<Map.Entry<K, V>> iterator() {
        return createEntryIterator();
      }
    };
  }
  







  Iterator<Map.Entry<K, V>> createEntryIterator()
  {
    return new EntryIterator();
  }
  
  private class EntryIterator implements Iterator<Map.Entry<K, V>>
  {
    final Iterator<Map.Entry<K, Collection<V>>> keyIterator;
    K key;
    Collection<V> collection;
    Iterator<V> valueIterator;
    
    EntryIterator() {
      keyIterator = map.entrySet().iterator();
      if (keyIterator.hasNext()) {
        findValueIteratorAndKey();
      } else {
        valueIterator = Iterators.emptyModifiableIterator();
      }
    }
    
    void findValueIteratorAndKey() {
      Map.Entry<K, Collection<V>> entry = (Map.Entry)keyIterator.next();
      key = entry.getKey();
      collection = ((Collection)entry.getValue());
      valueIterator = collection.iterator();
    }
    
    public boolean hasNext()
    {
      return (keyIterator.hasNext()) || (valueIterator.hasNext());
    }
    
    public Map.Entry<K, V> next()
    {
      if (!valueIterator.hasNext()) {
        findValueIteratorAndKey();
      }
      return Maps.immutableEntry(key, valueIterator.next());
    }
    
    public void remove()
    {
      valueIterator.remove();
      if (collection.isEmpty()) {
        keyIterator.remove();
      }
      AbstractMultimap.access$210(AbstractMultimap.this);
    }
  }
  


  public Map<K, Collection<V>> asMap()
  {
    Map<K, Collection<V>> result = asMap;
    return result == null ? (this.asMap = createAsMap()) : result;
  }
  
  private Map<K, Collection<V>> createAsMap() {
    return (map instanceof SortedMap) ? new SortedAsMap((SortedMap)map) : new AsMap(map);
  }
  

  private class AsMap
    extends AbstractMap<K, Collection<V>>
  {
    final transient Map<K, Collection<V>> submap;
    transient Set<Map.Entry<K, Collection<V>>> entrySet;
    
    AsMap()
    {
      this.submap = submap;
    }
    

    public Set<Map.Entry<K, Collection<V>>> entrySet()
    {
      Set<Map.Entry<K, Collection<V>>> result = entrySet;
      return result == null ? (this.entrySet = new AsMapEntries()) : result;
    }
    

    public boolean containsKey(Object key)
    {
      return Maps.safeContainsKey(submap, key);
    }
    
    public Collection<V> get(Object key) {
      Collection<V> collection = (Collection)Maps.safeGet(submap, key);
      if (collection == null) {
        return null;
      }
      
      K k = key;
      return AbstractMultimap.this.wrapCollection(k, collection);
    }
    
    public Set<K> keySet() {
      return AbstractMultimap.this.keySet();
    }
    
    public int size()
    {
      return submap.size();
    }
    
    public Collection<V> remove(Object key) {
      Collection<V> collection = (Collection)submap.remove(key);
      if (collection == null) {
        return null;
      }
      
      Collection<V> output = createCollection();
      output.addAll(collection);
      AbstractMultimap.access$220(AbstractMultimap.this, collection.size());
      collection.clear();
      return output;
    }
    
    public boolean equals(@Nullable Object object) {
      return (this == object) || (submap.equals(object));
    }
    
    public int hashCode() {
      return submap.hashCode();
    }
    
    public String toString() {
      return submap.toString();
    }
    
    public void clear()
    {
      if (submap == map) {
        AbstractMultimap.this.clear();
      }
      else
        Iterators.clear(new AsMapIterator());
    }
    
    class AsMapEntries extends Maps.EntrySet<K, Collection<V>> {
      AsMapEntries() {}
      
      Map<K, Collection<V>> map() {
        return AbstractMultimap.AsMap.this;
      }
      
      public Iterator<Map.Entry<K, Collection<V>>> iterator() {
        return new AbstractMultimap.AsMap.AsMapIterator(AbstractMultimap.AsMap.this);
      }
      

      public boolean contains(Object o)
      {
        return Collections2.safeContains(submap.entrySet(), o);
      }
      
      public boolean remove(Object o) {
        if (!contains(o)) {
          return false;
        }
        Map.Entry<?, ?> entry = (Map.Entry)o;
        AbstractMultimap.this.removeValuesForKey(entry.getKey());
        return true;
      }
    }
    
    class AsMapIterator implements Iterator<Map.Entry<K, Collection<V>>>
    {
      final Iterator<Map.Entry<K, Collection<V>>> delegateIterator = submap.entrySet().iterator();
      Collection<V> collection;
      
      AsMapIterator() {}
      
      public boolean hasNext() {
        return delegateIterator.hasNext();
      }
      
      public Map.Entry<K, Collection<V>> next()
      {
        Map.Entry<K, Collection<V>> entry = (Map.Entry)delegateIterator.next();
        K key = entry.getKey();
        collection = ((Collection)entry.getValue());
        return Maps.immutableEntry(key, AbstractMultimap.this.wrapCollection(key, collection));
      }
      
      public void remove()
      {
        delegateIterator.remove();
        AbstractMultimap.access$220(AbstractMultimap.this, collection.size());
        collection.clear();
      }
    }
  }
  
  private class SortedAsMap extends AbstractMultimap<K, V>.AsMap implements SortedMap<K, Collection<V>> {
    SortedSet<K> sortedKeySet;
    
    SortedAsMap() { super(submap); }
    
    SortedMap<K, Collection<V>> sortedMap()
    {
      return (SortedMap)submap;
    }
    
    public Comparator<? super K> comparator()
    {
      return sortedMap().comparator();
    }
    
    public K firstKey()
    {
      return sortedMap().firstKey();
    }
    
    public K lastKey()
    {
      return sortedMap().lastKey();
    }
    
    public SortedMap<K, Collection<V>> headMap(K toKey)
    {
      return new SortedAsMap(AbstractMultimap.this, sortedMap().headMap(toKey));
    }
    
    public SortedMap<K, Collection<V>> subMap(K fromKey, K toKey)
    {
      return new SortedAsMap(AbstractMultimap.this, sortedMap().subMap(fromKey, toKey));
    }
    
    public SortedMap<K, Collection<V>> tailMap(K fromKey)
    {
      return new SortedAsMap(AbstractMultimap.this, sortedMap().tailMap(fromKey));
    }
    



    public SortedSet<K> keySet()
    {
      SortedSet<K> result = sortedKeySet;
      return result == null ? (this.sortedKeySet = new AbstractMultimap.SortedKeySet(AbstractMultimap.this, sortedMap())) : result;
    }
  }
  


  public boolean equals(@Nullable Object object)
  {
    if (object == this) {
      return true;
    }
    if ((object instanceof Multimap)) {
      Multimap<?, ?> that = (Multimap)object;
      return map.equals(that.asMap());
    }
    return false;
  }
  







  public int hashCode()
  {
    return map.hashCode();
  }
  






  public String toString()
  {
    return map.toString();
  }
}
