package com.google.common.collect;

import com.google.common.annotations.Beta;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.AbstractCollection;
import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;



























































@Beta
public final class ArrayTable<R, C, V>
  implements Table<R, C, V>, Serializable
{
  private final ImmutableList<R> rowList;
  private final ImmutableList<C> columnList;
  private final ImmutableMap<R, Integer> rowKeyToIndex;
  private final ImmutableMap<C, Integer> columnKeyToIndex;
  private final V[][] array;
  private transient ArrayTable<R, C, V>.CellSet cellSet;
  private transient ArrayTable<R, C, V>.ColumnMap columnMap;
  private transient ArrayTable<R, C, V>.RowMap rowMap;
  private transient Collection<V> values;
  private static final long serialVersionUID = 0L;
  
  public static <R, C, V> ArrayTable<R, C, V> create(Iterable<? extends R> rowKeys, Iterable<? extends C> columnKeys)
  {
    return new ArrayTable(rowKeys, columnKeys);
  }
  



























  public static <R, C, V> ArrayTable<R, C, V> create(Table<R, C, V> table)
  {
    return new ArrayTable(table);
  }
  




  public static <R, C, V> ArrayTable<R, C, V> create(ArrayTable<R, C, V> table)
  {
    return new ArrayTable(table);
  }
  








  private ArrayTable(Iterable<? extends R> rowKeys, Iterable<? extends C> columnKeys)
  {
    rowList = ImmutableList.copyOf(rowKeys);
    columnList = ImmutableList.copyOf(columnKeys);
    Preconditions.checkArgument(!rowList.isEmpty());
    Preconditions.checkArgument(!columnList.isEmpty());
    





    ImmutableMap.Builder<R, Integer> rowBuilder = ImmutableMap.builder();
    for (int i = 0; i < rowList.size(); i++) {
      rowBuilder.put(rowList.get(i), Integer.valueOf(i));
    }
    rowKeyToIndex = rowBuilder.build();
    
    ImmutableMap.Builder<C, Integer> columnBuilder = ImmutableMap.builder();
    for (int i = 0; i < columnList.size(); i++) {
      columnBuilder.put(columnList.get(i), Integer.valueOf(i));
    }
    columnKeyToIndex = columnBuilder.build();
    

    V[][] tmpArray = (Object[][])new Object[rowList.size()][columnList.size()];
    
    array = tmpArray;
  }
  
  private ArrayTable(Table<R, C, V> table) {
    this(table.rowKeySet(), table.columnKeySet());
    putAll(table);
  }
  
  private ArrayTable(ArrayTable<R, C, V> table) {
    rowList = rowList;
    columnList = columnList;
    rowKeyToIndex = rowKeyToIndex;
    columnKeyToIndex = columnKeyToIndex;
    
    V[][] copy = (Object[][])new Object[rowList.size()][columnList.size()];
    array = copy;
    for (int i = 0; i < rowList.size(); i++) {
      System.arraycopy(array[i], 0, copy[i], 0, array[i].length);
    }
  }
  



  public ImmutableList<R> rowKeyList()
  {
    return rowList;
  }
  



  public ImmutableList<C> columnKeyList()
  {
    return columnList;
  }
  













  public V at(int rowIndex, int columnIndex)
  {
    return array[rowIndex][columnIndex];
  }
  














  public V set(int rowIndex, int columnIndex, @Nullable V value)
  {
    V oldValue = array[rowIndex][columnIndex];
    array[rowIndex][columnIndex] = value;
    return oldValue;
  }
  











  public V[][] toArray(Class<V> valueClass)
  {
    V[][] copy = (Object[][])Array.newInstance(valueClass, new int[] { rowList.size(), columnList.size() });
    
    for (int i = 0; i < rowList.size(); i++) {
      System.arraycopy(array[i], 0, copy[i], 0, array[i].length);
    }
    return copy;
  }
  





  @Deprecated
  public void clear()
  {
    throw new UnsupportedOperationException();
  }
  



  public void eraseAll()
  {
    for (V[] row : array) {
      Arrays.fill(row, null);
    }
  }
  




  public boolean contains(@Nullable Object rowKey, @Nullable Object columnKey)
  {
    return (containsRow(rowKey)) && (containsColumn(columnKey));
  }
  




  public boolean containsColumn(@Nullable Object columnKey)
  {
    return columnKeyToIndex.containsKey(columnKey);
  }
  




  public boolean containsRow(@Nullable Object rowKey)
  {
    return rowKeyToIndex.containsKey(rowKey);
  }
  
  public boolean containsValue(@Nullable Object value)
  {
    for (V[] row : array) {
      for (V element : row) {
        if (Objects.equal(value, element)) {
          return true;
        }
      }
    }
    return false;
  }
  
  public V get(@Nullable Object rowKey, @Nullable Object columnKey)
  {
    Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
    Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
    return getIndexed(rowIndex, columnIndex);
  }
  
  private V getIndexed(Integer rowIndex, Integer columnIndex) {
    return (rowIndex == null) || (columnIndex == null) ? null : array[rowIndex.intValue()][columnIndex.intValue()];
  }
  




  public boolean isEmpty()
  {
    return false;
  }
  






  public V put(R rowKey, C columnKey, @Nullable V value)
  {
    Preconditions.checkNotNull(rowKey);
    Preconditions.checkNotNull(columnKey);
    Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
    Preconditions.checkArgument(rowIndex != null, "Row %s not in %s", new Object[] { rowKey, rowList });
    Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
    Preconditions.checkArgument(columnIndex != null, "Column %s not in %s", new Object[] { columnKey, columnList });
    
    return set(rowIndex.intValue(), columnIndex.intValue(), value);
  }
  
















  public void putAll(Table<? extends R, ? extends C, ? extends V> table)
  {
    for (Table.Cell<? extends R, ? extends C, ? extends V> cell : table.cellSet()) {
      put(cell.getRowKey(), cell.getColumnKey(), cell.getValue());
    }
  }
  





  @Deprecated
  public V remove(Object rowKey, Object columnKey)
  {
    throw new UnsupportedOperationException();
  }
  












  public V erase(@Nullable Object rowKey, @Nullable Object columnKey)
  {
    Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
    Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
    if ((rowIndex == null) || (columnIndex == null)) {
      return null;
    }
    return set(rowIndex.intValue(), columnIndex.intValue(), null);
  }
  


  public int size()
  {
    return rowList.size() * columnList.size();
  }
  
  public boolean equals(@Nullable Object obj) {
    if ((obj instanceof Table)) {
      Table<?, ?, ?> other = (Table)obj;
      return cellSet().equals(other.cellSet());
    }
    return false;
  }
  
  public int hashCode() {
    return cellSet().hashCode();
  }
  


  public String toString()
  {
    return rowMap().toString();
  }
  















  public Set<Table.Cell<R, C, V>> cellSet()
  {
    ArrayTable<R, C, V>.CellSet set = cellSet;
    return set == null ? (this.cellSet = new CellSet(null)) : set;
  }
  
  private class CellSet extends AbstractSet<Table.Cell<R, C, V>> {
    private CellSet() {}
    
    public Iterator<Table.Cell<R, C, V>> iterator() { new AbstractIndexedListIterator(size()) {
        protected Table.Cell<R, C, V> get(final int index) {
          new Tables.AbstractCell() {
            final int rowIndex = index / columnList.size();
            final int columnIndex = index % columnList.size();
            
            public R getRowKey() {
              return rowList.get(rowIndex);
            }
            
            public C getColumnKey() {
              return columnList.get(columnIndex);
            }
            
            public V getValue() {
              return array[rowIndex][columnIndex];
            }
          };
        }
      }; }
    
    public int size()
    {
      return ArrayTable.this.size();
    }
    
    public boolean contains(Object obj) {
      if ((obj instanceof Table.Cell)) {
        Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
        Integer rowIndex = (Integer)rowKeyToIndex.get(cell.getRowKey());
        Integer columnIndex = (Integer)columnKeyToIndex.get(cell.getColumnKey());
        return (rowIndex != null) && (columnIndex != null) && (Objects.equal(array[rowIndex.intValue()][columnIndex.intValue()], cell.getValue()));
      }
      

      return false;
    }
  }
  












  public Map<R, V> column(C columnKey)
  {
    Preconditions.checkNotNull(columnKey);
    Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
    return columnIndex == null ? ImmutableMap.of() : new Column(columnIndex.intValue());
  }
  
  private class Column extends AbstractMap<R, V> {
    final int columnIndex;
    ArrayTable<R, C, V>.ColumnEntrySet entrySet;
    
    Column(int columnIndex) {
      this.columnIndex = columnIndex;
    }
    

    public Set<Map.Entry<R, V>> entrySet()
    {
      ArrayTable<R, C, V>.ColumnEntrySet set = entrySet;
      return set == null ? (this.entrySet = new ArrayTable.ColumnEntrySet(ArrayTable.this, columnIndex)) : set;
    }
    
    public V get(Object rowKey) {
      Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
      return ArrayTable.this.getIndexed(rowIndex, Integer.valueOf(columnIndex));
    }
    
    public boolean containsKey(Object rowKey) {
      return rowKeyToIndex.containsKey(rowKey);
    }
    
    public V put(R rowKey, V value) {
      Preconditions.checkNotNull(rowKey);
      Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
      Preconditions.checkArgument(rowIndex != null, "Row %s not in %s", new Object[] { rowKey, rowList });
      return set(rowIndex.intValue(), columnIndex, value);
    }
    
    public Set<R> keySet() {
      return rowKeySet();
    }
  }
  
  private class ColumnEntrySet extends AbstractSet<Map.Entry<R, V>> {
    final int columnIndex;
    
    ColumnEntrySet(int columnIndex) {
      this.columnIndex = columnIndex;
    }
    
    public Iterator<Map.Entry<R, V>> iterator() {
      new AbstractIndexedListIterator(size()) {
        protected Map.Entry<R, V> get(final int rowIndex) {
          new AbstractMapEntry() {
            public R getKey() {
              return rowList.get(rowIndex);
            }
            
            public V getValue() { return array[rowIndex][columnIndex]; }
            
            public V setValue(V value) {
              return set(rowIndex, columnIndex, value);
            }
          };
        }
      };
    }
    
    public int size() {
      return rowList.size();
    }
  }
  






  public ImmutableSet<C> columnKeySet()
  {
    return columnKeyToIndex.keySet();
  }
  


  public Map<C, Map<R, V>> columnMap()
  {
    ArrayTable<R, C, V>.ColumnMap map = columnMap;
    return map == null ? (this.columnMap = new ColumnMap(null)) : map;
  }
  
  private class ColumnMap extends AbstractMap<C, Map<R, V>> { transient ArrayTable<R, C, V>.ColumnMapEntrySet entrySet;
    
    private ColumnMap() {}
    
    public Set<Map.Entry<C, Map<R, V>>> entrySet() { ArrayTable<R, C, V>.ColumnMapEntrySet set = entrySet;
      return set == null ? (this.entrySet = new ArrayTable.ColumnMapEntrySet(ArrayTable.this, null)) : set;
    }
    
    public Map<R, V> get(Object columnKey) {
      Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
      return columnIndex == null ? null : new ArrayTable.Column(ArrayTable.this, columnIndex.intValue());
    }
    
    public boolean containsKey(Object columnKey) {
      return containsColumn(columnKey);
    }
    
    public Set<C> keySet() {
      return columnKeySet();
    }
    

    public Map<R, V> remove(Object columnKey) { throw new UnsupportedOperationException(); }
  }
  
  private class ColumnMapEntrySet extends AbstractSet<Map.Entry<C, Map<R, V>>> {
    private ColumnMapEntrySet() {}
    
    public Iterator<Map.Entry<C, Map<R, V>>> iterator() { new AbstractIndexedListIterator(size()) {
        protected Map.Entry<C, Map<R, V>> get(int index) {
          return Maps.immutableEntry(columnList.get(index), new ArrayTable.Column(ArrayTable.this, index));
        }
      }; }
    

    public int size()
    {
      return columnList.size();
    }
  }
  













  public Map<C, V> row(R rowKey)
  {
    Preconditions.checkNotNull(rowKey);
    Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
    return rowIndex == null ? ImmutableMap.of() : new Row(rowIndex.intValue());
  }
  
  private class Row extends AbstractMap<C, V> {
    final int rowIndex;
    ArrayTable<R, C, V>.RowEntrySet entrySet;
    
    Row(int rowIndex) { this.rowIndex = rowIndex; }
    


    public Set<Map.Entry<C, V>> entrySet()
    {
      ArrayTable<R, C, V>.RowEntrySet set = entrySet;
      return set == null ? (this.entrySet = new ArrayTable.RowEntrySet(ArrayTable.this, rowIndex)) : set;
    }
    
    public V get(Object columnKey) {
      Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
      return ArrayTable.this.getIndexed(Integer.valueOf(rowIndex), columnIndex);
    }
    
    public boolean containsKey(Object columnKey) {
      return containsColumn(columnKey);
    }
    
    public V put(C columnKey, V value) {
      Preconditions.checkNotNull(columnKey);
      Integer columnIndex = (Integer)columnKeyToIndex.get(columnKey);
      Preconditions.checkArgument(columnIndex != null, "Column %s not in %s", new Object[] { columnKey, columnList });
      
      return set(rowIndex, columnIndex.intValue(), value);
    }
    
    public Set<C> keySet() {
      return columnKeySet();
    }
  }
  
  private class RowEntrySet extends AbstractSet<Map.Entry<C, V>> {
    final int rowIndex;
    
    RowEntrySet(int rowIndex) {
      this.rowIndex = rowIndex;
    }
    
    public Iterator<Map.Entry<C, V>> iterator() {
      new AbstractIndexedListIterator(size()) {
        protected Map.Entry<C, V> get(final int columnIndex) {
          new AbstractMapEntry() {
            public C getKey() {
              return columnList.get(columnIndex);
            }
            
            public V getValue() { return array[rowIndex][columnIndex]; }
            
            public V setValue(V value) {
              return set(rowIndex, columnIndex, value);
            }
          };
        }
      };
    }
    
    public int size() {
      return columnList.size();
    }
  }
  






  public ImmutableSet<R> rowKeySet()
  {
    return rowKeyToIndex.keySet();
  }
  


  public Map<R, Map<C, V>> rowMap()
  {
    ArrayTable<R, C, V>.RowMap map = rowMap;
    return map == null ? (this.rowMap = new RowMap(null)) : map;
  }
  
  private class RowMap extends AbstractMap<R, Map<C, V>> { transient ArrayTable<R, C, V>.RowMapEntrySet entrySet;
    
    private RowMap() {}
    
    public Set<Map.Entry<R, Map<C, V>>> entrySet() { ArrayTable<R, C, V>.RowMapEntrySet set = entrySet;
      return set == null ? (this.entrySet = new ArrayTable.RowMapEntrySet(ArrayTable.this, null)) : set;
    }
    
    public Map<C, V> get(Object rowKey) {
      Integer rowIndex = (Integer)rowKeyToIndex.get(rowKey);
      return rowIndex == null ? null : new ArrayTable.Row(ArrayTable.this, rowIndex.intValue());
    }
    
    public boolean containsKey(Object rowKey) {
      return containsRow(rowKey);
    }
    
    public Set<R> keySet() {
      return rowKeySet();
    }
    

    public Map<C, V> remove(Object rowKey) { throw new UnsupportedOperationException(); }
  }
  
  private class RowMapEntrySet extends AbstractSet<Map.Entry<R, Map<C, V>>> {
    private RowMapEntrySet() {}
    
    public Iterator<Map.Entry<R, Map<C, V>>> iterator() { new AbstractIndexedListIterator(size()) {
        protected Map.Entry<R, Map<C, V>> get(int index) {
          return Maps.immutableEntry(rowList.get(index), new ArrayTable.Row(ArrayTable.this, index));
        }
      }; }
    

    public int size()
    {
      return rowList.size();
    }
  }
  











  public Collection<V> values()
  {
    Collection<V> v = values;
    return v == null ? (this.values = new Values(null)) : v;
  }
  
  private class Values extends AbstractCollection<V> { private Values() {}
    
    public Iterator<V> iterator() { new AbstractIndexedListIterator(size()) {
        protected V get(int index) {
          int rowIndex = index / columnList.size();
          int columnIndex = index % columnList.size();
          return array[rowIndex][columnIndex];
        }
      }; }
    
    public int size()
    {
      return ArrayTable.this.size();
    }
    
    public boolean contains(Object value) {
      return containsValue(value);
    }
  }
}
