package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import java.util.Comparator;
import javax.annotation.Nullable;





























@GwtCompatible
final class BstOperations
{
  @Nullable
  public static <K, N extends BstNode<K, N>> N seek(Comparator<? super K> comparator, @Nullable N tree, K key)
  {
    Preconditions.checkNotNull(comparator);
    if (tree == null) {
      return null;
    }
    int cmp = comparator.compare(key, tree.getKey());
    if (cmp == 0) {
      return tree;
    }
    BstSide side = cmp < 0 ? BstSide.LEFT : BstSide.RIGHT;
    return seek(comparator, tree.childOrNull(side), key);
  }
  















  public static <K, N extends BstNode<K, N>> BstMutationResult<K, N> mutate(Comparator<? super K> comparator, BstMutationRule<K, N> mutationRule, @Nullable N tree, K key)
  {
    Preconditions.checkNotNull(comparator);
    Preconditions.checkNotNull(mutationRule);
    Preconditions.checkNotNull(key);
    
    if (tree != null) {
      int cmp = comparator.compare(key, tree.getKey());
      if (cmp != 0) {
        BstSide side = cmp < 0 ? BstSide.LEFT : BstSide.RIGHT;
        BstMutationResult<K, N> mutation = mutate(comparator, mutationRule, tree.childOrNull(side), key);
        
        return mutation.lift(tree, side, mutationRule.getNodeFactory(), mutationRule.getBalancePolicy());
      }
    }
    
    return modify(tree, key, mutationRule);
  }
  



  public static <K, N extends BstNode<K, N>> BstMutationResult<K, N> mutate(BstInOrderPath<N> path, BstMutationRule<K, N> mutationRule)
  {
    Preconditions.checkNotNull(path);
    Preconditions.checkNotNull(mutationRule);
    BstBalancePolicy<N> balancePolicy = mutationRule.getBalancePolicy();
    BstNodeFactory<N> nodeFactory = mutationRule.getNodeFactory();
    BstModifier<K, N> modifier = mutationRule.getModifier();
    
    N target = path.getTip();
    K key = target.getKey();
    BstMutationResult<K, N> result = modify(target, key, mutationRule);
    while (path.hasPrefix()) {
      BstInOrderPath<N> prefix = (BstInOrderPath)path.getPrefix();
      result = result.lift(prefix.getTip(), path.getSideOfExtension(), nodeFactory, balancePolicy);
      path = prefix;
    }
    return result;
  }
  



  private static <K, N extends BstNode<K, N>> BstMutationResult<K, N> modify(@Nullable N tree, K key, BstMutationRule<K, N> mutationRule)
  {
    BstBalancePolicy<N> rebalancePolicy = mutationRule.getBalancePolicy();
    BstNodeFactory<N> nodeFactory = mutationRule.getNodeFactory();
    BstModifier<K, N> modifier = mutationRule.getModifier();
    
    N originalRoot = tree;
    
    N originalTarget = tree == null ? null : nodeFactory.createLeaf(tree);
    BstModificationResult<N> modResult = modifier.modify(key, originalTarget);
    N originalLeft = null;
    N originalRight = null;
    if (tree != null) {
      originalLeft = tree.childOrNull(BstSide.LEFT);
      originalRight = tree.childOrNull(BstSide.RIGHT); }
    N changedRoot;
    N changedRoot; switch (1.$SwitchMap$com$google$common$collect$BstModificationResult$ModificationType[modResult.getType().ordinal()]) {
    case 1: 
      changedRoot = tree;
      break;
    case 2: 
      if (modResult.getChangedTarget() != null) {
        changedRoot = nodeFactory.createNode(modResult.getChangedTarget(), originalLeft, originalRight);
      } else { N changedRoot;
        if (tree == null) {
          changedRoot = null;
        } else
          throw new AssertionError("Modification result is a REBUILDING_CHANGE, but rebalancing required");
      }
      break;
    case 3: 
      N changedRoot;
      if (modResult.getChangedTarget() != null) {
        changedRoot = rebalancePolicy.balance(nodeFactory, modResult.getChangedTarget(), originalLeft, originalRight);
      } else { N changedRoot;
        if (tree != null) {
          changedRoot = rebalancePolicy.combine(nodeFactory, originalLeft, originalRight);
        } else
          changedRoot = null;
      }
      break;
    default: 
      throw new AssertionError();
    }
    return BstMutationResult.mutationResult(key, originalRoot, changedRoot, modResult);
  }
  



  public static <K, N extends BstNode<K, N>> BstMutationResult<K, N> extractMin(N root, BstNodeFactory<N> nodeFactory, BstBalancePolicy<N> balancePolicy)
  {
    Preconditions.checkNotNull(root);
    Preconditions.checkNotNull(nodeFactory);
    Preconditions.checkNotNull(balancePolicy);
    if (root.hasChild(BstSide.LEFT)) {
      BstMutationResult<K, N> subResult = extractMin(root.getChild(BstSide.LEFT), nodeFactory, balancePolicy);
      
      return subResult.lift(root, BstSide.LEFT, nodeFactory, balancePolicy);
    }
    return BstMutationResult.mutationResult(root.getKey(), root, root.childOrNull(BstSide.RIGHT), BstModificationResult.rebalancingChange(root, null));
  }
  





  public static <K, N extends BstNode<K, N>> BstMutationResult<K, N> extractMax(N root, BstNodeFactory<N> nodeFactory, BstBalancePolicy<N> balancePolicy)
  {
    Preconditions.checkNotNull(root);
    Preconditions.checkNotNull(nodeFactory);
    Preconditions.checkNotNull(balancePolicy);
    if (root.hasChild(BstSide.RIGHT)) {
      BstMutationResult<K, N> subResult = extractMax(root.getChild(BstSide.RIGHT), nodeFactory, balancePolicy);
      
      return subResult.lift(root, BstSide.RIGHT, nodeFactory, balancePolicy);
    }
    return BstMutationResult.mutationResult(root.getKey(), root, root.childOrNull(BstSide.LEFT), BstModificationResult.rebalancingChange(root, null));
  }
  





  public static <N extends BstNode<?, N>> N insertMin(@Nullable N root, N entry, BstNodeFactory<N> nodeFactory, BstBalancePolicy<N> balancePolicy)
  {
    Preconditions.checkNotNull(entry);
    Preconditions.checkNotNull(nodeFactory);
    Preconditions.checkNotNull(balancePolicy);
    if (root == null) {
      return nodeFactory.createLeaf(entry);
    }
    return balancePolicy.balance(nodeFactory, root, insertMin(root.childOrNull(BstSide.LEFT), entry, nodeFactory, balancePolicy), root.childOrNull(BstSide.RIGHT));
  }
  







  public static <N extends BstNode<?, N>> N insertMax(@Nullable N root, N entry, BstNodeFactory<N> nodeFactory, BstBalancePolicy<N> balancePolicy)
  {
    Preconditions.checkNotNull(entry);
    Preconditions.checkNotNull(nodeFactory);
    Preconditions.checkNotNull(balancePolicy);
    if (root == null) {
      return nodeFactory.createLeaf(entry);
    }
    return balancePolicy.balance(nodeFactory, root, root.childOrNull(BstSide.LEFT), insertMax(root.childOrNull(BstSide.RIGHT), entry, nodeFactory, balancePolicy));
  }
}
