package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Function;
import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import com.google.common.base.Predicates;
import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;






























































@GwtCompatible
public final class Collections2
{
  public static <E> Collection<E> filter(Collection<E> unfiltered, Predicate<? super E> predicate)
  {
    if ((unfiltered instanceof FilteredCollection))
    {

      return ((FilteredCollection)unfiltered).createCombined(predicate);
    }
    
    return new FilteredCollection((Collection)Preconditions.checkNotNull(unfiltered), (Predicate)Preconditions.checkNotNull(predicate));
  }
  



  static boolean safeContains(Collection<?> collection, Object object)
  {
    try
    {
      return collection.contains(object);
    } catch (ClassCastException e) {}
    return false;
  }
  
  static class FilteredCollection<E> implements Collection<E>
  {
    final Collection<E> unfiltered;
    final Predicate<? super E> predicate;
    
    FilteredCollection(Collection<E> unfiltered, Predicate<? super E> predicate)
    {
      this.unfiltered = unfiltered;
      this.predicate = predicate;
    }
    
    FilteredCollection<E> createCombined(Predicate<? super E> newPredicate) {
      return new FilteredCollection(unfiltered, Predicates.and(predicate, newPredicate));
    }
    


    public boolean add(E element)
    {
      Preconditions.checkArgument(predicate.apply(element));
      return unfiltered.add(element);
    }
    
    public boolean addAll(Collection<? extends E> collection)
    {
      for (E element : collection) {
        Preconditions.checkArgument(predicate.apply(element));
      }
      return unfiltered.addAll(collection);
    }
    
    public void clear()
    {
      Iterables.removeIf(unfiltered, predicate);
    }
    


    public boolean contains(Object element)
    {
      try
      {
        E e = element;
        





        return (predicate.apply(e)) && (unfiltered.contains(element));
      } catch (NullPointerException e) {
        return false;
      } catch (ClassCastException e) {}
      return false;
    }
    

    public boolean containsAll(Collection<?> collection)
    {
      for (Object element : collection) {
        if (!contains(element)) {
          return false;
        }
      }
      return true;
    }
    
    public boolean isEmpty()
    {
      return !Iterators.any(unfiltered.iterator(), predicate);
    }
    
    public Iterator<E> iterator()
    {
      return Iterators.filter(unfiltered.iterator(), predicate);
    }
    


    public boolean remove(Object element)
    {
      try
      {
        E e = element;
        

        return (predicate.apply(e)) && (unfiltered.remove(element));
      } catch (NullPointerException e) {
        return false;
      } catch (ClassCastException e) {}
      return false;
    }
    

    public boolean removeAll(final Collection<?> collection)
    {
      Preconditions.checkNotNull(collection);
      Predicate<E> combinedPredicate = new Predicate()
      {
        public boolean apply(E input) {
          return (predicate.apply(input)) && (collection.contains(input));
        }
      };
      return Iterables.removeIf(unfiltered, combinedPredicate);
    }
    
    public boolean retainAll(final Collection<?> collection)
    {
      Preconditions.checkNotNull(collection);
      Predicate<E> combinedPredicate = new Predicate()
      {
        public boolean apply(E input)
        {
          return (predicate.apply(input)) && (!collection.contains(input));
        }
      };
      return Iterables.removeIf(unfiltered, combinedPredicate);
    }
    
    public int size()
    {
      return Iterators.size(iterator());
    }
    

    public Object[] toArray()
    {
      return Lists.newArrayList(iterator()).toArray();
    }
    
    public <T> T[] toArray(T[] array)
    {
      return Lists.newArrayList(iterator()).toArray(array);
    }
    
    public String toString() {
      return Iterators.toString(iterator());
    }
  }
  



















  public static <F, T> Collection<T> transform(Collection<F> fromCollection, Function<? super F, T> function)
  {
    return new TransformedCollection(fromCollection, function);
  }
  
  static class TransformedCollection<F, T> extends AbstractCollection<T>
  {
    final Collection<F> fromCollection;
    final Function<? super F, ? extends T> function;
    
    TransformedCollection(Collection<F> fromCollection, Function<? super F, ? extends T> function) {
      this.fromCollection = ((Collection)Preconditions.checkNotNull(fromCollection));
      this.function = ((Function)Preconditions.checkNotNull(function));
    }
    
    public void clear() {
      fromCollection.clear();
    }
    
    public boolean isEmpty() {
      return fromCollection.isEmpty();
    }
    
    public Iterator<T> iterator() {
      return Iterators.transform(fromCollection.iterator(), function);
    }
    
    public int size() {
      return fromCollection.size();
    }
  }
  











  static boolean containsAllImpl(Collection<?> self, Collection<?> c)
  {
    Preconditions.checkNotNull(self);
    for (Object o : c) {
      if (!self.contains(o)) {
        return false;
      }
    }
    return true;
  }
  


  static String toStringImpl(Collection<?> collection)
  {
    StringBuilder sb = newStringBuilderForCollection(collection.size()).append('[');
    
    STANDARD_JOINER.appendTo(sb, Iterables.transform(collection, new Function()
    {
      public Object apply(Object input) {
        return input == val$collection ? "(this Collection)" : input;
      }
    }));
    return ']';
  }
  


  static StringBuilder newStringBuilderForCollection(int size)
  {
    Preconditions.checkArgument(size >= 0, "size must be non-negative");
    return new StringBuilder((int)Math.min(size * 8L, 1073741824L));
  }
  


  static <T> Collection<T> cast(Iterable<T> iterable)
  {
    return (Collection)iterable;
  }
  
  static final Joiner STANDARD_JOINER = Joiner.on(", ");
}
