package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Objects;
import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import java.io.Serializable;
import java.util.Comparator;
import javax.annotation.Nullable;























@GwtCompatible(serializable=true)
final class GeneralRange<T>
  implements Serializable
{
  private final Comparator<? super T> comparator;
  private final Optional<T> lowerEndpoint;
  private final BoundType lowerBoundType;
  private final Optional<T> upperEndpoint;
  private final BoundType upperBoundType;
  private transient GeneralRange<T> reverse;
  
  static <T extends Comparable> GeneralRange<T> from(Range<T> range)
  {
    Optional<T> lowerEndpoint = range.hasLowerBound() ? Optional.of(range.lowerEndpoint()) : Optional.absent();
    
    BoundType lowerBoundType = range.hasLowerBound() ? range.lowerBoundType() : BoundType.OPEN;
    Optional<T> upperEndpoint = range.hasUpperBound() ? Optional.of(range.upperEndpoint()) : Optional.absent();
    
    BoundType upperBoundType = range.hasUpperBound() ? range.upperBoundType() : BoundType.OPEN;
    return new GeneralRange(Ordering.natural(), lowerEndpoint, lowerBoundType, upperEndpoint, upperBoundType);
  }
  



  static <T> GeneralRange<T> all(Comparator<? super T> comparator)
  {
    return new GeneralRange(comparator, Optional.absent(), BoundType.OPEN, Optional.absent(), BoundType.OPEN);
  }
  




  static <T> GeneralRange<T> downTo(Comparator<? super T> comparator, T endpoint, BoundType boundType)
  {
    return new GeneralRange(comparator, Optional.of(endpoint), boundType, Optional.absent(), BoundType.OPEN);
  }
  





  static <T> GeneralRange<T> upTo(Comparator<? super T> comparator, T endpoint, BoundType boundType)
  {
    return new GeneralRange(comparator, Optional.absent(), BoundType.OPEN, Optional.of(endpoint), boundType);
  }
  





  static <T> GeneralRange<T> range(Comparator<? super T> comparator, T lower, BoundType lowerType, T upper, BoundType upperType)
  {
    return new GeneralRange(comparator, Optional.of(lower), lowerType, Optional.of(upper), upperType);
  }
  







  private GeneralRange(Comparator<? super T> comparator, Optional<T> lowerEndpoint, BoundType lowerBoundType, Optional<T> upperEndpoint, BoundType upperBoundType)
  {
    this.comparator = ((Comparator)Preconditions.checkNotNull(comparator));
    this.lowerEndpoint = ((Optional)Preconditions.checkNotNull(lowerEndpoint));
    this.lowerBoundType = ((BoundType)Preconditions.checkNotNull(lowerBoundType));
    this.upperEndpoint = ((Optional)Preconditions.checkNotNull(upperEndpoint));
    this.upperBoundType = ((BoundType)Preconditions.checkNotNull(upperBoundType));
    if ((lowerEndpoint.isPresent()) && (upperEndpoint.isPresent())) {
      int cmp = comparator.compare(lowerEndpoint.get(), upperEndpoint.get());
      
      Preconditions.checkArgument(cmp <= 0, "lowerEndpoint (%s) > upperEndpoint (%s)", new Object[] { lowerEndpoint, upperEndpoint });
      
      if (cmp == 0) {
        Preconditions.checkArgument((lowerBoundType != BoundType.OPEN ? 1 : 0) | (upperBoundType != BoundType.OPEN ? 1 : 0));
      }
    }
  }
  
  Comparator<? super T> comparator() {
    return comparator;
  }
  
  boolean hasLowerBound() {
    return lowerEndpoint.isPresent();
  }
  
  boolean hasUpperBound() {
    return upperEndpoint.isPresent();
  }
  
  boolean isEmpty() {
    return ((hasUpperBound()) && (tooLow(upperEndpoint.get()))) || ((hasLowerBound()) && (tooHigh(lowerEndpoint.get())));
  }
  
  boolean tooLow(T t)
  {
    if (!hasLowerBound()) {
      return false;
    }
    T lbound = lowerEndpoint.get();
    int cmp = comparator.compare(t, lbound);
    return (cmp < 0 ? 1 : 0) | (cmp == 0 ? 1 : 0) & (lowerBoundType == BoundType.OPEN ? 1 : 0);
  }
  
  boolean tooHigh(T t) {
    if (!hasUpperBound()) {
      return false;
    }
    T ubound = upperEndpoint.get();
    int cmp = comparator.compare(t, ubound);
    return (cmp > 0 ? 1 : 0) | (cmp == 0 ? 1 : 0) & (upperBoundType == BoundType.OPEN ? 1 : 0);
  }
  
  boolean contains(T t) {
    Preconditions.checkNotNull(t);
    return (!tooLow(t)) && (!tooHigh(t));
  }
  


  GeneralRange<T> intersect(GeneralRange<T> other)
  {
    Preconditions.checkNotNull(other);
    Preconditions.checkArgument(comparator.equals(comparator));
    
    Optional<T> lowEnd = lowerEndpoint;
    BoundType lowType = lowerBoundType;
    if (!hasLowerBound()) {
      lowEnd = lowerEndpoint;
      lowType = lowerBoundType;
    } else if (other.hasLowerBound()) {
      int cmp = comparator.compare(lowerEndpoint.get(), lowerEndpoint.get());
      if ((cmp < 0) || ((cmp == 0) && (lowerBoundType == BoundType.OPEN))) {
        lowEnd = lowerEndpoint;
        lowType = lowerBoundType;
      }
    }
    
    Optional<T> upEnd = upperEndpoint;
    BoundType upType = upperBoundType;
    if (!hasUpperBound()) {
      upEnd = upperEndpoint;
      upType = upperBoundType;
    } else if (other.hasUpperBound()) {
      int cmp = comparator.compare(upperEndpoint.get(), upperEndpoint.get());
      if ((cmp > 0) || ((cmp == 0) && (upperBoundType == BoundType.OPEN))) {
        upEnd = upperEndpoint;
        upType = upperBoundType;
      }
    }
    
    if ((lowEnd.isPresent()) && (upEnd.isPresent())) {
      int cmp = comparator.compare(lowEnd.get(), upEnd.get());
      if ((cmp > 0) || ((cmp == 0) && (lowType == BoundType.OPEN) && (upType == BoundType.OPEN)))
      {
        lowEnd = upEnd;
        lowType = BoundType.OPEN;
        upType = BoundType.CLOSED;
      }
    }
    
    return new GeneralRange(comparator, lowEnd, lowType, upEnd, upType);
  }
  
  public boolean equals(@Nullable Object obj)
  {
    if ((obj instanceof GeneralRange)) {
      GeneralRange<?> r = (GeneralRange)obj;
      return (comparator.equals(comparator)) && (lowerEndpoint.equals(lowerEndpoint)) && (lowerBoundType.equals(lowerBoundType)) && (upperEndpoint.equals(upperEndpoint)) && (upperBoundType.equals(upperBoundType));
    }
    

    return false;
  }
  
  public int hashCode()
  {
    return Objects.hashCode(new Object[] { comparator, lowerEndpoint, lowerBoundType, upperEndpoint, upperBoundType });
  }
  





  public GeneralRange<T> reverse()
  {
    GeneralRange<T> result = reverse;
    if (result == null) {
      result = new GeneralRange(Ordering.from(comparator).reverse(), upperEndpoint, upperBoundType, lowerEndpoint, lowerBoundType);
      

      reverse = this;
      return this.reverse = result;
    }
    return result;
  }
  
  public String toString()
  {
    StringBuilder builder = new StringBuilder();
    builder.append(comparator).append(":");
    switch (1.$SwitchMap$com$google$common$collect$BoundType[lowerBoundType.ordinal()]) {
    case 1: 
      builder.append('[');
      break;
    case 2: 
      builder.append('(');
    }
    
    if (hasLowerBound()) {
      builder.append(lowerEndpoint.get());
    } else {
      builder.append("-∞");
    }
    builder.append(',');
    if (hasUpperBound()) {
      builder.append(upperEndpoint.get());
    } else {
      builder.append("∞");
    }
    switch (1.$SwitchMap$com$google$common$collect$BoundType[upperBoundType.ordinal()]) {
    case 1: 
      builder.append(']');
      break;
    case 2: 
      builder.append(')');
    }
    
    return builder.toString();
  }
}
