package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import javax.annotation.Nullable;
































@GwtCompatible(emulated=true)
public abstract class ImmutableCollection<E>
  implements Collection<E>, Serializable
{
  static final ImmutableCollection<Object> EMPTY_IMMUTABLE_COLLECTION = new EmptyImmutableCollection(null);
  

  private transient ImmutableList<E> asList;
  


  public abstract UnmodifiableIterator<E> iterator();
  


  public Object[] toArray()
  {
    return ObjectArrays.toArrayImpl(this);
  }
  
  public <T> T[] toArray(T[] other)
  {
    return ObjectArrays.toArrayImpl(this, other);
  }
  
  public boolean contains(@Nullable Object object)
  {
    return (object != null) && (Iterators.contains(iterator(), object));
  }
  
  public boolean containsAll(Collection<?> targets)
  {
    return Collections2.containsAllImpl(this, targets);
  }
  
  public boolean isEmpty()
  {
    return size() == 0;
  }
  
  public String toString() {
    return Collections2.toStringImpl(this);
  }
  





  public final boolean add(E e)
  {
    throw new UnsupportedOperationException();
  }
  





  public final boolean remove(Object object)
  {
    throw new UnsupportedOperationException();
  }
  





  public final boolean addAll(Collection<? extends E> newElements)
  {
    throw new UnsupportedOperationException();
  }
  





  public final boolean removeAll(Collection<?> oldElements)
  {
    throw new UnsupportedOperationException();
  }
  





  public final boolean retainAll(Collection<?> elementsToKeep)
  {
    throw new UnsupportedOperationException();
  }
  





  public final void clear()
  {
    throw new UnsupportedOperationException();
  }
  










  public ImmutableList<E> asList()
  {
    ImmutableList<E> list = asList;
    return list == null ? (this.asList = createAsList()) : list;
  }
  
  ImmutableList<E> createAsList() {
    switch (size()) {
    case 0: 
      return ImmutableList.of();
    case 1: 
      return ImmutableList.of(iterator().next());
    }
    return new ImmutableAsList(toArray(), this);
  }
  
  abstract boolean isPartialView();
  
  private static class EmptyImmutableCollection
    extends ImmutableCollection<Object>
  {
    public int size()
    {
      return 0;
    }
    
    public boolean isEmpty() {
      return true;
    }
    
    public boolean contains(@Nullable Object object) {
      return false;
    }
    
    public UnmodifiableIterator<Object> iterator() {
      return Iterators.EMPTY_ITERATOR;
    }
    
    private static final Object[] EMPTY_ARRAY = new Object[0];
    
    public Object[] toArray() {
      return EMPTY_ARRAY;
    }
    
    public <T> T[] toArray(T[] array) {
      if (array.length > 0) {
        array[0] = null;
      }
      return array;
    }
    
    ImmutableList<Object> createAsList() {
      return ImmutableList.of();
    }
    
    boolean isPartialView() {
      return false;
    }
  }
  

  private static class ArrayImmutableCollection<E>
    extends ImmutableCollection<E>
  {
    private final E[] elements;
    
    ArrayImmutableCollection(E[] elements)
    {
      this.elements = elements;
    }
    
    public int size()
    {
      return elements.length;
    }
    
    public boolean isEmpty() {
      return false;
    }
    
    public UnmodifiableIterator<E> iterator() {
      return Iterators.forArray(elements);
    }
    
    ImmutableList<E> createAsList() {
      return elements.length == 1 ? new SingletonImmutableList(elements[0]) : new RegularImmutableList(elements);
    }
    
    boolean isPartialView()
    {
      return false;
    }
  }
  
  private static class SerializedForm implements Serializable
  {
    final Object[] elements;
    private static final long serialVersionUID = 0L;
    
    SerializedForm(Object[] elements)
    {
      this.elements = elements;
    }
    
    Object readResolve() { return elements.length == 0 ? ImmutableCollection.EMPTY_IMMUTABLE_COLLECTION : new ImmutableCollection.ArrayImmutableCollection(Platform.clone(elements)); }
  }
  



  Object writeReplace()
  {
    return new SerializedForm(toArray());
  }
  















  public static abstract class Builder<E>
  {
    public abstract Builder<E> add(E paramE);
    














    public Builder<E> add(E... elements)
    {
      for (E element : elements) {
        add(element);
      }
      return this;
    }
    











    public Builder<E> addAll(Iterable<? extends E> elements)
    {
      for (E element : elements) {
        add(element);
      }
      return this;
    }
    











    public Builder<E> addAll(Iterator<? extends E> elements)
    {
      while (elements.hasNext()) {
        add(elements.next());
      }
      return this;
    }
    
    public abstract ImmutableCollection<E> build();
  }
}
