package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import javax.annotation.Nullable;






































































@GwtCompatible(serializable=true, emulated=true)
public abstract class ImmutableSortedSet<E>
  extends ImmutableSortedSetFauxverideShim<E>
  implements SortedSet<E>, SortedIterable<E>
{
  private static final Comparator<Comparable> NATURAL_ORDER = ;
  

  private static final ImmutableSortedSet<Comparable> NATURAL_EMPTY_SET = new EmptyImmutableSortedSet(NATURAL_ORDER);
  final transient Comparator<? super E> comparator;
  
  private static <E> ImmutableSortedSet<E> emptySet()
  {
    return NATURAL_EMPTY_SET;
  }
  
  static <E> ImmutableSortedSet<E> emptySet(Comparator<? super E> comparator)
  {
    if (NATURAL_ORDER.equals(comparator)) {
      return emptySet();
    }
    return new EmptyImmutableSortedSet(comparator);
  }
  



  public static <E> ImmutableSortedSet<E> of()
  {
    return emptySet();
  }
  



  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E element)
  {
    return new RegularImmutableSortedSet(ImmutableList.of(element), Ordering.natural());
  }
  









  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E e1, E e2)
  {
    return copyOf(Ordering.natural(), Arrays.asList(new Comparable[] { e1, e2 }));
  }
  








  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E e1, E e2, E e3)
  {
    return copyOf(Ordering.natural(), Arrays.asList(new Comparable[] { e1, e2, e3 }));
  }
  








  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E e1, E e2, E e3, E e4)
  {
    return copyOf(Ordering.natural(), Arrays.asList(new Comparable[] { e1, e2, e3, e4 }));
  }
  








  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E e1, E e2, E e3, E e4, E e5)
  {
    return copyOf(Ordering.natural(), Arrays.asList(new Comparable[] { e1, e2, e3, e4, e5 }));
  }
  









  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E e1, E e2, E e3, E e4, E e5, E e6, E... remaining)
  {
    int size = remaining.length + 6;
    List<E> all = new ArrayList(size);
    Collections.addAll(all, new Comparable[] { e1, e2, e3, e4, e5, e6 });
    Collections.addAll(all, remaining);
    return copyOf(Ordering.natural(), all);
  }
  













  @Deprecated
  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> of(E[] elements)
  {
    return copyOf(elements);
  }
  








  public static <E extends Comparable<? super E>> ImmutableSortedSet<E> copyOf(E[] elements)
  {
    return copyOf(Ordering.natural(), Arrays.asList(elements));
  }
  





























  public static <E> ImmutableSortedSet<E> copyOf(Iterable<? extends E> elements)
  {
    Ordering<E> naturalOrder = Ordering.natural();
    return copyOf(naturalOrder, elements);
  }
  
































  public static <E> ImmutableSortedSet<E> copyOf(Collection<? extends E> elements)
  {
    Ordering<E> naturalOrder = Ordering.natural();
    return copyOf(naturalOrder, elements);
  }
  














  public static <E> ImmutableSortedSet<E> copyOf(Iterator<? extends E> elements)
  {
    Ordering<E> naturalOrder = Ordering.natural();
    return copyOfInternal(naturalOrder, elements);
  }
  









  public static <E> ImmutableSortedSet<E> copyOf(Comparator<? super E> comparator, Iterator<? extends E> elements)
  {
    Preconditions.checkNotNull(comparator);
    return copyOfInternal(comparator, elements);
  }
  













  public static <E> ImmutableSortedSet<E> copyOf(Comparator<? super E> comparator, Iterable<? extends E> elements)
  {
    Preconditions.checkNotNull(comparator);
    return copyOfInternal(comparator, elements);
  }
  


















  public static <E> ImmutableSortedSet<E> copyOf(Comparator<? super E> comparator, Collection<? extends E> elements)
  {
    Preconditions.checkNotNull(comparator);
    return copyOfInternal(comparator, elements);
  }
  

















  public static <E> ImmutableSortedSet<E> copyOfSorted(SortedSet<E> sortedSet)
  {
    Comparator<? super E> comparator = sortedSet.comparator();
    if (comparator == null) {
      comparator = NATURAL_ORDER;
    }
    return copyOfInternal(comparator, sortedSet);
  }
  
  private static <E> ImmutableSortedSet<E> copyOfInternal(Comparator<? super E> comparator, Iterable<? extends E> elements)
  {
    boolean hasSameComparator = SortedIterables.hasSameComparator(comparator, elements);
    

    if ((hasSameComparator) && ((elements instanceof ImmutableSortedSet)))
    {
      ImmutableSortedSet<E> original = (ImmutableSortedSet)elements;
      if (!original.isPartialView()) {
        return original;
      }
    }
    ImmutableList<E> list = ImmutableList.copyOf(SortedIterables.sortedUnique(comparator, elements));
    
    return list.isEmpty() ? emptySet(comparator) : new RegularImmutableSortedSet(list, comparator);
  }
  


  private static <E> ImmutableSortedSet<E> copyOfInternal(Comparator<? super E> comparator, Iterator<? extends E> elements)
  {
    ImmutableList<E> list = ImmutableList.copyOf(SortedIterables.sortedUnique(comparator, elements));
    
    return list.isEmpty() ? emptySet(comparator) : new RegularImmutableSortedSet(list, comparator);
  }
  









  public static <E> Builder<E> orderedBy(Comparator<E> comparator)
  {
    return new Builder(comparator);
  }
  








  public static <E extends Comparable<E>> Builder<E> reverseOrder()
  {
    return new Builder(Ordering.natural().reverse());
  }
  











  public static <E extends Comparable<E>> Builder<E> naturalOrder()
  {
    return new Builder(Ordering.natural());
  }
  









  public static final class Builder<E>
    extends ImmutableSet.Builder<E>
  {
    private final Comparator<? super E> comparator;
    









    public Builder(Comparator<? super E> comparator)
    {
      this.comparator = ((Comparator)Preconditions.checkNotNull(comparator));
    }
    









    public Builder<E> add(E element)
    {
      super.add(element);
      return this;
    }
    







    public Builder<E> add(E... elements)
    {
      super.add(elements);
      return this;
    }
    







    public Builder<E> addAll(Iterable<? extends E> elements)
    {
      super.addAll(elements);
      return this;
    }
    







    public Builder<E> addAll(Iterator<? extends E> elements)
    {
      super.addAll(elements);
      return this;
    }
    



    public ImmutableSortedSet<E> build()
    {
      return ImmutableSortedSet.copyOfInternal(comparator, contents.iterator());
    }
  }
  
  int unsafeCompare(Object a, Object b) {
    return unsafeCompare(comparator, a, b);
  }
  




  static int unsafeCompare(Comparator<?> comparator, Object a, Object b)
  {
    Comparator<Object> unsafeComparator = comparator;
    return unsafeComparator.compare(a, b);
  }
  

  ImmutableSortedSet(Comparator<? super E> comparator)
  {
    this.comparator = comparator;
  }
  







  public Comparator<? super E> comparator()
  {
    return comparator;
  }
  






  public abstract UnmodifiableIterator<E> iterator();
  






  public ImmutableSortedSet<E> headSet(E toElement)
  {
    return headSet(toElement, false);
  }
  
  ImmutableSortedSet<E> headSet(E toElement, boolean inclusive) {
    return headSetImpl(Preconditions.checkNotNull(toElement), inclusive);
  }
  













  public ImmutableSortedSet<E> subSet(E fromElement, E toElement)
  {
    return subSet(fromElement, true, toElement, false);
  }
  
  ImmutableSortedSet<E> subSet(E fromElement, boolean fromInclusive, E toElement, boolean toInclusive)
  {
    Preconditions.checkNotNull(fromElement);
    Preconditions.checkNotNull(toElement);
    Preconditions.checkArgument(comparator.compare(fromElement, toElement) <= 0);
    return subSetImpl(fromElement, fromInclusive, toElement, toInclusive);
  }
  











  public ImmutableSortedSet<E> tailSet(E fromElement)
  {
    return tailSet(fromElement, true);
  }
  
  ImmutableSortedSet<E> tailSet(E fromElement, boolean inclusive) {
    return tailSetImpl(Preconditions.checkNotNull(fromElement), inclusive);
  }
  


  abstract ImmutableSortedSet<E> headSetImpl(E paramE, boolean paramBoolean);
  


  abstract ImmutableSortedSet<E> subSetImpl(E paramE1, boolean paramBoolean1, E paramE2, boolean paramBoolean2);
  

  abstract ImmutableSortedSet<E> tailSetImpl(E paramE, boolean paramBoolean);
  

  abstract int indexOf(@Nullable Object paramObject);
  

  private static class SerializedForm<E>
    implements Serializable
  {
    final Comparator<? super E> comparator;
    
    final Object[] elements;
    
    private static final long serialVersionUID = 0L;
    

    public SerializedForm(Comparator<? super E> comparator, Object[] elements)
    {
      this.comparator = comparator;
      this.elements = elements;
    }
    
    Object readResolve()
    {
      return new ImmutableSortedSet.Builder(comparator).add((Object[])elements).build();
    }
  }
  

  private void readObject(ObjectInputStream stream)
    throws InvalidObjectException
  {
    throw new InvalidObjectException("Use SerializedForm");
  }
  
  Object writeReplace() {
    return new SerializedForm(comparator, toArray());
  }
}
