package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.annotations.GwtIncompatible;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Preconditions;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;




















































@GwtCompatible(serializable=true, emulated=true)
public final class LinkedHashMultimap<K, V>
  extends AbstractSetMultimap<K, V>
{
  private static final int DEFAULT_VALUES_PER_KEY = 8;
  @VisibleForTesting
  transient int expectedValuesPerKey = 8;
  


  transient Collection<Map.Entry<K, V>> linkedEntries;
  


  @GwtIncompatible("java serialization not supported")
  private static final long serialVersionUID = 0L;
  


  public static <K, V> LinkedHashMultimap<K, V> create()
  {
    return new LinkedHashMultimap();
  }
  









  public static <K, V> LinkedHashMultimap<K, V> create(int expectedKeys, int expectedValuesPerKey)
  {
    return new LinkedHashMultimap(expectedKeys, expectedValuesPerKey);
  }
  









  public static <K, V> LinkedHashMultimap<K, V> create(Multimap<? extends K, ? extends V> multimap)
  {
    return new LinkedHashMultimap(multimap);
  }
  
  private LinkedHashMultimap() {
    super(new LinkedHashMap());
    linkedEntries = Sets.newLinkedHashSet();
  }
  
  private LinkedHashMultimap(int expectedKeys, int expectedValuesPerKey) {
    super(new LinkedHashMap(expectedKeys));
    Preconditions.checkArgument(expectedValuesPerKey >= 0);
    this.expectedValuesPerKey = expectedValuesPerKey;
    linkedEntries = new LinkedHashSet((int)Math.min(1073741824L, expectedKeys * expectedValuesPerKey));
  }
  

  private LinkedHashMultimap(Multimap<? extends K, ? extends V> multimap)
  {
    super(new LinkedHashMap(Maps.capacity(multimap.keySet().size())));
    
    linkedEntries = new LinkedHashSet(Maps.capacity(multimap.size()));
    
    putAll(multimap);
  }
  








  Set<V> createCollection()
  {
    return new LinkedHashSet(Maps.capacity(expectedValuesPerKey));
  }
  









  Collection<V> createCollection(@Nullable K key)
  {
    return new SetDecorator(key, createCollection());
  }
  
  private class SetDecorator extends ForwardingSet<V> {
    final Set<V> delegate;
    final K key;
    
    SetDecorator(Set<V> key) {
      this.delegate = delegate;
      this.key = key;
    }
    
    protected Set<V> delegate() {
      return delegate;
    }
    
    <E> Map.Entry<K, E> createEntry(@Nullable E value) {
      return Maps.immutableEntry(key, value);
    }
    
    <E> Collection<Map.Entry<K, E>> createEntries(Collection<E> values)
    {
      Collection<Map.Entry<K, E>> entries = Lists.newArrayListWithExpectedSize(values.size());
      
      for (E value : values) {
        entries.add(createEntry(value));
      }
      return entries;
    }
    
    public boolean add(@Nullable V value) {
      boolean changed = delegate.add(value);
      if (changed) {
        linkedEntries.add(createEntry(value));
      }
      return changed;
    }
    
    public boolean addAll(Collection<? extends V> values) {
      boolean changed = delegate.addAll(values);
      if (changed) {
        linkedEntries.addAll(createEntries(delegate()));
      }
      return changed;
    }
    
    public void clear() {
      for (V value : delegate) {
        linkedEntries.remove(createEntry(value));
      }
      delegate.clear();
    }
    
    public Iterator<V> iterator() {
      final Iterator<V> delegateIterator = delegate.iterator();
      new Iterator()
      {
        V value;
        
        public boolean hasNext() {
          return delegateIterator.hasNext();
        }
        
        public V next() {
          value = delegateIterator.next();
          return value;
        }
        
        public void remove() {
          delegateIterator.remove();
          linkedEntries.remove(createEntry(value));
        }
      };
    }
    
    public boolean remove(@Nullable Object value) {
      boolean changed = delegate.remove(value);
      if (changed)
      {



        linkedEntries.remove(createEntry(value));
      }
      return changed;
    }
    
    public boolean removeAll(Collection<?> values) {
      boolean changed = delegate.removeAll(values);
      if (changed) {
        linkedEntries.removeAll(createEntries(values));
      }
      return changed;
    }
    



    public boolean retainAll(Collection<?> values)
    {
      boolean changed = false;
      Iterator<V> iterator = delegate.iterator();
      while (iterator.hasNext()) {
        V value = iterator.next();
        if (!values.contains(value)) {
          iterator.remove();
          linkedEntries.remove(Maps.immutableEntry(key, value));
          changed = true;
        }
      }
      return changed;
    }
  }
  







  Iterator<Map.Entry<K, V>> createEntryIterator()
  {
    final Iterator<Map.Entry<K, V>> delegateIterator = linkedEntries.iterator();
    
    new Iterator()
    {
      Map.Entry<K, V> entry;
      
      public boolean hasNext() {
        return delegateIterator.hasNext();
      }
      
      public Map.Entry<K, V> next()
      {
        entry = ((Map.Entry)delegateIterator.next());
        return entry;
      }
      

      public void remove()
      {
        delegateIterator.remove();
        remove(entry.getKey(), entry.getValue());
      }
    };
  }
  








  public Set<V> replaceValues(@Nullable K key, Iterable<? extends V> values)
  {
    return super.replaceValues(key, values);
  }
  











  public Set<Map.Entry<K, V>> entries()
  {
    return super.entries();
  }
  






  public Collection<V> values()
  {
    return super.values();
  }
  






  @GwtIncompatible("java.io.ObjectOutputStream")
  private void writeObject(ObjectOutputStream stream)
    throws IOException
  {
    stream.defaultWriteObject();
    stream.writeInt(expectedValuesPerKey);
    Serialization.writeMultimap(this, stream);
    for (Map.Entry<K, V> entry : linkedEntries) {
      stream.writeObject(entry.getKey());
      stream.writeObject(entry.getValue());
    }
  }
  
  @GwtIncompatible("java.io.ObjectInputStream")
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException
  {
    stream.defaultReadObject();
    expectedValuesPerKey = stream.readInt();
    int distinctKeys = Serialization.readCount(stream);
    setMap(new LinkedHashMap(Maps.capacity(distinctKeys)));
    linkedEntries = new LinkedHashSet(distinctKeys * expectedValuesPerKey);
    
    Serialization.populateMultimap(this, stream, distinctKeys);
    linkedEntries.clear();
    for (int i = 0; i < size(); i++)
    {
      K key = stream.readObject();
      
      V value = stream.readObject();
      linkedEntries.add(Maps.immutableEntry(key, value));
    }
  }
}
