package com.google.common.collect;

import com.google.common.annotations.Beta;
import com.google.common.annotations.GwtCompatible;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Function;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.primitives.Ints;
import java.io.Serializable;
import java.util.AbstractList;
import java.util.AbstractSequentialList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.RandomAccess;
import javax.annotation.Nullable;













































@GwtCompatible
public final class Lists
{
  @GwtCompatible(serializable=true)
  public static <E> ArrayList<E> newArrayList()
  {
    return new ArrayList();
  }
  










  @GwtCompatible(serializable=true)
  public static <E> ArrayList<E> newArrayList(E... elements)
  {
    Preconditions.checkNotNull(elements);
    
    int capacity = computeArrayListCapacity(elements.length);
    ArrayList<E> list = new ArrayList(capacity);
    Collections.addAll(list, elements);
    return list;
  }
  
  @VisibleForTesting
  static int computeArrayListCapacity(int arraySize) { Preconditions.checkArgument(arraySize >= 0);
    

    return Ints.saturatedCast(5L + arraySize + arraySize / 10);
  }
  









  @GwtCompatible(serializable=true)
  public static <E> ArrayList<E> newArrayList(Iterable<? extends E> elements)
  {
    Preconditions.checkNotNull(elements);
    
    return (elements instanceof Collection) ? new ArrayList(Collections2.cast(elements)) : newArrayList(elements.iterator());
  }
  











  @GwtCompatible(serializable=true)
  public static <E> ArrayList<E> newArrayList(Iterator<? extends E> elements)
  {
    Preconditions.checkNotNull(elements);
    ArrayList<E> list = newArrayList();
    while (elements.hasNext()) {
      list.add(elements.next());
    }
    return list;
  }
  




















  @GwtCompatible(serializable=true)
  public static <E> ArrayList<E> newArrayListWithCapacity(int initialArraySize)
  {
    Preconditions.checkArgument(initialArraySize >= 0);
    return new ArrayList(initialArraySize);
  }
  















  @GwtCompatible(serializable=true)
  public static <E> ArrayList<E> newArrayListWithExpectedSize(int estimatedSize)
  {
    return new ArrayList(computeArrayListCapacity(estimatedSize));
  }
  









  @GwtCompatible(serializable=true)
  public static <E> LinkedList<E> newLinkedList()
  {
    return new LinkedList();
  }
  






  @GwtCompatible(serializable=true)
  public static <E> LinkedList<E> newLinkedList(Iterable<? extends E> elements)
  {
    LinkedList<E> list = newLinkedList();
    for (E element : elements) {
      list.add(element);
    }
    return list;
  }
  















  public static <E> List<E> asList(@Nullable E first, E[] rest)
  {
    return new OnePlusArrayList(first, rest);
  }
  
  private static class OnePlusArrayList<E> extends AbstractList<E> implements Serializable, RandomAccess
  {
    final E first;
    final E[] rest;
    private static final long serialVersionUID = 0L;
    
    OnePlusArrayList(@Nullable E first, E[] rest) {
      this.first = first;
      this.rest = ((Object[])Preconditions.checkNotNull(rest));
    }
    
    public int size() { return rest.length + 1; }
    
    public E get(int index)
    {
      Preconditions.checkElementIndex(index, size());
      return index == 0 ? first : rest[(index - 1)];
    }
  }
  


















  public static <E> List<E> asList(@Nullable E first, @Nullable E second, E[] rest)
  {
    return new TwoPlusArrayList(first, second, rest);
  }
  
  private static class TwoPlusArrayList<E> extends AbstractList<E> implements Serializable, RandomAccess
  {
    final E first;
    final E second;
    final E[] rest;
    private static final long serialVersionUID = 0L;
    
    TwoPlusArrayList(@Nullable E first, @Nullable E second, E[] rest) {
      this.first = first;
      this.second = second;
      this.rest = ((Object[])Preconditions.checkNotNull(rest));
    }
    
    public int size() { return rest.length + 2; }
    
    public E get(int index) {
      switch (index) {
      case 0: 
        return first;
      case 1: 
        return second;
      }
      
      Preconditions.checkElementIndex(index, size());
      return rest[(index - 2)];
    }
  }
  





























  public static <F, T> List<T> transform(List<F> fromList, Function<? super F, ? extends T> function)
  {
    return (fromList instanceof RandomAccess) ? new TransformingRandomAccessList(fromList, function) : new TransformingSequentialList(fromList, function);
  }
  

  private static class TransformingSequentialList<F, T>
    extends AbstractSequentialList<T>
    implements Serializable
  {
    final List<F> fromList;
    
    final Function<? super F, ? extends T> function;
    
    private static final long serialVersionUID = 0L;
    

    TransformingSequentialList(List<F> fromList, Function<? super F, ? extends T> function)
    {
      this.fromList = ((List)Preconditions.checkNotNull(fromList));
      this.function = ((Function)Preconditions.checkNotNull(function));
    }
    



    public void clear()
    {
      fromList.clear();
    }
    
    public int size() { return fromList.size(); }
    
    public ListIterator<T> listIterator(int index) {
      final ListIterator<F> delegate = fromList.listIterator(index);
      new ListIterator()
      {
        public void add(T e) {
          throw new UnsupportedOperationException();
        }
        
        public boolean hasNext()
        {
          return delegate.hasNext();
        }
        
        public boolean hasPrevious()
        {
          return delegate.hasPrevious();
        }
        
        public T next()
        {
          return function.apply(delegate.next());
        }
        
        public int nextIndex()
        {
          return delegate.nextIndex();
        }
        
        public T previous()
        {
          return function.apply(delegate.previous());
        }
        
        public int previousIndex()
        {
          return delegate.previousIndex();
        }
        
        public void remove()
        {
          delegate.remove();
        }
        
        public void set(T e)
        {
          throw new UnsupportedOperationException("not supported");
        }
      };
    }
  }
  


  private static class TransformingRandomAccessList<F, T>
    extends AbstractList<T>
    implements RandomAccess, Serializable
  {
    final List<F> fromList;
    

    final Function<? super F, ? extends T> function;
    

    private static final long serialVersionUID = 0L;
    

    TransformingRandomAccessList(List<F> fromList, Function<? super F, ? extends T> function)
    {
      this.fromList = ((List)Preconditions.checkNotNull(fromList));
      this.function = ((Function)Preconditions.checkNotNull(function));
    }
    
    public void clear() { fromList.clear(); }
    
    public T get(int index) {
      return function.apply(fromList.get(index));
    }
    
    public boolean isEmpty() { return fromList.isEmpty(); }
    
    public T remove(int index) {
      return function.apply(fromList.remove(index));
    }
    
    public int size() { return fromList.size(); }
  }
  



















  public static <T> List<List<T>> partition(List<T> list, int size)
  {
    Preconditions.checkNotNull(list);
    Preconditions.checkArgument(size > 0);
    return (list instanceof RandomAccess) ? new RandomAccessPartition(list, size) : new Partition(list, size);
  }
  
  private static class Partition<T> extends AbstractList<List<T>>
  {
    final List<T> list;
    final int size;
    
    Partition(List<T> list, int size)
    {
      this.list = list;
      this.size = size;
    }
    
    public List<T> get(int index) {
      int listSize = size();
      Preconditions.checkElementIndex(index, listSize);
      int start = index * size;
      int end = Math.min(start + size, list.size());
      return list.subList(start, end);
    }
    
    public int size()
    {
      int result = list.size() / size;
      if (result * size != list.size()) {
        result++;
      }
      return result;
    }
    
    public boolean isEmpty() {
      return list.isEmpty();
    }
  }
  
  private static class RandomAccessPartition<T> extends Lists.Partition<T> implements RandomAccess
  {
    RandomAccessPartition(List<T> list, int size) {
      super(size);
    }
  }
  




  @Beta
  public static ImmutableList<Character> charactersOf(String string)
  {
    return new StringAsImmutableList((String)Preconditions.checkNotNull(string));
  }
  
  private static final class StringAsImmutableList
    extends ImmutableList<Character>
  {
    private final String string;
    
    StringAsImmutableList(String string)
    {
      this.string = string;
    }
    
    public boolean contains(@Nullable Object object) {
      return indexOf(object) >= 0;
    }
    
    public int indexOf(@Nullable Object object) {
      return (object instanceof Character) ? string.indexOf(((Character)object).charValue()) : -1;
    }
    
    public int lastIndexOf(@Nullable Object object)
    {
      return (object instanceof Character) ? string.lastIndexOf(((Character)object).charValue()) : -1;
    }
    

    public UnmodifiableListIterator<Character> listIterator(int index)
    {
      new AbstractIndexedListIterator(size(), index) {
        protected Character get(int index) {
          return Character.valueOf(string.charAt(index));
        }
      };
    }
    
    public ImmutableList<Character> subList(int fromIndex, int toIndex)
    {
      return Lists.charactersOf(string.substring(fromIndex, toIndex));
    }
    
    boolean isPartialView() {
      return false;
    }
    
    public Character get(int index) {
      return Character.valueOf(string.charAt(index));
    }
    
    public int size() {
      return string.length();
    }
    
    public boolean equals(@Nullable Object obj) {
      if (!(obj instanceof List)) {
        return false;
      }
      List<?> list = (List)obj;
      int n = string.length();
      if (n != list.size()) {
        return false;
      }
      Iterator<?> iterator = list.iterator();
      for (int i = 0; i < n; i++) {
        Object elem = iterator.next();
        if ((!(elem instanceof Character)) || (((Character)elem).charValue() != string.charAt(i)))
        {
          return false;
        }
      }
      return true;
    }
    
    int hash = 0;
    
    public int hashCode() {
      int h = hash;
      if (h == 0) {
        h = 1;
        for (int i = 0; i < string.length(); i++) {
          h = h * 31 + string.charAt(i);
        }
        hash = h;
      }
      return h;
    }
  }
  









  @Beta
  public static List<Character> charactersOf(CharSequence sequence)
  {
    return new CharSequenceAsList((CharSequence)Preconditions.checkNotNull(sequence));
  }
  
  private static final class CharSequenceAsList extends AbstractList<Character>
  {
    private final CharSequence sequence;
    
    CharSequenceAsList(CharSequence sequence) {
      this.sequence = sequence;
    }
    
    public Character get(int index) {
      return Character.valueOf(sequence.charAt(index));
    }
    
    public boolean contains(@Nullable Object o) {
      return indexOf(o) >= 0;
    }
    
    public int indexOf(@Nullable Object o) {
      if ((o instanceof Character)) {
        char c = ((Character)o).charValue();
        for (int i = 0; i < sequence.length(); i++) {
          if (sequence.charAt(i) == c) {
            return i;
          }
        }
      }
      return -1;
    }
    
    public int lastIndexOf(@Nullable Object o) {
      if ((o instanceof Character)) {
        char c = ((Character)o).charValue();
        for (int i = sequence.length() - 1; i >= 0; i--) {
          if (sequence.charAt(i) == c) {
            return i;
          }
        }
      }
      return -1;
    }
    
    public int size() {
      return sequence.length();
    }
    
    public List<Character> subList(int fromIndex, int toIndex) {
      return Lists.charactersOf(sequence.subSequence(fromIndex, toIndex));
    }
    
    public int hashCode() {
      int hash = 1;
      for (int i = 0; i < sequence.length(); i++) {
        hash = hash * 31 + sequence.charAt(i);
      }
      return hash;
    }
    
    public boolean equals(@Nullable Object o) {
      if (!(o instanceof List)) {
        return false;
      }
      List<?> list = (List)o;
      int n = sequence.length();
      if (n != list.size()) {
        return false;
      }
      Iterator<?> iterator = list.iterator();
      for (int i = 0; i < n; i++) {
        Object elem = iterator.next();
        if ((!(elem instanceof Character)) || (((Character)elem).charValue() != sequence.charAt(i)))
        {
          return false;
        }
      }
      return true;
    }
  }
  











  public static <T> List<T> reverse(List<T> list)
  {
    if ((list instanceof ReverseList))
      return ((ReverseList)list).getForwardList();
    if ((list instanceof RandomAccess)) {
      return new RandomAccessReverseList(list);
    }
    return new ReverseList(list);
  }
  
  private static class ReverseList<T> extends AbstractList<T>
  {
    private final List<T> forwardList;
    
    ReverseList(List<T> forwardList) {
      this.forwardList = ((List)Preconditions.checkNotNull(forwardList));
    }
    
    List<T> getForwardList() {
      return forwardList;
    }
    
    private int reverseIndex(int index) {
      int size = size();
      Preconditions.checkElementIndex(index, size);
      return size - 1 - index;
    }
    
    private int reversePosition(int index) {
      int size = size();
      Preconditions.checkPositionIndex(index, size);
      return size - index;
    }
    
    public void add(int index, @Nullable T element) {
      forwardList.add(reversePosition(index), element);
    }
    
    public void clear() {
      forwardList.clear();
    }
    
    public T remove(int index) {
      return forwardList.remove(reverseIndex(index));
    }
    
    protected void removeRange(int fromIndex, int toIndex) {
      subList(fromIndex, toIndex).clear();
    }
    
    public T set(int index, @Nullable T element) {
      return forwardList.set(reverseIndex(index), element);
    }
    
    public T get(int index) {
      return forwardList.get(reverseIndex(index));
    }
    
    public boolean isEmpty() {
      return forwardList.isEmpty();
    }
    
    public int size() {
      return forwardList.size();
    }
    
    public boolean contains(@Nullable Object o) {
      return forwardList.contains(o);
    }
    
    public boolean containsAll(Collection<?> c) {
      return forwardList.containsAll(c);
    }
    
    public List<T> subList(int fromIndex, int toIndex) {
      Preconditions.checkPositionIndexes(fromIndex, toIndex, size());
      return Lists.reverse(forwardList.subList(reversePosition(toIndex), reversePosition(fromIndex)));
    }
    
    public int indexOf(@Nullable Object o)
    {
      int index = forwardList.lastIndexOf(o);
      return index >= 0 ? reverseIndex(index) : -1;
    }
    
    public int lastIndexOf(@Nullable Object o) {
      int index = forwardList.indexOf(o);
      return index >= 0 ? reverseIndex(index) : -1;
    }
    
    public Iterator<T> iterator() {
      return listIterator();
    }
    
    public ListIterator<T> listIterator(int index) {
      int start = reversePosition(index);
      final ListIterator<T> forwardIterator = forwardList.listIterator(start);
      new ListIterator()
      {
        boolean canRemove;
        boolean canSet;
        
        public void add(T e) {
          forwardIterator.add(e);
          forwardIterator.previous();
          canSet = (this.canRemove = 0);
        }
        
        public boolean hasNext() {
          return forwardIterator.hasPrevious();
        }
        
        public boolean hasPrevious() {
          return forwardIterator.hasNext();
        }
        
        public T next() {
          if (!hasNext()) {
            throw new NoSuchElementException();
          }
          canSet = (this.canRemove = 1);
          return forwardIterator.previous();
        }
        
        public int nextIndex() {
          return Lists.ReverseList.this.reversePosition(forwardIterator.nextIndex());
        }
        
        public T previous() {
          if (!hasPrevious()) {
            throw new NoSuchElementException();
          }
          canSet = (this.canRemove = 1);
          return forwardIterator.next();
        }
        
        public int previousIndex() {
          return nextIndex() - 1;
        }
        
        public void remove() {
          Preconditions.checkState(canRemove);
          forwardIterator.remove();
          canRemove = (this.canSet = 0);
        }
        
        public void set(T e) {
          Preconditions.checkState(canSet);
          forwardIterator.set(e);
        }
      };
    }
  }
  
  private static class RandomAccessReverseList<T> extends Lists.ReverseList<T> implements RandomAccess
  {
    RandomAccessReverseList(List<T> forwardList) {
      super();
    }
  }
  


  static int hashCodeImpl(List<?> list)
  {
    int hashCode = 1;
    for (Object o : list) {
      hashCode = 31 * hashCode + (o == null ? 0 : o.hashCode());
    }
    return hashCode;
  }
  


  static boolean equalsImpl(List<?> list, @Nullable Object object)
  {
    if (object == Preconditions.checkNotNull(list)) {
      return true;
    }
    if (!(object instanceof List)) {
      return false;
    }
    
    List<?> o = (List)object;
    
    return (list.size() == o.size()) && (Iterators.elementsEqual(list.iterator(), o.iterator()));
  }
  




  static <E> boolean addAllImpl(List<E> list, int index, Iterable<? extends E> elements)
  {
    boolean changed = false;
    ListIterator<E> listIterator = list.listIterator(index);
    for (E e : elements) {
      listIterator.add(e);
      changed = true;
    }
    return changed;
  }
  


  static int indexOfImpl(List<?> list, @Nullable Object element)
  {
    ListIterator<?> listIterator = list.listIterator();
    while (listIterator.hasNext()) {
      if (Objects.equal(element, listIterator.next())) {
        return listIterator.previousIndex();
      }
    }
    return -1;
  }
  


  static int lastIndexOfImpl(List<?> list, @Nullable Object element)
  {
    ListIterator<?> listIterator = list.listIterator(list.size());
    while (listIterator.hasPrevious()) {
      if (Objects.equal(element, listIterator.previous())) {
        return listIterator.nextIndex();
      }
    }
    return -1;
  }
  


  static <E> ListIterator<E> listIteratorImpl(List<E> list, int index)
  {
    return new AbstractListWrapper(list).listIterator(index);
  }
  

  static <E> List<E> subListImpl(List<E> list, int fromIndex, int toIndex)
  {
    List<E> wrapper;
    
    List<E> wrapper;
    if ((list instanceof RandomAccess)) {
      wrapper = new RandomAccessListWrapper(list) { private static final long serialVersionUID = 0L;
        
        public ListIterator<E> listIterator(int index) { return backingList.listIterator(index);
        }

      };
    }
    else {
      wrapper = new AbstractListWrapper(list) { private static final long serialVersionUID = 0L;
        
        public ListIterator<E> listIterator(int index) { return backingList.listIterator(index); }
      };
    }
    


    return wrapper.subList(fromIndex, toIndex);
  }
  
  private static class AbstractListWrapper<E> extends AbstractList<E> {
    final List<E> backingList;
    
    AbstractListWrapper(List<E> backingList) {
      this.backingList = ((List)Preconditions.checkNotNull(backingList));
    }
    
    public void add(int index, E element) {
      backingList.add(index, element);
    }
    
    public boolean addAll(int index, Collection<? extends E> c) {
      return backingList.addAll(index, c);
    }
    
    public E get(int index) {
      return backingList.get(index);
    }
    
    public E remove(int index) {
      return backingList.remove(index);
    }
    
    public E set(int index, E element) {
      return backingList.set(index, element);
    }
    
    public boolean contains(Object o) {
      return backingList.contains(o);
    }
    
    public int size() {
      return backingList.size();
    }
  }
  
  private static class RandomAccessListWrapper<E> extends Lists.AbstractListWrapper<E> implements RandomAccess
  {
    RandomAccessListWrapper(List<E> backingList) {
      super();
    }
  }
}
