package com.google.common.collect;

import com.google.common.annotations.Beta;
import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import java.io.Serializable;
import java.util.Comparator;
import java.util.NoSuchElementException;
import java.util.SortedSet;
import javax.annotation.Nullable;































































































@GwtCompatible
@Beta
public final class Range<C extends Comparable>
  implements Predicate<C>, Serializable
{
  final Cut<C> lowerBound;
  final Cut<C> upperBound;
  private static final long serialVersionUID = 0L;
  
  Range(Cut<C> lowerBound, Cut<C> upperBound)
  {
    if (lowerBound.compareTo(upperBound) > 0) {
      throw new IllegalArgumentException("Invalid range: " + toString(lowerBound, upperBound));
    }
    
    this.lowerBound = lowerBound;
    this.upperBound = upperBound;
  }
  


  public boolean hasLowerBound()
  {
    return lowerBound != Cut.belowAll();
  }
  





  public C lowerEndpoint()
  {
    return lowerBound.endpoint();
  }
  







  public BoundType lowerBoundType()
  {
    return lowerBound.typeAsLowerBound();
  }
  


  public boolean hasUpperBound()
  {
    return upperBound != Cut.aboveAll();
  }
  





  public C upperEndpoint()
  {
    return upperBound.endpoint();
  }
  







  public BoundType upperBoundType()
  {
    return upperBound.typeAsUpperBound();
  }
  








  public boolean isEmpty()
  {
    return lowerBound.equals(upperBound);
  }
  




  public boolean contains(C value)
  {
    Preconditions.checkNotNull(value);
    
    return (lowerBound.isLessThan(value)) && (!upperBound.isLessThan(value));
  }
  




  public boolean apply(C input)
  {
    return contains(input);
  }
  



  public boolean containsAll(Iterable<? extends C> values)
  {
    if (Iterables.isEmpty(values)) {
      return true;
    }
    

    if ((values instanceof SortedSet)) {
      SortedSet<? extends C> set = cast(values);
      Comparator<?> comparator = set.comparator();
      if ((Ordering.natural().equals(comparator)) || (comparator == null)) {
        return (contains((Comparable)set.first())) && (contains((Comparable)set.last()));
      }
    }
    
    for (C value : values) {
      if (!contains(value)) {
        return false;
      }
    }
    return true;
  }
  





































  public boolean encloses(Range<C> other)
  {
    return (lowerBound.compareTo(lowerBound) <= 0) && (upperBound.compareTo(upperBound) >= 0);
  }
  



























  public Range<C> intersection(Range<C> other)
  {
    Cut<C> newLower = (Cut)Ordering.natural().max(lowerBound, lowerBound);
    Cut<C> newUpper = (Cut)Ordering.natural().min(upperBound, upperBound);
    return Ranges.create(newLower, newUpper);
  }
  
























  public boolean isConnected(Range<C> other)
  {
    return (lowerBound.compareTo(upperBound) <= 0) && (lowerBound.compareTo(upperBound) <= 0);
  }
  




















  public Range<C> span(Range<C> other)
  {
    Cut<C> newLower = (Cut)Ordering.natural().min(lowerBound, lowerBound);
    Cut<C> newUpper = (Cut)Ordering.natural().max(upperBound, upperBound);
    return Ranges.create(newLower, newUpper);
  }
  




















  @GwtCompatible(serializable=false)
  public ContiguousSet<C> asSet(DiscreteDomain<C> domain)
  {
    Preconditions.checkNotNull(domain);
    Range<C> effectiveRange = this;
    try {
      if (!hasLowerBound()) {
        effectiveRange = effectiveRange.intersection(Ranges.atLeast(domain.minValue()));
      }
      
      if (!hasUpperBound()) {
        effectiveRange = effectiveRange.intersection(Ranges.atMost(domain.maxValue()));
      }
    }
    catch (NoSuchElementException e) {
      throw new IllegalArgumentException(e);
    }
    

    boolean empty = (effectiveRange.isEmpty()) || (compareOrThrow(lowerBound.leastValueAbove(domain), upperBound.greatestValueBelow(domain)) > 0);
    



    return empty ? new EmptyContiguousSet(domain) : new RegularContiguousSet(effectiveRange, domain);
  }
  


























  public Range<C> canonical(DiscreteDomain<C> domain)
  {
    Preconditions.checkNotNull(domain);
    Cut<C> lower = lowerBound.canonical(domain);
    Cut<C> upper = upperBound.canonical(domain);
    return (lower == lowerBound) && (upper == upperBound) ? this : Ranges.create(lower, upper);
  }
  









  public boolean equals(@Nullable Object object)
  {
    if ((object instanceof Range)) {
      Range<?> other = (Range)object;
      return (lowerBound.equals(lowerBound)) && (upperBound.equals(upperBound));
    }
    
    return false;
  }
  
  public int hashCode()
  {
    return lowerBound.hashCode() * 31 + upperBound.hashCode();
  }
  



  public String toString()
  {
    return toString(lowerBound, upperBound);
  }
  
  private static String toString(Cut<?> lowerBound, Cut<?> upperBound) {
    StringBuilder sb = new StringBuilder(16);
    lowerBound.describeAsLowerBound(sb);
    sb.append('‥');
    upperBound.describeAsUpperBound(sb);
    return sb.toString();
  }
  


  private static <T> SortedSet<T> cast(Iterable<T> iterable)
  {
    return (SortedSet)iterable;
  }
  
  static int compareOrThrow(Comparable left, Comparable right)
  {
    return left.compareTo(right);
  }
}
