package com.google.common.collect;

import com.google.common.annotations.Beta;
import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import java.util.Iterator;































































@GwtCompatible
@Beta
public final class Ranges
{
  static <C extends Comparable<?>> Range<C> create(Cut<C> lowerBound, Cut<C> upperBound)
  {
    return new Range(lowerBound, upperBound);
  }
  






  public static <C extends Comparable<?>> Range<C> open(C lower, C upper)
  {
    return create(Cut.aboveValue(lower), Cut.belowValue(upper));
  }
  






  public static <C extends Comparable<?>> Range<C> closed(C lower, C upper)
  {
    return create(Cut.belowValue(lower), Cut.aboveValue(upper));
  }
  







  public static <C extends Comparable<?>> Range<C> closedOpen(C lower, C upper)
  {
    return create(Cut.belowValue(lower), Cut.belowValue(upper));
  }
  







  public static <C extends Comparable<?>> Range<C> openClosed(C lower, C upper)
  {
    return create(Cut.aboveValue(lower), Cut.aboveValue(upper));
  }
  








  public static <C extends Comparable<?>> Range<C> range(C lower, BoundType lowerType, C upper, BoundType upperType)
  {
    Preconditions.checkNotNull(lowerType);
    Preconditions.checkNotNull(upperType);
    
    Cut<C> lowerBound = lowerType == BoundType.OPEN ? Cut.aboveValue(lower) : Cut.belowValue(lower);
    

    Cut<C> upperBound = upperType == BoundType.OPEN ? Cut.belowValue(upper) : Cut.aboveValue(upper);
    

    return create(lowerBound, upperBound);
  }
  



  public static <C extends Comparable<?>> Range<C> lessThan(C endpoint)
  {
    return create(Cut.belowAll(), Cut.belowValue(endpoint));
  }
  



  public static <C extends Comparable<?>> Range<C> atMost(C endpoint)
  {
    return create(Cut.belowAll(), Cut.aboveValue(endpoint));
  }
  




  public static <C extends Comparable<?>> Range<C> upTo(C endpoint, BoundType boundType)
  {
    switch (1.$SwitchMap$com$google$common$collect$BoundType[boundType.ordinal()]) {
    case 1: 
      return lessThan(endpoint);
    case 2: 
      return atMost(endpoint);
    }
    throw new AssertionError();
  }
  




  public static <C extends Comparable<?>> Range<C> greaterThan(C endpoint)
  {
    return create(Cut.aboveValue(endpoint), Cut.aboveAll());
  }
  



  public static <C extends Comparable<?>> Range<C> atLeast(C endpoint)
  {
    return create(Cut.belowValue(endpoint), Cut.aboveAll());
  }
  




  public static <C extends Comparable<?>> Range<C> downTo(C endpoint, BoundType boundType)
  {
    switch (1.$SwitchMap$com$google$common$collect$BoundType[boundType.ordinal()]) {
    case 1: 
      return greaterThan(endpoint);
    case 2: 
      return atLeast(endpoint);
    }
    throw new AssertionError();
  }
  

  public static <C extends Comparable<?>> Range<C> all()
  {
    return create(Cut.belowAll(), Cut.aboveAll());
  }
  




  public static <C extends Comparable<?>> Range<C> singleton(C value)
  {
    return closed(value, value);
  }
  










  public static <C extends Comparable<?>> Range<C> encloseAll(Iterable<C> values)
  {
    Preconditions.checkNotNull(values);
    if ((values instanceof ContiguousSet)) {
      return ((ContiguousSet)values).range();
    }
    Iterator<C> valueIterator = values.iterator();
    C min = (Comparable)Preconditions.checkNotNull(valueIterator.next());
    C max = min;
    while (valueIterator.hasNext()) {
      C value = (Comparable)Preconditions.checkNotNull(valueIterator.next());
      min = (Comparable)Ordering.natural().min(min, value);
      max = (Comparable)Ordering.natural().max(max, value);
    }
    return closed(min, max);
  }
}
