package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import java.util.List;
import javax.annotation.Nullable;






















@GwtCompatible(serializable=true, emulated=true)
class RegularImmutableList<E>
  extends ImmutableList<E>
{
  private final transient int offset;
  private final transient int size;
  private final transient Object[] array;
  
  RegularImmutableList(Object[] array, int offset, int size)
  {
    this.offset = offset;
    this.size = size;
    this.array = array;
  }
  
  RegularImmutableList(Object[] array) {
    this(array, 0, array.length);
  }
  
  public int size()
  {
    return size;
  }
  
  public boolean isEmpty() {
    return false;
  }
  
  boolean isPartialView() {
    return (offset != 0) || (size != array.length);
  }
  
  public boolean contains(@Nullable Object target) {
    return indexOf(target) != -1;
  }
  

  public UnmodifiableIterator<E> iterator()
  {
    return Iterators.forArray(array, offset, size);
  }
  
  public Object[] toArray() {
    Object[] newArray = new Object[size()];
    System.arraycopy(array, offset, newArray, 0, size);
    return newArray;
  }
  
  public <T> T[] toArray(T[] other) {
    if (other.length < size) {
      other = ObjectArrays.newArray(other, size);
    } else if (other.length > size) {
      other[size] = null;
    }
    System.arraycopy(array, offset, other, 0, size);
    return other;
  }
  


  public E get(int index)
  {
    Preconditions.checkElementIndex(index, size);
    return array[(index + offset)];
  }
  
  public int indexOf(@Nullable Object target) {
    if (target != null) {
      for (int i = offset; i < offset + size; i++) {
        if (array[i].equals(target)) {
          return i - offset;
        }
      }
    }
    return -1;
  }
  
  public int lastIndexOf(@Nullable Object target) {
    if (target != null) {
      for (int i = offset + size - 1; i >= offset; i--) {
        if (array[i].equals(target)) {
          return i - offset;
        }
      }
    }
    return -1;
  }
  
  public ImmutableList<E> subList(int fromIndex, int toIndex) {
    Preconditions.checkPositionIndexes(fromIndex, toIndex, size);
    return fromIndex == toIndex ? ImmutableList.of() : new RegularImmutableList(array, offset + fromIndex, toIndex - fromIndex);
  }
  


  public UnmodifiableListIterator<E> listIterator(int start)
  {
    new AbstractIndexedListIterator(size, start)
    {
      protected E get(int index)
      {
        return array[(index + offset)];
      }
    };
  }
  
  public boolean equals(@Nullable Object object)
  {
    if (object == this) {
      return true;
    }
    if (!(object instanceof List)) {
      return false;
    }
    
    List<?> that = (List)object;
    if (size() != that.size()) {
      return false;
    }
    
    int index = offset;
    if ((object instanceof RegularImmutableList)) {
      RegularImmutableList<?> other = (RegularImmutableList)object;
      for (int i = offset; i < offset + size; i++) {
        if (!array[(index++)].equals(array[i])) {
          return false;
        }
      }
    } else {
      for (Object element : that) {
        if (!array[(index++)].equals(element)) {
          return false;
        }
      }
    }
    return true;
  }
  

  public int hashCode()
  {
    int hashCode = 1;
    for (int i = offset; i < offset + size; i++) {
      hashCode = 31 * hashCode + array[i].hashCode();
    }
    return hashCode;
  }
  
  public String toString() {
    StringBuilder sb = Collections2.newStringBuilderForCollection(size()).append('[').append(array[offset]);
    
    for (int i = offset + 1; i < offset + size; i++) {
      sb.append(", ").append(array[i]);
    }
    return ']';
  }
}
