package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import java.util.Map.Entry;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

























@GwtCompatible(serializable=true, emulated=true)
final class RegularImmutableMap<K, V>
  extends ImmutableMap<K, V>
{
  private final transient LinkedEntry<K, V>[] entries;
  private final transient LinkedEntry<K, V>[] table;
  private final transient int mask;
  private final transient int keySetHashCode;
  private transient ImmutableSet<Map.Entry<K, V>> entrySet;
  private transient ImmutableSet<K> keySet;
  private transient ImmutableCollection<V> values;
  private static final long serialVersionUID = 0L;
  
  RegularImmutableMap(Map.Entry<?, ?>... immutableEntries)
  {
    int size = immutableEntries.length;
    entries = createEntryArray(size);
    
    int tableSize = chooseTableSize(size);
    table = createEntryArray(tableSize);
    mask = (tableSize - 1);
    
    int keySetHashCodeMutable = 0;
    for (int entryIndex = 0; entryIndex < size; entryIndex++)
    {

      Map.Entry<K, V> entry = immutableEntries[entryIndex];
      K key = entry.getKey();
      int keyHashCode = key.hashCode();
      keySetHashCodeMutable += keyHashCode;
      int tableIndex = Hashing.smear(keyHashCode) & mask;
      LinkedEntry<K, V> existing = table[tableIndex];
      
      LinkedEntry<K, V> linkedEntry = newLinkedEntry(key, entry.getValue(), existing);
      
      table[tableIndex] = linkedEntry;
      entries[entryIndex] = linkedEntry;
      while (existing != null) {
        Preconditions.checkArgument(!key.equals(existing.getKey()), "duplicate key: %s", new Object[] { key });
        existing = existing.next();
      }
    }
    keySetHashCode = keySetHashCodeMutable;
  }
  
  private static int chooseTableSize(int size)
  {
    int tableSize = Integer.highestOneBit(size) << 1;
    Preconditions.checkArgument(tableSize > 0, "table too large: %s", new Object[] { Integer.valueOf(size) });
    return tableSize;
  }
  





  private LinkedEntry<K, V>[] createEntryArray(int size)
  {
    return new LinkedEntry[size];
  }
  
  private static <K, V> LinkedEntry<K, V> newLinkedEntry(K key, V value, @Nullable LinkedEntry<K, V> next)
  {
    return next == null ? new TerminalEntry(key, value) : new NonTerminalEntry(key, value, next);
  }
  
  private static abstract interface LinkedEntry<K, V>
    extends Map.Entry<K, V>
  {
    @Nullable
    public abstract LinkedEntry<K, V> next();
  }
  
  @Immutable
  private static final class NonTerminalEntry<K, V>
    extends ImmutableEntry<K, V> implements RegularImmutableMap.LinkedEntry<K, V>
  {
    final RegularImmutableMap.LinkedEntry<K, V> next;
    
    NonTerminalEntry(K key, V value, RegularImmutableMap.LinkedEntry<K, V> next)
    {
      super(value);
      this.next = next;
    }
    
    public RegularImmutableMap.LinkedEntry<K, V> next() {
      return next;
    }
  }
  


  @Immutable
  private static final class TerminalEntry<K, V>
    extends ImmutableEntry<K, V>
    implements RegularImmutableMap.LinkedEntry<K, V>
  {
    TerminalEntry(K key, V value)
    {
      super(value);
    }
    
    @Nullable
    public RegularImmutableMap.LinkedEntry<K, V> next() { return null; }
  }
  
  public V get(@Nullable Object key)
  {
    if (key == null) {
      return null;
    }
    int index = Hashing.smear(key.hashCode()) & mask;
    for (LinkedEntry<K, V> entry = table[index]; 
        entry != null; 
        entry = entry.next()) {
      K candidateKey = entry.getKey();
      






      if (key.equals(candidateKey)) {
        return entry.getValue();
      }
    }
    return null;
  }
  
  public int size()
  {
    return entries.length;
  }
  
  public boolean isEmpty() {
    return false;
  }
  
  public boolean containsValue(@Nullable Object value) {
    if (value == null) {
      return false;
    }
    for (Map.Entry<K, V> entry : entries) {
      if (entry.getValue().equals(value)) {
        return true;
      }
    }
    return false;
  }
  
  boolean isPartialView() {
    return false;
  }
  

  public ImmutableSet<Map.Entry<K, V>> entrySet()
  {
    ImmutableSet<Map.Entry<K, V>> es = entrySet;
    return es == null ? (this.entrySet = new EntrySet(this)) : es;
  }
  
  private static class EntrySet<K, V> extends ImmutableSet.ArrayImmutableSet<Map.Entry<K, V>>
  {
    final transient RegularImmutableMap<K, V> map;
    
    EntrySet(RegularImmutableMap<K, V> map) {
      super();
      this.map = map;
    }
    
    public boolean contains(Object target) {
      if ((target instanceof Map.Entry)) {
        Map.Entry<?, ?> entry = (Map.Entry)target;
        V mappedValue = map.get(entry.getKey());
        return (mappedValue != null) && (mappedValue.equals(entry.getValue()));
      }
      return false;
    }
  }
  

  public ImmutableSet<K> keySet()
  {
    ImmutableSet<K> ks = keySet;
    return ks == null ? (this.keySet = new KeySet(this)) : ks;
  }
  
  private static class KeySet<K, V> extends ImmutableSet.TransformedImmutableSet<Map.Entry<K, V>, K>
  {
    final RegularImmutableMap<K, V> map;
    
    KeySet(RegularImmutableMap<K, V> map)
    {
      super(keySetHashCode);
      this.map = map;
    }
    
    K transform(Map.Entry<K, V> element) {
      return element.getKey();
    }
    
    public boolean contains(Object target) {
      return map.containsKey(target);
    }
    
    boolean isPartialView() {
      return true;
    }
  }
  

  public ImmutableCollection<V> values()
  {
    ImmutableCollection<V> v = values;
    return v == null ? (this.values = new Values(this)) : v;
  }
  
  private static class Values<V> extends ImmutableCollection<V>
  {
    final RegularImmutableMap<?, V> map;
    
    Values(RegularImmutableMap<?, V> map) {
      this.map = map;
    }
    
    public int size()
    {
      return map.entries.length;
    }
    
    public UnmodifiableIterator<V> iterator() {
      new AbstractIndexedListIterator(map.entries.length) {
        protected V get(int index) {
          return map.entries[index].getValue();
        }
      };
    }
    
    public boolean contains(Object target) {
      return map.containsValue(target);
    }
    
    boolean isPartialView() {
      return true;
    }
  }
  
  public String toString() {
    StringBuilder result = Collections2.newStringBuilderForCollection(size()).append('{');
    
    Collections2.STANDARD_JOINER.appendTo(result, entries);
    return '}';
  }
}
