package com.google.common.collect;

import com.google.common.base.Preconditions;
import com.google.common.primitives.Ints;
import java.util.Comparator;
import java.util.List;
import javax.annotation.Nullable;




















final class RegularImmutableSortedMultiset<E>
  extends ImmutableSortedMultiset<E>
{
  final transient ImmutableList<CumulativeCountEntry<E>> entries;
  
  private static final class CumulativeCountEntry<E>
    extends Multisets.AbstractEntry<E>
  {
    final E element;
    final int count;
    final long cumulativeCount;
    
    CumulativeCountEntry(E element, int count, @Nullable CumulativeCountEntry<E> previous)
    {
      this.element = element;
      this.count = count;
      cumulativeCount = (count + (previous == null ? 0L : cumulativeCount));
    }
    
    public E getElement()
    {
      return element;
    }
    
    public int getCount()
    {
      return count;
    }
  }
  
  static <E> RegularImmutableSortedMultiset<E> createFromSorted(Comparator<? super E> comparator, List<? extends Multiset.Entry<E>> entries)
  {
    List<CumulativeCountEntry<E>> newEntries = Lists.newArrayListWithCapacity(entries.size());
    CumulativeCountEntry<E> previous = null;
    for (Multiset.Entry<E> entry : entries) {
      newEntries.add(previous = new CumulativeCountEntry(entry.getElement(), entry.getCount(), previous));
    }
    
    return new RegularImmutableSortedMultiset(comparator, ImmutableList.copyOf(newEntries));
  }
  


  RegularImmutableSortedMultiset(Comparator<? super E> comparator, ImmutableList<CumulativeCountEntry<E>> entries)
  {
    super(comparator);
    this.entries = entries;
    assert (!entries.isEmpty());
  }
  
  ImmutableList<E> elementList() {
    new TransformedImmutableList(entries)
    {
      E transform(RegularImmutableSortedMultiset.CumulativeCountEntry<E> entry) {
        return entry.getElement();
      }
    };
  }
  
  ImmutableSortedSet<E> createElementSet()
  {
    return new RegularImmutableSortedSet(elementList(), comparator());
  }
  
  ImmutableSortedSet<E> createDescendingElementSet()
  {
    return new RegularImmutableSortedSet(elementList().reverse(), reverseComparator());
  }
  

  UnmodifiableIterator<Multiset.Entry<E>> entryIterator()
  {
    return entries.iterator();
  }
  

  UnmodifiableIterator<Multiset.Entry<E>> descendingEntryIterator()
  {
    return entries.reverse().iterator();
  }
  
  public CumulativeCountEntry<E> firstEntry()
  {
    return (CumulativeCountEntry)entries.get(0);
  }
  
  public CumulativeCountEntry<E> lastEntry()
  {
    return (CumulativeCountEntry)entries.get(entries.size() - 1);
  }
  
  public int size()
  {
    CumulativeCountEntry<E> firstEntry = firstEntry();
    CumulativeCountEntry<E> lastEntry = lastEntry();
    return Ints.saturatedCast(cumulativeCount - cumulativeCount + count);
  }
  

  int distinctElements()
  {
    return entries.size();
  }
  
  boolean isPartialView()
  {
    return entries.isPartialView();
  }
  

  public int count(@Nullable Object element)
  {
    if (element == null) {
      return 0;
    }
    try {
      int index = SortedLists.binarySearch(elementList(), element, comparator(), SortedLists.KeyPresentBehavior.ANY_PRESENT, SortedLists.KeyAbsentBehavior.INVERTED_INSERTION_INDEX);
      
      return index >= 0 ? ((CumulativeCountEntry)entries.get(index)).getCount() : 0;
    } catch (ClassCastException e) {}
    return 0;
  }
  

  public ImmutableSortedMultiset<E> headMultiset(E upperBound, BoundType boundType)
  {
    int index;
    switch (2.$SwitchMap$com$google$common$collect$BoundType[boundType.ordinal()]) {
    case 1: 
      index = SortedLists.binarySearch(elementList(), Preconditions.checkNotNull(upperBound), comparator(), SortedLists.KeyPresentBehavior.ANY_PRESENT, SortedLists.KeyAbsentBehavior.NEXT_HIGHER);
      
      break;
    case 2: 
      index = SortedLists.binarySearch(elementList(), Preconditions.checkNotNull(upperBound), comparator(), SortedLists.KeyPresentBehavior.ANY_PRESENT, SortedLists.KeyAbsentBehavior.NEXT_LOWER) + 1;
      
      break;
    default: 
      throw new AssertionError();
    }
    return createSubMultiset(0, index);
  }
  
  public ImmutableSortedMultiset<E> tailMultiset(E lowerBound, BoundType boundType)
  {
    int index;
    switch (2.$SwitchMap$com$google$common$collect$BoundType[boundType.ordinal()]) {
    case 1: 
      index = SortedLists.binarySearch(elementList(), Preconditions.checkNotNull(lowerBound), comparator(), SortedLists.KeyPresentBehavior.ANY_PRESENT, SortedLists.KeyAbsentBehavior.NEXT_LOWER) + 1;
      
      break;
    case 2: 
      index = SortedLists.binarySearch(elementList(), Preconditions.checkNotNull(lowerBound), comparator(), SortedLists.KeyPresentBehavior.ANY_PRESENT, SortedLists.KeyAbsentBehavior.NEXT_HIGHER);
      
      break;
    default: 
      throw new AssertionError();
    }
    return createSubMultiset(index, distinctElements());
  }
  
  private ImmutableSortedMultiset<E> createSubMultiset(int newFromIndex, int newToIndex) {
    if ((newFromIndex == 0) && (newToIndex == entries.size()))
      return this;
    if (newFromIndex >= newToIndex) {
      return emptyMultiset(comparator());
    }
    return new RegularImmutableSortedMultiset(comparator(), entries.subList(newFromIndex, newToIndex));
  }
}
