package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.annotations.GwtIncompatible;
import com.google.common.base.Function;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import com.google.common.base.Predicates;
import com.google.common.primitives.Ints;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.AbstractSet;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.TreeSet;
import javax.annotation.Nullable;














































@GwtCompatible(emulated=true)
public final class Sets
{
  @GwtCompatible(serializable=true)
  public static <E extends Enum<E>> ImmutableSet<E> immutableEnumSet(E anElement, E... otherElements)
  {
    return new ImmutableEnumSet(EnumSet.of(anElement, otherElements));
  }
  












  @GwtCompatible(serializable=true)
  public static <E extends Enum<E>> ImmutableSet<E> immutableEnumSet(Iterable<E> elements)
  {
    Iterator<E> iterator = elements.iterator();
    if (!iterator.hasNext()) {
      return ImmutableSet.of();
    }
    if ((elements instanceof EnumSet)) {
      EnumSet<E> enumSetClone = EnumSet.copyOf((EnumSet)elements);
      return new ImmutableEnumSet(enumSetClone);
    }
    E first = (Enum)iterator.next();
    EnumSet<E> set = EnumSet.of(first);
    while (iterator.hasNext()) {
      set.add(iterator.next());
    }
    return new ImmutableEnumSet(set);
  }
  



















  public static <E extends Enum<E>> EnumSet<E> newEnumSet(Iterable<E> iterable, Class<E> elementType)
  {
    Preconditions.checkNotNull(iterable);
    EnumSet<E> set = EnumSet.noneOf(elementType);
    Iterables.addAll(set, iterable);
    return set;
  }
  












  public static <E> HashSet<E> newHashSet()
  {
    return new HashSet();
  }
  













  public static <E> HashSet<E> newHashSet(E... elements)
  {
    HashSet<E> set = newHashSetWithExpectedSize(elements.length);
    Collections.addAll(set, elements);
    return set;
  }
  












  public static <E> HashSet<E> newHashSetWithExpectedSize(int expectedSize)
  {
    return new HashSet(Maps.capacity(expectedSize));
  }
  












  public static <E> HashSet<E> newHashSet(Iterable<? extends E> elements)
  {
    return (elements instanceof Collection) ? new HashSet(Collections2.cast(elements)) : newHashSet(elements.iterator());
  }
  














  public static <E> HashSet<E> newHashSet(Iterator<? extends E> elements)
  {
    HashSet<E> set = newHashSet();
    while (elements.hasNext()) {
      set.add(elements.next());
    }
    return set;
  }
  









  public static <E> LinkedHashSet<E> newLinkedHashSet()
  {
    return new LinkedHashSet();
  }
  











  public static <E> LinkedHashSet<E> newLinkedHashSet(Iterable<? extends E> elements)
  {
    if ((elements instanceof Collection)) {
      return new LinkedHashSet(Collections2.cast(elements));
    }
    LinkedHashSet<E> set = newLinkedHashSet();
    for (E element : elements) {
      set.add(element);
    }
    return set;
  }
  










  public static <E extends Comparable> TreeSet<E> newTreeSet()
  {
    return new TreeSet();
  }
  















  public static <E extends Comparable> TreeSet<E> newTreeSet(Iterable<? extends E> elements)
  {
    TreeSet<E> set = newTreeSet();
    for (E element : elements) {
      set.add(element);
    }
    return set;
  }
  










  public static <E> TreeSet<E> newTreeSet(Comparator<? super E> comparator)
  {
    return new TreeSet((Comparator)Preconditions.checkNotNull(comparator));
  }
  









  public static <E> Set<E> newIdentityHashSet()
  {
    return newSetFromMap(Maps.newIdentityHashMap());
  }
  















  public static <E extends Enum<E>> EnumSet<E> complementOf(Collection<E> collection)
  {
    if ((collection instanceof EnumSet)) {
      return EnumSet.complementOf((EnumSet)collection);
    }
    Preconditions.checkArgument(!collection.isEmpty(), "collection is empty; use the other version of this method");
    
    Class<E> type = ((Enum)collection.iterator().next()).getDeclaringClass();
    return makeComplementByHand(collection, type);
  }
  












  public static <E extends Enum<E>> EnumSet<E> complementOf(Collection<E> collection, Class<E> type)
  {
    Preconditions.checkNotNull(collection);
    return (collection instanceof EnumSet) ? EnumSet.complementOf((EnumSet)collection) : makeComplementByHand(collection, type);
  }
  


  private static <E extends Enum<E>> EnumSet<E> makeComplementByHand(Collection<E> collection, Class<E> type)
  {
    EnumSet<E> result = EnumSet.allOf(type);
    result.removeAll(collection);
    return result;
  }
  








































  public static <E> Set<E> newSetFromMap(Map<E, Boolean> map) { return new SetFromMap(map); }
  
  private static class SetFromMap<E> extends AbstractSet<E> implements Set<E>, Serializable {
    private final Map<E, Boolean> m;
    private transient Set<E> s;
    @GwtIncompatible("not needed in emulated source")
    private static final long serialVersionUID = 0L;
    
    SetFromMap(Map<E, Boolean> map) {
      Preconditions.checkArgument(map.isEmpty(), "Map is non-empty");
      m = map;
      s = map.keySet();
    }
    
    public void clear() {
      m.clear();
    }
    
    public int size() { return m.size(); }
    
    public boolean isEmpty() {
      return m.isEmpty();
    }
    
    public boolean contains(Object o) { return m.containsKey(o); }
    
    public boolean remove(Object o) {
      return m.remove(o) != null;
    }
    
    public boolean add(E e) { return m.put(e, Boolean.TRUE) == null; }
    
    public Iterator<E> iterator() {
      return s.iterator();
    }
    
    public Object[] toArray() { return s.toArray(); }
    
    public <T> T[] toArray(T[] a) {
      return s.toArray(a);
    }
    
    public String toString() { return s.toString(); }
    
    public int hashCode() {
      return s.hashCode();
    }
    
    public boolean equals(@Nullable Object object) { return (this == object) || (s.equals(object)); }
    
    public boolean containsAll(Collection<?> c) {
      return s.containsAll(c);
    }
    
    public boolean removeAll(Collection<?> c) { return s.removeAll(c); }
    
    public boolean retainAll(Collection<?> c) {
      return s.retainAll(c);
    }
    



    @GwtIncompatible("java.io.ObjectInputStream")
    private void readObject(ObjectInputStream stream)
      throws IOException, ClassNotFoundException
    {
      stream.defaultReadObject();
      s = m.keySet();
    }
  }
  


















  public static abstract class SetView<E>
    extends AbstractSet<E>
  {
    public ImmutableSet<E> immutableCopy()
    {
      return ImmutableSet.copyOf(this);
    }
    








    public <S extends Set<E>> S copyInto(S set)
    {
      set.addAll(this);
      return set;
    }
  }
  















  public static <E> SetView<E> union(Set<? extends E> set1, final Set<? extends E> set2)
  {
    Preconditions.checkNotNull(set1, "set1");
    Preconditions.checkNotNull(set2, "set2");
    
    final Set<? extends E> set2minus1 = difference(set2, set1);
    
    new SetView(set1) {
      public int size() {
        return val$set1.size() + set2minus1.size();
      }
      
      public boolean isEmpty() { return (val$set1.isEmpty()) && (set2.isEmpty()); }
      
      public Iterator<E> iterator() {
        return Iterators.unmodifiableIterator(Iterators.concat(val$set1.iterator(), set2minus1.iterator()));
      }
      

      public boolean contains(Object object) { return (val$set1.contains(object)) || (set2.contains(object)); }
      
      public <S extends Set<E>> S copyInto(S set) {
        set.addAll(val$set1);
        set.addAll(set2);
        return set;
      }
      
      public ImmutableSet<E> immutableCopy() { return new ImmutableSet.Builder().addAll(val$set1).addAll(set2).build(); }
    };
  }
  




























  public static <E> SetView<E> intersection(Set<E> set1, final Set<?> set2)
  {
    Preconditions.checkNotNull(set1, "set1");
    Preconditions.checkNotNull(set2, "set2");
    
    final Predicate<Object> inSet2 = Predicates.in(set2);
    new SetView(set1) {
      public Iterator<E> iterator() {
        return Iterators.filter(val$set1.iterator(), inSet2);
      }
      
      public int size() { return Iterators.size(iterator()); }
      
      public boolean isEmpty() {
        return !iterator().hasNext();
      }
      
      public boolean contains(Object object) { return (val$set1.contains(object)) && (set2.contains(object)); }
      
      public boolean containsAll(Collection<?> collection) {
        return (val$set1.containsAll(collection)) && (set2.containsAll(collection));
      }
    };
  }
  












  public static <E> SetView<E> difference(Set<E> set1, final Set<?> set2)
  {
    Preconditions.checkNotNull(set1, "set1");
    Preconditions.checkNotNull(set2, "set2");
    
    final Predicate<Object> notInSet2 = Predicates.not(Predicates.in(set2));
    new SetView(set1) {
      public Iterator<E> iterator() {
        return Iterators.filter(val$set1.iterator(), notInSet2);
      }
      
      public int size() { return Iterators.size(iterator()); }
      
      public boolean isEmpty() {
        return set2.containsAll(val$set1);
      }
      
      public boolean contains(Object element) { return (val$set1.contains(element)) && (!set2.contains(element)); }
    };
  }
  













  public static <E> SetView<E> symmetricDifference(Set<? extends E> set1, Set<? extends E> set2)
  {
    Preconditions.checkNotNull(set1, "set1");
    Preconditions.checkNotNull(set2, "set2");
    

    return difference(union(set1, set2), intersection(set1, set2));
  }
  



























  public static <E> Set<E> filter(Set<E> unfiltered, Predicate<? super E> predicate)
  {
    if ((unfiltered instanceof FilteredSet))
    {

      FilteredSet<E> filtered = (FilteredSet)unfiltered;
      Predicate<E> combinedPredicate = Predicates.and(predicate, predicate);
      
      return new FilteredSet((Set)unfiltered, combinedPredicate);
    }
    

    return new FilteredSet((Set)Preconditions.checkNotNull(unfiltered), (Predicate)Preconditions.checkNotNull(predicate));
  }
  
  private static class FilteredSet<E> extends Collections2.FilteredCollection<E> implements Set<E>
  {
    FilteredSet(Set<E> unfiltered, Predicate<? super E> predicate)
    {
      super(predicate);
    }
    
    public boolean equals(@Nullable Object object) {
      return Sets.equalsImpl(this, object);
    }
    
    public int hashCode() {
      return Sets.hashCodeImpl(this);
    }
  }
  













































  public static <B> Set<List<B>> cartesianProduct(List<? extends Set<? extends B>> sets)
  {
    for (Set<? extends B> set : sets) {
      if (set.isEmpty()) {
        return ImmutableSet.of();
      }
    }
    CartesianSet<B> cartesianSet = new CartesianSet(sets);
    return cartesianSet;
  }
  













































  public static <B> Set<List<B>> cartesianProduct(Set<? extends B>... sets)
  {
    return cartesianProduct(Arrays.asList(sets));
  }
  
  private static class CartesianSet<B> extends AbstractSet<List<B>> {
    final ImmutableList<CartesianSet<B>.Axis> axes;
    final int size;
    
    CartesianSet(List<? extends Set<? extends B>> sets) {
      long dividend = 1L;
      ImmutableList.Builder<CartesianSet<B>.Axis> builder = ImmutableList.builder();
      for (Set<? extends B> set : sets) {
        CartesianSet<B>.Axis axis = new Axis(set, (int)dividend);
        builder.add(axis);
        dividend *= axis.size();
        Preconditions.checkArgument(dividend <= 2147483647L, "cartesian product is too big");
      }
      
      axes = builder.build();
      size = Ints.checkedCast(dividend);
    }
    
    public int size() {
      return size;
    }
    
    public UnmodifiableIterator<List<B>> iterator() {
      new UnmodifiableIterator()
      {
        int index;
        
        public boolean hasNext() {
          return index < size;
        }
        
        public List<B> next()
        {
          if (!hasNext()) {
            throw new NoSuchElementException();
          }
          
          Object[] tuple = new Object[axes.size()];
          for (int i = 0; i < tuple.length; i++) {
            tuple[i] = ((Sets.CartesianSet.Axis)axes.get(i)).getForIndex(index);
          }
          index += 1;
          

          List<B> result = ImmutableList.copyOf(tuple);
          return result;
        }
      };
    }
    
    public boolean contains(Object element) {
      if (!(element instanceof List)) {
        return false;
      }
      List<?> tuple = (List)element;
      int dimensions = axes.size();
      if (tuple.size() != dimensions) {
        return false;
      }
      for (int i = 0; i < dimensions; i++) {
        if (!((Axis)axes.get(i)).contains(tuple.get(i))) {
          return false;
        }
      }
      return true;
    }
    

    public boolean equals(@Nullable Object object)
    {
      if ((object instanceof CartesianSet)) {
        CartesianSet<?> that = (CartesianSet)object;
        return axes.equals(axes);
      }
      return super.equals(object);
    }
    



    public int hashCode()
    {
      int adjust = size - 1;
      for (int i = 0; i < axes.size(); i++) {
        adjust *= 31;
      }
      return axes.hashCode() + adjust;
    }
    
    private class Axis {
      final ImmutableSet<? extends B> choices;
      final ImmutableList<? extends B> choicesList;
      final int dividend;
      
      Axis(int set) {
        choices = ImmutableSet.copyOf(set);
        choicesList = choices.asList();
        this.dividend = dividend;
      }
      
      int size() {
        return choices.size();
      }
      
      B getForIndex(int index) {
        return choicesList.get(index / dividend % size());
      }
      
      boolean contains(Object target) {
        return choices.contains(target);
      }
      
      public boolean equals(Object obj) {
        if ((obj instanceof Axis)) {
          Axis that = (Axis)obj;
          return choices.equals(choices);
        }
        
        return false;
      }
      


      public int hashCode()
      {
        return size / choices.size() * choices.hashCode();
      }
    }
  }
  




























  @GwtCompatible(serializable=false)
  public static <E> Set<Set<E>> powerSet(Set<E> set)
  {
    ImmutableSet<E> input = ImmutableSet.copyOf(set);
    Preconditions.checkArgument(input.size() <= 30, "Too many elements to create power set: %s > 30", new Object[] { Integer.valueOf(input.size()) });
    
    return new PowerSet(input);
  }
  
  private static final class PowerSet<E> extends AbstractSet<Set<E>> {
    final ImmutableSet<E> inputSet;
    final ImmutableList<E> inputList;
    final int powerSetSize;
    
    PowerSet(ImmutableSet<E> input) {
      inputSet = input;
      inputList = input.asList();
      powerSetSize = (1 << input.size());
    }
    
    public int size() {
      return powerSetSize;
    }
    
    public boolean isEmpty() {
      return false;
    }
    
    public Iterator<Set<E>> iterator() {
      new AbstractIndexedListIterator(powerSetSize) {
        protected Set<E> get(final int setBits) {
          new AbstractSet() {
            public int size() {
              return Integer.bitCount(setBits);
            }
            
            public Iterator<E> iterator() { return new Sets.PowerSet.BitFilteredSetIterator(inputList, setBits); }
          };
        }
      };
    }
    
    private static final class BitFilteredSetIterator<E> extends UnmodifiableIterator<E>
    {
      final ImmutableList<E> input;
      int remainingSetBits;
      
      BitFilteredSetIterator(ImmutableList<E> input, int allSetBits)
      {
        this.input = input;
        remainingSetBits = allSetBits;
      }
      
      public boolean hasNext() {
        return remainingSetBits != 0;
      }
      
      public E next() {
        int index = Integer.numberOfTrailingZeros(remainingSetBits);
        if (index == 32) {
          throw new NoSuchElementException();
        }
        
        int currentElementMask = 1 << index;
        remainingSetBits &= (currentElementMask ^ 0xFFFFFFFF);
        return input.get(index);
      }
    }
    
    public boolean contains(@Nullable Object obj) {
      if ((obj instanceof Set)) {
        Set<?> set = (Set)obj;
        return inputSet.containsAll(set);
      }
      return false;
    }
    
    public boolean equals(@Nullable Object obj) {
      if ((obj instanceof PowerSet)) {
        PowerSet<?> that = (PowerSet)obj;
        return inputSet.equals(inputSet);
      }
      return super.equals(obj);
    }
    




    public int hashCode()
    {
      return inputSet.hashCode() << inputSet.size() - 1;
    }
    
    public String toString() {
      return "powerSet(" + inputSet + ")";
    }
  }
  


  static int hashCodeImpl(Set<?> s)
  {
    int hashCode = 0;
    for (Object o : s) {
      hashCode += (o != null ? o.hashCode() : 0);
    }
    return hashCode;
  }
  


  static boolean equalsImpl(Set<?> s, @Nullable Object object)
  {
    if (s == object) {
      return true;
    }
    if ((object instanceof Set)) {
      Set<?> o = (Set)object;
      try
      {
        return (s.size() == o.size()) && (s.containsAll(o));
      } catch (NullPointerException ignored) {
        return false;
      } catch (ClassCastException ignored) {
        return false;
      }
    }
    return false;
  }
  












  static <A, B> Set<B> transform(Set<A> set, InvertibleFunction<A, B> bijection)
  {
    return new TransformedSet((Set)Preconditions.checkNotNull(set, "set"), (InvertibleFunction)Preconditions.checkNotNull(bijection, "bijection"));
  }
  


  static abstract class InvertibleFunction<A, B>
    implements Function<A, B>
  {
    abstract A invert(B paramB);
    

    public InvertibleFunction<B, A> inverse()
    {
      new InvertibleFunction() {
        public A apply(B b) {
          return Sets.InvertibleFunction.this.invert(b);
        }
        
        B invert(A a) {
          return Sets.InvertibleFunction.this.apply(a);
        }
        
        public Sets.InvertibleFunction<A, B> inverse()
        {
          return Sets.InvertibleFunction.this;
        }
      };
    }
  }
  
  private static class TransformedSet<A, B> extends AbstractSet<B> {
    final Set<A> delegate;
    final Sets.InvertibleFunction<A, B> bijection;
    
    TransformedSet(Set<A> delegate, Sets.InvertibleFunction<A, B> bijection) {
      this.delegate = delegate;
      this.bijection = bijection;
    }
    
    public Iterator<B> iterator() {
      return Iterators.transform(delegate.iterator(), bijection);
    }
    
    public int size() {
      return delegate.size();
    }
    
    public boolean contains(Object o)
    {
      B b = o;
      A a = bijection.invert(b);
      




















      return (delegate.contains(a)) && (Objects.equal(bijection.apply(a), o));
    }
    
    public boolean add(B b) {
      return delegate.add(bijection.invert(b));
    }
    
    public boolean remove(Object o)
    {
      return (contains(o)) && (delegate.remove(bijection.invert(o)));
    }
    
    public void clear() {
      delegate.clear();
    }
  }
}
