package com.google.common.collect;

import com.google.common.base.Preconditions;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.concurrent.atomic.AtomicReference;
import javax.annotation.Nullable;








































final class SortedTreeMultiset<E>
  extends AbstractSortedMultiset<E>
{
  private final GeneralRange<E> range;
  private final AtomicReference<SortedTreeMultiset<E>.Node> rootReference;
  
  public static <E extends Comparable> SortedTreeMultiset<E> create()
  {
    return new SortedTreeMultiset(Ordering.natural());
  }
  


  public static <E> SortedTreeMultiset<E> create(Comparator<? super E> comparator)
  {
    return new SortedTreeMultiset(comparator);
  }
  




  public static <E extends Comparable> SortedTreeMultiset<E> create(Iterable<? extends E> elements)
  {
    SortedTreeMultiset<E> multiset = create();
    Iterables.addAll(multiset, elements);
    return multiset;
  }
  
  private SortedTreeMultiset(Comparator<? super E> comparator) {
    super(comparator);
    range = GeneralRange.all(comparator);
    rootReference = new AtomicReference();
  }
  
  private SortedTreeMultiset(GeneralRange<E> range, AtomicReference<SortedTreeMultiset<E>.Node> root) {
    super(range.comparator());
    this.range = range;
    rootReference = root;
  }
  
  E checkElement(Object o)
  {
    Preconditions.checkNotNull(o);
    return o;
  }
  




  int distinctElements()
  {
    SortedTreeMultiset<E>.Node root = (Node)rootReference.get();
    return BstRangeOps.totalInRange(distinctAggregate, range, root);
  }
  
  public int size()
  {
    SortedTreeMultiset<E>.Node root = (Node)rootReference.get();
    return BstRangeOps.totalInRange(sizeAggregate, range, root);
  }
  
  public int count(@Nullable Object element)
  {
    if (element == null) {
      return 0;
    }
    try {
      E e = checkElement(element);
      if (range.contains(e)) {
        SortedTreeMultiset<E>.Node node = (Node)BstOperations.seek(comparator(), (BstNode)rootReference.get(), e);
        return node == null ? 0 : elemOccurrences;
      }
      return 0;
    } catch (ClassCastException e) {}
    return 0;
  }
  
  private int mutate(E e, SortedTreeMultiset<E>.MultisetModifier modifier)
  {
    BstMutationRule<E, SortedTreeMultiset<E>.Node> mutationRule = BstMutationRule.createRule(modifier, BstCountBasedBalancePolicies.singleRebalancePolicy(distinctAggregate), nodeFactory);
    

    BstMutationResult<E, SortedTreeMultiset<E>.Node> mutationResult = BstOperations.mutate(comparator(), mutationRule, (BstNode)rootReference.get(), e);
    
    if (!rootReference.compareAndSet(mutationResult.getOriginalRoot(), mutationResult.getChangedRoot()))
    {
      throw new ConcurrentModificationException();
    }
    SortedTreeMultiset<E>.Node original = (Node)mutationResult.getOriginalTarget();
    return original == null ? 0 : elemOccurrences;
  }
  
  public int add(E element, int occurrences)
  {
    Preconditions.checkNotNull(element);
    if (occurrences == 0) {
      return count(element);
    }
    Preconditions.checkArgument(range.contains(element));
    return mutate(element, new AddModifier(occurrences, null));
  }
  
  public int remove(@Nullable Object element, int occurrences)
  {
    if (element == null)
      return 0;
    if (occurrences == 0) {
      return count(element);
    }
    try {
      E e = checkElement(element);
      return range.contains(e) ? mutate(e, new RemoveModifier(occurrences, null)) : 0;
    } catch (ClassCastException e) {}
    return 0;
  }
  

  public boolean setCount(E element, int oldCount, int newCount)
  {
    Preconditions.checkNotNull(element);
    Preconditions.checkArgument(range.contains(element));
    return mutate(element, new ConditionalSetCountModifier(oldCount, newCount, null)) == oldCount;
  }
  
  public int setCount(E element, int count)
  {
    Preconditions.checkNotNull(element);
    Preconditions.checkArgument(range.contains(element));
    return mutate(element, new SetCountModifier(count, null));
  }
  
  private final transient BstPathFactory<SortedTreeMultiset<E>.Node, BstInOrderPath<SortedTreeMultiset<E>.Node>> pathFactory = BstInOrderPath.inOrderFactory();
  

  Iterator<Multiset.Entry<E>> entryIterator()
  {
    SortedTreeMultiset<E>.Node root = (Node)rootReference.get();
    BstInOrderPath<SortedTreeMultiset<E>.Node> startingPath = (BstInOrderPath)BstRangeOps.furthestPath(range, BstSide.LEFT, pathFactory, root);
    
    return iteratorInDirection(startingPath, BstSide.RIGHT);
  }
  
  Iterator<Multiset.Entry<E>> descendingEntryIterator()
  {
    SortedTreeMultiset<E>.Node root = (Node)rootReference.get();
    BstInOrderPath<SortedTreeMultiset<E>.Node> startingPath = (BstInOrderPath)BstRangeOps.furthestPath(range, BstSide.RIGHT, pathFactory, root);
    
    return iteratorInDirection(startingPath, BstSide.LEFT);
  }
  
  private Iterator<Multiset.Entry<E>> iteratorInDirection(@Nullable BstInOrderPath<SortedTreeMultiset<E>.Node> start, final BstSide direction)
  {
    final Iterator<BstInOrderPath<SortedTreeMultiset<E>.Node>> pathIterator = new AbstractLinkedIterator(start)
    {
      protected BstInOrderPath<SortedTreeMultiset<E>.Node> computeNext(BstInOrderPath<SortedTreeMultiset<E>.Node> previous)
      {
        if (!previous.hasNext(direction)) {
          return null;
        }
        BstInOrderPath<SortedTreeMultiset<E>.Node> next = previous.next(direction);
        
        return range.contains(((SortedTreeMultiset.Node)next.getTip()).getKey()) ? next : null;
      }
    };
    new Iterator() {
      E toRemove = null;
      
      public boolean hasNext()
      {
        return pathIterator.hasNext();
      }
      
      public Multiset.Entry<E> next()
      {
        BstInOrderPath<SortedTreeMultiset<E>.Node> path = (BstInOrderPath)pathIterator.next();
        return Multisets.immutableEntry(this.toRemove = ((SortedTreeMultiset.Node)path.getTip()).getKey(), getTip()elemOccurrences);
      }
      

      public void remove()
      {
        Preconditions.checkState(toRemove != null);
        setCount(toRemove, 0);
        toRemove = null;
      }
    };
  }
  
  public void clear()
  {
    SortedTreeMultiset<E>.Node root = (Node)rootReference.get();
    SortedTreeMultiset<E>.Node cleared = (Node)BstRangeOps.minusRange(range, BstCountBasedBalancePolicies.fullRebalancePolicy(distinctAggregate), nodeFactory, root);
    

    if (!rootReference.compareAndSet(root, cleared)) {
      throw new ConcurrentModificationException();
    }
  }
  
  public SortedMultiset<E> headMultiset(E upperBound, BoundType boundType)
  {
    Preconditions.checkNotNull(upperBound);
    return new SortedTreeMultiset(range.intersect(GeneralRange.upTo(comparator, upperBound, boundType)), rootReference);
  }
  

  public SortedMultiset<E> tailMultiset(E lowerBound, BoundType boundType)
  {
    Preconditions.checkNotNull(lowerBound);
    return new SortedTreeMultiset(range.intersect(GeneralRange.downTo(comparator, lowerBound, boundType)), rootReference);
  }
  
  private final class Node extends BstNode<E, SortedTreeMultiset<E>.Node>
  {
    private final int elemOccurrences;
    private final int size;
    private final int distinct;
    
    private Node(int key, @Nullable SortedTreeMultiset<E>.Node elemCount, @Nullable SortedTreeMultiset<E>.Node left) {
      super(left, right);
      Preconditions.checkArgument(elemCount > 0);
      elemOccurrences = elemCount;
      size = (elemCount + SortedTreeMultiset.this.sizeOrZero(left) + SortedTreeMultiset.this.sizeOrZero(right));
      distinct = (1 + SortedTreeMultiset.this.distinctOrZero(left) + SortedTreeMultiset.this.distinctOrZero(right));
    }
    
    private Node(int key) {
      this(key, elemCount, null, null);
    }
  }
  
  private int sizeOrZero(@Nullable SortedTreeMultiset<E>.Node node) {
    return node == null ? 0 : size;
  }
  
  private int distinctOrZero(@Nullable SortedTreeMultiset<E>.Node node) {
    return node == null ? 0 : distinct;
  }
  
  private final transient BstAggregate<SortedTreeMultiset<E>.Node> distinctAggregate = new BstAggregate()
  {
    public int entryValue(SortedTreeMultiset<E>.Node entry) {
      return 1;
    }
    
    public int treeValue(@Nullable SortedTreeMultiset<E>.Node tree)
    {
      return SortedTreeMultiset.this.distinctOrZero(tree);
    }
  };
  
  private final transient BstAggregate<SortedTreeMultiset<E>.Node> sizeAggregate = new BstAggregate()
  {
    public int entryValue(SortedTreeMultiset<E>.Node entry) {
      return elemOccurrences;
    }
    
    public int treeValue(@Nullable SortedTreeMultiset<E>.Node tree)
    {
      return SortedTreeMultiset.this.sizeOrZero(tree);
    }
  };
  
  private final transient BstNodeFactory<SortedTreeMultiset<E>.Node> nodeFactory = new BstNodeFactory()
  {

    public SortedTreeMultiset<E>.Node createNode(SortedTreeMultiset<E>.Node source, @Nullable SortedTreeMultiset<E>.Node left, @Nullable SortedTreeMultiset<E>.Node right) { return new SortedTreeMultiset.Node(SortedTreeMultiset.this, source.getKey(), elemOccurrences, left, right, null); }
  };
  
  private abstract class MultisetModifier implements BstModifier<E, SortedTreeMultiset<E>.Node> {
    private MultisetModifier() {}
    
    abstract int newCount(int paramInt);
    
    @Nullable
    public BstModificationResult<SortedTreeMultiset<E>.Node> modify(E key, @Nullable SortedTreeMultiset<E>.Node originalEntry) {
      int oldCount = originalEntry == null ? 0 : SortedTreeMultiset.Node.access$000(originalEntry);
      int newCount = newCount(oldCount);
      if (oldCount == newCount)
        return BstModificationResult.identity(originalEntry);
      if (newCount == 0)
        return BstModificationResult.rebalancingChange(originalEntry, null);
      if (oldCount == 0) {
        return BstModificationResult.rebalancingChange(null, new SortedTreeMultiset.Node(SortedTreeMultiset.this, key, newCount, null));
      }
      return BstModificationResult.rebuildingChange(originalEntry, new SortedTreeMultiset.Node(SortedTreeMultiset.this, key, newCount, null));
    }
  }
  
  private final class AddModifier extends SortedTreeMultiset.MultisetModifier {
    private final int countToAdd;
    
    private AddModifier(int countToAdd) {
      super(null);
      Preconditions.checkArgument(countToAdd > 0);
      this.countToAdd = countToAdd;
    }
    


    int newCount(int oldCount) { return oldCount + countToAdd; }
  }
  
  private final class RemoveModifier extends SortedTreeMultiset.MultisetModifier {
    private final int countToRemove;
    
    private RemoveModifier(int countToRemove) {
      super(null);
      Preconditions.checkArgument(countToRemove > 0);
      this.countToRemove = countToRemove;
    }
    


    int newCount(int oldCount) { return Math.max(0, oldCount - countToRemove); }
  }
  
  private final class SetCountModifier extends SortedTreeMultiset.MultisetModifier {
    private final int countToSet;
    
    private SetCountModifier(int countToSet) {
      super(null);
      Preconditions.checkArgument(countToSet >= 0);
      this.countToSet = countToSet;
    }
    


    int newCount(int oldCount) { return countToSet; }
  }
  
  private final class ConditionalSetCountModifier extends SortedTreeMultiset.MultisetModifier {
    private final int expectedCount;
    private final int setCount;
    
    private ConditionalSetCountModifier(int expectedCount, int setCount) {
      super(null);
      Preconditions.checkArgument((setCount >= 0 ? 1 : 0) & (expectedCount >= 0 ? 1 : 0));
      this.expectedCount = expectedCount;
      this.setCount = setCount;
    }
    
    int newCount(int oldCount)
    {
      return oldCount == expectedCount ? setCount : oldCount;
    }
  }
}
