package com.google.common.collect;

import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import com.google.common.base.Predicates;
import com.google.common.base.Supplier;
import java.io.Serializable;
import java.util.AbstractCollection;
import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;



































@GwtCompatible
class StandardTable<R, C, V>
  implements Table<R, C, V>, Serializable
{
  final Map<R, Map<C, V>> backingMap;
  final Supplier<? extends Map<C, V>> factory;
  private transient StandardTable<R, C, V>.CellSet cellSet;
  private transient StandardTable<R, C, V>.RowKeySet rowKeySet;
  private transient Set<C> columnKeySet;
  private transient StandardTable<R, C, V>.Values values;
  private transient StandardTable<R, C, V>.RowMap rowMap;
  private transient StandardTable<R, C, V>.ColumnMap columnMap;
  private static final long serialVersionUID = 0L;
  
  StandardTable(Map<R, Map<C, V>> backingMap, Supplier<? extends Map<C, V>> factory)
  {
    this.backingMap = backingMap;
    this.factory = factory;
  }
  


  public boolean contains(@Nullable Object rowKey, @Nullable Object columnKey)
  {
    if ((rowKey == null) || (columnKey == null)) {
      return false;
    }
    Map<C, V> map = (Map)Maps.safeGet(backingMap, rowKey);
    return (map != null) && (Maps.safeContainsKey(map, columnKey));
  }
  
  public boolean containsColumn(@Nullable Object columnKey) {
    if (columnKey == null) {
      return false;
    }
    for (Map<C, V> map : backingMap.values()) {
      if (Maps.safeContainsKey(map, columnKey)) {
        return true;
      }
    }
    return false;
  }
  
  public boolean containsRow(@Nullable Object rowKey) {
    return (rowKey != null) && (Maps.safeContainsKey(backingMap, rowKey));
  }
  
  public boolean containsValue(@Nullable Object value) {
    if (value == null) {
      return false;
    }
    for (Map<C, V> map : backingMap.values()) {
      if (map.containsValue(value)) {
        return true;
      }
    }
    return false;
  }
  
  public V get(@Nullable Object rowKey, @Nullable Object columnKey) {
    if ((rowKey == null) || (columnKey == null)) {
      return null;
    }
    Map<C, V> map = (Map)Maps.safeGet(backingMap, rowKey);
    return map == null ? null : Maps.safeGet(map, columnKey);
  }
  
  public boolean isEmpty() {
    return backingMap.isEmpty();
  }
  
  public int size() {
    int size = 0;
    for (Map<C, V> map : backingMap.values()) {
      size += map.size();
    }
    return size;
  }
  
  public boolean equals(@Nullable Object obj) {
    if (obj == this) {
      return true;
    }
    if ((obj instanceof Table)) {
      Table<?, ?, ?> other = (Table)obj;
      return cellSet().equals(other.cellSet());
    }
    return false;
  }
  
  public int hashCode() {
    return cellSet().hashCode();
  }
  


  public String toString()
  {
    return rowMap().toString();
  }
  

  public void clear()
  {
    backingMap.clear();
  }
  
  private Map<C, V> getOrCreate(R rowKey) {
    Map<C, V> map = (Map)backingMap.get(rowKey);
    if (map == null) {
      map = (Map)factory.get();
      backingMap.put(rowKey, map);
    }
    return map;
  }
  
  public V put(R rowKey, C columnKey, V value) {
    Preconditions.checkNotNull(rowKey);
    Preconditions.checkNotNull(columnKey);
    Preconditions.checkNotNull(value);
    return getOrCreate(rowKey).put(columnKey, value);
  }
  
  public void putAll(Table<? extends R, ? extends C, ? extends V> table)
  {
    for (Table.Cell<? extends R, ? extends C, ? extends V> cell : table.cellSet()) {
      put(cell.getRowKey(), cell.getColumnKey(), cell.getValue());
    }
  }
  
  public V remove(@Nullable Object rowKey, @Nullable Object columnKey)
  {
    if ((rowKey == null) || (columnKey == null)) {
      return null;
    }
    Map<C, V> map = (Map)Maps.safeGet(backingMap, rowKey);
    if (map == null) {
      return null;
    }
    V value = map.remove(columnKey);
    if (map.isEmpty()) {
      backingMap.remove(rowKey);
    }
    return value;
  }
  
  private Map<R, V> removeColumn(Object column) {
    Map<R, V> output = new LinkedHashMap();
    Iterator<Map.Entry<R, Map<C, V>>> iterator = backingMap.entrySet().iterator();
    
    while (iterator.hasNext()) {
      Map.Entry<R, Map<C, V>> entry = (Map.Entry)iterator.next();
      V value = ((Map)entry.getValue()).remove(column);
      if (value != null) {
        output.put(entry.getKey(), value);
        if (((Map)entry.getValue()).isEmpty()) {
          iterator.remove();
        }
      }
    }
    return output;
  }
  
  private boolean containsMapping(Object rowKey, Object columnKey, Object value)
  {
    return (value != null) && (value.equals(get(rowKey, columnKey)));
  }
  
  private boolean removeMapping(Object rowKey, Object columnKey, Object value)
  {
    if (containsMapping(rowKey, columnKey, value)) {
      remove(rowKey, columnKey);
      return true;
    }
    return false;
  }
  

  private abstract class TableCollection<T>
    extends AbstractCollection<T>
  {
    private TableCollection() {}
    
    public boolean isEmpty()
    {
      return backingMap.isEmpty();
    }
    
    public void clear() {
      backingMap.clear();
    }
  }
  
  private abstract class TableSet<T> extends AbstractSet<T>
  {
    private TableSet() {}
    
    public boolean isEmpty()
    {
      return backingMap.isEmpty();
    }
    
    public void clear() {
      backingMap.clear();
    }
  }
  











  public Set<Table.Cell<R, C, V>> cellSet()
  {
    StandardTable<R, C, V>.CellSet result = cellSet;
    return result == null ? (this.cellSet = new CellSet(null)) : result;
  }
  
  private class CellSet extends StandardTable<R, C, V>.TableSet<Table.Cell<R, C, V>> { private CellSet() { super(null); }
    
    public Iterator<Table.Cell<R, C, V>> iterator() { return new StandardTable.CellIterator(StandardTable.this, null); }
    
    public int size()
    {
      return StandardTable.this.size();
    }
    
    public boolean contains(Object obj) {
      if ((obj instanceof Table.Cell)) {
        Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
        return StandardTable.this.containsMapping(cell.getRowKey(), cell.getColumnKey(), cell.getValue());
      }
      
      return false;
    }
    
    public boolean remove(Object obj) {
      if ((obj instanceof Table.Cell)) {
        Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
        return StandardTable.this.removeMapping(cell.getRowKey(), cell.getColumnKey(), cell.getValue());
      }
      
      return false;
    }
  }
  
  private class CellIterator implements Iterator<Table.Cell<R, C, V>> {
    final Iterator<Map.Entry<R, Map<C, V>>> rowIterator = backingMap.entrySet().iterator();
    
    Map.Entry<R, Map<C, V>> rowEntry;
    Iterator<Map.Entry<C, V>> columnIterator = Iterators.emptyModifiableIterator();
    
    private CellIterator() {}
    
    public boolean hasNext() { return (rowIterator.hasNext()) || (columnIterator.hasNext()); }
    
    public Table.Cell<R, C, V> next()
    {
      if (!columnIterator.hasNext()) {
        rowEntry = ((Map.Entry)rowIterator.next());
        columnIterator = ((Map)rowEntry.getValue()).entrySet().iterator();
      }
      Map.Entry<C, V> columnEntry = (Map.Entry)columnIterator.next();
      return Tables.immutableCell(rowEntry.getKey(), columnEntry.getKey(), columnEntry.getValue());
    }
    
    public void remove()
    {
      columnIterator.remove();
      if (((Map)rowEntry.getValue()).isEmpty()) {
        rowIterator.remove();
      }
    }
  }
  

  public Map<C, V> row(R rowKey) { return new Row(rowKey); }
  
  class Row extends AbstractMap<C, V> { final R rowKey;
    Map<C, V> backingRowMap;
    Set<C> keySet;
    Set<Map.Entry<C, V>> entrySet;
    
    Row() { this.rowKey = Preconditions.checkNotNull(rowKey); }
    


    Map<C, V> backingRowMap()
    {
      return (backingRowMap == null) || ((backingRowMap.isEmpty()) && (backingMap.containsKey(rowKey))) ? (this.backingRowMap = computeBackingRowMap()) : backingRowMap;
    }
    


    Map<C, V> computeBackingRowMap()
    {
      return (Map)backingMap.get(rowKey);
    }
    
    void maintainEmptyInvariant()
    {
      if ((backingRowMap() != null) && (backingRowMap.isEmpty())) {
        backingMap.remove(rowKey);
        backingRowMap = null;
      }
    }
    
    public boolean containsKey(Object key)
    {
      Map<C, V> backingRowMap = backingRowMap();
      return (key != null) && (backingRowMap != null) && (Maps.safeContainsKey(backingRowMap, key));
    }
    

    public V get(Object key)
    {
      Map<C, V> backingRowMap = backingRowMap();
      return (key != null) && (backingRowMap != null) ? Maps.safeGet(backingRowMap, key) : null;
    }
    


    public V put(C key, V value)
    {
      Preconditions.checkNotNull(key);
      Preconditions.checkNotNull(value);
      if ((backingRowMap != null) && (!backingRowMap.isEmpty())) {
        return backingRowMap.put(key, value);
      }
      return put(rowKey, key, value);
    }
    
    public V remove(Object key)
    {
      try {
        Map<C, V> backingRowMap = backingRowMap();
        if (backingRowMap == null) {
          return null;
        }
        V result = backingRowMap.remove(key);
        maintainEmptyInvariant();
        return result;
      } catch (ClassCastException e) {}
      return null;
    }
    

    public void clear()
    {
      Map<C, V> backingRowMap = backingRowMap();
      if (backingRowMap != null) {
        backingRowMap.clear();
      }
      maintainEmptyInvariant();
    }
    


    public Set<C> keySet()
    {
      Set<C> result = keySet;
      if (result == null) {
        this.keySet = new Maps.KeySet()
        {
          Map<C, V> map() {
            return StandardTable.Row.this;
          }
        };
      }
      return result;
    }
    


    public Set<Map.Entry<C, V>> entrySet()
    {
      Set<Map.Entry<C, V>> result = entrySet;
      if (result == null) {
        return this.entrySet = new RowEntrySet(null);
      }
      return result;
    }
    
    private class RowEntrySet extends Maps.EntrySet<C, V> {
      private RowEntrySet() {}
      
      Map<C, V> map() { return StandardTable.Row.this; }
      

      public int size()
      {
        Map<C, V> map = backingRowMap();
        return map == null ? 0 : map.size();
      }
      
      public Iterator<Map.Entry<C, V>> iterator()
      {
        Map<C, V> map = backingRowMap();
        if (map == null) {
          return Iterators.emptyModifiableIterator();
        }
        final Iterator<Map.Entry<C, V>> iterator = map.entrySet().iterator();
        new Iterator()
        {
          public boolean hasNext() { return iterator.hasNext(); }
          
          public Map.Entry<C, V> next() {
            final Map.Entry<C, V> entry = (Map.Entry)iterator.next();
            new ForwardingMapEntry() {
              protected Map.Entry<C, V> delegate() {
                return entry;
              }
              
              public V setValue(V value) { return super.setValue(Preconditions.checkNotNull(value)); }
              

              public boolean equals(Object object)
              {
                return standardEquals(object);
              }
            };
          }
          
          public void remove()
          {
            iterator.remove();
            maintainEmptyInvariant();
          }
        };
      }
    }
  }
  







  public Map<R, V> column(C columnKey) { return new Column(columnKey); }
  
  private class Column extends Maps.ImprovedAbstractMap<R, V> {
    final C columnKey;
    StandardTable<R, C, V>.Column.Values columnValues;
    StandardTable<R, C, V>.Column.KeySet keySet;
    
    Column() { this.columnKey = Preconditions.checkNotNull(columnKey); }
    
    public V put(R key, V value)
    {
      return put(key, columnKey, value);
    }
    
    public V get(Object key) {
      return get(key, columnKey);
    }
    
    public boolean containsKey(Object key) {
      return contains(key, columnKey);
    }
    
    public V remove(Object key) {
      return remove(key, columnKey);
    }
    
    public Set<Map.Entry<R, V>> createEntrySet() {
      return new EntrySet();
    }
    

    public Collection<V> values()
    {
      StandardTable<R, C, V>.Column.Values result = columnValues;
      return result == null ? (this.columnValues = new Values()) : result;
    }
    



    boolean removePredicate(Predicate<? super Map.Entry<R, V>> predicate)
    {
      boolean changed = false;
      Iterator<Map.Entry<R, Map<C, V>>> iterator = backingMap.entrySet().iterator();
      
      while (iterator.hasNext()) {
        Map.Entry<R, Map<C, V>> entry = (Map.Entry)iterator.next();
        Map<C, V> map = (Map)entry.getValue();
        V value = map.get(columnKey);
        if ((value != null) && (predicate.apply(new ImmutableEntry(entry.getKey(), value))))
        {

          map.remove(columnKey);
          changed = true;
          if (map.isEmpty()) {
            iterator.remove();
          }
        }
      }
      return changed;
    }
    
    class EntrySet extends AbstractSet<Map.Entry<R, V>> { EntrySet() {}
      
      public Iterator<Map.Entry<R, V>> iterator() { return new StandardTable.Column.EntrySetIterator(StandardTable.Column.this); }
      
      public int size()
      {
        int size = 0;
        for (Map<C, V> map : backingMap.values()) {
          if (map.containsKey(columnKey)) {
            size++;
          }
        }
        return size;
      }
      
      public boolean isEmpty() {
        return !containsColumn(columnKey);
      }
      
      public void clear() {
        Predicate<Map.Entry<R, V>> predicate = Predicates.alwaysTrue();
        removePredicate(predicate);
      }
      
      public boolean contains(Object o) {
        if ((o instanceof Map.Entry)) {
          Map.Entry<?, ?> entry = (Map.Entry)o;
          return StandardTable.this.containsMapping(entry.getKey(), columnKey, entry.getValue());
        }
        return false;
      }
      
      public boolean remove(Object obj) {
        if ((obj instanceof Map.Entry)) {
          Map.Entry<?, ?> entry = (Map.Entry)obj;
          return StandardTable.this.removeMapping(entry.getKey(), columnKey, entry.getValue());
        }
        return false;
      }
      
      public boolean removeAll(Collection<?> c) {
        boolean changed = false;
        for (Object obj : c) {
          changed |= remove(obj);
        }
        return changed;
      }
      

      public boolean retainAll(Collection<?> c) { return removePredicate(Predicates.not(Predicates.in(c))); }
    }
    
    class EntrySetIterator extends AbstractIterator<Map.Entry<R, V>> { EntrySetIterator() {}
      
      final Iterator<Map.Entry<R, Map<C, V>>> iterator = backingMap.entrySet().iterator();
      
      protected Map.Entry<R, V> computeNext() {
        while (iterator.hasNext()) {
          final Map.Entry<R, Map<C, V>> entry = (Map.Entry)iterator.next();
          if (((Map)entry.getValue()).containsKey(columnKey)) {
            new AbstractMapEntry() {
              public R getKey() {
                return entry.getKey();
              }
              
              public V getValue() { return ((Map)entry.getValue()).get(columnKey); }
              
              public V setValue(V value) {
                return ((Map)entry.getValue()).put(columnKey, Preconditions.checkNotNull(value));
              }
            };
          }
        }
        return (Map.Entry)endOfData();
      }
    }
    

    public Set<R> keySet()
    {
      StandardTable<R, C, V>.Column.KeySet result = keySet;
      return result == null ? (this.keySet = new KeySet()) : result;
    }
    
    class KeySet extends AbstractSet<R> { KeySet() {}
      
      public Iterator<R> iterator() { return StandardTable.keyIteratorImpl(StandardTable.Column.this); }
      
      public int size()
      {
        return entrySet().size();
      }
      
      public boolean isEmpty() {
        return !containsColumn(columnKey);
      }
      
      public boolean contains(Object obj) {
        return contains(obj, columnKey);
      }
      
      public boolean remove(Object obj) {
        return remove(obj, columnKey) != null;
      }
      
      public void clear() {
        entrySet().clear();
      }
      
      public boolean removeAll(Collection<?> c) {
        boolean changed = false;
        for (Object obj : c) {
          changed |= remove(obj);
        }
        return changed;
      }
      
      public boolean retainAll(final Collection<?> c) {
        Preconditions.checkNotNull(c);
        Predicate<Map.Entry<R, V>> predicate = new Predicate()
        {
          public boolean apply(Map.Entry<R, V> entry) {
            return !c.contains(entry.getKey());
          }
        };
        return removePredicate(predicate);
      }
    }
    
    class Values extends AbstractCollection<V> { Values() {}
      
      public Iterator<V> iterator() { return StandardTable.valueIteratorImpl(StandardTable.Column.this); }
      
      public int size()
      {
        return entrySet().size();
      }
      
      public boolean isEmpty() {
        return !containsColumn(columnKey);
      }
      
      public void clear() {
        entrySet().clear();
      }
      
      public boolean remove(Object obj) {
        if (obj == null) {
          return false;
        }
        Iterator<Map<C, V>> iterator = backingMap.values().iterator();
        while (iterator.hasNext()) {
          Map<C, V> map = (Map)iterator.next();
          if (map.entrySet().remove(new ImmutableEntry(columnKey, obj)))
          {
            if (map.isEmpty()) {
              iterator.remove();
            }
            return true;
          }
        }
        return false;
      }
      
      public boolean removeAll(final Collection<?> c) {
        Preconditions.checkNotNull(c);
        Predicate<Map.Entry<R, V>> predicate = new Predicate()
        {
          public boolean apply(Map.Entry<R, V> entry) {
            return c.contains(entry.getValue());
          }
        };
        return removePredicate(predicate);
      }
      
      public boolean retainAll(final Collection<?> c) {
        Preconditions.checkNotNull(c);
        Predicate<Map.Entry<R, V>> predicate = new Predicate()
        {
          public boolean apply(Map.Entry<R, V> entry) {
            return !c.contains(entry.getValue());
          }
        };
        return removePredicate(predicate);
      }
    }
  }
  

  public Set<R> rowKeySet()
  {
    Set<R> result = rowKeySet;
    return result == null ? (this.rowKeySet = new RowKeySet()) : result;
  }
  
  class RowKeySet extends StandardTable<R, C, V>.TableSet<R> { RowKeySet() { super(null); }
    
    public Iterator<R> iterator() { return StandardTable.keyIteratorImpl(rowMap()); }
    
    public int size()
    {
      return backingMap.size();
    }
    
    public boolean contains(Object obj) {
      return containsRow(obj);
    }
    
    public boolean remove(Object obj) {
      return (obj != null) && (backingMap.remove(obj) != null);
    }
  }
  











  public Set<C> columnKeySet()
  {
    Set<C> result = columnKeySet;
    return result == null ? (this.columnKeySet = new ColumnKeySet(null)) : result;
  }
  
  private class ColumnKeySet extends StandardTable<R, C, V>.TableSet<C> { private ColumnKeySet() { super(null); }
    
    public Iterator<C> iterator() { return createColumnKeyIterator(); }
    
    public int size()
    {
      return Iterators.size(iterator());
    }
    
    public boolean remove(Object obj) {
      if (obj == null) {
        return false;
      }
      boolean changed = false;
      Iterator<Map<C, V>> iterator = backingMap.values().iterator();
      while (iterator.hasNext()) {
        Map<C, V> map = (Map)iterator.next();
        if (map.keySet().remove(obj)) {
          changed = true;
          if (map.isEmpty()) {
            iterator.remove();
          }
        }
      }
      return changed;
    }
    
    public boolean removeAll(Collection<?> c) {
      Preconditions.checkNotNull(c);
      boolean changed = false;
      Iterator<Map<C, V>> iterator = backingMap.values().iterator();
      while (iterator.hasNext()) {
        Map<C, V> map = (Map)iterator.next();
        

        if (Iterators.removeAll(map.keySet().iterator(), c)) {
          changed = true;
          if (map.isEmpty()) {
            iterator.remove();
          }
        }
      }
      return changed;
    }
    
    public boolean retainAll(Collection<?> c) {
      Preconditions.checkNotNull(c);
      boolean changed = false;
      Iterator<Map<C, V>> iterator = backingMap.values().iterator();
      while (iterator.hasNext()) {
        Map<C, V> map = (Map)iterator.next();
        if (map.keySet().retainAll(c)) {
          changed = true;
          if (map.isEmpty()) {
            iterator.remove();
          }
        }
      }
      return changed;
    }
    
    public boolean contains(Object obj) {
      if (obj == null) {
        return false;
      }
      for (Map<C, V> map : backingMap.values()) {
        if (map.containsKey(obj)) {
          return true;
        }
      }
      return false;
    }
  }
  



  Iterator<C> createColumnKeyIterator()
  {
    return new ColumnKeyIterator(null);
  }
  
  private class ColumnKeyIterator
    extends AbstractIterator<C>
  {
    final Map<C, V> seen = (Map)factory.get();
    final Iterator<Map<C, V>> mapIterator = backingMap.values().iterator();
    Iterator<Map.Entry<C, V>> entryIterator = Iterators.emptyIterator();
    
    private ColumnKeyIterator() {}
    
    protected C computeNext() { for (;;) { if (entryIterator.hasNext()) {
          Map.Entry<C, V> entry = (Map.Entry)entryIterator.next();
          if (!seen.containsKey(entry.getKey())) {
            seen.put(entry.getKey(), entry.getValue());
            return entry.getKey();
          }
        } else { if (!mapIterator.hasNext()) break;
          entryIterator = ((Map)mapIterator.next()).entrySet().iterator();
        } }
      return endOfData();
    }
  }
  









  public Collection<V> values()
  {
    StandardTable<R, C, V>.Values result = values;
    return result == null ? (this.values = new Values(null)) : result;
  }
  
  private class Values extends StandardTable<R, C, V>.TableCollection<V> { private Values() { super(null); }
    
    public Iterator<V> iterator() { final Iterator<Table.Cell<R, C, V>> cellIterator = cellSet().iterator();
      new Iterator() {
        public boolean hasNext() {
          return cellIterator.hasNext();
        }
        
        public V next() { return ((Table.Cell)cellIterator.next()).getValue(); }
        
        public void remove() {
          cellIterator.remove();
        }
      };
    }
    
    public int size() {
      return StandardTable.this.size();
    }
  }
  

  public Map<R, Map<C, V>> rowMap()
  {
    StandardTable<R, C, V>.RowMap result = rowMap;
    return result == null ? (this.rowMap = new RowMap()) : result;
  }
  
  class RowMap extends Maps.ImprovedAbstractMap<R, Map<C, V>> { RowMap() {}
    
    public boolean containsKey(Object key) { return containsRow(key); }
    


    public Map<C, V> get(Object key)
    {
      return containsRow(key) ? row(key) : null;
    }
    
    public Set<R> keySet() {
      return rowKeySet();
    }
    
    public Map<C, V> remove(Object key) {
      return key == null ? null : (Map)backingMap.remove(key);
    }
    

    protected Set<Map.Entry<R, Map<C, V>>> createEntrySet() { return new EntrySet(); }
    
    class EntrySet extends StandardTable<R, C, V>.TableSet<Map.Entry<R, Map<C, V>>> {
      EntrySet() { super(null); }
      
      public Iterator<Map.Entry<R, Map<C, V>>> iterator() { return new StandardTable.RowMap.EntryIterator(StandardTable.RowMap.this); }
      
      public int size()
      {
        return backingMap.size();
      }
      
      public boolean contains(Object obj) {
        if ((obj instanceof Map.Entry)) {
          Map.Entry<?, ?> entry = (Map.Entry)obj;
          return (entry.getKey() != null) && ((entry.getValue() instanceof Map)) && (Collections2.safeContains(backingMap.entrySet(), entry));
        }
        

        return false;
      }
      
      public boolean remove(Object obj) {
        if ((obj instanceof Map.Entry)) {
          Map.Entry<?, ?> entry = (Map.Entry)obj;
          return (entry.getKey() != null) && ((entry.getValue() instanceof Map)) && (backingMap.entrySet().remove(entry));
        }
        

        return false;
      } }
    
    class EntryIterator implements Iterator<Map.Entry<R, Map<C, V>>> { EntryIterator() {}
      
      final Iterator<R> delegate = backingMap.keySet().iterator();
      
      public boolean hasNext() {
        return delegate.hasNext();
      }
      
      public Map.Entry<R, Map<C, V>> next() {
        R rowKey = delegate.next();
        return new ImmutableEntry(rowKey, row(rowKey));
      }
      
      public void remove() {
        delegate.remove();
      }
    }
  }
  

  public Map<C, Map<R, V>> columnMap()
  {
    StandardTable<R, C, V>.ColumnMap result = columnMap;
    return result == null ? (this.columnMap = new ColumnMap(null)) : result;
  }
  
  private class ColumnMap extends Maps.ImprovedAbstractMap<C, Map<R, V>> {
    StandardTable<R, C, V>.ColumnMap.ColumnMapValues columnMapValues;
    
    private ColumnMap() {}
    
    public Map<R, V> get(Object key) { return containsColumn(key) ? column(key) : null; }
    
    public boolean containsKey(Object key)
    {
      return containsColumn(key);
    }
    
    public Map<R, V> remove(Object key) {
      return containsColumn(key) ? StandardTable.this.removeColumn(key) : null;
    }
    
    public Set<Map.Entry<C, Map<R, V>>> createEntrySet() {
      return new ColumnMapEntrySet();
    }
    
    public Set<C> keySet() {
      return columnKeySet();
    }
    

    public Collection<Map<R, V>> values()
    {
      StandardTable<R, C, V>.ColumnMap.ColumnMapValues result = columnMapValues;
      return result == null ? (this.columnMapValues = new ColumnMapValues(null)) : result;
    }
    
    class ColumnMapEntrySet extends StandardTable<R, C, V>.TableSet<Map.Entry<C, Map<R, V>>> {
      ColumnMapEntrySet() { super(null); }
      
      public Iterator<Map.Entry<C, Map<R, V>>> iterator() { final Iterator<C> columnIterator = columnKeySet().iterator();
        new UnmodifiableIterator()
        {
          public boolean hasNext() { return columnIterator.hasNext(); }
          
          public Map.Entry<C, Map<R, V>> next() {
            C columnKey = columnIterator.next();
            return new ImmutableEntry(columnKey, column(columnKey));
          }
        };
      }
      
      public int size()
      {
        return columnKeySet().size();
      }
      
      public boolean contains(Object obj) {
        if ((obj instanceof Map.Entry)) {
          Map.Entry<?, ?> entry = (Map.Entry)obj;
          if (containsColumn(entry.getKey()))
          {


            C columnKey = entry.getKey();
            return get(columnKey).equals(entry.getValue());
          }
        }
        return false;
      }
      
      public boolean remove(Object obj) {
        if (contains(obj)) {
          Map.Entry<?, ?> entry = (Map.Entry)obj;
          StandardTable.this.removeColumn(entry.getKey());
          return true;
        }
        return false;
      }
      
      public boolean removeAll(Collection<?> c) {
        boolean changed = false;
        for (Object obj : c) {
          changed |= remove(obj);
        }
        return changed;
      }
      
      public boolean retainAll(Collection<?> c) {
        boolean changed = false;
        for (C columnKey : Lists.newArrayList(columnKeySet().iterator())) {
          if (!c.contains(new ImmutableEntry(columnKey, column(columnKey))))
          {
            StandardTable.this.removeColumn(columnKey);
            changed = true;
          }
        }
        return changed;
      }
    }
    
    private class ColumnMapValues extends StandardTable<R, C, V>.TableCollection<Map<R, V>> { private ColumnMapValues() { super(null); }
      
      public Iterator<Map<R, V>> iterator() { return StandardTable.valueIteratorImpl(StandardTable.ColumnMap.this); }
      
      public boolean remove(Object obj)
      {
        for (Map.Entry<C, Map<R, V>> entry : entrySet()) {
          if (((Map)entry.getValue()).equals(obj)) {
            StandardTable.this.removeColumn(entry.getKey());
            return true;
          }
        }
        return false;
      }
      
      public boolean removeAll(Collection<?> c) {
        Preconditions.checkNotNull(c);
        boolean changed = false;
        for (C columnKey : Lists.newArrayList(columnKeySet().iterator())) {
          if (c.contains(column(columnKey))) {
            StandardTable.this.removeColumn(columnKey);
            changed = true;
          }
        }
        return changed;
      }
      
      public boolean retainAll(Collection<?> c) {
        Preconditions.checkNotNull(c);
        boolean changed = false;
        for (C columnKey : Lists.newArrayList(columnKeySet().iterator())) {
          if (!c.contains(column(columnKey))) {
            StandardTable.this.removeColumn(columnKey);
            changed = true;
          }
        }
        return changed;
      }
      
      public int size() {
        return columnKeySet().size();
      }
    }
  }
  







  static <K, V> Iterator<K> keyIteratorImpl(Map<K, V> map)
  {
    Iterator<Map.Entry<K, V>> entryIterator = map.entrySet().iterator();
    new Iterator() {
      public boolean hasNext() {
        return val$entryIterator.hasNext();
      }
      
      public K next() { return ((Map.Entry)val$entryIterator.next()).getKey(); }
      
      public void remove() {
        val$entryIterator.remove();
      }
    };
  }
  



  static <K, V> Iterator<V> valueIteratorImpl(Map<K, V> map)
  {
    Iterator<Map.Entry<K, V>> entryIterator = map.entrySet().iterator();
    new Iterator() {
      public boolean hasNext() {
        return val$entryIterator.hasNext();
      }
      
      public V next() { return ((Map.Entry)val$entryIterator.next()).getValue(); }
      
      public void remove() {
        val$entryIterator.remove();
      }
    };
  }
}
