package com.google.common.collect;

import com.google.common.annotations.Beta;
import com.google.common.annotations.GwtCompatible;
import com.google.common.base.Function;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.base.Supplier;
import java.io.Serializable;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;








































@GwtCompatible
@Beta
public final class Tables
{
  public static <R, C, V> Table.Cell<R, C, V> immutableCell(@Nullable R rowKey, @Nullable C columnKey, @Nullable V value)
  {
    return new ImmutableCell(rowKey, columnKey, value);
  }
  
  static final class ImmutableCell<R, C, V> extends Tables.AbstractCell<R, C, V> implements Serializable
  {
    private final R rowKey;
    private final C columnKey;
    private final V value;
    private static final long serialVersionUID = 0L;
    
    ImmutableCell(@Nullable R rowKey, @Nullable C columnKey, @Nullable V value) {
      this.rowKey = rowKey;
      this.columnKey = columnKey;
      this.value = value;
    }
    
    public R getRowKey()
    {
      return rowKey;
    }
    
    public C getColumnKey() {
      return columnKey;
    }
    
    public V getValue() {
      return value;
    }
  }
  


  static abstract class AbstractCell<R, C, V>
    implements Table.Cell<R, C, V>
  {
    public boolean equals(Object obj)
    {
      if (obj == this) {
        return true;
      }
      if ((obj instanceof Table.Cell)) {
        Table.Cell<?, ?, ?> other = (Table.Cell)obj;
        return (Objects.equal(getRowKey(), other.getRowKey())) && (Objects.equal(getColumnKey(), other.getColumnKey())) && (Objects.equal(getValue(), other.getValue()));
      }
      

      return false;
    }
    
    public int hashCode() {
      return Objects.hashCode(new Object[] { getRowKey(), getColumnKey(), getValue() });
    }
    
    public String toString() {
      return "(" + getRowKey() + "," + getColumnKey() + ")=" + getValue();
    }
  }
  













  public static <R, C, V> Table<C, R, V> transpose(Table<R, C, V> table)
  {
    return (table instanceof TransposeTable) ? original : new TransposeTable(table);
  }
  
  private static class TransposeTable<C, R, V> implements Table<C, R, V>
  {
    final Table<R, C, V> original;
    
    TransposeTable(Table<R, C, V> original)
    {
      this.original = ((Table)Preconditions.checkNotNull(original));
    }
    
    public void clear()
    {
      original.clear();
    }
    
    public Map<C, V> column(R columnKey)
    {
      return original.row(columnKey);
    }
    
    public Set<R> columnKeySet()
    {
      return original.rowKeySet();
    }
    
    public Map<R, Map<C, V>> columnMap()
    {
      return original.rowMap();
    }
    

    public boolean contains(@Nullable Object rowKey, @Nullable Object columnKey)
    {
      return original.contains(columnKey, rowKey);
    }
    
    public boolean containsColumn(@Nullable Object columnKey)
    {
      return original.containsRow(columnKey);
    }
    
    public boolean containsRow(@Nullable Object rowKey)
    {
      return original.containsColumn(rowKey);
    }
    
    public boolean containsValue(@Nullable Object value)
    {
      return original.containsValue(value);
    }
    
    public V get(@Nullable Object rowKey, @Nullable Object columnKey)
    {
      return original.get(columnKey, rowKey);
    }
    
    public boolean isEmpty()
    {
      return original.isEmpty();
    }
    
    public V put(C rowKey, R columnKey, V value)
    {
      return original.put(columnKey, rowKey, value);
    }
    
    public void putAll(Table<? extends C, ? extends R, ? extends V> table)
    {
      original.putAll(Tables.transpose(table));
    }
    
    public V remove(@Nullable Object rowKey, @Nullable Object columnKey)
    {
      return original.remove(columnKey, rowKey);
    }
    
    public Map<R, V> row(C rowKey)
    {
      return original.column(rowKey);
    }
    
    public Set<C> rowKeySet()
    {
      return original.columnKeySet();
    }
    
    public Map<C, Map<R, V>> rowMap()
    {
      return original.columnMap();
    }
    
    public int size()
    {
      return original.size();
    }
    
    public Collection<V> values()
    {
      return original.values();
    }
    
    public boolean equals(@Nullable Object obj) {
      if (obj == this) {
        return true;
      }
      if ((obj instanceof Table)) {
        Table<?, ?, ?> other = (Table)obj;
        return cellSet().equals(other.cellSet());
      }
      return false;
    }
    
    public int hashCode() {
      return cellSet().hashCode();
    }
    
    public String toString() {
      return rowMap().toString();
    }
    

    private static final Function<Table.Cell<?, ?, ?>, Table.Cell<?, ?, ?>> TRANSPOSE_CELL = new Function()
    {
      public Table.Cell<?, ?, ?> apply(Table.Cell<?, ?, ?> cell)
      {
        return Tables.immutableCell(cell.getColumnKey(), cell.getRowKey(), cell.getValue());
      }
    };
    
    TransposeTable<C, R, V>.CellSet cellSet;
    

    public Set<Table.Cell<C, R, V>> cellSet()
    {
      TransposeTable<C, R, V>.CellSet result = cellSet;
      return result == null ? (this.cellSet = new CellSet()) : result;
    }
    
    class CellSet
      extends Collections2.TransformedCollection<Table.Cell<R, C, V>, Table.Cell<C, R, V>> implements Set<Table.Cell<C, R, V>>
    {
      CellSet()
      {
        super(Tables.TransposeTable.TRANSPOSE_CELL);
      }
      
      public boolean equals(Object obj) {
        if (obj == this) {
          return true;
        }
        if (!(obj instanceof Set)) {
          return false;
        }
        Set<?> os = (Set)obj;
        if (os.size() != size()) {
          return false;
        }
        return containsAll(os);
      }
      
      public int hashCode() {
        return Sets.hashCodeImpl(this);
      }
      
      public boolean contains(Object obj) {
        if ((obj instanceof Table.Cell)) {
          Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
          return original.cellSet().contains(Tables.immutableCell(cell.getColumnKey(), cell.getRowKey(), cell.getValue()));
        }
        
        return false;
      }
      
      public boolean remove(Object obj) {
        if ((obj instanceof Table.Cell)) {
          Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
          return original.cellSet().remove(Tables.immutableCell(cell.getColumnKey(), cell.getRowKey(), cell.getValue()));
        }
        
        return false;
      }
    }
  }
  









































  public static <R, C, V> Table<R, C, V> newCustomTable(Map<R, Map<C, V>> backingMap, Supplier<? extends Map<C, V>> factory)
  {
    Preconditions.checkArgument(backingMap.isEmpty());
    Preconditions.checkNotNull(factory);
    
    return new StandardTable(backingMap, factory);
  }
  




























  public static <R, C, V1, V2> Table<R, C, V2> transformValues(Table<R, C, V1> fromTable, Function<? super V1, V2> function) { return new TransformedTable(fromTable, function); }
  
  private static class TransformedTable<R, C, V1, V2> implements Table<R, C, V2> {
    final Table<R, C, V1> fromTable;
    final Function<? super V1, V2> function;
    TransformedTable<R, C, V1, V2>.CellSet cellSet;
    Collection<V2> values;
    Map<R, Map<C, V2>> rowMap;
    Map<C, Map<R, V2>> columnMap;
    
    TransformedTable(Table<R, C, V1> fromTable, Function<? super V1, V2> function) { this.fromTable = ((Table)Preconditions.checkNotNull(fromTable));
      this.function = ((Function)Preconditions.checkNotNull(function));
    }
    
    public boolean contains(Object rowKey, Object columnKey) {
      return fromTable.contains(rowKey, columnKey);
    }
    
    public boolean containsRow(Object rowKey) {
      return fromTable.containsRow(rowKey);
    }
    
    public boolean containsColumn(Object columnKey) {
      return fromTable.containsColumn(columnKey);
    }
    
    public boolean containsValue(Object value) {
      return values().contains(value);
    }
    

    public V2 get(Object rowKey, Object columnKey)
    {
      return contains(rowKey, columnKey) ? function.apply(fromTable.get(rowKey, columnKey)) : null;
    }
    
    public boolean isEmpty()
    {
      return fromTable.isEmpty();
    }
    
    public int size() {
      return fromTable.size();
    }
    
    public void clear() {
      fromTable.clear();
    }
    
    public V2 put(R rowKey, C columnKey, V2 value) {
      throw new UnsupportedOperationException();
    }
    
    public void putAll(Table<? extends R, ? extends C, ? extends V2> table)
    {
      throw new UnsupportedOperationException();
    }
    
    public V2 remove(Object rowKey, Object columnKey) {
      return contains(rowKey, columnKey) ? function.apply(fromTable.remove(rowKey, columnKey)) : null;
    }
    
    public Map<C, V2> row(R rowKey)
    {
      return Maps.transformValues(fromTable.row(rowKey), function);
    }
    
    public Map<R, V2> column(C columnKey) {
      return Maps.transformValues(fromTable.column(columnKey), function);
    }
    
    Function<Table.Cell<R, C, V1>, Table.Cell<R, C, V2>> cellFunction() {
      new Function() {
        public Table.Cell<R, C, V2> apply(Table.Cell<R, C, V1> cell) {
          return Tables.immutableCell(cell.getRowKey(), cell.getColumnKey(), function.apply(cell.getValue()));
        }
      };
    }
    
    class CellSet
      extends Collections2.TransformedCollection<Table.Cell<R, C, V1>, Table.Cell<R, C, V2>> implements Set<Table.Cell<R, C, V2>>
    {
      CellSet()
      {
        super(cellFunction());
      }
      
      public boolean equals(Object obj) { return Sets.equalsImpl(this, obj); }
      

      public int hashCode() { return Sets.hashCodeImpl(this); }
      
      public boolean contains(Object obj) {
        if ((obj instanceof Table.Cell)) {
          Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
          if (!Objects.equal(cell.getValue(), get(cell.getRowKey(), cell.getColumnKey())))
          {
            return false;
          }
          return (cell.getValue() != null) || (fromTable.contains(cell.getRowKey(), cell.getColumnKey()));
        }
        
        return false;
      }
      
      public boolean remove(Object obj) { if (contains(obj)) {
          Table.Cell<?, ?, ?> cell = (Table.Cell)obj;
          fromTable.remove(cell.getRowKey(), cell.getColumnKey());
          return true;
        }
        return false;
      }
    }
    

    public Set<Table.Cell<R, C, V2>> cellSet()
    {
      return cellSet == null ? (this.cellSet = new CellSet()) : cellSet;
    }
    
    public Set<R> rowKeySet() {
      return fromTable.rowKeySet();
    }
    
    public Set<C> columnKeySet() {
      return fromTable.columnKeySet();
    }
    

    public Collection<V2> values()
    {
      return values == null ? (this.values = Collections2.transform(fromTable.values(), function)) : values;
    }
    

    Map<R, Map<C, V2>> createRowMap()
    {
      Function<Map<C, V1>, Map<C, V2>> rowFunction = new Function()
      {
        public Map<C, V2> apply(Map<C, V1> row) {
          return Maps.transformValues(row, function);
        }
      };
      return Maps.transformValues(fromTable.rowMap(), rowFunction);
    }
    

    public Map<R, Map<C, V2>> rowMap()
    {
      return rowMap == null ? (this.rowMap = createRowMap()) : rowMap;
    }
    
    Map<C, Map<R, V2>> createColumnMap() {
      Function<Map<R, V1>, Map<R, V2>> columnFunction = new Function()
      {
        public Map<R, V2> apply(Map<R, V1> column) {
          return Maps.transformValues(column, function);
        }
      };
      return Maps.transformValues(fromTable.columnMap(), columnFunction);
    }
    

    public Map<C, Map<R, V2>> columnMap()
    {
      return columnMap == null ? (this.columnMap = createColumnMap()) : columnMap;
    }
    
    public boolean equals(@Nullable Object obj) {
      if (obj == this) {
        return true;
      }
      if ((obj instanceof Table)) {
        Table<?, ?, ?> other = (Table)obj;
        return cellSet().equals(other.cellSet());
      }
      return false;
    }
    
    public int hashCode() {
      return cellSet().hashCode();
    }
    
    public String toString() {
      return rowMap().toString();
    }
  }
}
