package com.google.common.net;

import com.google.common.annotations.Beta;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import java.util.regex.Matcher;
import java.util.regex.Pattern;






















































@Beta
public final class HostAndPort
{
  private static final int NO_PORT = -1;
  private final String host;
  private final int port;
  private final boolean hasBracketlessColons;
  
  private HostAndPort(String host, int port, boolean hasBracketlessColons)
  {
    this.host = host;
    this.port = port;
    this.hasBracketlessColons = hasBracketlessColons;
  }
  






  public String getHostText()
  {
    return host;
  }
  
  public boolean hasPort()
  {
    return port >= 0;
  }
  






  public int getPort()
  {
    Preconditions.checkState(hasPort());
    return port;
  }
  


  public int getPortOrDefault(int defaultPort)
  {
    return hasPort() ? port : defaultPort;
  }
  











  public static HostAndPort fromParts(String host, int port)
  {
    Preconditions.checkArgument(isValidPort(port));
    HostAndPort parsedHost = fromString(host);
    Preconditions.checkArgument(!parsedHost.hasPort());
    return new HostAndPort(host, port, hasBracketlessColons);
  }
  
  private static final Pattern BRACKET_PATTERN = Pattern.compile("^\\[(.*:.*)\\](?::(\\d*))?$");
  









  public static HostAndPort fromString(String hostPortString)
  {
    Preconditions.checkNotNull(hostPortString);
    
    String portString = null;
    boolean hasBracketlessColons = false;
    String host;
    if (hostPortString.startsWith("["))
    {
      Matcher matcher = BRACKET_PATTERN.matcher(hostPortString);
      Preconditions.checkArgument(matcher.matches(), "Invalid bracketed host/port: %s", new Object[] { hostPortString });
      
      String host = matcher.group(1);
      portString = matcher.group(2);
    } else {
      int colonPos = hostPortString.indexOf(':');
      if ((colonPos >= 0) && (hostPortString.indexOf(':', colonPos + 1) == -1))
      {
        String host = hostPortString.substring(0, colonPos);
        portString = hostPortString.substring(colonPos + 1);
      }
      else {
        host = hostPortString;
        hasBracketlessColons = colonPos >= 0;
      }
    }
    
    int port = -1;
    if (portString != null)
    {
      try {
        port = Integer.parseInt(portString);
      } catch (NumberFormatException e) {
        throw new IllegalArgumentException("Unparseable port number: " + hostPortString);
      }
      Preconditions.checkArgument(isValidPort(port), "Port number out of range: %s", new Object[] { hostPortString });
    }
    

    return new HostAndPort(host, port, hasBracketlessColons);
  }
  









  public HostAndPort withDefaultPort(int defaultPort)
  {
    Preconditions.checkArgument(isValidPort(defaultPort));
    if ((hasPort()) || (port == defaultPort)) {
      return this;
    }
    return new HostAndPort(host, defaultPort, hasBracketlessColons);
  }
  














  public HostAndPort requireBracketsForIPv6()
  {
    Preconditions.checkArgument(!hasBracketlessColons, "Possible bracketless IPv6 literal: %s", new Object[] { host });
    
    return this;
  }
  
  public boolean equals(Object other)
  {
    if (this == other) {
      return true;
    }
    if ((other instanceof HostAndPort)) {
      HostAndPort that = (HostAndPort)other;
      return (Objects.equal(host, host)) && (port == port) && (hasBracketlessColons == hasBracketlessColons);
    }
    

    return false;
  }
  
  public int hashCode()
  {
    return Objects.hashCode(new Object[] { host, Integer.valueOf(port), Boolean.valueOf(hasBracketlessColons) });
  }
  

  public String toString()
  {
    StringBuilder builder = new StringBuilder(host.length() + 7);
    if (host.indexOf(':') >= 0) {
      builder.append('[').append(host).append(']');
    } else {
      builder.append(host);
    }
    if (hasPort()) {
      builder.append(':').append(port);
    }
    return builder.toString();
  }
  
  private static boolean isValidPort(int port)
  {
    return (port >= 0) && (port <= 65535);
  }
}
