package com.google.common.util.concurrent;

import com.google.common.base.Preconditions;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;






















abstract class AbstractListeningExecutorService
  implements ListeningExecutorService
{
  public ListenableFuture<?> submit(Runnable task)
  {
    ListenableFutureTask<Void> ftask = ListenableFutureTask.create(task, null);
    execute(ftask);
    return ftask;
  }
  
  public <T> ListenableFuture<T> submit(Runnable task, T result) {
    ListenableFutureTask<T> ftask = ListenableFutureTask.create(task, result);
    execute(ftask);
    return ftask;
  }
  
  public <T> ListenableFuture<T> submit(Callable<T> task) {
    ListenableFutureTask<T> ftask = ListenableFutureTask.create(task);
    execute(ftask);
    return ftask;
  }
  


  private <T> T doInvokeAny(Collection<? extends Callable<T>> tasks, boolean timed, long nanos)
    throws InterruptedException, ExecutionException, TimeoutException
  {
    int ntasks = tasks.size();
    Preconditions.checkArgument(ntasks > 0);
    List<Future<T>> futures = new ArrayList(ntasks);
    ExecutorCompletionService<T> ecs = new ExecutorCompletionService(this);
    







    try
    {
      ExecutionException ee = null;
      long lastTime = timed ? System.nanoTime() : 0L;
      Iterator<? extends Callable<T>> it = tasks.iterator();
      

      futures.add(ecs.submit((Callable)it.next()));
      ntasks--;
      int active = 1;
      for (;;)
      {
        Future<T> f = ecs.poll();
        if (f == null)
          if (ntasks > 0) {
            ntasks--;
            futures.add(ecs.submit((Callable)it.next()));
            active++;
          } else { if (active == 0)
              break;
            if (timed) {
              f = ecs.poll(nanos, TimeUnit.NANOSECONDS);
              if (f == null) {
                throw new TimeoutException();
              }
              long now = System.nanoTime();
              nanos -= now - lastTime;
              lastTime = now;
            } else {
              f = ecs.take();
            }
          }
        if (f != null) {
          active--;
          try { Iterator i$;
            Future<T> f; return f.get();
          } catch (ExecutionException eex) {
            ee = eex;
          } catch (RuntimeException rex) {
            ee = new ExecutionException(rex);
          }
        }
      }
      
      if (ee == null) {
        ee = new ExecutionException(null);
      }
      throw ee;
    } finally {
      for (Future<T> f : futures) {
        f.cancel(true);
      }
    }
  }
  
  public <T> T invokeAny(Collection<? extends Callable<T>> tasks) throws InterruptedException, ExecutionException
  {
    try {
      return doInvokeAny(tasks, false, 0L);
    } catch (TimeoutException cannotHappen) {
      throw new AssertionError();
    }
  }
  
  public <T> T invokeAny(Collection<? extends Callable<T>> tasks, long timeout, TimeUnit unit)
    throws InterruptedException, ExecutionException, TimeoutException
  {
    return doInvokeAny(tasks, true, unit.toNanos(timeout));
  }
  
  public <T> List<Future<T>> invokeAll(Collection<? extends Callable<T>> tasks) throws InterruptedException
  {
    if (tasks == null) {
      throw new NullPointerException();
    }
    List<Future<T>> futures = new ArrayList(tasks.size());
    boolean done = false;
    try {
      for (Callable<T> t : tasks) {
        ListenableFutureTask<T> f = ListenableFutureTask.create(t);
        futures.add(f);
        execute(f);
      }
      for (Future<T> f : futures) {
        if (!f.isDone()) {
          try {
            f.get();
          }
          catch (CancellationException ignore) {}catch (ExecutionException ignore) {}
        }
      }
      
      done = true;
      Iterator i$; Future<T> f; return futures;
    } finally {
      if (!done) {
        for (Future<T> f : futures) {
          f.cancel(true);
        }
      }
    }
  }
  
  public <T> List<Future<T>> invokeAll(Collection<? extends Callable<T>> tasks, long timeout, TimeUnit unit)
    throws InterruptedException
  {
    if ((tasks == null) || (unit == null)) {
      throw new NullPointerException();
    }
    long nanos = unit.toNanos(timeout);
    List<Future<T>> futures = new ArrayList(tasks.size());
    boolean done = false;
    try {
      for (Callable<T> t : tasks) {
        futures.add(ListenableFutureTask.create(t));
      }
      
      long lastTime = System.nanoTime();
      


      Iterator<Future<T>> it = futures.iterator();
      List<Future<T>> localList1; while (it.hasNext()) {
        execute((Runnable)it.next());
        long now = System.nanoTime();
        nanos -= now - lastTime;
        lastTime = now;
        if (nanos <= 0L) { Iterator i$;
          Future<T> f; return futures;
        }
      }
      
      for (Future<T> f : futures) {
        if (!f.isDone()) { Iterator i$;
          if (nanos <= 0L) { Future<T> f;
            return futures;
          }
          try {
            f.get(nanos, TimeUnit.NANOSECONDS);
          } catch (CancellationException ignore) {}catch (ExecutionException ignore) {}catch (TimeoutException toe) {
            Iterator i$;
            Future<T> f;
            return futures;
          }
          long now = System.nanoTime();
          nanos -= now - lastTime;
          lastTime = now;
        }
      }
      done = true;
      Iterator i$; Object f; return futures;
    } finally {
      if (!done) {
        for (Future<T> f : futures) {
          f.cancel(true);
        }
      }
    }
  }
}
