package com.google.common.util.concurrent;

import com.google.common.annotations.Beta;
import com.google.common.base.Preconditions;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.ReentrantLock;





















@Beta
public abstract class AbstractService
  implements Service
{
  private final ReentrantLock lock;
  private final Transition startup;
  private final Transition shutdown;
  private Service.State state;
  private boolean shutdownWhenStartupFinishes;
  
  public AbstractService()
  {
    lock = new ReentrantLock();
    
    startup = new Transition(null);
    shutdown = new Transition(null);
    




    state = Service.State.NEW;
    




    shutdownWhenStartupFinishes = false;
  }
  







  protected abstract void doStart();
  







  protected abstract void doStop();
  






  public final ListenableFuture<Service.State> start()
  {
    lock.lock();
    try {
      if (state == Service.State.NEW) {
        state = Service.State.STARTING;
        doStart();
      }
    }
    catch (Throwable startupFailure) {
      notifyFailed(startupFailure);
    } finally {
      lock.unlock();
    }
    
    return startup;
  }
  
  public final ListenableFuture<Service.State> stop()
  {
    lock.lock();
    try {
      if (state == Service.State.NEW) {
        state = Service.State.TERMINATED;
        startup.set(Service.State.TERMINATED);
        shutdown.set(Service.State.TERMINATED);
      } else if (state == Service.State.STARTING) {
        shutdownWhenStartupFinishes = true;
        startup.set(Service.State.STOPPING);
      } else if (state == Service.State.RUNNING) {
        state = Service.State.STOPPING;
        doStop();
      }
    }
    catch (Throwable shutdownFailure) {
      notifyFailed(shutdownFailure);
    } finally {
      lock.unlock();
    }
    
    return shutdown;
  }
  
  public Service.State startAndWait()
  {
    return (Service.State)Futures.getUnchecked(start());
  }
  
  public Service.State stopAndWait()
  {
    return (Service.State)Futures.getUnchecked(stop());
  }
  







  protected final void notifyStarted()
  {
    lock.lock();
    try {
      if (state != Service.State.STARTING) {
        IllegalStateException failure = new IllegalStateException("Cannot notifyStarted() when the service is " + state);
        
        notifyFailed(failure);
        throw failure;
      }
      
      state = Service.State.RUNNING;
      if (shutdownWhenStartupFinishes) {
        stop();
      } else {
        startup.set(Service.State.RUNNING);
      }
    } finally {
      lock.unlock();
    }
  }
  







  protected final void notifyStopped()
  {
    lock.lock();
    try {
      if ((state != Service.State.STOPPING) && (state != Service.State.RUNNING)) {
        IllegalStateException failure = new IllegalStateException("Cannot notifyStopped() when the service is " + state);
        
        notifyFailed(failure);
        throw failure;
      }
      
      state = Service.State.TERMINATED;
      shutdown.set(Service.State.TERMINATED);
    } finally {
      lock.unlock();
    }
  }
  





  protected final void notifyFailed(Throwable cause)
  {
    Preconditions.checkNotNull(cause);
    
    lock.lock();
    try {
      if (state == Service.State.STARTING) {
        startup.setException(cause);
        shutdown.setException(new Exception("Service failed to start.", cause));
      }
      else if (state == Service.State.STOPPING) {
        shutdown.setException(cause);
      }
      
      state = Service.State.FAILED;
    } finally {
      lock.unlock();
    }
  }
  
  public final boolean isRunning()
  {
    return state() == Service.State.RUNNING;
  }
  
  public final Service.State state()
  {
    lock.lock();
    try { Service.State localState;
      if ((shutdownWhenStartupFinishes) && (state == Service.State.STARTING)) {
        return Service.State.STOPPING;
      }
      return state;
    }
    finally {
      lock.unlock();
    }
  }
  
  public String toString() {
    return getClass().getSimpleName() + " [" + state() + "]";
  }
  
  private class Transition extends AbstractFuture<Service.State>
  {
    private Transition() {}
    
    public Service.State get(long timeout, TimeUnit unit) throws InterruptedException, TimeoutException, ExecutionException
    {
      try
      {
        return (Service.State)super.get(timeout, unit);
      } catch (TimeoutException e) {
        throw new TimeoutException(toString());
      }
    }
  }
}
