package com.google.common.util.concurrent;

import com.google.common.annotations.Beta;
import com.google.common.base.Preconditions;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;














































public final class MoreExecutors
{
  @Beta
  public static ExecutorService getExitingExecutorService(ThreadPoolExecutor executor, long terminationTimeout, TimeUnit timeUnit)
  {
    executor.setThreadFactory(new ThreadFactoryBuilder().setDaemon(true).setThreadFactory(executor.getThreadFactory()).build());
    



    ExecutorService service = Executors.unconfigurableExecutorService(executor);
    
    addDelayedShutdownHook(service, terminationTimeout, timeUnit);
    
    return service;
  }
  

















  @Beta
  public static ScheduledExecutorService getExitingScheduledExecutorService(ScheduledThreadPoolExecutor executor, long terminationTimeout, TimeUnit timeUnit)
  {
    executor.setThreadFactory(new ThreadFactoryBuilder().setDaemon(true).setThreadFactory(executor.getThreadFactory()).build());
    



    ScheduledExecutorService service = Executors.unconfigurableScheduledExecutorService(executor);
    

    addDelayedShutdownHook(service, terminationTimeout, timeUnit);
    
    return service;
  }
  












  @Beta
  public static void addDelayedShutdownHook(ExecutorService service, final long terminationTimeout, TimeUnit timeUnit)
  {
    Runtime.getRuntime().addShutdownHook(new Thread(new Runnable()
    {

      public void run()
      {

        try
        {

          val$service.shutdown();
          val$service.awaitTermination(terminationTimeout, val$timeUnit);
        }
        catch (InterruptedException ignored) {}
      }
    }));
  }
  
















  @Beta
  public static ExecutorService getExitingExecutorService(ThreadPoolExecutor executor)
  {
    return getExitingExecutorService(executor, 120L, TimeUnit.SECONDS);
  }
  















  @Beta
  public static ScheduledExecutorService getExitingScheduledExecutorService(ScheduledThreadPoolExecutor executor)
  {
    return getExitingScheduledExecutorService(executor, 120L, TimeUnit.SECONDS);
  }
  
































  public static ListeningExecutorService sameThreadExecutor()
  {
    return new SameThreadExecutorService(null);
  }
  




  private static class SameThreadExecutorService
    extends AbstractListeningExecutorService
  {
    private final Lock lock = new ReentrantLock();
    

    private final Condition termination = lock.newCondition();
    







    private int runningTasks = 0;
    private boolean shutdown = false;
    
    public void execute(Runnable command)
    {
      startTask();
      try {
        command.run();
      } finally {
        endTask();
      }
    }
    
    public boolean isShutdown()
    {
      lock.lock();
      try {
        return shutdown;
      } finally {
        lock.unlock();
      }
    }
    
    public void shutdown()
    {
      lock.lock();
      try {
        shutdown = true;
      } finally {
        lock.unlock();
      }
    }
    

    public List<Runnable> shutdownNow()
    {
      shutdown();
      return Collections.emptyList();
    }
    
    public boolean isTerminated()
    {
      lock.lock();
      try {
        return (shutdown) && (runningTasks == 0);
      } finally {
        lock.unlock();
      }
    }
    
    /* Error */
    public boolean awaitTermination(long timeout, TimeUnit unit)
      throws InterruptedException
    {
      // Byte code:
      //   0: aload_3
      //   1: lload_1
      //   2: invokevirtual 83	java/util/concurrent/TimeUnit:toNanos	(J)J
      //   5: lstore 4
      //   7: aload_0
      //   8: getfield 27	com/google/common/util/concurrent/MoreExecutors$SameThreadExecutorService:lock	Ljava/util/concurrent/locks/Lock;
      //   11: invokeinterface 60 1 0
      //   16: aload_0
      //   17: invokevirtual 85	com/google/common/util/concurrent/MoreExecutors$SameThreadExecutorService:isTerminated	()Z
      //   20: ifeq +18 -> 38
      //   23: iconst_1
      //   24: istore 6
      //   26: aload_0
      //   27: getfield 27	com/google/common/util/concurrent/MoreExecutors$SameThreadExecutorService:lock	Ljava/util/concurrent/locks/Lock;
      //   30: invokeinterface 63 1 0
      //   35: iload 6
      //   37: ireturn
      //   38: lload 4
      //   40: lconst_0
      //   41: lcmp
      //   42: ifgt +18 -> 60
      //   45: iconst_0
      //   46: istore 6
      //   48: aload_0
      //   49: getfield 27	com/google/common/util/concurrent/MoreExecutors$SameThreadExecutorService:lock	Ljava/util/concurrent/locks/Lock;
      //   52: invokeinterface 63 1 0
      //   57: iload 6
      //   59: ireturn
      //   60: aload_0
      //   61: getfield 35	com/google/common/util/concurrent/MoreExecutors$SameThreadExecutorService:termination	Ljava/util/concurrent/locks/Condition;
      //   64: lload 4
      //   66: invokeinterface 90 3 0
      //   71: lstore 4
      //   73: goto -57 -> 16
      //   76: astore 7
      //   78: aload_0
      //   79: getfield 27	com/google/common/util/concurrent/MoreExecutors$SameThreadExecutorService:lock	Ljava/util/concurrent/locks/Lock;
      //   82: invokeinterface 63 1 0
      //   87: aload 7
      //   89: athrow
      // Line number table:
      //   Java source line #299	-> byte code offset #0
      //   Java source line #300	-> byte code offset #7
      //   Java source line #303	-> byte code offset #16
      //   Java source line #304	-> byte code offset #23
      //   Java source line #312	-> byte code offset #26
      //   Java source line #305	-> byte code offset #38
      //   Java source line #306	-> byte code offset #45
      //   Java source line #312	-> byte code offset #48
      //   Java source line #308	-> byte code offset #60
      //   Java source line #312	-> byte code offset #76
      // Local variable table:
      //   start	length	slot	name	signature
      //   0	90	0	this	SameThreadExecutorService
      //   0	90	1	timeout	long
      //   0	90	3	unit	TimeUnit
      //   5	67	4	nanos	long
      //   24	34	6	bool	boolean
      //   76	12	7	localObject	Object
      // Exception table:
      //   from	to	target	type
      //   16	26	76	finally
      //   38	48	76	finally
      //   60	78	76	finally
    }
    
    private void startTask()
    {
      lock.lock();
      try {
        if (isShutdown()) {
          throw new RejectedExecutionException("Executor already shutdown");
        }
        runningTasks += 1;
      } finally {
        lock.unlock();
      }
    }
    


    private void endTask()
    {
      lock.lock();
      try {
        runningTasks -= 1;
        if (isTerminated()) {
          termination.signalAll();
        }
      } finally {
        lock.unlock();
      }
    }
  }
  


















  public static ListeningExecutorService listeningDecorator(ExecutorService delegate)
  {
    return (delegate instanceof ScheduledExecutorService) ? new ScheduledListeningDecorator((ScheduledExecutorService)delegate) : (delegate instanceof ListeningExecutorService) ? (ListeningExecutorService)delegate : new ListeningDecorator(delegate);
  }
  























  public static ListeningScheduledExecutorService listeningDecorator(ScheduledExecutorService delegate)
  {
    return (delegate instanceof ListeningScheduledExecutorService) ? (ListeningScheduledExecutorService)delegate : new ScheduledListeningDecorator(delegate);
  }
  
  private static class ListeningDecorator
    extends AbstractListeningExecutorService
  {
    final ExecutorService delegate;
    
    ListeningDecorator(ExecutorService delegate)
    {
      this.delegate = ((ExecutorService)Preconditions.checkNotNull(delegate));
    }
    
    public boolean awaitTermination(long timeout, TimeUnit unit)
      throws InterruptedException
    {
      return delegate.awaitTermination(timeout, unit);
    }
    
    public boolean isShutdown()
    {
      return delegate.isShutdown();
    }
    
    public boolean isTerminated()
    {
      return delegate.isTerminated();
    }
    
    public void shutdown()
    {
      delegate.shutdown();
    }
    
    public List<Runnable> shutdownNow()
    {
      return delegate.shutdownNow();
    }
    
    public void execute(Runnable command)
    {
      delegate.execute(command);
    }
  }
  
  private static class ScheduledListeningDecorator extends MoreExecutors.ListeningDecorator implements ListeningScheduledExecutorService
  {
    final ScheduledExecutorService delegate;
    
    ScheduledListeningDecorator(ScheduledExecutorService delegate) {
      super();
      this.delegate = ((ScheduledExecutorService)Preconditions.checkNotNull(delegate));
    }
    

    public ScheduledFuture<?> schedule(Runnable command, long delay, TimeUnit unit)
    {
      return delegate.schedule(command, delay, unit);
    }
    

    public <V> ScheduledFuture<V> schedule(Callable<V> callable, long delay, TimeUnit unit)
    {
      return delegate.schedule(callable, delay, unit);
    }
    

    public ScheduledFuture<?> scheduleAtFixedRate(Runnable command, long initialDelay, long period, TimeUnit unit)
    {
      return delegate.scheduleAtFixedRate(command, initialDelay, period, unit);
    }
    

    public ScheduledFuture<?> scheduleWithFixedDelay(Runnable command, long initialDelay, long delay, TimeUnit unit)
    {
      return delegate.scheduleWithFixedDelay(command, initialDelay, delay, unit);
    }
  }
}
