package net.minecraft.util.com.google.common.collect;

import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import javax.annotation.Nullable;
import net.minecraft.util.com.google.common.annotations.Beta;
import net.minecraft.util.com.google.common.annotations.GwtCompatible;
import net.minecraft.util.com.google.common.base.Function;
import net.minecraft.util.com.google.common.base.Joiner;
import net.minecraft.util.com.google.common.base.Preconditions;
import net.minecraft.util.com.google.common.base.Predicate;
import net.minecraft.util.com.google.common.base.Predicates;
import net.minecraft.util.com.google.common.math.IntMath;
import net.minecraft.util.com.google.common.math.LongMath;
































































@GwtCompatible
public final class Collections2
{
  public static <E> Collection<E> filter(Collection<E> unfiltered, Predicate<? super E> predicate)
  {
    if ((unfiltered instanceof FilteredCollection))
    {

      return ((FilteredCollection)unfiltered).createCombined(predicate);
    }
    
    return new FilteredCollection((Collection)Preconditions.checkNotNull(unfiltered), (Predicate)Preconditions.checkNotNull(predicate));
  }
  






  static boolean safeContains(Collection<?> collection, @Nullable Object object)
  {
    Preconditions.checkNotNull(collection);
    try {
      return collection.contains(object);
    } catch (ClassCastException e) {
      return false;
    } catch (NullPointerException e) {}
    return false;
  }
  





  static boolean safeRemove(Collection<?> collection, @Nullable Object object)
  {
    Preconditions.checkNotNull(collection);
    try {
      return collection.remove(object);
    } catch (ClassCastException e) {
      return false;
    } catch (NullPointerException e) {}
    return false;
  }
  
  static class FilteredCollection<E> extends AbstractCollection<E>
  {
    final Collection<E> unfiltered;
    final Predicate<? super E> predicate;
    
    FilteredCollection(Collection<E> unfiltered, Predicate<? super E> predicate)
    {
      this.unfiltered = unfiltered;
      this.predicate = predicate;
    }
    
    FilteredCollection<E> createCombined(Predicate<? super E> newPredicate) {
      return new FilteredCollection(unfiltered, Predicates.and(predicate, newPredicate));
    }
    


    public boolean add(E element)
    {
      Preconditions.checkArgument(predicate.apply(element));
      return unfiltered.add(element);
    }
    
    public boolean addAll(Collection<? extends E> collection)
    {
      for (E element : collection) {
        Preconditions.checkArgument(predicate.apply(element));
      }
      return unfiltered.addAll(collection);
    }
    
    public void clear()
    {
      Iterables.removeIf(unfiltered, predicate);
    }
    
    public boolean contains(@Nullable Object element)
    {
      if (Collections2.safeContains(unfiltered, element))
      {
        E e = element;
        return predicate.apply(e);
      }
      return false;
    }
    
    public boolean containsAll(Collection<?> collection)
    {
      return Collections2.containsAllImpl(this, collection);
    }
    
    public boolean isEmpty()
    {
      return !Iterables.any(unfiltered, predicate);
    }
    
    public Iterator<E> iterator()
    {
      return Iterators.filter(unfiltered.iterator(), predicate);
    }
    
    public boolean remove(Object element)
    {
      return (contains(element)) && (unfiltered.remove(element));
    }
    
    public boolean removeAll(Collection<?> collection)
    {
      return Iterables.removeIf(unfiltered, Predicates.and(predicate, Predicates.in(collection)));
    }
    
    public boolean retainAll(Collection<?> collection)
    {
      return Iterables.removeIf(unfiltered, Predicates.and(predicate, Predicates.not(Predicates.in(collection))));
    }
    
    public int size()
    {
      return Iterators.size(iterator());
    }
    

    public Object[] toArray()
    {
      return Lists.newArrayList(iterator()).toArray();
    }
    
    public <T> T[] toArray(T[] array)
    {
      return Lists.newArrayList(iterator()).toArray(array);
    }
  }
  



















  public static <F, T> Collection<T> transform(Collection<F> fromCollection, Function<? super F, T> function)
  {
    return new TransformedCollection(fromCollection, function);
  }
  
  static class TransformedCollection<F, T> extends AbstractCollection<T>
  {
    final Collection<F> fromCollection;
    final Function<? super F, ? extends T> function;
    
    TransformedCollection(Collection<F> fromCollection, Function<? super F, ? extends T> function) {
      this.fromCollection = ((Collection)Preconditions.checkNotNull(fromCollection));
      this.function = ((Function)Preconditions.checkNotNull(function));
    }
    
    public void clear() {
      fromCollection.clear();
    }
    
    public boolean isEmpty() {
      return fromCollection.isEmpty();
    }
    
    public Iterator<T> iterator() {
      return Iterators.transform(fromCollection.iterator(), function);
    }
    
    public int size() {
      return fromCollection.size();
    }
  }
  











  static boolean containsAllImpl(Collection<?> self, Collection<?> c)
  {
    return Iterables.all(c, Predicates.in(self));
  }
  


  static String toStringImpl(Collection<?> collection)
  {
    StringBuilder sb = newStringBuilderForCollection(collection.size()).append('[');
    
    STANDARD_JOINER.appendTo(sb, Iterables.transform(collection, new Function()
    {
      public Object apply(Object input) {
        return input == val$collection ? "(this Collection)" : input;
      }
    }));
    return ']';
  }
  


  static StringBuilder newStringBuilderForCollection(int size)
  {
    Preconditions.checkArgument(size >= 0, "size must be non-negative");
    return new StringBuilder((int)Math.min(size * 8L, 1073741824L));
  }
  


  static <T> Collection<T> cast(Iterable<T> iterable)
  {
    return (Collection)iterable;
  }
  
  static final Joiner STANDARD_JOINER = Joiner.on(", ").useForNull("null");
  


























  @Beta
  public static <E extends Comparable<? super E>> Collection<List<E>> orderedPermutations(Iterable<E> elements)
  {
    return orderedPermutations(elements, Ordering.natural());
  }
  














































  @Beta
  public static <E> Collection<List<E>> orderedPermutations(Iterable<E> elements, Comparator<? super E> comparator)
  {
    return new OrderedPermutationCollection(elements, comparator);
  }
  
  private static final class OrderedPermutationCollection<E> extends AbstractCollection<List<E>>
  {
    final ImmutableList<E> inputList;
    final Comparator<? super E> comparator;
    final int size;
    
    OrderedPermutationCollection(Iterable<E> input, Comparator<? super E> comparator)
    {
      inputList = Ordering.from(comparator).immutableSortedCopy(input);
      this.comparator = comparator;
      size = calculateSize(inputList, comparator);
    }
    









    private static <E> int calculateSize(List<E> sortedInputList, Comparator<? super E> comparator)
    {
      long permutations = 1L;
      int n = 1;
      int r = 1;
      while (n < sortedInputList.size()) {
        int comparison = comparator.compare(sortedInputList.get(n - 1), sortedInputList.get(n));
        
        if (comparison < 0)
        {
          permutations *= LongMath.binomial(n, r);
          r = 0;
          if (!Collections2.isPositiveInt(permutations)) {
            return 2147483647;
          }
        }
        n++;
        r++;
      }
      permutations *= LongMath.binomial(n, r);
      if (!Collections2.isPositiveInt(permutations)) {
        return 2147483647;
      }
      return (int)permutations;
    }
    
    public int size() {
      return size;
    }
    
    public boolean isEmpty() {
      return false;
    }
    
    public Iterator<List<E>> iterator() {
      return new Collections2.OrderedPermutationIterator(inputList, comparator);
    }
    
    public boolean contains(@Nullable Object obj) {
      if ((obj instanceof List)) {
        List<?> list = (List)obj;
        return Collections2.isPermutation(inputList, list);
      }
      return false;
    }
    
    public String toString() {
      return "orderedPermutationCollection(" + inputList + ")";
    }
  }
  
  private static final class OrderedPermutationIterator<E>
    extends AbstractIterator<List<E>>
  {
    List<E> nextPermutation;
    final Comparator<? super E> comparator;
    
    OrderedPermutationIterator(List<E> list, Comparator<? super E> comparator)
    {
      nextPermutation = Lists.newArrayList(list);
      this.comparator = comparator;
    }
    
    protected List<E> computeNext() {
      if (nextPermutation == null) {
        return (List)endOfData();
      }
      ImmutableList<E> next = ImmutableList.copyOf(nextPermutation);
      calculateNextPermutation();
      return next;
    }
    
    void calculateNextPermutation() {
      int j = findNextJ();
      if (j == -1) {
        nextPermutation = null;
        return;
      }
      
      int l = findNextL(j);
      Collections.swap(nextPermutation, j, l);
      int n = nextPermutation.size();
      Collections.reverse(nextPermutation.subList(j + 1, n));
    }
    
    int findNextJ() {
      for (int k = nextPermutation.size() - 2; k >= 0; k--) {
        if (comparator.compare(nextPermutation.get(k), nextPermutation.get(k + 1)) < 0)
        {
          return k;
        }
      }
      return -1;
    }
    
    int findNextL(int j) {
      E ak = nextPermutation.get(j);
      for (int l = nextPermutation.size() - 1; l > j; l--) {
        if (comparator.compare(ak, nextPermutation.get(l)) < 0) {
          return l;
        }
      }
      throw new AssertionError("this statement should be unreachable");
    }
  }
  



















  @Beta
  public static <E> Collection<List<E>> permutations(Collection<E> elements)
  {
    return new PermutationCollection(ImmutableList.copyOf(elements));
  }
  
  private static final class PermutationCollection<E> extends AbstractCollection<List<E>>
  {
    final ImmutableList<E> inputList;
    
    PermutationCollection(ImmutableList<E> input) {
      inputList = input;
    }
    
    public int size() {
      return IntMath.factorial(inputList.size());
    }
    
    public boolean isEmpty() {
      return false;
    }
    
    public Iterator<List<E>> iterator() {
      return new Collections2.PermutationIterator(inputList);
    }
    
    public boolean contains(@Nullable Object obj) {
      if ((obj instanceof List)) {
        List<?> list = (List)obj;
        return Collections2.isPermutation(inputList, list);
      }
      return false;
    }
    
    public String toString() {
      return "permutations(" + inputList + ")";
    }
  }
  
  private static class PermutationIterator<E> extends AbstractIterator<List<E>>
  {
    final List<E> list;
    final int[] c;
    final int[] o;
    int j;
    
    PermutationIterator(List<E> list) {
      this.list = new ArrayList(list);
      int n = list.size();
      c = new int[n];
      o = new int[n];
      Arrays.fill(c, 0);
      Arrays.fill(o, 1);
      j = 2147483647;
    }
    
    protected List<E> computeNext() {
      if (j <= 0) {
        return (List)endOfData();
      }
      ImmutableList<E> next = ImmutableList.copyOf(list);
      calculateNextPermutation();
      return next;
    }
    
    void calculateNextPermutation() {
      j = (list.size() - 1);
      int s = 0;
      


      if (j == -1) {
        return;
      }
      int q;
      for (;;) {
        q = c[j] + o[j];
        if (q < 0) {
          switchDirection();
        }
        else {
          if (q != j + 1) break;
          if (j == 0) {
            return;
          }
          s++;
          switchDirection();
        }
      }
      
      Collections.swap(list, j - c[j] + s, j - q + s);
      c[j] = q;
    }
    

    void switchDirection()
    {
      o[j] = (-o[j]);
      j -= 1;
    }
  }
  



  private static boolean isPermutation(List<?> first, List<?> second)
  {
    if (first.size() != second.size()) {
      return false;
    }
    Multiset<?> firstMultiset = HashMultiset.create(first);
    Multiset<?> secondMultiset = HashMultiset.create(second);
    return firstMultiset.equals(secondMultiset);
  }
  
  private static boolean isPositiveInt(long n) {
    return (n >= 0L) && (n <= 2147483647L);
  }
}
