package net.minecraft.util.com.google.common.io;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.RandomAccessFile;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileChannel.MapMode;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import net.minecraft.util.com.google.common.annotations.Beta;
import net.minecraft.util.com.google.common.base.Joiner;
import net.minecraft.util.com.google.common.base.Preconditions;
import net.minecraft.util.com.google.common.base.Predicate;
import net.minecraft.util.com.google.common.base.Splitter;
import net.minecraft.util.com.google.common.collect.ImmutableSet;
import net.minecraft.util.com.google.common.collect.Lists;
import net.minecraft.util.com.google.common.collect.TreeTraverser;
import net.minecraft.util.com.google.common.hash.HashCode;
import net.minecraft.util.com.google.common.hash.HashFunction;










































@Beta
public final class Files
{
  private static final int TEMP_DIR_ATTEMPTS = 10000;
  
  public static BufferedReader newReader(File file, Charset charset)
    throws FileNotFoundException
  {
    Preconditions.checkNotNull(file);
    Preconditions.checkNotNull(charset);
    return new BufferedReader(new InputStreamReader(new FileInputStream(file), charset));
  }
  









  public static BufferedWriter newWriter(File file, Charset charset)
    throws FileNotFoundException
  {
    Preconditions.checkNotNull(file);
    Preconditions.checkNotNull(charset);
    return new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), charset));
  }
  





  public static ByteSource asByteSource(File file)
  {
    return new FileByteSource(file, null);
  }
  
  private static final class FileByteSource extends ByteSource
  {
    private final File file;
    
    private FileByteSource(File file) {
      this.file = ((File)Preconditions.checkNotNull(file));
    }
    
    public FileInputStream openStream() throws IOException
    {
      return new FileInputStream(file);
    }
    
    public long size() throws IOException
    {
      if (!file.isFile()) {
        throw new FileNotFoundException(file.toString());
      }
      return file.length();
    }
    
    public byte[] read() throws IOException
    {
      long size = file.length();
      

      if (size == 0L) {
        return super.read();
      }
      


      if (size > 2147483647L)
      {
        throw new OutOfMemoryError("file is too large to fit in a byte array: " + size + " bytes");
      }
      


      byte[] bytes = new byte[(int)size];
      
      Closer closer = Closer.create();
      try {
        InputStream in = (InputStream)closer.register(openStream());
        int off = 0;
        int read = 0;
        


        while ((off < size) && ((read = in.read(bytes, off, (int)size - off)) != -1)) {
          off += read;
        }
        
        if (off < size)
        {
          return Arrays.copyOf(bytes, off);
        }
        


        int b = in.read();
        if (b == -1)
        {
          return bytes;
        }
        

        Files.InternalByteArrayOutputStream out = new Files.InternalByteArrayOutputStream(null);
        
        out.write(b);
        ByteStreams.copy(in, out);
        
        byte[] result = new byte[bytes.length + out.size()];
        System.arraycopy(bytes, 0, result, 0, bytes.length);
        out.writeTo(result, bytes.length);
        return result;
      } catch (Throwable e) {
        throw closer.rethrow(e);
      } finally {
        closer.close();
      }
    }
    
    public String toString()
    {
      return "Files.asByteSource(" + file + ")";
    }
  }
  





  private static final class InternalByteArrayOutputStream
    extends ByteArrayOutputStream
  {
    void writeTo(byte[] b, int off)
    {
      System.arraycopy(buf, 0, b, off, count);
    }
  }
  








  public static ByteSink asByteSink(File file, FileWriteMode... modes)
  {
    return new FileByteSink(file, modes, null);
  }
  
  private static final class FileByteSink extends ByteSink
  {
    private final File file;
    private final ImmutableSet<FileWriteMode> modes;
    
    private FileByteSink(File file, FileWriteMode... modes) {
      this.file = ((File)Preconditions.checkNotNull(file));
      this.modes = ImmutableSet.copyOf(modes);
    }
    
    public FileOutputStream openStream() throws IOException
    {
      return new FileOutputStream(file, modes.contains(FileWriteMode.APPEND));
    }
    
    public String toString()
    {
      return "Files.asByteSink(" + file + ", " + modes + ")";
    }
  }
  





  public static CharSource asCharSource(File file, Charset charset)
  {
    return asByteSource(file).asCharSource(charset);
  }
  










  public static CharSink asCharSink(File file, Charset charset, FileWriteMode... modes)
  {
    return asByteSink(file, modes).asCharSink(charset);
  }
  







  public static InputSupplier<FileInputStream> newInputStreamSupplier(File file)
  {
    return ByteStreams.asInputSupplier(asByteSource(file));
  }
  







  public static OutputSupplier<FileOutputStream> newOutputStreamSupplier(File file)
  {
    return newOutputStreamSupplier(file, false);
  }
  









  public static OutputSupplier<FileOutputStream> newOutputStreamSupplier(File file, boolean append)
  {
    return ByteStreams.asOutputSupplier(asByteSink(file, modes(append)));
  }
  
  private static FileWriteMode[] modes(boolean append) {
    return append ? new FileWriteMode[] { FileWriteMode.APPEND } : new FileWriteMode[0];
  }
  











  public static InputSupplier<InputStreamReader> newReaderSupplier(File file, Charset charset)
  {
    return CharStreams.asInputSupplier(asCharSource(file, charset));
  }
  









  public static OutputSupplier<OutputStreamWriter> newWriterSupplier(File file, Charset charset)
  {
    return newWriterSupplier(file, charset, false);
  }
  











  public static OutputSupplier<OutputStreamWriter> newWriterSupplier(File file, Charset charset, boolean append)
  {
    return CharStreams.asOutputSupplier(asCharSink(file, charset, modes(append)));
  }
  







  public static byte[] toByteArray(File file)
    throws IOException
  {
    return asByteSource(file).read();
  }
  








  public static String toString(File file, Charset charset)
    throws IOException
  {
    return asCharSource(file, charset).read();
  }
  







  public static void copy(InputSupplier<? extends InputStream> from, File to)
    throws IOException
  {
    ByteStreams.asByteSource(from).copyTo(asByteSink(to, new FileWriteMode[0]));
  }
  





  public static void write(byte[] from, File to)
    throws IOException
  {
    asByteSink(to, new FileWriteMode[0]).write(from);
  }
  







  public static void copy(File from, OutputSupplier<? extends OutputStream> to)
    throws IOException
  {
    asByteSource(from).copyTo(ByteStreams.asByteSink(to));
  }
  





  public static void copy(File from, OutputStream to)
    throws IOException
  {
    asByteSource(from).copyTo(to);
  }
  











  public static void copy(File from, File to)
    throws IOException
  {
    Preconditions.checkArgument(!from.equals(to), "Source %s and destination %s must be different", new Object[] { from, to });
    
    asByteSource(from).copyTo(asByteSink(to, new FileWriteMode[0]));
  }
  










  public static <R extends Readable,  extends Closeable> void copy(InputSupplier<R> from, File to, Charset charset)
    throws IOException
  {
    CharStreams.asCharSource(from).copyTo(asCharSink(to, charset, new FileWriteMode[0]));
  }
  









  public static void write(CharSequence from, File to, Charset charset)
    throws IOException
  {
    asCharSink(to, charset, new FileWriteMode[0]).write(from);
  }
  









  public static void append(CharSequence from, File to, Charset charset)
    throws IOException
  {
    write(from, to, charset, true);
  }
  










  private static void write(CharSequence from, File to, Charset charset, boolean append)
    throws IOException
  {
    asCharSink(to, charset, modes(append)).write(from);
  }
  










  public static <W extends Appendable,  extends Closeable> void copy(File from, Charset charset, OutputSupplier<W> to)
    throws IOException
  {
    asCharSource(from, charset).copyTo(CharStreams.asCharSink(to));
  }
  









  public static void copy(File from, Charset charset, Appendable to)
    throws IOException
  {
    asCharSource(from, charset).copyTo(to);
  }
  



  public static boolean equal(File file1, File file2)
    throws IOException
  {
    Preconditions.checkNotNull(file1);
    Preconditions.checkNotNull(file2);
    if ((file1 == file2) || (file1.equals(file2))) {
      return true;
    }
    





    long len1 = file1.length();
    long len2 = file2.length();
    if ((len1 != 0L) && (len2 != 0L) && (len1 != len2)) {
      return false;
    }
    return asByteSource(file1).contentEquals(asByteSource(file2));
  }
  


















  public static File createTempDir()
  {
    File baseDir = new File(System.getProperty("java.io.tmpdir"));
    String baseName = System.currentTimeMillis() + "-";
    
    for (int counter = 0; counter < 10000; counter++) {
      File tempDir = new File(baseDir, baseName + counter);
      if (tempDir.mkdir()) {
        return tempDir;
      }
    }
    throw new IllegalStateException("Failed to create directory within 10000 attempts (tried " + baseName + "0 to " + baseName + 9999 + ')');
  }
  







  public static void touch(File file)
    throws IOException
  {
    Preconditions.checkNotNull(file);
    if ((!file.createNewFile()) && (!file.setLastModified(System.currentTimeMillis())))
    {
      throw new IOException("Unable to update modification time of " + file);
    }
  }
  








  public static void createParentDirs(File file)
    throws IOException
  {
    Preconditions.checkNotNull(file);
    File parent = file.getCanonicalFile().getParentFile();
    if (parent == null)
    {






      return;
    }
    parent.mkdirs();
    if (!parent.isDirectory()) {
      throw new IOException("Unable to create parent directories of " + file);
    }
  }
  







  public static void move(File from, File to)
    throws IOException
  {
    Preconditions.checkNotNull(from);
    Preconditions.checkNotNull(to);
    Preconditions.checkArgument(!from.equals(to), "Source %s and destination %s must be different", new Object[] { from, to });
    

    if (!from.renameTo(to)) {
      copy(from, to);
      if (!from.delete()) {
        if (!to.delete()) {
          throw new IOException("Unable to delete " + to);
        }
        throw new IOException("Unable to delete " + from);
      }
    }
  }
  










  public static String readFirstLine(File file, Charset charset)
    throws IOException
  {
    return asCharSource(file, charset).readFirstLine();
  }
  
















  public static List<String> readLines(File file, Charset charset)
    throws IOException
  {
    (List)readLines(file, charset, new LineProcessor() {
      final List<String> result = Lists.newArrayList();
      
      public boolean processLine(String line)
      {
        result.add(line);
        return true;
      }
      
      public List<String> getResult()
      {
        return result;
      }
    });
  }
  










  public static <T> T readLines(File file, Charset charset, LineProcessor<T> callback)
    throws IOException
  {
    return CharStreams.readLines(newReaderSupplier(file, charset), callback);
  }
  










  public static <T> T readBytes(File file, ByteProcessor<T> processor)
    throws IOException
  {
    return ByteStreams.readBytes(newInputStreamSupplier(file), processor);
  }
  








  public static HashCode hash(File file, HashFunction hashFunction)
    throws IOException
  {
    return asByteSource(file).hash(hashFunction);
  }
  














  public static MappedByteBuffer map(File file)
    throws IOException
  {
    Preconditions.checkNotNull(file);
    return map(file, FileChannel.MapMode.READ_ONLY);
  }
  

















  public static MappedByteBuffer map(File file, FileChannel.MapMode mode)
    throws IOException
  {
    Preconditions.checkNotNull(file);
    Preconditions.checkNotNull(mode);
    if (!file.exists()) {
      throw new FileNotFoundException(file.toString());
    }
    return map(file, mode, file.length());
  }
  




















  public static MappedByteBuffer map(File file, FileChannel.MapMode mode, long size)
    throws FileNotFoundException, IOException
  {
    Preconditions.checkNotNull(file);
    Preconditions.checkNotNull(mode);
    
    Closer closer = Closer.create();
    try {
      RandomAccessFile raf = (RandomAccessFile)closer.register(new RandomAccessFile(file, mode == FileChannel.MapMode.READ_ONLY ? "r" : "rw"));
      
      return map(raf, mode, size);
    } catch (Throwable e) {
      throw closer.rethrow(e);
    } finally {
      closer.close();
    }
  }
  
  private static MappedByteBuffer map(RandomAccessFile raf, FileChannel.MapMode mode, long size) throws IOException
  {
    Closer closer = Closer.create();
    try {
      FileChannel channel = (FileChannel)closer.register(raf.getChannel());
      return channel.map(mode, 0L, size);
    } catch (Throwable e) {
      throw closer.rethrow(e);
    } finally {
      closer.close();
    }
  }
  




















  public static String simplifyPath(String pathname)
  {
    Preconditions.checkNotNull(pathname);
    if (pathname.length() == 0) {
      return ".";
    }
    

    Iterable<String> components = Splitter.on('/').omitEmptyStrings().split(pathname);
    
    List<String> path = new ArrayList();
    

    for (String component : components) {
      if (!component.equals("."))
      {
        if (component.equals("..")) {
          if ((path.size() > 0) && (!((String)path.get(path.size() - 1)).equals(".."))) {
            path.remove(path.size() - 1);
          } else {
            path.add("..");
          }
        } else {
          path.add(component);
        }
      }
    }
    
    String result = Joiner.on('/').join(path);
    if (pathname.charAt(0) == '/') {
      result = "/" + result;
    }
    
    while (result.startsWith("/../")) {
      result = result.substring(3);
    }
    if (result.equals("/..")) {
      result = "/";
    } else if ("".equals(result)) {
      result = ".";
    }
    
    return result;
  }
  






  public static String getFileExtension(String fullName)
  {
    Preconditions.checkNotNull(fullName);
    String fileName = new File(fullName).getName();
    int dotIndex = fileName.lastIndexOf('.');
    return dotIndex == -1 ? "" : fileName.substring(dotIndex + 1);
  }
  









  public static String getNameWithoutExtension(String file)
  {
    Preconditions.checkNotNull(file);
    String fileName = new File(file).getName();
    int dotIndex = fileName.lastIndexOf('.');
    return dotIndex == -1 ? fileName : fileName.substring(0, dotIndex);
  }
  









  public static TreeTraverser<File> fileTreeTraverser()
  {
    return FILE_TREE_TRAVERSER;
  }
  
  private static final TreeTraverser<File> FILE_TREE_TRAVERSER = new TreeTraverser()
  {
    public Iterable<File> children(File file)
    {
      if (file.isDirectory()) {
        File[] files = file.listFiles();
        if (files != null) {
          return Collections.unmodifiableList(Arrays.asList(files));
        }
      }
      
      return Collections.emptyList();
    }
    
    public String toString()
    {
      return "Files.fileTreeTraverser()";
    }
  };
  




  public static Predicate<File> isDirectory()
  {
    return FilePredicate.IS_DIRECTORY;
  }
  




  public static Predicate<File> isFile()
  {
    return FilePredicate.IS_FILE;
  }
  
  private static abstract enum FilePredicate implements Predicate<File> {
    IS_DIRECTORY, 
    










    IS_FILE;
    
    private FilePredicate() {}
  }
}
