package net.minecraft.util.com.google.common.reflect;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.LinkedHashMap;
import java.util.Map.Entry;
import java.util.Set;
import java.util.jar.Attributes;
import java.util.jar.Attributes.Name;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import net.minecraft.util.com.google.common.annotations.Beta;
import net.minecraft.util.com.google.common.annotations.VisibleForTesting;
import net.minecraft.util.com.google.common.base.Preconditions;
import net.minecraft.util.com.google.common.base.Splitter;
import net.minecraft.util.com.google.common.collect.FluentIterable;
import net.minecraft.util.com.google.common.collect.ImmutableMap;
import net.minecraft.util.com.google.common.collect.ImmutableSet;
import net.minecraft.util.com.google.common.collect.ImmutableSet.Builder;
import net.minecraft.util.com.google.common.collect.ImmutableSortedSet;
import net.minecraft.util.com.google.common.collect.ImmutableSortedSet.Builder;
import net.minecraft.util.com.google.common.collect.Maps;
import net.minecraft.util.com.google.common.collect.Ordering;
import net.minecraft.util.com.google.common.collect.Sets;























@Beta
public final class ClassPath
{
  private static final Logger logger = Logger.getLogger(ClassPath.class.getName());
  

  private static final Splitter CLASS_PATH_ATTRIBUTE_SEPARATOR = Splitter.on(" ").omitEmptyStrings();
  
  private static final String CLASS_FILE_NAME_EXTENSION = ".class";
  
  private final ImmutableSet<ResourceInfo> resources;
  
  private ClassPath(ImmutableSet<ResourceInfo> resources)
  {
    this.resources = resources;
  }
  







  public static ClassPath from(ClassLoader classloader)
    throws IOException
  {
    Scanner scanner = new Scanner();
    for (Map.Entry<URI, ClassLoader> entry : getClassPathEntries(classloader).entrySet()) {
      scanner.scan((URI)entry.getKey(), (ClassLoader)entry.getValue());
    }
    return new ClassPath(scanner.getResources());
  }
  



  public ImmutableSet<ResourceInfo> getResources()
  {
    return resources;
  }
  
  public ImmutableSet<ClassInfo> getTopLevelClasses()
  {
    return FluentIterable.from(resources).filter(ClassInfo.class).toSet();
  }
  
  public ImmutableSet<ClassInfo> getTopLevelClasses(String packageName)
  {
    Preconditions.checkNotNull(packageName);
    ImmutableSet.Builder<ClassInfo> builder = ImmutableSet.builder();
    for (ClassInfo classInfo : getTopLevelClasses()) {
      if (classInfo.getPackageName().equals(packageName)) {
        builder.add(classInfo);
      }
    }
    return builder.build();
  }
  



  public ImmutableSet<ClassInfo> getTopLevelClassesRecursive(String packageName)
  {
    Preconditions.checkNotNull(packageName);
    String packagePrefix = packageName + '.';
    ImmutableSet.Builder<ClassInfo> builder = ImmutableSet.builder();
    for (ClassInfo classInfo : getTopLevelClasses()) {
      if (classInfo.getName().startsWith(packagePrefix)) {
        builder.add(classInfo);
      }
    }
    return builder.build();
  }
  


  @Beta
  public static class ResourceInfo
  {
    private final String resourceName;
    
    final ClassLoader loader;
    

    static ResourceInfo of(String resourceName, ClassLoader loader)
    {
      if ((resourceName.endsWith(".class")) && (!resourceName.contains("$"))) {
        return new ClassPath.ClassInfo(resourceName, loader);
      }
      return new ResourceInfo(resourceName, loader);
    }
    
    ResourceInfo(String resourceName, ClassLoader loader)
    {
      this.resourceName = ((String)Preconditions.checkNotNull(resourceName));
      this.loader = ((ClassLoader)Preconditions.checkNotNull(loader));
    }
    
    public final URL url()
    {
      return (URL)Preconditions.checkNotNull(loader.getResource(resourceName), "Failed to load resource: %s", new Object[] { resourceName });
    }
    

    public final String getResourceName()
    {
      return resourceName;
    }
    
    public int hashCode() {
      return resourceName.hashCode();
    }
    
    public boolean equals(Object obj) {
      if ((obj instanceof ResourceInfo)) {
        ResourceInfo that = (ResourceInfo)obj;
        return (resourceName.equals(resourceName)) && (loader == loader);
      }
      
      return false;
    }
    
    public String toString()
    {
      return resourceName;
    }
  }
  

  @Beta
  public static final class ClassInfo
    extends ClassPath.ResourceInfo
  {
    private final String className;
    

    ClassInfo(String resourceName, ClassLoader loader)
    {
      super(loader);
      className = ClassPath.getClassName(resourceName);
    }
    
    public String getPackageName()
    {
      return Reflection.getPackageName(className);
    }
    
    public String getSimpleName()
    {
      String packageName = getPackageName();
      if (packageName.isEmpty()) {
        return className;
      }
      
      return className.substring(packageName.length() + 1);
    }
    
    public String getName()
    {
      return className;
    }
    




    public Class<?> load()
    {
      try
      {
        return loader.loadClass(className);
      }
      catch (ClassNotFoundException e) {
        throw new IllegalStateException(e);
      }
    }
    
    public String toString() {
      return className;
    }
  }
  
  @VisibleForTesting
  static ImmutableMap<URI, ClassLoader> getClassPathEntries(ClassLoader classloader) {
    LinkedHashMap<URI, ClassLoader> entries = Maps.newLinkedHashMap();
    
    ClassLoader parent = classloader.getParent();
    if (parent != null) {
      entries.putAll(getClassPathEntries(parent));
    }
    if ((classloader instanceof URLClassLoader)) {
      URLClassLoader urlClassLoader = (URLClassLoader)classloader;
      for (URL entry : urlClassLoader.getURLs()) {
        URI uri;
        try {
          uri = entry.toURI();
        } catch (URISyntaxException e) {
          throw new IllegalArgumentException(e);
        }
        if (!entries.containsKey(uri)) {
          entries.put(uri, classloader);
        }
      }
    }
    return ImmutableMap.copyOf(entries);
  }
  
  @VisibleForTesting
  static final class Scanner {
    private final ImmutableSortedSet.Builder<ClassPath.ResourceInfo> resources = new ImmutableSortedSet.Builder(Ordering.usingToString());
    
    private final Set<URI> scannedUris = Sets.newHashSet();
    
    ImmutableSortedSet<ClassPath.ResourceInfo> getResources() {
      return resources.build();
    }
    
    void scan(URI uri, ClassLoader classloader) throws IOException {
      if ((uri.getScheme().equals("file")) && (scannedUris.add(uri))) {
        scanFrom(new File(uri), classloader);
      }
    }
    
    @VisibleForTesting
    void scanFrom(File file, ClassLoader classloader) throws IOException {
      if (!file.exists()) {
        return;
      }
      if (file.isDirectory()) {
        scanDirectory(file, classloader);
      } else {
        scanJar(file, classloader);
      }
    }
    
    private void scanDirectory(File directory, ClassLoader classloader) throws IOException {
      scanDirectory(directory, classloader, "", ImmutableSet.of());
    }
    
    private void scanDirectory(File directory, ClassLoader classloader, String packagePrefix, ImmutableSet<File> ancestors)
      throws IOException
    {
      File canonical = directory.getCanonicalFile();
      if (ancestors.contains(canonical))
      {
        return;
      }
      File[] files = directory.listFiles();
      if (files == null) {
        ClassPath.logger.warning("Cannot read directory " + directory);
        
        return;
      }
      ImmutableSet<File> newAncestors = ImmutableSet.builder().addAll(ancestors).add(canonical).build();
      


      for (File f : files) {
        String name = f.getName();
        if (f.isDirectory()) {
          scanDirectory(f, classloader, packagePrefix + name + "/", newAncestors);
        } else {
          String resourceName = packagePrefix + name;
          if (!resourceName.equals("META-INF/MANIFEST.MF")) {
            resources.add(ClassPath.ResourceInfo.of(resourceName, classloader));
          }
        }
      }
    }
    
    private void scanJar(File file, ClassLoader classloader) throws IOException {
      JarFile jarFile;
      try {
        jarFile = new JarFile(file);
      }
      catch (IOException e) {
        return;
      }
      try {
        for (URI uri : getClassPathFromManifest(file, jarFile.getManifest())) {
          scan(uri, classloader);
        }
        Enumeration<JarEntry> entries = jarFile.entries();
        while (entries.hasMoreElements()) {
          JarEntry entry = (JarEntry)entries.nextElement();
          if ((!entry.isDirectory()) && (!entry.getName().equals("META-INF/MANIFEST.MF")))
          {

            resources.add(ClassPath.ResourceInfo.of(entry.getName(), classloader)); }
        }
        return;
      } finally {
        try { jarFile.close();
        }
        catch (IOException ignored) {}
      }
    }
    




    @VisibleForTesting
    static ImmutableSet<URI> getClassPathFromManifest(File jarFile, @Nullable Manifest manifest)
    {
      if (manifest == null) {
        return ImmutableSet.of();
      }
      ImmutableSet.Builder<URI> builder = ImmutableSet.builder();
      String classpathAttribute = manifest.getMainAttributes().getValue(Attributes.Name.CLASS_PATH.toString());
      
      if (classpathAttribute != null) {
        for (String path : ClassPath.CLASS_PATH_ATTRIBUTE_SEPARATOR.split(classpathAttribute)) {
          URI uri;
          try {
            uri = getClassPathEntry(jarFile, path);
          }
          catch (URISyntaxException e) {
            ClassPath.logger.warning("Invalid Class-Path entry: " + path); }
          continue;
          
          builder.add(uri);
        }
      }
      return builder.build();
    }
    




    @VisibleForTesting
    static URI getClassPathEntry(File jarFile, String path)
      throws URISyntaxException
    {
      URI uri = new URI(path);
      if (uri.isAbsolute()) {
        return uri;
      }
      return new File(jarFile.getParentFile(), path.replace('/', File.separatorChar)).toURI();
    }
  }
  
  @VisibleForTesting
  static String getClassName(String filename) {
    int classNameEnd = filename.length() - ".class".length();
    return filename.substring(0, classNameEnd).replace('/', '.');
  }
}
