package net.minecraft.util.com.google.common.util.concurrent;

import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import javax.annotation.Nullable;
import javax.annotation.concurrent.GuardedBy;
import javax.annotation.concurrent.Immutable;
import net.minecraft.util.com.google.common.annotations.Beta;
import net.minecraft.util.com.google.common.base.Preconditions;
import net.minecraft.util.com.google.common.collect.Lists;





































@Beta
public abstract class AbstractService
  implements Service
{
  private final Monitor monitor = new Monitor();
  
  private final Transition startup = new Transition(null);
  private final Transition shutdown = new Transition(null);
  
  private final Monitor.Guard isStartable = new Monitor.Guard(monitor) {
    public boolean isSatisfied() {
      return state() == Service.State.NEW;
    }
  };
  
  private final Monitor.Guard isStoppable = new Monitor.Guard(monitor) {
    public boolean isSatisfied() {
      return state().compareTo(Service.State.RUNNING) <= 0;
    }
  };
  
  private final Monitor.Guard hasReachedRunning = new Monitor.Guard(monitor) {
    public boolean isSatisfied() {
      return state().compareTo(Service.State.RUNNING) >= 0;
    }
  };
  
  private final Monitor.Guard isStopped = new Monitor.Guard(monitor) {
    public boolean isSatisfied() {
      return state().isTerminal();
    }
  };
  


  @GuardedBy("monitor")
  private final List<ListenerExecutorPair> listeners = Lists.newArrayList();
  







  private final ExecutionQueue queuedListeners = new ExecutionQueue();
  








  @GuardedBy("monitor")
  private volatile StateSnapshot snapshot = new StateSnapshot(Service.State.NEW);
  



  protected AbstractService()
  {
    addListener(new Service.Listener()
    {
      public void running() {
        startup.set(Service.State.RUNNING);
      }
      
      public void stopping(Service.State from) {
        if (from == Service.State.STARTING) {
          startup.set(Service.State.STOPPING);
        }
      }
      
      public void terminated(Service.State from) {
        if (from == Service.State.NEW) {
          startup.set(Service.State.TERMINATED);
        }
        shutdown.set(Service.State.TERMINATED);
      }
      
      public void failed(Service.State from, Throwable failure) {
        switch (AbstractService.11.$SwitchMap$com$google$common$util$concurrent$Service$State[from.ordinal()]) {
        case 1: 
          startup.setException(failure);
          shutdown.setException(new Exception("Service failed to start.", failure));
          break;
        case 2: 
          shutdown.setException(new Exception("Service failed while running", failure));
          break;
        case 3: 
          shutdown.setException(failure);
          break;
        case 4: 
        case 5: 
        case 6: 
        default: 
          throw new AssertionError("Unexpected from state: " + from); } } }, MoreExecutors.sameThreadExecutor());
  }
  








  protected abstract void doStart();
  








  protected abstract void doStop();
  







  public final Service startAsync()
  {
    if (monitor.enterIf(isStartable)) {
      try {
        snapshot = new StateSnapshot(Service.State.STARTING);
        starting();
        doStart();
      }
      catch (Throwable startupFailure) {
        notifyFailed(startupFailure);
      } finally {
        monitor.leave();
        executeListeners();
      }
    } else {
      throw new IllegalStateException("Service " + this + " has already been started");
    }
    return this;
  }
  
  @Deprecated
  public final ListenableFuture<Service.State> start()
  {
    if (monitor.enterIf(isStartable)) {
      try {
        snapshot = new StateSnapshot(Service.State.STARTING);
        starting();
        doStart();
      } catch (Throwable startupFailure) {
        notifyFailed(startupFailure);
      } finally {
        monitor.leave();
        executeListeners();
      }
    }
    return startup;
  }
  
  public final Service stopAsync() {
    stop();
    return this;
  }
  
  @Deprecated
  public final ListenableFuture<Service.State> stop()
  {
    if (monitor.enterIf(isStoppable)) {
      try {
        Service.State previous = state();
        switch (11.$SwitchMap$com$google$common$util$concurrent$Service$State[previous.ordinal()]) {
        case 6: 
          snapshot = new StateSnapshot(Service.State.TERMINATED);
          terminated(Service.State.NEW);
          break;
        case 1: 
          snapshot = new StateSnapshot(Service.State.STARTING, true, null);
          stopping(Service.State.STARTING);
          break;
        case 2: 
          snapshot = new StateSnapshot(Service.State.STOPPING);
          stopping(Service.State.RUNNING);
          doStop();
          break;
        
        case 3: 
        case 4: 
        case 5: 
          throw new AssertionError("isStoppable is incorrectly implemented, saw: " + previous);
        default: 
          throw new AssertionError("Unexpected state: " + previous);
        }
      }
      catch (Throwable shutdownFailure)
      {
        notifyFailed(shutdownFailure);
      } finally {
        monitor.leave();
        executeListeners();
      }
    }
    return shutdown;
  }
  
  @Deprecated
  public Service.State startAndWait()
  {
    return (Service.State)Futures.getUnchecked(start());
  }
  
  @Deprecated
  public Service.State stopAndWait()
  {
    return (Service.State)Futures.getUnchecked(stop());
  }
  
  public final void awaitRunning() {
    monitor.enterWhenUninterruptibly(hasReachedRunning);
    try {
      checkCurrentState(Service.State.RUNNING);
    } finally {
      monitor.leave();
    }
  }
  
  public final void awaitRunning(long timeout, TimeUnit unit) throws TimeoutException {
    if (monitor.enterWhenUninterruptibly(hasReachedRunning, timeout, unit)) {
      try {
        checkCurrentState(Service.State.RUNNING);
      } finally {
        monitor.leave();

      }
      
    }
    else
    {
      throw new TimeoutException("Timed out waiting for " + this + " to reach the RUNNING state. " + "Current state: " + state());
    }
  }
  
  public final void awaitTerminated()
  {
    monitor.enterWhenUninterruptibly(isStopped);
    try {
      checkCurrentState(Service.State.TERMINATED);
    } finally {
      monitor.leave();
    }
  }
  
  public final void awaitTerminated(long timeout, TimeUnit unit) throws TimeoutException {
    if (monitor.enterWhenUninterruptibly(isStopped, timeout, unit)) {
      try {
        Service.State state = state();
        checkCurrentState(Service.State.TERMINATED);
      } finally {
        monitor.leave();

      }
      
    }
    else
    {
      throw new TimeoutException("Timed out waiting for " + this + " to reach a terminal state. " + "Current state: " + state());
    }
  }
  

  @GuardedBy("monitor")
  private void checkCurrentState(Service.State expected)
  {
    Service.State actual = state();
    if (actual != expected) {
      if (actual == Service.State.FAILED)
      {
        throw new IllegalStateException("Expected the service to be " + expected + ", but the service has FAILED", failureCause());
      }
      
      throw new IllegalStateException("Expected the service to be " + expected + ", but was " + actual);
    }
  }
  






  protected final void notifyStarted()
  {
    monitor.enter();
    
    try
    {
      if (snapshot.state != Service.State.STARTING) {
        IllegalStateException failure = new IllegalStateException("Cannot notifyStarted() when the service is " + snapshot.state);
        
        notifyFailed(failure);
        throw failure;
      }
      
      if (snapshot.shutdownWhenStartupFinishes) {
        snapshot = new StateSnapshot(Service.State.STOPPING);
        

        doStop();
      } else {
        snapshot = new StateSnapshot(Service.State.RUNNING);
        running();
      }
    } finally {
      monitor.leave();
      executeListeners();
    }
  }
  






  protected final void notifyStopped()
  {
    monitor.enter();
    
    try
    {
      Service.State previous = snapshot.state;
      if ((previous != Service.State.STOPPING) && (previous != Service.State.RUNNING)) {
        IllegalStateException failure = new IllegalStateException("Cannot notifyStopped() when the service is " + previous);
        
        notifyFailed(failure);
        throw failure;
      }
      snapshot = new StateSnapshot(Service.State.TERMINATED);
      terminated(previous);
    } finally {
      monitor.leave();
      executeListeners();
    }
  }
  




  protected final void notifyFailed(Throwable cause)
  {
    Preconditions.checkNotNull(cause);
    
    monitor.enter();
    try {
      Service.State previous = state();
      switch (11.$SwitchMap$com$google$common$util$concurrent$Service$State[previous.ordinal()]) {
      case 4: 
      case 6: 
        throw new IllegalStateException("Failed while in state:" + previous, cause);
      case 1: 
      case 2: 
      case 3: 
        snapshot = new StateSnapshot(Service.State.FAILED, false, cause);
        failed(previous, cause);
        break;
      case 5: 
        break;
      
      default: 
        throw new AssertionError("Unexpected state: " + previous);
      }
    } finally {
      monitor.leave();
      executeListeners();
    }
  }
  
  public final boolean isRunning()
  {
    return state() == Service.State.RUNNING;
  }
  
  public final Service.State state()
  {
    return snapshot.externalState();
  }
  



  public final Throwable failureCause()
  {
    return snapshot.failureCause();
  }
  



  public final void addListener(Service.Listener listener, Executor executor)
  {
    Preconditions.checkNotNull(listener, "listener");
    Preconditions.checkNotNull(executor, "executor");
    monitor.enter();
    try {
      Service.State currentState = state();
      if ((currentState != Service.State.TERMINATED) && (currentState != Service.State.FAILED)) {
        listeners.add(new ListenerExecutorPair(listener, executor));
      }
    } finally {
      monitor.leave();
    }
  }
  
  public String toString() {
    return getClass().getSimpleName() + " [" + state() + "]";
  }
  
  private class Transition extends AbstractFuture<Service.State>
  {
    private Transition() {}
    
    public Service.State get(long timeout, TimeUnit unit) throws InterruptedException, TimeoutException, ExecutionException
    {
      try
      {
        return (Service.State)super.get(timeout, unit);
      } catch (TimeoutException e) {
        throw new TimeoutException(toString());
      }
    }
  }
  



  private void executeListeners()
  {
    if (!monitor.isOccupiedByCurrentThread()) {
      queuedListeners.execute();
    }
  }
  
  @GuardedBy("monitor")
  private void starting() {
    for (final ListenerExecutorPair pair : listeners) {
      queuedListeners.add(new Runnable()
      {
        public void run() { pairlistener.starting(); } }, executor);
    }
  }
  

  @GuardedBy("monitor")
  private void running()
  {
    for (final ListenerExecutorPair pair : listeners) {
      queuedListeners.add(new Runnable()
      {
        public void run() { pairlistener.running(); } }, executor);
    }
  }
  

  @GuardedBy("monitor")
  private void stopping(final Service.State from)
  {
    for (final ListenerExecutorPair pair : listeners) {
      queuedListeners.add(new Runnable()
      {
        public void run() { pairlistener.stopping(from); } }, executor);
    }
  }
  

  @GuardedBy("monitor")
  private void terminated(final Service.State from)
  {
    for (final ListenerExecutorPair pair : listeners) {
      queuedListeners.add(new Runnable()
      {
        public void run() { pairlistener.terminated(from); } }, executor);
    }
    


    listeners.clear();
  }
  
  @GuardedBy("monitor")
  private void failed(final Service.State from, final Throwable cause) {
    for (final ListenerExecutorPair pair : listeners) {
      queuedListeners.add(new Runnable()
      {
        public void run() { pairlistener.failed(from, cause); } }, executor);
    }
    


    listeners.clear();
  }
  
  private static class ListenerExecutorPair
  {
    final Service.Listener listener;
    final Executor executor;
    
    ListenerExecutorPair(Service.Listener listener, Executor executor) {
      this.listener = listener;
      this.executor = executor;
    }
  }
  





  @Immutable
  private static final class StateSnapshot
  {
    final Service.State state;
    



    final boolean shutdownWhenStartupFinishes;
    



    @Nullable
    final Throwable failure;
    




    StateSnapshot(Service.State internalState)
    {
      this(internalState, false, null);
    }
    
    StateSnapshot(Service.State internalState, boolean shutdownWhenStartupFinishes, @Nullable Throwable failure)
    {
      Preconditions.checkArgument((!shutdownWhenStartupFinishes) || (internalState == Service.State.STARTING), "shudownWhenStartupFinishes can only be set if state is STARTING. Got %s instead.", new Object[] { internalState });
      

      Preconditions.checkArgument(((failure != null ? 1 : 0) ^ (internalState == Service.State.FAILED ? 1 : 0)) == 0, "A failure cause should be set if and only if the state is failed.  Got %s and %s instead.", new Object[] { internalState, failure });
      

      state = internalState;
      this.shutdownWhenStartupFinishes = shutdownWhenStartupFinishes;
      this.failure = failure;
    }
    
    Service.State externalState()
    {
      if ((shutdownWhenStartupFinishes) && (state == Service.State.STARTING)) {
        return Service.State.STOPPING;
      }
      return state;
    }
    

    Throwable failureCause()
    {
      Preconditions.checkState(state == Service.State.FAILED, "failureCause() is only valid if the service has failed, service is %s", new Object[] { state });
      
      return failure;
    }
  }
}
