package net.minecraft.util.io.netty.buffer;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import net.minecraft.util.io.netty.util.CharsetUtil;




















public final class ByteBufUtil
{
  private static final char[] HEXDUMP_TABLE = new char[1024];
  
  static {
    char[] DIGITS = "0123456789abcdef".toCharArray();
    for (int i = 0; i < 256; i++) {
      HEXDUMP_TABLE[(i << 1)] = DIGITS[(i >>> 4 & 0xF)];
      HEXDUMP_TABLE[((i << 1) + 1)] = DIGITS[(i & 0xF)];
    }
  }
  



  public static String hexDump(ByteBuf buffer)
  {
    return hexDump(buffer, buffer.readerIndex(), buffer.readableBytes());
  }
  



  public static String hexDump(ByteBuf buffer, int fromIndex, int length)
  {
    if (length < 0) {
      throw new IllegalArgumentException("length: " + length);
    }
    if (length == 0) {
      return "";
    }
    
    int endIndex = fromIndex + length;
    char[] buf = new char[length << 1];
    
    int srcIdx = fromIndex;
    for (int dstIdx = 0; 
        srcIdx < endIndex; dstIdx += 2) {
      System.arraycopy(HEXDUMP_TABLE, buffer.getUnsignedByte(srcIdx) << 1, buf, dstIdx, 2);srcIdx++;
    }
    


    return new String(buf);
  }
  



  public static int hashCode(ByteBuf buffer)
  {
    int aLen = buffer.readableBytes();
    int intCount = aLen >>> 2;
    int byteCount = aLen & 0x3;
    
    int hashCode = 1;
    int arrayIndex = buffer.readerIndex();
    if (buffer.order() == ByteOrder.BIG_ENDIAN) {
      for (int i = intCount; i > 0; i--) {
        hashCode = 31 * hashCode + buffer.getInt(arrayIndex);
        arrayIndex += 4;
      }
    } else {
      for (int i = intCount; i > 0; i--) {
        hashCode = 31 * hashCode + swapInt(buffer.getInt(arrayIndex));
        arrayIndex += 4;
      }
    }
    
    for (int i = byteCount; i > 0; i--) {
      hashCode = 31 * hashCode + buffer.getByte(arrayIndex++);
    }
    
    if (hashCode == 0) {
      hashCode = 1;
    }
    
    return hashCode;
  }
  




  public static boolean equals(ByteBuf bufferA, ByteBuf bufferB)
  {
    int aLen = bufferA.readableBytes();
    if (aLen != bufferB.readableBytes()) {
      return false;
    }
    
    int longCount = aLen >>> 3;
    int byteCount = aLen & 0x7;
    
    int aIndex = bufferA.readerIndex();
    int bIndex = bufferB.readerIndex();
    
    if (bufferA.order() == bufferB.order()) {
      for (int i = longCount; i > 0; i--) {
        if (bufferA.getLong(aIndex) != bufferB.getLong(bIndex)) {
          return false;
        }
        aIndex += 8;
        bIndex += 8;
      }
    } else {
      for (int i = longCount; i > 0; i--) {
        if (bufferA.getLong(aIndex) != swapLong(bufferB.getLong(bIndex))) {
          return false;
        }
        aIndex += 8;
        bIndex += 8;
      }
    }
    
    for (int i = byteCount; i > 0; i--) {
      if (bufferA.getByte(aIndex) != bufferB.getByte(bIndex)) {
        return false;
      }
      aIndex++;
      bIndex++;
    }
    
    return true;
  }
  



  public static int compare(ByteBuf bufferA, ByteBuf bufferB)
  {
    int aLen = bufferA.readableBytes();
    int bLen = bufferB.readableBytes();
    int minLength = Math.min(aLen, bLen);
    int uintCount = minLength >>> 2;
    int byteCount = minLength & 0x3;
    
    int aIndex = bufferA.readerIndex();
    int bIndex = bufferB.readerIndex();
    
    if (bufferA.order() == bufferB.order()) {
      for (int i = uintCount; i > 0; i--) {
        long va = bufferA.getUnsignedInt(aIndex);
        long vb = bufferB.getUnsignedInt(bIndex);
        if (va > vb) {
          return 1;
        }
        if (va < vb) {
          return -1;
        }
        aIndex += 4;
        bIndex += 4;
      }
    } else {
      for (int i = uintCount; i > 0; i--) {
        long va = bufferA.getUnsignedInt(aIndex);
        long vb = swapInt(bufferB.getInt(bIndex)) & 0xFFFFFFFF;
        if (va > vb) {
          return 1;
        }
        if (va < vb) {
          return -1;
        }
        aIndex += 4;
        bIndex += 4;
      }
    }
    
    for (int i = byteCount; i > 0; i--) {
      short va = bufferA.getUnsignedByte(aIndex);
      short vb = bufferB.getUnsignedByte(bIndex);
      if (va > vb) {
        return 1;
      }
      if (va < vb) {
        return -1;
      }
      aIndex++;
      bIndex++;
    }
    
    return aLen - bLen;
  }
  



  public static int indexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value)
  {
    if (fromIndex <= toIndex) {
      return firstIndexOf(buffer, fromIndex, toIndex, value);
    }
    return lastIndexOf(buffer, fromIndex, toIndex, value);
  }
  



  public static short swapShort(short value)
  {
    return Short.reverseBytes(value);
  }
  


  public static int swapMedium(int value)
  {
    int swapped = value << 16 & 0xFF0000 | value & 0xFF00 | value >>> 16 & 0xFF;
    if ((swapped & 0x800000) != 0) {
      swapped |= 0xFF000000;
    }
    return swapped;
  }
  


  public static int swapInt(int value)
  {
    return Integer.reverseBytes(value);
  }
  


  public static long swapLong(long value)
  {
    return Long.reverseBytes(value);
  }
  


  public static ByteBuf readBytes(ByteBufAllocator alloc, ByteBuf buffer, int length)
  {
    boolean release = true;
    ByteBuf dst = alloc.buffer(length);
    try {
      buffer.readBytes(dst);
      release = false;
      return dst;
    } finally {
      if (release) {
        dst.release();
      }
    }
  }
  
  private static int firstIndexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value) {
    fromIndex = Math.max(fromIndex, 0);
    if ((fromIndex >= toIndex) || (buffer.capacity() == 0)) {
      return -1;
    }
    
    for (int i = fromIndex; i < toIndex; i++) {
      if (buffer.getByte(i) == value) {
        return i;
      }
    }
    
    return -1;
  }
  
  private static int lastIndexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value) {
    fromIndex = Math.min(fromIndex, buffer.capacity());
    if ((fromIndex < 0) || (buffer.capacity() == 0)) {
      return -1;
    }
    
    for (int i = fromIndex - 1; i >= toIndex; i--) {
      if (buffer.getByte(i) == value) {
        return i;
      }
    }
    
    return -1;
  }
  
  static ByteBuffer encodeString(CharBuffer src, Charset charset) {
    CharsetEncoder encoder = CharsetUtil.getEncoder(charset);
    ByteBuffer dst = ByteBuffer.allocate((int)(src.remaining() * encoder.maxBytesPerChar()));
    try
    {
      CoderResult cr = encoder.encode(src, dst, true);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
      cr = encoder.flush(dst);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
    } catch (CharacterCodingException x) {
      throw new IllegalStateException(x);
    }
    dst.flip();
    return dst;
  }
  
  static String decodeString(ByteBuffer src, Charset charset) {
    CharsetDecoder decoder = CharsetUtil.getDecoder(charset);
    CharBuffer dst = CharBuffer.allocate((int)(src.remaining() * decoder.maxCharsPerByte()));
    try
    {
      CoderResult cr = decoder.decode(src, dst, true);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
      cr = decoder.flush(dst);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
    } catch (CharacterCodingException x) {
      throw new IllegalStateException(x);
    }
    return dst.flip().toString();
  }
}
