package net.minecraft.util.io.netty.buffer;

import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicInteger;
import net.minecraft.util.io.netty.util.internal.PlatformDependent;
import net.minecraft.util.io.netty.util.internal.StringUtil;
import net.minecraft.util.io.netty.util.internal.SystemPropertyUtil;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;

















public class PooledByteBufAllocator
  extends AbstractByteBufAllocator
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(PooledByteBufAllocator.class);
  
  private static final int DEFAULT_NUM_HEAP_ARENA;
  
  private static final int DEFAULT_NUM_DIRECT_ARENA;
  private static final int DEFAULT_PAGE_SIZE;
  private static final int DEFAULT_MAX_ORDER;
  private static final int MIN_PAGE_SIZE = 4096;
  private static final int MAX_CHUNK_SIZE = 1073741824;
  
  static
  {
    int defaultPageSize = SystemPropertyUtil.getInt("io.netty.allocator.pageSize", 8192);
    Throwable pageSizeFallbackCause = null;
    try {
      validateAndCalculatePageShifts(defaultPageSize);
    } catch (Throwable t) {
      pageSizeFallbackCause = t;
      defaultPageSize = 8192;
    }
    DEFAULT_PAGE_SIZE = defaultPageSize;
    
    int defaultMaxOrder = SystemPropertyUtil.getInt("io.netty.allocator.maxOrder", 11);
    Throwable maxOrderFallbackCause = null;
    try {
      validateAndCalculateChunkSize(DEFAULT_PAGE_SIZE, defaultMaxOrder);
    } catch (Throwable t) {
      maxOrderFallbackCause = t;
      defaultMaxOrder = 11;
    }
    DEFAULT_MAX_ORDER = defaultMaxOrder;
    


    Runtime runtime = Runtime.getRuntime();
    int defaultChunkSize = DEFAULT_PAGE_SIZE << DEFAULT_MAX_ORDER;
    DEFAULT_NUM_HEAP_ARENA = Math.max(0, SystemPropertyUtil.getInt("io.netty.allocator.numHeapArenas", (int)Math.min(runtime.availableProcessors(), Runtime.getRuntime().maxMemory() / defaultChunkSize / 2L / 3L)));
    




    DEFAULT_NUM_DIRECT_ARENA = Math.max(0, SystemPropertyUtil.getInt("io.netty.allocator.numDirectArenas", (int)Math.min(runtime.availableProcessors(), PlatformDependent.maxDirectMemory() / defaultChunkSize / 2L / 3L)));
    





    if (logger.isDebugEnabled()) {
      logger.debug("-Dio.netty.allocator.numHeapArenas: {}", Integer.valueOf(DEFAULT_NUM_HEAP_ARENA));
      logger.debug("-Dio.netty.allocator.numDirectArenas: {}", Integer.valueOf(DEFAULT_NUM_DIRECT_ARENA));
      if (pageSizeFallbackCause == null) {
        logger.debug("-Dio.netty.allocator.pageSize: {}", Integer.valueOf(DEFAULT_PAGE_SIZE));
      } else {
        logger.debug("-Dio.netty.allocator.pageSize: {}", Integer.valueOf(DEFAULT_PAGE_SIZE), pageSizeFallbackCause);
      }
      if (maxOrderFallbackCause == null) {
        logger.debug("-Dio.netty.allocator.maxOrder: {}", Integer.valueOf(DEFAULT_MAX_ORDER));
      } else {
        logger.debug("-Dio.netty.allocator.maxOrder: {}", Integer.valueOf(DEFAULT_MAX_ORDER), maxOrderFallbackCause);
      }
      logger.debug("-Dio.netty.allocator.chunkSize: {}", Integer.valueOf(DEFAULT_PAGE_SIZE << DEFAULT_MAX_ORDER));
    }
  }
  
  public static final PooledByteBufAllocator DEFAULT = new PooledByteBufAllocator(PlatformDependent.directBufferPreferred());
  
  private final PoolArena<byte[]>[] heapArenas;
  
  private final PoolArena<ByteBuffer>[] directArenas;
  
  final ThreadLocal<PoolThreadCache> threadCache = new ThreadLocal() {
    private final AtomicInteger index = new AtomicInteger();
    
    protected PoolThreadCache initialValue() {
      int idx = index.getAndIncrement();
      
      PoolArena<byte[]> heapArena;
      PoolArena<byte[]> heapArena;
      if (heapArenas != null) {
        heapArena = heapArenas[Math.abs(idx % heapArenas.length)];
      } else
        heapArena = null;
      PoolArena<ByteBuffer> directArena;
      PoolArena<ByteBuffer> directArena;
      if (directArenas != null) {
        directArena = directArenas[Math.abs(idx % directArenas.length)];
      } else {
        directArena = null;
      }
      
      return new PoolThreadCache(heapArena, directArena);
    }
  };
  
  public PooledByteBufAllocator() {
    this(false);
  }
  
  public PooledByteBufAllocator(boolean preferDirect) {
    this(preferDirect, DEFAULT_NUM_HEAP_ARENA, DEFAULT_NUM_DIRECT_ARENA, DEFAULT_PAGE_SIZE, DEFAULT_MAX_ORDER);
  }
  
  public PooledByteBufAllocator(int nHeapArena, int nDirectArena, int pageSize, int maxOrder) {
    this(false, nHeapArena, nDirectArena, pageSize, maxOrder);
  }
  
  public PooledByteBufAllocator(boolean preferDirect, int nHeapArena, int nDirectArena, int pageSize, int maxOrder) {
    super(preferDirect);
    
    int chunkSize = validateAndCalculateChunkSize(pageSize, maxOrder);
    
    if (nHeapArena < 0) {
      throw new IllegalArgumentException("nHeapArena: " + nHeapArena + " (expected: >= 0)");
    }
    if (nDirectArena < 0) {
      throw new IllegalArgumentException("nDirectArea: " + nDirectArena + " (expected: >= 0)");
    }
    
    int pageShifts = validateAndCalculatePageShifts(pageSize);
    
    if (nHeapArena > 0) {
      heapArenas = newArenaArray(nHeapArena);
      for (int i = 0; i < heapArenas.length; i++) {
        heapArenas[i] = new PoolArena.HeapArena(this, pageSize, maxOrder, pageShifts, chunkSize);
      }
    } else {
      heapArenas = null;
    }
    
    if (nDirectArena > 0) {
      directArenas = newArenaArray(nDirectArena);
      for (int i = 0; i < directArenas.length; i++) {
        directArenas[i] = new PoolArena.DirectArena(this, pageSize, maxOrder, pageShifts, chunkSize);
      }
    } else {
      directArenas = null;
    }
  }
  
  private static <T> PoolArena<T>[] newArenaArray(int size)
  {
    return new PoolArena[size];
  }
  
  private static int validateAndCalculatePageShifts(int pageSize) {
    if (pageSize < 4096) {
      throw new IllegalArgumentException("pageSize: " + pageSize + " (expected: 4096+)");
    }
    

    boolean found1 = false;
    int pageShifts = 0;
    for (int i = pageSize; i != 0; i >>= 1) {
      if ((i & 0x1) != 0) {
        if (!found1) {
          found1 = true;
        } else {
          throw new IllegalArgumentException("pageSize: " + pageSize + " (expected: power of 2");
        }
      }
      else if (!found1) {
        pageShifts++;
      }
    }
    
    return pageShifts;
  }
  
  private static int validateAndCalculateChunkSize(int pageSize, int maxOrder) {
    if (maxOrder > 14) {
      throw new IllegalArgumentException("maxOrder: " + maxOrder + " (expected: 0-14)");
    }
    

    int chunkSize = pageSize;
    for (int i = maxOrder; i > 0; i--) {
      if (chunkSize > 536870912) {
        throw new IllegalArgumentException(String.format("pageSize (%d) << maxOrder (%d) must not exceed %d", new Object[] { Integer.valueOf(pageSize), Integer.valueOf(maxOrder), Integer.valueOf(1073741824) }));
      }
      
      chunkSize <<= 1;
    }
    return chunkSize;
  }
  
  protected ByteBuf newHeapBuffer(int initialCapacity, int maxCapacity)
  {
    PoolThreadCache cache = (PoolThreadCache)threadCache.get();
    PoolArena<byte[]> heapArena = heapArena;
    if (heapArena != null) {
      return heapArena.allocate(cache, initialCapacity, maxCapacity);
    }
    return new UnpooledHeapByteBuf(this, initialCapacity, maxCapacity);
  }
  

  protected ByteBuf newDirectBuffer(int initialCapacity, int maxCapacity)
  {
    PoolThreadCache cache = (PoolThreadCache)threadCache.get();
    PoolArena<ByteBuffer> directArena = directArena;
    if (directArena != null) {
      return directArena.allocate(cache, initialCapacity, maxCapacity);
    }
    if (PlatformDependent.hasUnsafe()) {
      return new UnpooledUnsafeDirectByteBuf(this, initialCapacity, maxCapacity);
    }
    return new UnpooledDirectByteBuf(this, initialCapacity, maxCapacity);
  }
  


  public boolean isDirectBufferPooled()
  {
    return directArenas != null;
  }
  
  public String toString() {
    StringBuilder buf = new StringBuilder();
    buf.append(heapArenas.length);
    buf.append(" heap arena(s):");
    buf.append(StringUtil.NEWLINE);
    for (PoolArena<byte[]> a : heapArenas) {
      buf.append(a);
    }
    buf.append(directArenas.length);
    buf.append(" direct arena(s):");
    buf.append(StringUtil.NEWLINE);
    for (PoolArena<ByteBuffer> a : directArenas) {
      buf.append(a);
    }
    return buf.toString();
  }
}
