package net.minecraft.util.io.netty.channel;

import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;
import java.util.concurrent.atomic.AtomicLongFieldUpdater;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.buffer.ByteBufAllocator;
import net.minecraft.util.io.netty.buffer.ByteBufHolder;
import net.minecraft.util.io.netty.util.Recycler;
import net.minecraft.util.io.netty.util.Recycler.Handle;
import net.minecraft.util.io.netty.util.ReferenceCountUtil;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;


























public final class ChannelOutboundBuffer
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(ChannelOutboundBuffer.class);
  
  private static final int INITIAL_CAPACITY = 32;
  
  private static final Recycler<ChannelOutboundBuffer> RECYCLER = new Recycler()
  {

    protected ChannelOutboundBuffer newObject(Recycler.Handle handle) { return new ChannelOutboundBuffer(handle, null); }
  };
  private final Recycler.Handle handle;
  private AbstractChannel channel;
  
  static ChannelOutboundBuffer newInstance(AbstractChannel channel) { ChannelOutboundBuffer buffer = (ChannelOutboundBuffer)RECYCLER.get();
    channel = channel;
    totalPendingSize = 0L;
    writable = 1;
    return buffer;
  }
  


  private Entry[] buffer;
  
  private int flushed;
  
  private int unflushed;
  
  private int tail;
  
  private ByteBuffer[] nioBuffers;
  
  private int nioBufferCount;
  
  private long nioBufferSize;
  
  private boolean inFail;
  
  private static final AtomicLongFieldUpdater<ChannelOutboundBuffer> TOTAL_PENDING_SIZE_UPDATER = AtomicLongFieldUpdater.newUpdater(ChannelOutboundBuffer.class, "totalPendingSize");
  

  private volatile long totalPendingSize;
  
  private static final AtomicIntegerFieldUpdater<ChannelOutboundBuffer> WRITABLE_UPDATER = AtomicIntegerFieldUpdater.newUpdater(ChannelOutboundBuffer.class, "writable");
  

  private volatile int writable = 1;
  
  private ChannelOutboundBuffer(Recycler.Handle handle) {
    this.handle = handle;
    
    buffer = new Entry[32];
    for (int i = 0; i < buffer.length; i++) {
      buffer[i] = new Entry(null);
    }
    
    nioBuffers = new ByteBuffer[32];
  }
  
  void addMessage(Object msg, ChannelPromise promise) {
    int size = channel.estimatorHandle().size(msg);
    if (size < 0) {
      size = 0;
    }
    
    Entry e = buffer[(tail++)];
    msg = msg;
    pendingSize = size;
    promise = promise;
    total = total(msg);
    
    tail &= buffer.length - 1;
    
    if (tail == flushed) {
      addCapacity();
    }
    


    incrementPendingOutboundBytes(size);
  }
  
  private void addCapacity() {
    int p = flushed;
    int n = buffer.length;
    int r = n - p;
    int s = size();
    
    int newCapacity = n << 1;
    if (newCapacity < 0) {
      throw new IllegalStateException();
    }
    
    Entry[] e = new Entry[newCapacity];
    System.arraycopy(buffer, p, e, 0, r);
    System.arraycopy(buffer, 0, e, r, p);
    for (int i = n; i < e.length; i++) {
      e[i] = new Entry(null);
    }
    
    buffer = e;
    flushed = 0;
    unflushed = s;
    tail = n;
  }
  
  void addFlush() {
    unflushed = tail;
  }
  





  void incrementPendingOutboundBytes(int size)
  {
    Channel channel = this.channel;
    if ((size == 0) || (channel == null)) {
      return;
    }
    
    long oldValue = totalPendingSize;
    long newWriteBufferSize = oldValue + size;
    while (!TOTAL_PENDING_SIZE_UPDATER.compareAndSet(this, oldValue, newWriteBufferSize)) {
      oldValue = totalPendingSize;
      newWriteBufferSize = oldValue + size;
    }
    
    int highWaterMark = channel.config().getWriteBufferHighWaterMark();
    
    if ((newWriteBufferSize > highWaterMark) && 
      (WRITABLE_UPDATER.compareAndSet(this, 1, 0))) {
      channel.pipeline().fireChannelWritabilityChanged();
    }
  }
  






  void decrementPendingOutboundBytes(int size)
  {
    Channel channel = this.channel;
    if ((size == 0) || (channel == null)) {
      return;
    }
    
    long oldValue = totalPendingSize;
    long newWriteBufferSize = oldValue - size;
    while (!TOTAL_PENDING_SIZE_UPDATER.compareAndSet(this, oldValue, newWriteBufferSize)) {
      oldValue = totalPendingSize;
      newWriteBufferSize = oldValue - size;
    }
    
    int lowWaterMark = channel.config().getWriteBufferLowWaterMark();
    
    if (((newWriteBufferSize == 0L) || (newWriteBufferSize < lowWaterMark)) && 
      (WRITABLE_UPDATER.compareAndSet(this, 0, 1))) {
      channel.pipeline().fireChannelWritabilityChanged();
    }
  }
  
  private static long total(Object msg)
  {
    if ((msg instanceof ByteBuf)) {
      return ((ByteBuf)msg).readableBytes();
    }
    if ((msg instanceof FileRegion)) {
      return ((FileRegion)msg).count();
    }
    if ((msg instanceof ByteBufHolder)) {
      return ((ByteBufHolder)msg).content().readableBytes();
    }
    return -1L;
  }
  
  public Object current() {
    if (isEmpty()) {
      return null;
    }
    return buffer[flushed].msg;
  }
  




  public void current(Object msg)
  {
    Entry entry = buffer[flushed];
    safeRelease(msg);
    msg = msg;
  }
  
  public void progress(long amount) {
    Entry e = buffer[flushed];
    ChannelPromise p = promise;
    if ((p instanceof ChannelProgressivePromise)) {
      long progress = progress + amount;
      progress = progress;
      ((ChannelProgressivePromise)p).tryProgress(progress, total);
    }
  }
  
  public boolean remove() {
    if (isEmpty()) {
      return false;
    }
    
    Entry e = buffer[flushed];
    Object msg = msg;
    if (msg == null) {
      return false;
    }
    
    ChannelPromise promise = promise;
    int size = pendingSize;
    
    e.clear();
    
    flushed = (flushed + 1 & buffer.length - 1);
    
    safeRelease(msg);
    
    promise.trySuccess();
    decrementPendingOutboundBytes(size);
    
    return true;
  }
  
  public boolean remove(Throwable cause) {
    if (isEmpty()) {
      return false;
    }
    
    Entry e = buffer[flushed];
    Object msg = msg;
    if (msg == null) {
      return false;
    }
    
    ChannelPromise promise = promise;
    int size = pendingSize;
    
    e.clear();
    
    flushed = (flushed + 1 & buffer.length - 1);
    
    safeRelease(msg);
    
    safeFail(promise, cause);
    decrementPendingOutboundBytes(size);
    
    return true;
  }
  










  public ByteBuffer[] nioBuffers()
  {
    long nioBufferSize = 0L;
    int nioBufferCount = 0;
    int mask = buffer.length - 1;
    ByteBufAllocator alloc = channel.alloc();
    ByteBuffer[] nioBuffers = this.nioBuffers;
    
    int i = flushed;
    Object m; while ((i != unflushed) && ((m = buffer[i].msg) != null)) {
      if (!(m instanceof ByteBuf)) {
        this.nioBufferCount = 0;
        this.nioBufferSize = 0L;
        return null;
      }
      
      Entry entry = buffer[i];
      ByteBuf buf = (ByteBuf)m;
      int readerIndex = buf.readerIndex();
      int readableBytes = buf.writerIndex() - readerIndex;
      
      if (readableBytes > 0) {
        nioBufferSize += readableBytes;
        int count = count;
        if (count == -1) {
          count = (count = buf.nioBufferCount());
        }
        if (nioBufferCount + count > nioBuffers.length) {
          this.nioBuffers = (nioBuffers = doubleNioBufferArray(nioBuffers, nioBufferCount));
        }
        if ((buf.isDirect()) || (!alloc.isDirectBufferPooled())) {
          if (count == 1) {
            ByteBuffer nioBuf = buf;
            if (nioBuf == null)
            {

              buf = (nioBuf = buf.internalNioBuffer(readerIndex, readableBytes));
            }
            nioBuffers[(nioBufferCount++)] = nioBuf;
          } else {
            ByteBuffer[] nioBufs = buffers;
            if (nioBufs == null)
            {
              buffers = (nioBufs = buf.nioBuffers());
            }
            nioBufferCount = fillBufferArray(nioBufs, nioBuffers, nioBufferCount);
          }
        } else {
          nioBufferCount = fillBufferArrayNonDirect(entry, buf, readerIndex, readableBytes, alloc, nioBuffers, nioBufferCount);
        }
      }
      
      i = i + 1 & mask;
    }
    this.nioBufferCount = nioBufferCount;
    this.nioBufferSize = nioBufferSize;
    
    return nioBuffers;
  }
  
  private static int fillBufferArray(ByteBuffer[] nioBufs, ByteBuffer[] nioBuffers, int nioBufferCount) {
    for (ByteBuffer nioBuf : nioBufs) {
      if (nioBuf == null) {
        break;
      }
      nioBuffers[(nioBufferCount++)] = nioBuf;
    }
    return nioBufferCount;
  }
  
  private static int fillBufferArrayNonDirect(Entry entry, ByteBuf buf, int readerIndex, int readableBytes, ByteBufAllocator alloc, ByteBuffer[] nioBuffers, int nioBufferCount)
  {
    ByteBuf directBuf = alloc.directBuffer(readableBytes);
    directBuf.writeBytes(buf, readerIndex, readableBytes);
    buf.release();
    msg = directBuf;
    
    ByteBuffer nioBuf = entry.buf = directBuf.internalNioBuffer(0, readableBytes);
    count = 1;
    nioBuffers[(nioBufferCount++)] = nioBuf;
    return nioBufferCount;
  }
  
  private static ByteBuffer[] doubleNioBufferArray(ByteBuffer[] array, int size) {
    int newCapacity = array.length << 1;
    if (newCapacity < 0) {
      throw new IllegalStateException();
    }
    
    ByteBuffer[] newArray = new ByteBuffer[newCapacity];
    System.arraycopy(array, 0, newArray, 0, size);
    
    return newArray;
  }
  
  public int nioBufferCount() {
    return nioBufferCount;
  }
  
  public long nioBufferSize() {
    return nioBufferSize;
  }
  
  boolean getWritable() {
    return writable != 0;
  }
  
  public int size() {
    return unflushed - flushed & buffer.length - 1;
  }
  
  public boolean isEmpty() {
    return unflushed == flushed;
  }
  




  void failFlushed(Throwable cause)
  {
    if (inFail) {
      return;
    }
    try
    {
      inFail = true;
      for (;;) {
        if (!remove(cause)) {
          break;
        }
      }
    } finally {
      inFail = false;
    }
  }
  
  void close(final ClosedChannelException cause) {
    if (inFail) {
      channel.eventLoop().execute(new Runnable()
      {
        public void run() {
          close(cause);
        }
      });
      return;
    }
    
    inFail = true;
    
    if (channel.isOpen()) {
      throw new IllegalStateException("close() must be invoked after the channel is closed.");
    }
    
    if (!isEmpty()) {
      throw new IllegalStateException("close() must be invoked after all flushed writes are handled.");
    }
    

    int unflushedCount = tail - unflushed & buffer.length - 1;
    try {
      for (int i = 0; i < unflushedCount; i++) {
        Entry e = buffer[(unflushed + i & buffer.length - 1)];
        safeRelease(msg);
        msg = null;
        safeFail(promise, cause);
        promise = null;
        

        int size = pendingSize;
        long oldValue = totalPendingSize;
        long newWriteBufferSize = oldValue - size;
        while (!TOTAL_PENDING_SIZE_UPDATER.compareAndSet(this, oldValue, newWriteBufferSize)) {
          oldValue = totalPendingSize;
          newWriteBufferSize = oldValue - size;
        }
        
        pendingSize = 0;
      }
    } finally {
      tail = unflushed;
      inFail = false;
    }
    
    recycle();
  }
  
  private static void safeRelease(Object message) {
    try {
      ReferenceCountUtil.release(message);
    } catch (Throwable t) {
      logger.warn("Failed to release a message.", t);
    }
  }
  
  private static void safeFail(ChannelPromise promise, Throwable cause) {
    if ((!(promise instanceof VoidChannelPromise)) && (!promise.tryFailure(cause))) {
      logger.warn("Promise done already: {} - new exception is:", promise, cause);
    }
  }
  
  public void recycle() {
    if (buffer.length > 32) {
      Entry[] e = new Entry[32];
      System.arraycopy(buffer, 0, e, 0, 32);
      buffer = e;
    }
    
    if (nioBuffers.length > 32) {
      nioBuffers = new ByteBuffer[32];
    }
    else
    {
      Arrays.fill(nioBuffers, null);
    }
    


    flushed = 0;
    unflushed = 0;
    tail = 0;
    

    channel = null;
    
    RECYCLER.recycle(this, handle);
  }
  
  private static final class Entry {
    Object msg;
    ByteBuffer[] buffers;
    ByteBuffer buf;
    ChannelPromise promise;
    long progress;
    long total;
    int pendingSize;
    int count = -1;
    
    public void clear() {
      buffers = null;
      buf = null;
      msg = null;
      promise = null;
      progress = 0L;
      total = 0L;
      pendingSize = 0;
      count = -1;
    }
  }
}
