package net.minecraft.util.io.netty.channel.local;

import java.net.SocketAddress;
import java.nio.channels.AlreadyConnectedException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ConnectionPendingException;
import java.nio.channels.NotYetConnectedException;
import java.util.ArrayDeque;
import java.util.Queue;
import net.minecraft.util.io.netty.channel.AbstractChannel;
import net.minecraft.util.io.netty.channel.AbstractChannel.AbstractUnsafe;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.Channel.Unsafe;
import net.minecraft.util.io.netty.channel.ChannelConfig;
import net.minecraft.util.io.netty.channel.ChannelException;
import net.minecraft.util.io.netty.channel.ChannelMetadata;
import net.minecraft.util.io.netty.channel.ChannelOutboundBuffer;
import net.minecraft.util.io.netty.channel.ChannelPipeline;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.channel.DefaultChannelConfig;
import net.minecraft.util.io.netty.channel.EventLoop;
import net.minecraft.util.io.netty.channel.SingleThreadEventLoop;
import net.minecraft.util.io.netty.util.ReferenceCountUtil;
import net.minecraft.util.io.netty.util.concurrent.SingleThreadEventExecutor;

















public class LocalChannel
  extends AbstractChannel
{
  private static final ChannelMetadata METADATA = new ChannelMetadata(false);
  
  private static final int MAX_READER_STACK_DEPTH = 8;
  private static final ThreadLocal<Integer> READER_STACK_DEPTH = new ThreadLocal()
  {
    protected Integer initialValue() {
      return Integer.valueOf(0);
    }
  };
  
  private final ChannelConfig config = new DefaultChannelConfig(this);
  private final Queue<Object> inboundBuffer = new ArrayDeque();
  private final Runnable readTask = new Runnable()
  {
    public void run() {
      ChannelPipeline pipeline = pipeline();
      for (;;) {
        Object m = inboundBuffer.poll();
        if (m == null) {
          break;
        }
        pipeline.fireChannelRead(m);
      }
      pipeline.fireChannelReadComplete();
    }
  };
  
  private final Runnable shutdownHook = new Runnable()
  {
    public void run() {
      unsafe().close(unsafe().voidPromise());
    }
  };
  private volatile int state;
  private volatile LocalChannel peer;
  private volatile LocalAddress localAddress;
  private volatile LocalAddress remoteAddress;
  private volatile ChannelPromise connectPromise;
  private volatile boolean readInProgress;
  
  public LocalChannel()
  {
    super(null);
  }
  
  LocalChannel(LocalServerChannel parent, LocalChannel peer) {
    super(parent);
    this.peer = peer;
    localAddress = parent.localAddress();
    remoteAddress = peer.localAddress();
  }
  
  public ChannelMetadata metadata()
  {
    return METADATA;
  }
  
  public ChannelConfig config()
  {
    return config;
  }
  
  public LocalServerChannel parent()
  {
    return (LocalServerChannel)super.parent();
  }
  
  public LocalAddress localAddress()
  {
    return (LocalAddress)super.localAddress();
  }
  
  public LocalAddress remoteAddress()
  {
    return (LocalAddress)super.remoteAddress();
  }
  
  public boolean isOpen()
  {
    return state < 3;
  }
  
  public boolean isActive()
  {
    return state == 2;
  }
  
  protected AbstractChannel.AbstractUnsafe newUnsafe()
  {
    return new LocalUnsafe(null);
  }
  
  protected boolean isCompatible(EventLoop loop)
  {
    return loop instanceof SingleThreadEventLoop;
  }
  
  protected SocketAddress localAddress0()
  {
    return localAddress;
  }
  
  protected SocketAddress remoteAddress0()
  {
    return remoteAddress;
  }
  
  protected void doRegister() throws Exception
  {
    if (peer != null) {
      state = 2;
      
      peer.remoteAddress = parent().localAddress();
      peer.state = 2;
      




      peer.eventLoop().execute(new Runnable()
      {
        public void run() {
          peer.pipeline().fireChannelActive();
          access$200connectPromise.setSuccess();
        }
      });
    }
    ((SingleThreadEventExecutor)eventLoop()).addShutdownHook(shutdownHook);
  }
  
  protected void doBind(SocketAddress localAddress) throws Exception
  {
    this.localAddress = LocalChannelRegistry.register(this, this.localAddress, localAddress);
    

    state = 1;
  }
  
  protected void doDisconnect() throws Exception
  {
    doClose();
  }
  
  protected void doClose() throws Exception
  {
    if (state <= 2)
    {
      if (localAddress != null) {
        if (parent() == null) {
          LocalChannelRegistry.unregister(localAddress);
        }
        localAddress = null;
      }
      state = 3;
    }
    
    final LocalChannel peer = this.peer;
    if ((peer != null) && (peer.isActive()))
    {

      EventLoop eventLoop = peer.eventLoop();
      if (eventLoop.inEventLoop()) {
        peer.unsafe().close(unsafe().voidPromise());
      } else {
        peer.eventLoop().execute(new Runnable()
        {
          public void run() {
            peer.unsafe().close(unsafe().voidPromise());
          }
        });
      }
      this.peer = null;
    }
  }
  
  protected void doDeregister() throws Exception
  {
    if (isOpen()) {
      unsafe().close(unsafe().voidPromise());
    }
    ((SingleThreadEventExecutor)eventLoop()).removeShutdownHook(shutdownHook);
  }
  
  protected void doBeginRead() throws Exception
  {
    if (readInProgress) {
      return;
    }
    
    ChannelPipeline pipeline = pipeline();
    Queue<Object> inboundBuffer = this.inboundBuffer;
    if (inboundBuffer.isEmpty()) {
      readInProgress = true;
      return;
    }
    
    Integer stackDepth = (Integer)READER_STACK_DEPTH.get();
    if (stackDepth.intValue() < 8) {
      READER_STACK_DEPTH.set(Integer.valueOf(stackDepth.intValue() + 1));
      try {
        for (;;) {
          Object received = inboundBuffer.poll();
          if (received == null) {
            break;
          }
          pipeline.fireChannelRead(received);
        }
        pipeline.fireChannelReadComplete();
      } finally {
        READER_STACK_DEPTH.set(stackDepth);
      }
    } else {
      eventLoop().execute(readTask);
    }
  }
  
  protected void doWrite(ChannelOutboundBuffer in) throws Exception
  {
    if (state < 2) {
      throw new NotYetConnectedException();
    }
    if (state > 2) {
      throw new ClosedChannelException();
    }
    
    final LocalChannel peer = this.peer;
    final ChannelPipeline peerPipeline = peer.pipeline();
    EventLoop peerLoop = peer.eventLoop();
    
    if (peerLoop == eventLoop()) {
      for (;;) {
        Object msg = in.current();
        if (msg == null) {
          break;
        }
        inboundBuffer.add(msg);
        ReferenceCountUtil.retain(msg);
        in.remove();
      }
      finishPeerRead(peer, peerPipeline);
    }
    else {
      final Object[] msgsCopy = new Object[in.size()];
      for (int i = 0; i < msgsCopy.length; i++) {
        msgsCopy[i] = ReferenceCountUtil.retain(in.current());
        in.remove();
      }
      
      peerLoop.execute(new Runnable()
      {
        public void run() {
          for (int i = 0; i < msgsCopy.length; i++) {
            peerinboundBuffer.add(msgsCopy[i]);
          }
          LocalChannel.finishPeerRead(peer, peerPipeline);
        }
      });
    }
  }
  
  private static void finishPeerRead(LocalChannel peer, ChannelPipeline peerPipeline) {
    if (readInProgress) {
      readInProgress = false;
      for (;;) {
        Object received = inboundBuffer.poll();
        if (received == null) {
          break;
        }
        peerPipeline.fireChannelRead(received);
      }
      peerPipeline.fireChannelReadComplete();
    }
  }
  
  private class LocalUnsafe extends AbstractChannel.AbstractUnsafe { private LocalUnsafe() { super(); }
    

    public void connect(SocketAddress remoteAddress, SocketAddress localAddress, ChannelPromise promise)
    {
      if (!ensureOpen(promise)) {
        return;
      }
      
      if (state == 2) {
        Exception cause = new AlreadyConnectedException();
        promise.setFailure(cause);
        pipeline().fireExceptionCaught(cause);
        return;
      }
      
      if (connectPromise != null) {
        throw new ConnectionPendingException();
      }
      
      connectPromise = promise;
      
      if (state != 1)
      {
        if (localAddress == null) {
          localAddress = new LocalAddress(LocalChannel.this);
        }
      }
      
      if (localAddress != null) {
        try {
          doBind(localAddress);
        } catch (Throwable t) {
          promise.setFailure(t);
          pipeline().fireExceptionCaught(t);
          close(voidPromise());
          return;
        }
      }
      
      Channel boundChannel = LocalChannelRegistry.get(remoteAddress);
      if (!(boundChannel instanceof LocalServerChannel)) {
        Exception cause = new ChannelException("connection refused");
        promise.setFailure(cause);
        close(voidPromise());
        return;
      }
      
      LocalServerChannel serverChannel = (LocalServerChannel)boundChannel;
      peer = serverChannel.serve(LocalChannel.this);
    }
  }
}
