package net.minecraft.util.io.netty.channel.nio;

import java.io.IOException;
import java.net.ConnectException;
import java.net.SocketAddress;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import net.minecraft.util.io.netty.channel.AbstractChannel;
import net.minecraft.util.io.netty.channel.AbstractChannel.AbstractUnsafe;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.Channel.Unsafe;
import net.minecraft.util.io.netty.channel.ChannelConfig;
import net.minecraft.util.io.netty.channel.ChannelException;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelPipeline;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.channel.ConnectTimeoutException;
import net.minecraft.util.io.netty.channel.EventLoop;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;















public abstract class AbstractNioChannel
  extends AbstractChannel
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(AbstractNioChannel.class);
  

  private final SelectableChannel ch;
  

  protected final int readInterestOp;
  

  private volatile SelectionKey selectionKey;
  

  private volatile boolean inputShutdown;
  

  private ChannelPromise connectPromise;
  
  private ScheduledFuture<?> connectTimeoutFuture;
  
  private SocketAddress requestedRemoteAddress;
  

  protected AbstractNioChannel(Channel parent, SelectableChannel ch, int readInterestOp)
  {
    super(parent);
    this.ch = ch;
    this.readInterestOp = readInterestOp;
    try {
      ch.configureBlocking(false);
    } catch (IOException e) {
      try {
        ch.close();
      } catch (IOException e2) {
        if (logger.isWarnEnabled()) {
          logger.warn("Failed to close a partially initialized socket.", e2);
        }
      }
      

      throw new ChannelException("Failed to enter non-blocking mode.", e);
    }
  }
  
  public boolean isOpen()
  {
    return ch.isOpen();
  }
  
  public NioUnsafe unsafe()
  {
    return (NioUnsafe)super.unsafe();
  }
  
  protected SelectableChannel javaChannel() {
    return ch;
  }
  
  public NioEventLoop eventLoop()
  {
    return (NioEventLoop)super.eventLoop();
  }
  


  protected SelectionKey selectionKey()
  {
    assert (selectionKey != null);
    return selectionKey;
  }
  


  protected boolean isInputShutdown()
  {
    return inputShutdown;
  }
  


  void setInputShutdown()
  {
    inputShutdown = true;
  }
  
















  protected abstract class AbstractNioUnsafe
    extends AbstractChannel.AbstractUnsafe
    implements AbstractNioChannel.NioUnsafe
  {
    protected AbstractNioUnsafe()
    {
      super();
    }
    
    public SelectableChannel ch() {
      return javaChannel();
    }
    

    public void connect(final SocketAddress remoteAddress, SocketAddress localAddress, ChannelPromise promise)
    {
      if (!ensureOpen(promise)) {
        return;
      }
      try
      {
        if (connectPromise != null) {
          throw new IllegalStateException("connection attempt already made");
        }
        
        boolean wasActive = isActive();
        if (doConnect(remoteAddress, localAddress)) {
          promise.setSuccess();
          if ((!wasActive) && (isActive())) {
            pipeline().fireChannelActive();
          }
        } else {
          connectPromise = promise;
          requestedRemoteAddress = remoteAddress;
          

          int connectTimeoutMillis = config().getConnectTimeoutMillis();
          if (connectTimeoutMillis > 0) {
            connectTimeoutFuture = eventLoop().schedule(new Runnable()
            {
              public void run() {
                ChannelPromise connectPromise = AbstractNioChannel.this.connectPromise;
                ConnectTimeoutException cause = new ConnectTimeoutException("connection timed out: " + remoteAddress);
                
                if ((connectPromise != null) && (connectPromise.tryFailure(cause)))
                  close(voidPromise()); } }, connectTimeoutMillis, TimeUnit.MILLISECONDS);
          }
          



          promise.addListener(new ChannelFutureListener()
          {
            public void operationComplete(ChannelFuture future) throws Exception {
              if (future.isCancelled()) {
                if (connectTimeoutFuture != null) {
                  connectTimeoutFuture.cancel(false);
                }
                connectPromise = null;
                close(voidPromise());
              }
            }
          });
        }
      } catch (Throwable t) {
        if ((t instanceof ConnectException)) {
          Throwable newT = new ConnectException(t.getMessage() + ": " + remoteAddress);
          newT.setStackTrace(t.getStackTrace());
          t = newT;
        }
        closeIfClosed();
        promise.tryFailure(t);
      }
    }
    



    public void finishConnect()
    {
      assert (eventLoop().inEventLoop());
      assert (connectPromise != null);
      try
      {
        boolean wasActive = isActive();
        doFinishConnect();
        connectPromise.setSuccess();
        if ((!wasActive) && (isActive())) {
          pipeline().fireChannelActive();
        }
      } catch (Throwable t) {
        if ((t instanceof ConnectException)) {
          Throwable newT = new ConnectException(t.getMessage() + ": " + requestedRemoteAddress);
          newT.setStackTrace(t.getStackTrace());
          t = newT;
        }
        
        connectPromise.setFailure(t);
        closeIfClosed();
      }
      finally
      {
        if (connectTimeoutFuture != null) {
          connectTimeoutFuture.cancel(false);
        }
        connectPromise = null;
      }
    }
    



    protected void flush0()
    {
      if (isFlushPending()) {
        return;
      }
      super.flush0();
    }
    

    public void forceFlush()
    {
      super.flush0();
    }
    
    private boolean isFlushPending() {
      SelectionKey selectionKey = selectionKey();
      return (selectionKey.isValid()) && ((selectionKey.interestOps() & 0x4) != 0);
    }
  }
  
  protected boolean isCompatible(EventLoop loop)
  {
    return loop instanceof NioEventLoop;
  }
  
  protected void doRegister() throws Exception
  {
    boolean selected = false;
    for (;;) {
      try {
        selectionKey = javaChannel().register(eventLoop()selector, 0, this);
        return;
      } catch (CancelledKeyException e) {
        if (!selected)
        {

          eventLoop().selectNow();
          selected = true;
        }
        else
        {
          throw e;
        }
      }
    }
  }
  
  protected void doDeregister() throws Exception
  {
    eventLoop().cancel(selectionKey());
  }
  
  protected void doBeginRead() throws Exception
  {
    if (inputShutdown) {
      return;
    }
    
    SelectionKey selectionKey = this.selectionKey;
    if (!selectionKey.isValid()) {
      return;
    }
    
    int interestOps = selectionKey.interestOps();
    if ((interestOps & readInterestOp) == 0) {
      selectionKey.interestOps(interestOps | readInterestOp);
    }
  }
  
  protected abstract boolean doConnect(SocketAddress paramSocketAddress1, SocketAddress paramSocketAddress2)
    throws Exception;
  
  protected abstract void doFinishConnect()
    throws Exception;
  
  public static abstract interface NioUnsafe
    extends Channel.Unsafe
  {
    public abstract SelectableChannel ch();
    
    public abstract void finishConnect();
    
    public abstract void read();
    
    public abstract void forceFlush();
  }
}
