package net.minecraft.util.io.netty.handler.codec;

import java.nio.ByteOrder;
import java.util.List;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.buffer.ByteBufAllocator;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;




























































































































































































public class LengthFieldBasedFrameDecoder
  extends ByteToMessageDecoder
{
  private final ByteOrder byteOrder;
  private final int maxFrameLength;
  private final int lengthFieldOffset;
  private final int lengthFieldLength;
  private final int lengthFieldEndOffset;
  private final int lengthAdjustment;
  private final int initialBytesToStrip;
  private final boolean failFast;
  private boolean discardingTooLongFrame;
  private long tooLongFrameLength;
  private long bytesToDiscard;
  
  public LengthFieldBasedFrameDecoder(int maxFrameLength, int lengthFieldOffset, int lengthFieldLength)
  {
    this(maxFrameLength, lengthFieldOffset, lengthFieldLength, 0, 0);
  }
  


















  public LengthFieldBasedFrameDecoder(int maxFrameLength, int lengthFieldOffset, int lengthFieldLength, int lengthAdjustment, int initialBytesToStrip)
  {
    this(maxFrameLength, lengthFieldOffset, lengthFieldLength, lengthAdjustment, initialBytesToStrip, true);
  }
  



























  public LengthFieldBasedFrameDecoder(int maxFrameLength, int lengthFieldOffset, int lengthFieldLength, int lengthAdjustment, int initialBytesToStrip, boolean failFast)
  {
    this(ByteOrder.BIG_ENDIAN, maxFrameLength, lengthFieldOffset, lengthFieldLength, lengthAdjustment, initialBytesToStrip, failFast);
  }
  




























  public LengthFieldBasedFrameDecoder(ByteOrder byteOrder, int maxFrameLength, int lengthFieldOffset, int lengthFieldLength, int lengthAdjustment, int initialBytesToStrip, boolean failFast)
  {
    if (byteOrder == null) {
      throw new NullPointerException("byteOrder");
    }
    
    if (maxFrameLength <= 0) {
      throw new IllegalArgumentException("maxFrameLength must be a positive integer: " + maxFrameLength);
    }
    


    if (lengthFieldOffset < 0) {
      throw new IllegalArgumentException("lengthFieldOffset must be a non-negative integer: " + lengthFieldOffset);
    }
    


    if (initialBytesToStrip < 0) {
      throw new IllegalArgumentException("initialBytesToStrip must be a non-negative integer: " + initialBytesToStrip);
    }
    


    if ((lengthFieldLength != 1) && (lengthFieldLength != 2) && (lengthFieldLength != 3) && (lengthFieldLength != 4) && (lengthFieldLength != 8))
    {

      throw new IllegalArgumentException("lengthFieldLength must be either 1, 2, 3, 4, or 8: " + lengthFieldLength);
    }
    


    if (lengthFieldOffset > maxFrameLength - lengthFieldLength) {
      throw new IllegalArgumentException("maxFrameLength (" + maxFrameLength + ") " + "must be equal to or greater than " + "lengthFieldOffset (" + lengthFieldOffset + ") + " + "lengthFieldLength (" + lengthFieldLength + ").");
    }
    




    this.byteOrder = byteOrder;
    this.maxFrameLength = maxFrameLength;
    this.lengthFieldOffset = lengthFieldOffset;
    this.lengthFieldLength = lengthFieldLength;
    this.lengthAdjustment = lengthAdjustment;
    lengthFieldEndOffset = (lengthFieldOffset + lengthFieldLength);
    this.initialBytesToStrip = initialBytesToStrip;
    this.failFast = failFast;
  }
  
  protected final void decode(ChannelHandlerContext ctx, ByteBuf in, List<Object> out) throws Exception
  {
    Object decoded = decode(ctx, in);
    if (decoded != null) {
      out.add(decoded);
    }
  }
  






  protected Object decode(ChannelHandlerContext ctx, ByteBuf in)
    throws Exception
  {
    if (discardingTooLongFrame) {
      long bytesToDiscard = this.bytesToDiscard;
      int localBytesToDiscard = (int)Math.min(bytesToDiscard, in.readableBytes());
      in.skipBytes(localBytesToDiscard);
      bytesToDiscard -= localBytesToDiscard;
      this.bytesToDiscard = bytesToDiscard;
      
      failIfNecessary(false);
    }
    
    if (in.readableBytes() < lengthFieldEndOffset) {
      return null;
    }
    
    int actualLengthFieldOffset = in.readerIndex() + lengthFieldOffset;
    long frameLength = getFrameLength(in, actualLengthFieldOffset);
    
    if (frameLength < 0L) {
      in.skipBytes(lengthFieldEndOffset);
      throw new CorruptedFrameException("negative pre-adjustment length field: " + frameLength);
    }
    

    frameLength += lengthAdjustment + lengthFieldEndOffset;
    
    if (frameLength < lengthFieldEndOffset) {
      in.skipBytes(lengthFieldEndOffset);
      throw new CorruptedFrameException("Adjusted frame length (" + frameLength + ") is less " + "than lengthFieldEndOffset: " + lengthFieldEndOffset);
    }
    


    if (frameLength > maxFrameLength) {
      long discard = frameLength - in.readableBytes();
      tooLongFrameLength = frameLength;
      
      if (discard < 0L)
      {
        in.skipBytes((int)frameLength);
      }
      else {
        discardingTooLongFrame = true;
        this.bytesToDiscard = discard;
        in.skipBytes(in.readableBytes());
      }
      failIfNecessary(true);
      return null;
    }
    

    int frameLengthInt = (int)frameLength;
    if (in.readableBytes() < frameLengthInt) {
      return null;
    }
    
    if (initialBytesToStrip > frameLengthInt) {
      in.skipBytes(frameLengthInt);
      throw new CorruptedFrameException("Adjusted frame length (" + frameLength + ") is less " + "than initialBytesToStrip: " + initialBytesToStrip);
    }
    

    in.skipBytes(initialBytesToStrip);
    

    int readerIndex = in.readerIndex();
    int actualFrameLength = frameLengthInt - initialBytesToStrip;
    ByteBuf frame = extractFrame(ctx, in, readerIndex, actualFrameLength);
    in.readerIndex(readerIndex + actualFrameLength);
    return frame;
  }
  
  private long getFrameLength(ByteBuf in, int actualLengthFieldOffset) {
    in = in.order(byteOrder);
    long frameLength;
    switch (lengthFieldLength) {
    case 1: 
      frameLength = in.getUnsignedByte(actualLengthFieldOffset);
      break;
    case 2: 
      frameLength = in.getUnsignedShort(actualLengthFieldOffset);
      break;
    case 3: 
      frameLength = in.getUnsignedMedium(actualLengthFieldOffset);
      break;
    case 4: 
      frameLength = in.getUnsignedInt(actualLengthFieldOffset);
      break;
    case 8: 
      frameLength = in.getLong(actualLengthFieldOffset);
      break;
    case 5: case 6: case 7: default: 
      throw new Error("should not reach here");
    }
    return frameLength;
  }
  
  private void failIfNecessary(boolean firstDetectionOfTooLongFrame) {
    if (bytesToDiscard == 0L)
    {

      long tooLongFrameLength = this.tooLongFrameLength;
      this.tooLongFrameLength = 0L;
      discardingTooLongFrame = false;
      if ((!failFast) || ((failFast) && (firstDetectionOfTooLongFrame)))
      {
        fail(tooLongFrameLength);
      }
      
    }
    else if ((failFast) && (firstDetectionOfTooLongFrame)) {
      fail(this.tooLongFrameLength);
    }
  }
  











  protected ByteBuf extractFrame(ChannelHandlerContext ctx, ByteBuf buffer, int index, int length)
  {
    ByteBuf frame = ctx.alloc().buffer(length);
    frame.writeBytes(buffer, index, length);
    return frame;
  }
  
  private void fail(long frameLength) {
    if (frameLength > 0L) {
      throw new TooLongFrameException("Adjusted frame length exceeds " + maxFrameLength + ": " + frameLength + " - discarded");
    }
    

    throw new TooLongFrameException("Adjusted frame length exceeds " + maxFrameLength + " - discarding");
  }
}
