package net.minecraft.util.io.netty.handler.codec.compression;

import java.util.concurrent.TimeUnit;
import java.util.zip.CRC32;
import java.util.zip.Deflater;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.buffer.ByteBufAllocator;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.channel.ChannelPromiseNotifier;
import net.minecraft.util.io.netty.util.concurrent.EventExecutor;



















public class JdkZlibEncoder
  extends ZlibEncoder
{
  private final byte[] encodeBuf = new byte[8192];
  
  private final Deflater deflater;
  
  private volatile boolean finished;
  
  private volatile ChannelHandlerContext ctx;
  
  private final boolean gzip;
  private final CRC32 crc = new CRC32();
  private static final byte[] gzipHeader = { 31, -117, 8, 0, 0, 0, 0, 0, 0, 0 };
  private boolean writeHeader = true;
  





  public JdkZlibEncoder()
  {
    this(6);
  }
  










  public JdkZlibEncoder(int compressionLevel)
  {
    this(ZlibWrapper.ZLIB, compressionLevel);
  }
  





  public JdkZlibEncoder(ZlibWrapper wrapper)
  {
    this(wrapper, 6);
  }
  










  public JdkZlibEncoder(ZlibWrapper wrapper, int compressionLevel)
  {
    if ((compressionLevel < 0) || (compressionLevel > 9)) {
      throw new IllegalArgumentException("compressionLevel: " + compressionLevel + " (expected: 0-9)");
    }
    
    if (wrapper == null) {
      throw new NullPointerException("wrapper");
    }
    if (wrapper == ZlibWrapper.ZLIB_OR_NONE) {
      throw new IllegalArgumentException("wrapper '" + ZlibWrapper.ZLIB_OR_NONE + "' is not " + "allowed for compression.");
    }
    


    gzip = (wrapper == ZlibWrapper.GZIP);
    deflater = new Deflater(compressionLevel, wrapper != ZlibWrapper.ZLIB);
  }
  









  public JdkZlibEncoder(byte[] dictionary)
  {
    this(6, dictionary);
  }
  













  public JdkZlibEncoder(int compressionLevel, byte[] dictionary)
  {
    if ((compressionLevel < 0) || (compressionLevel > 9)) {
      throw new IllegalArgumentException("compressionLevel: " + compressionLevel + " (expected: 0-9)");
    }
    
    if (dictionary == null) {
      throw new NullPointerException("dictionary");
    }
    
    gzip = false;
    deflater = new Deflater(compressionLevel);
    deflater.setDictionary(dictionary);
  }
  
  public ChannelFuture close()
  {
    return close(ctx().newPromise());
  }
  
  public ChannelFuture close(final ChannelPromise promise)
  {
    ChannelHandlerContext ctx = ctx();
    EventExecutor executor = ctx.executor();
    if (executor.inEventLoop()) {
      return finishEncode(ctx, promise);
    }
    final ChannelPromise p = ctx.newPromise();
    executor.execute(new Runnable()
    {
      public void run() {
        ChannelFuture f = JdkZlibEncoder.this.finishEncode(JdkZlibEncoder.access$000(JdkZlibEncoder.this), p);
        f.addListener(new ChannelPromiseNotifier(new ChannelPromise[] { promise }));
      }
    });
    return p;
  }
  
  private ChannelHandlerContext ctx()
  {
    ChannelHandlerContext ctx = this.ctx;
    if (ctx == null) {
      throw new IllegalStateException("not added to a pipeline");
    }
    return ctx;
  }
  
  public boolean isClosed()
  {
    return finished;
  }
  
  protected void encode(ChannelHandlerContext ctx, ByteBuf uncompressed, ByteBuf out) throws Exception
  {
    if (finished) {
      out.writeBytes(uncompressed);
      return;
    }
    
    byte[] inAry = new byte[uncompressed.readableBytes()];
    uncompressed.readBytes(inAry);
    
    int sizeEstimate = (int)Math.ceil(inAry.length * 1.001D) + 12;
    out.ensureWritable(sizeEstimate);
    
    if (gzip) {
      crc.update(inAry);
      if (writeHeader) {
        out.writeBytes(gzipHeader);
        writeHeader = false;
      }
    }
    
    deflater.setInput(inAry);
    while (!deflater.needsInput()) {
      int numBytes = deflater.deflate(encodeBuf, 0, encodeBuf.length, 2);
      out.writeBytes(encodeBuf, 0, numBytes);
    }
  }
  
  public void close(final ChannelHandlerContext ctx, final ChannelPromise promise) throws Exception
  {
    ChannelFuture f = finishEncode(ctx, ctx.newPromise());
    f.addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture f) throws Exception {
        ctx.close(promise);
      }
    });
    
    if (!f.isDone())
    {
      ctx.executor().schedule(new Runnable()
      {

        public void run() { ctx.close(promise); } }, 10L, TimeUnit.SECONDS);
    }
  }
  

  private ChannelFuture finishEncode(ChannelHandlerContext ctx, ChannelPromise promise)
  {
    if (finished) {
      promise.setSuccess();
      return promise;
    }
    
    finished = true;
    ByteBuf footer = ctx.alloc().buffer();
    deflater.finish();
    while (!deflater.finished()) {
      int numBytes = deflater.deflate(encodeBuf, 0, encodeBuf.length);
      footer.writeBytes(encodeBuf, 0, numBytes);
    }
    if (gzip) {
      int crcValue = (int)crc.getValue();
      int uncBytes = deflater.getTotalIn();
      footer.writeByte(crcValue);
      footer.writeByte(crcValue >>> 8);
      footer.writeByte(crcValue >>> 16);
      footer.writeByte(crcValue >>> 24);
      footer.writeByte(uncBytes);
      footer.writeByte(uncBytes >>> 8);
      footer.writeByte(uncBytes >>> 16);
      footer.writeByte(uncBytes >>> 24);
    }
    deflater.end();
    return ctx.writeAndFlush(footer, promise);
  }
  
  public void handlerAdded(ChannelHandlerContext ctx) throws Exception
  {
    this.ctx = ctx;
  }
}
