package net.minecraft.util.io.netty.handler.codec.http.websocketx;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.channel.ChannelPipeline;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.handler.codec.http.FullHttpRequest;
import net.minecraft.util.io.netty.handler.codec.http.FullHttpResponse;
import net.minecraft.util.io.netty.handler.codec.http.HttpContentCompressor;
import net.minecraft.util.io.netty.handler.codec.http.HttpHeaders;
import net.minecraft.util.io.netty.handler.codec.http.HttpObjectAggregator;
import net.minecraft.util.io.netty.handler.codec.http.HttpRequestDecoder;
import net.minecraft.util.io.netty.handler.codec.http.HttpResponseEncoder;
import net.minecraft.util.io.netty.handler.codec.http.HttpServerCodec;
import net.minecraft.util.io.netty.util.internal.EmptyArrays;
import net.minecraft.util.io.netty.util.internal.StringUtil;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;



















public abstract class WebSocketServerHandshaker
{
  protected static final InternalLogger logger = InternalLoggerFactory.getInstance(WebSocketServerHandshaker.class);
  



  private final String uri;
  



  private final String[] subprotocols;
  


  private final WebSocketVersion version;
  


  private final int maxFramePayloadLength;
  


  private String selectedSubprotocol;
  


  public static final String SUB_PROTOCOL_WILDCARD = "*";
  



  protected WebSocketServerHandshaker(WebSocketVersion version, String uri, String subprotocols, int maxFramePayloadLength)
  {
    this.version = version;
    this.uri = uri;
    if (subprotocols != null) {
      String[] subprotocolArray = StringUtil.split(subprotocols, ',');
      for (int i = 0; i < subprotocolArray.length; i++) {
        subprotocolArray[i] = subprotocolArray[i].trim();
      }
      this.subprotocols = subprotocolArray;
    } else {
      this.subprotocols = EmptyArrays.EMPTY_STRINGS;
    }
    this.maxFramePayloadLength = maxFramePayloadLength;
  }
  


  public String uri()
  {
    return uri;
  }
  


  public Set<String> subprotocols()
  {
    Set<String> ret = new LinkedHashSet();
    Collections.addAll(ret, subprotocols);
    return ret;
  }
  


  public WebSocketVersion version()
  {
    return version;
  }
  




  public int maxFramePayloadLength()
  {
    return maxFramePayloadLength;
  }
  










  public ChannelFuture handshake(Channel channel, FullHttpRequest req)
  {
    return handshake(channel, req, null, channel.newPromise());
  }
  

















  public final ChannelFuture handshake(Channel channel, FullHttpRequest req, HttpHeaders responseHeaders, final ChannelPromise promise)
  {
    if (logger.isDebugEnabled()) {
      logger.debug(String.format("%s WS Version %s server handshake", new Object[] { channel, version() }));
    }
    FullHttpResponse response = newHandshakeResponse(req, responseHeaders);
    channel.writeAndFlush(response).addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture future) throws Exception {
        if (future.isSuccess()) {
          ChannelPipeline p = future.channel().pipeline();
          if (p.get(HttpObjectAggregator.class) != null) {
            p.remove(HttpObjectAggregator.class);
          }
          if (p.get(HttpContentCompressor.class) != null) {
            p.remove(HttpContentCompressor.class);
          }
          ChannelHandlerContext ctx = p.context(HttpRequestDecoder.class);
          if (ctx == null)
          {
            ctx = p.context(HttpServerCodec.class);
            if (ctx == null) {
              promise.setFailure(new IllegalStateException("No HttpDecoder and no HttpServerCodec in the pipeline"));
              
              return;
            }
            p.addBefore(ctx.name(), "wsdecoder", newWebsocketDecoder());
            p.replace(ctx.name(), "wsencoder", newWebSocketEncoder());
          } else {
            p.replace(ctx.name(), "wsdecoder", newWebsocketDecoder());
            
            p.replace(HttpResponseEncoder.class, "wsencoder", newWebSocketEncoder());
          }
          promise.setSuccess();
        } else {
          promise.setFailure(future.cause());
        }
      }
    });
    return promise;
  }
  





  protected abstract FullHttpResponse newHandshakeResponse(FullHttpRequest paramFullHttpRequest, HttpHeaders paramHttpHeaders);
  





  public ChannelFuture close(Channel channel, CloseWebSocketFrame frame)
  {
    if (channel == null) {
      throw new NullPointerException("channel");
    }
    return close(channel, frame, channel.newPromise());
  }
  









  public ChannelFuture close(Channel channel, CloseWebSocketFrame frame, ChannelPromise promise)
  {
    if (channel == null) {
      throw new NullPointerException("channel");
    }
    return channel.writeAndFlush(frame, promise).addListener(ChannelFutureListener.CLOSE);
  }
  






  protected String selectSubprotocol(String requestedSubprotocols)
  {
    if ((requestedSubprotocols == null) || (subprotocols.length == 0)) {
      return null;
    }
    
    String[] requestedSubprotocolArray = StringUtil.split(requestedSubprotocols, ',');
    for (String p : requestedSubprotocolArray) {
      String requestedSubprotocol = p.trim();
      
      for (String supportedSubprotocol : subprotocols) {
        if (("*".equals(supportedSubprotocol)) || (requestedSubprotocol.equals(supportedSubprotocol)))
        {
          selectedSubprotocol = requestedSubprotocol;
          return requestedSubprotocol;
        }
      }
    }
    

    return null;
  }
  





  public String selectedSubprotocol()
  {
    return selectedSubprotocol;
  }
  
  protected abstract WebSocketFrameDecoder newWebsocketDecoder();
  
  protected abstract WebSocketFrameEncoder newWebSocketEncoder();
}
