package net.minecraft.util.io.netty.handler.codec.spdy;

import java.util.List;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.buffer.ByteBufAllocator;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.handler.codec.ByteToMessageDecoder;









































public class SpdyFrameDecoder
  extends ByteToMessageDecoder
{
  private static final SpdyProtocolException INVALID_FRAME = new SpdyProtocolException("Received invalid frame");
  
  private final int spdyVersion;
  
  private final int maxChunkSize;
  
  private final SpdyHeaderBlockDecoder headerBlockDecoder;
  
  private State state;
  
  private SpdySettingsFrame spdySettingsFrame;
  private SpdyHeadersFrame spdyHeadersFrame;
  private byte flags;
  private int length;
  private int version;
  private int type;
  private int streamId;
  
  private static enum State
  {
    READ_COMMON_HEADER, 
    READ_CONTROL_FRAME, 
    READ_SETTINGS_FRAME, 
    READ_HEADER_BLOCK_FRAME, 
    READ_HEADER_BLOCK, 
    READ_DATA_FRAME, 
    DISCARD_FRAME, 
    FRAME_ERROR;
    

    private State() {}
  }
  
  public SpdyFrameDecoder(int version)
  {
    this(version, 8192, 16384);
  }
  


  public SpdyFrameDecoder(int version, int maxChunkSize, int maxHeaderSize)
  {
    this(version, maxChunkSize, SpdyHeaderBlockDecoder.newInstance(version, maxHeaderSize));
  }
  
  protected SpdyFrameDecoder(int version, int maxChunkSize, SpdyHeaderBlockDecoder headerBlockDecoder)
  {
    if ((version < 2) || (version > 3)) {
      throw new IllegalArgumentException("unsupported version: " + version);
    }
    
    if (maxChunkSize <= 0) {
      throw new IllegalArgumentException("maxChunkSize must be a positive integer: " + maxChunkSize);
    }
    
    spdyVersion = version;
    this.maxChunkSize = maxChunkSize;
    this.headerBlockDecoder = headerBlockDecoder;
    state = State.READ_COMMON_HEADER;
  }
  
  public void decodeLast(ChannelHandlerContext ctx, ByteBuf in, List<Object> out) throws Exception
  {
    try {
      decode(ctx, in, out);
    } finally {
      headerBlockDecoder.end();
    }
  }
  
  protected void decode(ChannelHandlerContext ctx, ByteBuf buffer, List<Object> out) throws Exception
  {
    switch (1.$SwitchMap$io$netty$handler$codec$spdy$SpdyFrameDecoder$State[state.ordinal()]) {
    case 1: 
      state = readCommonHeader(buffer);
      if (state == State.FRAME_ERROR) {
        if (version != spdyVersion) {
          fireProtocolException(ctx, "Unsupported version: " + version);
        } else {
          fireInvalidFrameException(ctx);
        }
      }
      


      if (length == 0) {
        if (state == State.READ_DATA_FRAME) {
          SpdyDataFrame spdyDataFrame = new DefaultSpdyDataFrame(streamId);
          spdyDataFrame.setLast((flags & 0x1) != 0);
          state = State.READ_COMMON_HEADER;
          out.add(spdyDataFrame);
          return;
        }
        
        state = State.READ_COMMON_HEADER;
      }
      
      return;
    case 2: 
      try
      {
        Object frame = readControlFrame(buffer);
        if (frame != null) {
          state = State.READ_COMMON_HEADER;
          out.add(frame);
        }
        return;
      } catch (IllegalArgumentException e) {
        state = State.FRAME_ERROR;
        fireInvalidFrameException(ctx);
        
        return;
      }
    case 3: 
      if (spdySettingsFrame == null)
      {
        if (buffer.readableBytes() < 4) {
          return;
        }
        int numEntries = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex());
        buffer.skipBytes(4);
        length -= 4;
        

        if (((length & 0x7) != 0) || (length >> 3 != numEntries)) {
          state = State.FRAME_ERROR;
          fireInvalidFrameException(ctx);
          return;
        }
        
        spdySettingsFrame = new DefaultSpdySettingsFrame();
        
        boolean clear = (flags & 0x1) != 0;
        spdySettingsFrame.setClearPreviouslyPersistedSettings(clear);
      }
      
      int readableEntries = Math.min(buffer.readableBytes() >> 3, length >> 3);
      for (int i = 0; i < readableEntries; i++) { byte ID_flags;
        byte ID_flags;
        int ID;
        if (version < 3)
        {


          int ID = buffer.readByte() & 0xFF | (buffer.readByte() & 0xFF) << 8 | (buffer.readByte() & 0xFF) << 16;
          

          ID_flags = buffer.readByte();
        } else {
          ID_flags = buffer.readByte();
          ID = SpdyCodecUtil.getUnsignedMedium(buffer, buffer.readerIndex());
          buffer.skipBytes(3);
        }
        int value = SpdyCodecUtil.getSignedInt(buffer, buffer.readerIndex());
        buffer.skipBytes(4);
        

        if (ID == 0) {
          state = State.FRAME_ERROR;
          spdySettingsFrame = null;
          fireInvalidFrameException(ctx);
          return;
        }
        
        if (!spdySettingsFrame.isSet(ID)) {
          boolean persistVal = (ID_flags & 0x1) != 0;
          boolean persisted = (ID_flags & 0x2) != 0;
          spdySettingsFrame.setValue(ID, value, persistVal, persisted);
        }
      }
      
      length -= 8 * readableEntries;
      if (length == 0) {
        state = State.READ_COMMON_HEADER;
        Object frame = spdySettingsFrame;
        spdySettingsFrame = null;
        out.add(frame);
        return;
      }
      return;
    case 4: 
      try
      {
        spdyHeadersFrame = readHeaderBlockFrame(buffer);
        if (spdyHeadersFrame != null) {
          if (length == 0) {
            state = State.READ_COMMON_HEADER;
            Object frame = spdyHeadersFrame;
            spdyHeadersFrame = null;
            out.add(frame);
            return;
          }
          state = State.READ_HEADER_BLOCK;
        }
        return;
      } catch (IllegalArgumentException e) {
        state = State.FRAME_ERROR;
        fireInvalidFrameException(ctx);
        return;
      }
    
    case 5: 
      int compressedBytes = Math.min(buffer.readableBytes(), length);
      ByteBuf compressed = buffer.slice(buffer.readerIndex(), compressedBytes);
      try
      {
        headerBlockDecoder.decode(compressed, spdyHeadersFrame);
      } catch (Exception e) {
        state = State.FRAME_ERROR;
        spdyHeadersFrame = null;
        ctx.fireExceptionCaught(e);
        return;
      }
      
      int readBytes = compressedBytes - compressed.readableBytes();
      buffer.skipBytes(readBytes);
      length -= readBytes;
      
      if ((spdyHeadersFrame != null) && ((spdyHeadersFrame.isInvalid()) || (spdyHeadersFrame.isTruncated())))
      {

        Object frame = spdyHeadersFrame;
        spdyHeadersFrame = null;
        if (length == 0) {
          headerBlockDecoder.reset();
          state = State.READ_COMMON_HEADER;
        }
        out.add(frame);
        return;
      }
      
      if (length == 0) {
        Object frame = spdyHeadersFrame;
        spdyHeadersFrame = null;
        headerBlockDecoder.reset();
        state = State.READ_COMMON_HEADER;
        if (frame != null) {
          out.add(frame);
        }
      }
      return;
    
    case 6: 
      if (streamId == 0) {
        state = State.FRAME_ERROR;
        fireProtocolException(ctx, "Received invalid data frame");
        return;
      }
      

      int dataLength = Math.min(maxChunkSize, length);
      

      if (buffer.readableBytes() < dataLength) {
        return;
      }
      
      ByteBuf data = ctx.alloc().buffer(dataLength);
      data.writeBytes(buffer, dataLength);
      SpdyDataFrame spdyDataFrame = new DefaultSpdyDataFrame(streamId, data);
      length -= dataLength;
      
      if (length == 0) {
        spdyDataFrame.setLast((flags & 0x1) != 0);
        state = State.READ_COMMON_HEADER;
      }
      out.add(spdyDataFrame);
      return;
    
    case 7: 
      int numBytes = Math.min(buffer.readableBytes(), length);
      buffer.skipBytes(numBytes);
      length -= numBytes;
      if (length == 0) {
        state = State.READ_COMMON_HEADER;
      }
      return;
    
    case 8: 
      buffer.skipBytes(buffer.readableBytes());
      return;
    }
    
    throw new Error("Shouldn't reach here.");
  }
  

  private State readCommonHeader(ByteBuf buffer)
  {
    if (buffer.readableBytes() < 8) {
      return State.READ_COMMON_HEADER;
    }
    
    int frameOffset = buffer.readerIndex();
    int flagsOffset = frameOffset + 4;
    int lengthOffset = frameOffset + 5;
    buffer.skipBytes(8);
    

    boolean control = (buffer.getByte(frameOffset) & 0x80) != 0;
    flags = buffer.getByte(flagsOffset);
    length = SpdyCodecUtil.getUnsignedMedium(buffer, lengthOffset);
    
    if (control)
    {
      version = (SpdyCodecUtil.getUnsignedShort(buffer, frameOffset) & 0x7FFF);
      
      int typeOffset = frameOffset + 2;
      type = SpdyCodecUtil.getUnsignedShort(buffer, typeOffset);
      
      streamId = 0;
    }
    else {
      version = spdyVersion;
      
      type = 0;
      
      streamId = SpdyCodecUtil.getUnsignedInt(buffer, frameOffset);
    }
    
    if ((version != spdyVersion) || (!isValidFrameHeader())) {
      return State.FRAME_ERROR;
    }
    
    State nextState;
    
    if (willGenerateFrame()) { State nextState;
      switch (type) {
      case 0: 
        nextState = State.READ_DATA_FRAME;
        break;
      
      case 1: 
      case 2: 
      case 8: 
        nextState = State.READ_HEADER_BLOCK_FRAME;
        break;
      
      case 4: 
        nextState = State.READ_SETTINGS_FRAME;
        break;
      case 3: case 5: case 6: 
      case 7: default: 
        nextState = State.READ_CONTROL_FRAME; break; }
    } else { State nextState;
      if (length != 0) {
        nextState = State.DISCARD_FRAME;
      } else
        nextState = State.READ_COMMON_HEADER;
    }
    return nextState;
  }
  
  private Object readControlFrame(ByteBuf buffer) {
    int streamId;
    int statusCode;
    switch (type) {
    case 3: 
      if (buffer.readableBytes() < 8) {
        return null;
      }
      
      streamId = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex());
      statusCode = SpdyCodecUtil.getSignedInt(buffer, buffer.readerIndex() + 4);
      buffer.skipBytes(8);
      
      return new DefaultSpdyRstStreamFrame(streamId, statusCode);
    
    case 6: 
      if (buffer.readableBytes() < 4) {
        return null;
      }
      
      int ID = SpdyCodecUtil.getSignedInt(buffer, buffer.readerIndex());
      buffer.skipBytes(4);
      
      return new DefaultSpdyPingFrame(ID);
    
    case 7: 
      int minLength = version < 3 ? 4 : 8;
      if (buffer.readableBytes() < minLength) {
        return null;
      }
      
      int lastGoodStreamId = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex());
      buffer.skipBytes(4);
      
      if (version < 3) {
        return new DefaultSpdyGoAwayFrame(lastGoodStreamId);
      }
      
      statusCode = SpdyCodecUtil.getSignedInt(buffer, buffer.readerIndex());
      buffer.skipBytes(4);
      
      return new DefaultSpdyGoAwayFrame(lastGoodStreamId, statusCode);
    
    case 9: 
      if (buffer.readableBytes() < 8) {
        return null;
      }
      
      streamId = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex());
      int deltaWindowSize = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex() + 4);
      buffer.skipBytes(8);
      
      return new DefaultSpdyWindowUpdateFrame(streamId, deltaWindowSize);
    }
    
    throw new Error("Shouldn't reach here.");
  }
  
  private SpdyHeadersFrame readHeaderBlockFrame(ByteBuf buffer)
  {
    int minLength;
    int streamId;
    switch (type) {
    case 1: 
      minLength = version < 3 ? 12 : 10;
      if (buffer.readableBytes() < minLength) {
        return null;
      }
      
      int offset = buffer.readerIndex();
      streamId = SpdyCodecUtil.getUnsignedInt(buffer, offset);
      int associatedToStreamId = SpdyCodecUtil.getUnsignedInt(buffer, offset + 4);
      byte priority = (byte)(buffer.getByte(offset + 8) >> 5 & 0x7);
      if (version < 3) {
        priority = (byte)(priority >> 1);
      }
      buffer.skipBytes(10);
      length -= 10;
      

      if ((version < 3) && (length == 2) && (buffer.getShort(buffer.readerIndex()) == 0)) {
        buffer.skipBytes(2);
        length = 0;
      }
      
      SpdySynStreamFrame spdySynStreamFrame = new DefaultSpdySynStreamFrame(streamId, associatedToStreamId, priority);
      
      spdySynStreamFrame.setLast((flags & 0x1) != 0);
      spdySynStreamFrame.setUnidirectional((flags & 0x2) != 0);
      
      return spdySynStreamFrame;
    
    case 2: 
      minLength = version < 3 ? 8 : 4;
      if (buffer.readableBytes() < minLength) {
        return null;
      }
      
      streamId = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex());
      buffer.skipBytes(4);
      length -= 4;
      

      if (version < 3) {
        buffer.skipBytes(2);
        length -= 2;
      }
      

      if ((version < 3) && (length == 2) && (buffer.getShort(buffer.readerIndex()) == 0)) {
        buffer.skipBytes(2);
        length = 0;
      }
      
      SpdySynReplyFrame spdySynReplyFrame = new DefaultSpdySynReplyFrame(streamId);
      spdySynReplyFrame.setLast((flags & 0x1) != 0);
      
      return spdySynReplyFrame;
    
    case 8: 
      if (buffer.readableBytes() < 4) {
        return null;
      }
      

      if ((version < 3) && (length > 4) && (buffer.readableBytes() < 8)) {
        return null;
      }
      
      streamId = SpdyCodecUtil.getUnsignedInt(buffer, buffer.readerIndex());
      buffer.skipBytes(4);
      length -= 4;
      

      if ((version < 3) && (length != 0)) {
        buffer.skipBytes(2);
        length -= 2;
      }
      

      if ((version < 3) && (length == 2) && (buffer.getShort(buffer.readerIndex()) == 0)) {
        buffer.skipBytes(2);
        length = 0;
      }
      
      SpdyHeadersFrame spdyHeadersFrame = new DefaultSpdyHeadersFrame(streamId);
      spdyHeadersFrame.setLast((flags & 0x1) != 0);
      
      return spdyHeadersFrame;
    }
    
    throw new Error("Shouldn't reach here.");
  }
  
  private boolean isValidFrameHeader()
  {
    switch (type) {
    case 0: 
      return streamId != 0;
    
    case 1: 
      return length >= 12;
    
    case 2: 
      return length >= 8;
    
    case 3: 
      return (flags == 0) && (length == 8);
    
    case 4: 
      return length >= 4;
    
    case 6: 
      return length == 4;
    
    case 7: 
      return length == 4;
    
    case 8: 
      if (version < 3) {
        return (length == 4) || (length >= 8);
      }
      return length >= 4;
    

    case 9: 
      return length == 8;
    }
    
    return true;
  }
  
  private boolean willGenerateFrame()
  {
    switch (type) {
    case 0: 
    case 1: 
    case 2: 
    case 3: 
    case 4: 
    case 6: 
    case 7: 
    case 8: 
    case 9: 
      return true;
    }
    
    return false;
  }
  
  private static void fireInvalidFrameException(ChannelHandlerContext ctx)
  {
    ctx.fireExceptionCaught(INVALID_FRAME);
  }
  
  private static void fireProtocolException(ChannelHandlerContext ctx, String message) {
    ctx.fireExceptionCaught(new SpdyProtocolException(message));
  }
}
