package net.minecraft.util.io.netty.handler.codec.spdy;

import java.util.Set;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.handler.codec.MessageToByteEncoder;
import net.minecraft.util.io.netty.handler.codec.UnsupportedMessageTypeException;
























public class SpdyFrameEncoder
  extends MessageToByteEncoder<SpdyFrame>
{
  private final int version;
  private final SpdyHeaderBlockEncoder headerBlockEncoder;
  
  public SpdyFrameEncoder(int version)
  {
    this(version, 6, 15, 8);
  }
  


  public SpdyFrameEncoder(int version, int compressionLevel, int windowBits, int memLevel)
  {
    this(version, SpdyHeaderBlockEncoder.newInstance(version, compressionLevel, windowBits, memLevel));
  }
  
  protected SpdyFrameEncoder(int version, SpdyHeaderBlockEncoder headerBlockEncoder)
  {
    if ((version < 2) || (version > 3)) {
      throw new IllegalArgumentException("unknown version: " + version);
    }
    
    this.version = version;
    this.headerBlockEncoder = headerBlockEncoder;
  }
  
  public void handlerAdded(ChannelHandlerContext ctx) throws Exception
  {
    ctx.channel().closeFuture().addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture future) throws Exception {
        headerBlockEncoder.end();
      }
    });
  }
  
  protected void encode(ChannelHandlerContext ctx, SpdyFrame msg, ByteBuf out) throws Exception
  {
    if ((msg instanceof SpdyDataFrame))
    {
      SpdyDataFrame spdyDataFrame = (SpdyDataFrame)msg;
      ByteBuf data = spdyDataFrame.content();
      byte flags = spdyDataFrame.isLast() ? 1 : 0;
      out.ensureWritable(8 + data.readableBytes());
      out.writeInt(spdyDataFrame.getStreamId() & 0x7FFFFFFF);
      out.writeByte(flags);
      out.writeMedium(data.readableBytes());
      out.writeBytes(data, data.readerIndex(), data.readableBytes());
    }
    else if ((msg instanceof SpdySynStreamFrame))
    {
      SpdySynStreamFrame spdySynStreamFrame = (SpdySynStreamFrame)msg;
      ByteBuf data = headerBlockEncoder.encode(ctx, spdySynStreamFrame);
      try {
        byte flags = spdySynStreamFrame.isLast() ? 1 : 0;
        if (spdySynStreamFrame.isUnidirectional()) {
          flags = (byte)(flags | 0x2);
        }
        int headerBlockLength = data.readableBytes();
        int length;
        int length; if (version < 3) {
          length = headerBlockLength == 0 ? 12 : 10 + headerBlockLength;
        } else {
          length = 10 + headerBlockLength;
        }
        out.ensureWritable(8 + length);
        out.writeShort(version | 0x8000);
        out.writeShort(1);
        out.writeByte(flags);
        out.writeMedium(length);
        out.writeInt(spdySynStreamFrame.getStreamId());
        out.writeInt(spdySynStreamFrame.getAssociatedToStreamId());
        if (version < 3)
        {
          byte priority = spdySynStreamFrame.getPriority();
          if (priority > 3) {
            priority = 3;
          }
          out.writeShort((priority & 0xFF) << 14);
        } else {
          out.writeShort((spdySynStreamFrame.getPriority() & 0xFF) << 13);
        }
        if ((version < 3) && (data.readableBytes() == 0)) {
          out.writeShort(0);
        }
        out.writeBytes(data, data.readerIndex(), headerBlockLength);
      } finally {
        data.release();
      }
    }
    else if ((msg instanceof SpdySynReplyFrame))
    {
      SpdySynReplyFrame spdySynReplyFrame = (SpdySynReplyFrame)msg;
      ByteBuf data = headerBlockEncoder.encode(ctx, spdySynReplyFrame);
      try {
        byte flags = spdySynReplyFrame.isLast() ? 1 : 0;
        int headerBlockLength = data.readableBytes();
        int length;
        int length; if (version < 3) {
          length = headerBlockLength == 0 ? 8 : 6 + headerBlockLength;
        } else {
          length = 4 + headerBlockLength;
        }
        out.ensureWritable(8 + length);
        out.writeShort(version | 0x8000);
        out.writeShort(2);
        out.writeByte(flags);
        out.writeMedium(length);
        out.writeInt(spdySynReplyFrame.getStreamId());
        if (version < 3) {
          if (headerBlockLength == 0) {
            out.writeInt(0);
          } else {
            out.writeShort(0);
          }
        }
        out.writeBytes(data, data.readerIndex(), headerBlockLength);
      } finally {
        data.release();
      }
    }
    else if ((msg instanceof SpdyRstStreamFrame))
    {
      SpdyRstStreamFrame spdyRstStreamFrame = (SpdyRstStreamFrame)msg;
      out.ensureWritable(16);
      out.writeShort(version | 0x8000);
      out.writeShort(3);
      out.writeInt(8);
      out.writeInt(spdyRstStreamFrame.getStreamId());
      out.writeInt(spdyRstStreamFrame.getStatus().getCode());
    } else { SpdySettingsFrame spdySettingsFrame;
      if ((msg instanceof SpdySettingsFrame))
      {
        spdySettingsFrame = (SpdySettingsFrame)msg;
        byte flags = spdySettingsFrame.clearPreviouslyPersistedSettings() ? 1 : 0;
        
        Set<Integer> IDs = spdySettingsFrame.getIds();
        int numEntries = IDs.size();
        int length = 4 + numEntries * 8;
        out.ensureWritable(8 + length);
        out.writeShort(version | 0x8000);
        out.writeShort(4);
        out.writeByte(flags);
        out.writeMedium(length);
        out.writeInt(numEntries);
        for (Integer ID : IDs) {
          int id = ID.intValue();
          byte ID_flags = 0;
          if (spdySettingsFrame.isPersistValue(id)) {
            ID_flags = (byte)(ID_flags | 0x1);
          }
          if (spdySettingsFrame.isPersisted(id)) {
            ID_flags = (byte)(ID_flags | 0x2);
          }
          if (version < 3)
          {


            out.writeByte(id & 0xFF);
            out.writeByte(id >> 8 & 0xFF);
            out.writeByte(id >> 16 & 0xFF);
            out.writeByte(ID_flags);
          } else {
            out.writeByte(ID_flags);
            out.writeMedium(id);
          }
          out.writeInt(spdySettingsFrame.getValue(id));
        }
      }
      else if ((msg instanceof SpdyPingFrame))
      {
        SpdyPingFrame spdyPingFrame = (SpdyPingFrame)msg;
        out.ensureWritable(12);
        out.writeShort(version | 0x8000);
        out.writeShort(6);
        out.writeInt(4);
        out.writeInt(spdyPingFrame.getId());
      }
      else if ((msg instanceof SpdyGoAwayFrame))
      {
        SpdyGoAwayFrame spdyGoAwayFrame = (SpdyGoAwayFrame)msg;
        int length = version < 3 ? 4 : 8;
        out.ensureWritable(8 + length);
        out.writeShort(version | 0x8000);
        out.writeShort(7);
        out.writeInt(length);
        out.writeInt(spdyGoAwayFrame.getLastGoodStreamId());
        if (version >= 3) {
          out.writeInt(spdyGoAwayFrame.getStatus().getCode());
        }
      }
      else if ((msg instanceof SpdyHeadersFrame))
      {
        SpdyHeadersFrame spdyHeadersFrame = (SpdyHeadersFrame)msg;
        ByteBuf data = headerBlockEncoder.encode(ctx, spdyHeadersFrame);
        try {
          byte flags = spdyHeadersFrame.isLast() ? 1 : 0;
          int headerBlockLength = data.readableBytes();
          int length;
          int length; if (version < 3) {
            length = headerBlockLength == 0 ? 4 : 6 + headerBlockLength;
          } else {
            length = 4 + headerBlockLength;
          }
          out.ensureWritable(8 + length);
          out.writeShort(version | 0x8000);
          out.writeShort(8);
          out.writeByte(flags);
          out.writeMedium(length);
          out.writeInt(spdyHeadersFrame.getStreamId());
          if ((version < 3) && (headerBlockLength != 0)) {
            out.writeShort(0);
          }
          out.writeBytes(data, data.readerIndex(), headerBlockLength);
        } finally {
          data.release();
        }
      }
      else if ((msg instanceof SpdyWindowUpdateFrame))
      {
        SpdyWindowUpdateFrame spdyWindowUpdateFrame = (SpdyWindowUpdateFrame)msg;
        out.ensureWritable(16);
        out.writeShort(version | 0x8000);
        out.writeShort(9);
        out.writeInt(8);
        out.writeInt(spdyWindowUpdateFrame.getStreamId());
        out.writeInt(spdyWindowUpdateFrame.getDeltaWindowSize());
      } else {
        throw new UnsupportedMessageTypeException(msg, new Class[0]);
      }
    }
  }
}
