package net.minecraft.util.io.netty.handler.codec.spdy;

import java.util.Comparator;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.util.internal.PlatformDependent;













final class SpdySession
{
  private final Map<Integer, StreamState> activeStreams;
  
  SpdySession()
  {
    activeStreams = PlatformDependent.newConcurrentHashMap();
  }
  
  int numActiveStreams() { return activeStreams.size(); }
  
  boolean noActiveStreams()
  {
    return activeStreams.isEmpty();
  }
  
  boolean isActiveStream(int streamId) {
    return activeStreams.containsKey(Integer.valueOf(streamId));
  }
  
  Set<Integer> getActiveStreams()
  {
    TreeSet<Integer> streamIds = new TreeSet(new PriorityComparator(null));
    streamIds.addAll(activeStreams.keySet());
    return streamIds;
  }
  

  void acceptStream(int streamId, byte priority, boolean remoteSideClosed, boolean localSideClosed, int sendWindowSize, int receiveWindowSize)
  {
    if ((!remoteSideClosed) || (!localSideClosed)) {
      activeStreams.put(Integer.valueOf(streamId), new StreamState(priority, remoteSideClosed, localSideClosed, sendWindowSize, receiveWindowSize));
    }
  }
  
  void removeStream(int streamId, Throwable cause)
  {
    StreamState state = (StreamState)activeStreams.remove(Integer.valueOf(streamId));
    if (state != null) {
      state.clearPendingWrites(cause);
    }
  }
  
  boolean isRemoteSideClosed(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return (state == null) || (state.isRemoteSideClosed());
  }
  
  void closeRemoteSide(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    if (state != null) {
      state.closeRemoteSide();
      if (state.isLocalSideClosed()) {
        activeStreams.remove(Integer.valueOf(streamId));
      }
    }
  }
  
  boolean isLocalSideClosed(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return (state == null) || (state.isLocalSideClosed());
  }
  
  void closeLocalSide(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    if (state != null) {
      state.closeLocalSide();
      if (state.isRemoteSideClosed()) {
        activeStreams.remove(Integer.valueOf(streamId));
      }
    }
  }
  



  boolean hasReceivedReply(int streamId)
  {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return (state != null) && (state.hasReceivedReply());
  }
  
  void receivedReply(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    if (state != null) {
      state.receivedReply();
    }
  }
  
  int getSendWindowSize(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return state != null ? state.getSendWindowSize() : -1;
  }
  
  int updateSendWindowSize(int streamId, int deltaWindowSize) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return state != null ? state.updateSendWindowSize(deltaWindowSize) : -1;
  }
  
  int updateReceiveWindowSize(int streamId, int deltaWindowSize) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    if (deltaWindowSize > 0) {
      state.setReceiveWindowSizeLowerBound(0);
    }
    return state != null ? state.updateReceiveWindowSize(deltaWindowSize) : -1;
  }
  
  int getReceiveWindowSizeLowerBound(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return state != null ? state.getReceiveWindowSizeLowerBound() : 0;
  }
  
  void updateAllSendWindowSizes(int deltaWindowSize) {
    for (StreamState state : activeStreams.values()) {
      state.updateSendWindowSize(deltaWindowSize);
    }
  }
  
  void updateAllReceiveWindowSizes(int deltaWindowSize) {
    for (StreamState state : activeStreams.values()) {
      state.updateReceiveWindowSize(deltaWindowSize);
      if (deltaWindowSize < 0) {
        state.setReceiveWindowSizeLowerBound(deltaWindowSize);
      }
    }
  }
  
  boolean putPendingWrite(int streamId, PendingWrite pendingWrite) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return (state != null) && (state.putPendingWrite(pendingWrite));
  }
  
  PendingWrite getPendingWrite(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return state != null ? state.getPendingWrite() : null;
  }
  
  PendingWrite removePendingWrite(int streamId) {
    StreamState state = (StreamState)activeStreams.get(Integer.valueOf(streamId));
    return state != null ? state.removePendingWrite() : null;
  }
  
  private static final class StreamState
  {
    private final byte priority;
    private boolean remoteSideClosed;
    private boolean localSideClosed;
    private boolean receivedReply;
    private final AtomicInteger sendWindowSize;
    private final AtomicInteger receiveWindowSize;
    private int receiveWindowSizeLowerBound;
    private final Queue<SpdySession.PendingWrite> pendingWriteQueue = new ConcurrentLinkedQueue();
    

    StreamState(byte priority, boolean remoteSideClosed, boolean localSideClosed, int sendWindowSize, int receiveWindowSize)
    {
      this.priority = priority;
      this.remoteSideClosed = remoteSideClosed;
      this.localSideClosed = localSideClosed;
      this.sendWindowSize = new AtomicInteger(sendWindowSize);
      this.receiveWindowSize = new AtomicInteger(receiveWindowSize);
    }
    
    byte getPriority() {
      return priority;
    }
    
    boolean isRemoteSideClosed() {
      return remoteSideClosed;
    }
    
    void closeRemoteSide() {
      remoteSideClosed = true;
    }
    
    boolean isLocalSideClosed() {
      return localSideClosed;
    }
    
    void closeLocalSide() {
      localSideClosed = true;
    }
    
    boolean hasReceivedReply() {
      return receivedReply;
    }
    
    void receivedReply() {
      receivedReply = true;
    }
    
    int getSendWindowSize() {
      return sendWindowSize.get();
    }
    
    int updateSendWindowSize(int deltaWindowSize) {
      return sendWindowSize.addAndGet(deltaWindowSize);
    }
    
    int updateReceiveWindowSize(int deltaWindowSize) {
      return receiveWindowSize.addAndGet(deltaWindowSize);
    }
    
    int getReceiveWindowSizeLowerBound() {
      return receiveWindowSizeLowerBound;
    }
    
    void setReceiveWindowSizeLowerBound(int receiveWindowSizeLowerBound) {
      this.receiveWindowSizeLowerBound = receiveWindowSizeLowerBound;
    }
    
    boolean putPendingWrite(SpdySession.PendingWrite msg) {
      return pendingWriteQueue.offer(msg);
    }
    
    SpdySession.PendingWrite getPendingWrite() {
      return (SpdySession.PendingWrite)pendingWriteQueue.peek();
    }
    
    SpdySession.PendingWrite removePendingWrite() {
      return (SpdySession.PendingWrite)pendingWriteQueue.poll();
    }
    
    void clearPendingWrites(Throwable cause) {
      for (;;) {
        SpdySession.PendingWrite pendingWrite = (SpdySession.PendingWrite)pendingWriteQueue.poll();
        if (pendingWrite == null) {
          break;
        }
        pendingWrite.fail(cause);
      }
    }
  }
  
  private final class PriorityComparator implements Comparator<Integer> {
    private PriorityComparator() {}
    
    public int compare(Integer id1, Integer id2) { SpdySession.StreamState state1 = (SpdySession.StreamState)activeStreams.get(id1);
      SpdySession.StreamState state2 = (SpdySession.StreamState)activeStreams.get(id2);
      return state1.getPriority() - state2.getPriority();
    }
  }
  
  public static final class PendingWrite {
    final SpdyDataFrame spdyDataFrame;
    final ChannelPromise promise;
    
    PendingWrite(SpdyDataFrame spdyDataFrame, ChannelPromise promise) {
      this.spdyDataFrame = spdyDataFrame;
      this.promise = promise;
    }
    
    void fail(Throwable cause) {
      spdyDataFrame.release();
      promise.setFailure(cause);
    }
  }
}
