package net.minecraft.util.io.netty.handler.codec.spdy;

import java.util.concurrent.atomic.AtomicInteger;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.ChannelDuplexHandler;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.util.internal.EmptyArrays;


















public class SpdySessionHandler
  extends ChannelDuplexHandler
{
  private static final SpdyProtocolException PROTOCOL_EXCEPTION = new SpdyProtocolException();
  private static final SpdyProtocolException STREAM_CLOSED = new SpdyProtocolException("Stream closed");
  
  static {
    PROTOCOL_EXCEPTION.setStackTrace(EmptyArrays.EMPTY_STACK_TRACE);
    STREAM_CLOSED.setStackTrace(EmptyArrays.EMPTY_STACK_TRACE);
  }
  
  private final SpdySession spdySession = new SpdySession();
  
  private int lastGoodStreamId;
  private static final int DEFAULT_MAX_CONCURRENT_STREAMS = 2147483647;
  private int remoteConcurrentStreams = 2147483647;
  private int localConcurrentStreams = 2147483647;
  private int maxConcurrentStreams = 2147483647;
  
  private static final int DEFAULT_WINDOW_SIZE = 65536;
  private int initialSendWindowSize = 65536;
  private int initialReceiveWindowSize = 65536;
  
  private final Object flowControlLock = new Object();
  
  private final AtomicInteger pings = new AtomicInteger();
  

  private boolean sentGoAwayFrame;
  

  private boolean receivedGoAwayFrame;
  

  private ChannelFutureListener closeSessionFutureListener;
  

  private final boolean server;
  

  private final boolean flowControl;
  

  public SpdySessionHandler(int version, boolean server)
  {
    if ((version < 2) || (version > 3)) {
      throw new IllegalArgumentException("unsupported version: " + version);
    }
    
    this.server = server;
    flowControl = (version >= 3);
  }
  
  public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception
  {
    if ((msg instanceof SpdyDataFrame))
    {






















      SpdyDataFrame spdyDataFrame = (SpdyDataFrame)msg;
      int streamId = spdyDataFrame.getStreamId();
      


      if (!spdySession.isActiveStream(streamId)) {
        spdyDataFrame.release();
        if (streamId <= lastGoodStreamId) {
          issueStreamError(ctx, streamId, SpdyStreamStatus.PROTOCOL_ERROR);
        } else if (!sentGoAwayFrame) {
          issueStreamError(ctx, streamId, SpdyStreamStatus.INVALID_STREAM);
        }
        return;
      }
      


      if (spdySession.isRemoteSideClosed(streamId)) {
        spdyDataFrame.release();
        issueStreamError(ctx, streamId, SpdyStreamStatus.STREAM_ALREADY_CLOSED);
        return;
      }
      

      if ((!isRemoteInitiatedID(streamId)) && (!spdySession.hasReceivedReply(streamId))) {
        spdyDataFrame.release();
        issueStreamError(ctx, streamId, SpdyStreamStatus.PROTOCOL_ERROR);
        return;
      }
      






      if (flowControl)
      {
        int deltaWindowSize = -1 * spdyDataFrame.content().readableBytes();
        int newWindowSize = spdySession.updateReceiveWindowSize(streamId, deltaWindowSize);
        





        if (newWindowSize < spdySession.getReceiveWindowSizeLowerBound(streamId)) {
          spdyDataFrame.release();
          issueStreamError(ctx, streamId, SpdyStreamStatus.FLOW_CONTROL_ERROR);
          return;
        }
        


        if (newWindowSize < 0) {
          while (spdyDataFrame.content().readableBytes() > initialReceiveWindowSize) {
            SpdyDataFrame partialDataFrame = new DefaultSpdyDataFrame(streamId, spdyDataFrame.content().readSlice(initialReceiveWindowSize).retain());
            
            ctx.writeAndFlush(partialDataFrame);
          }
        }
        

        if ((newWindowSize <= initialReceiveWindowSize / 2) && (!spdyDataFrame.isLast())) {
          deltaWindowSize = initialReceiveWindowSize - newWindowSize;
          spdySession.updateReceiveWindowSize(streamId, deltaWindowSize);
          SpdyWindowUpdateFrame spdyWindowUpdateFrame = new DefaultSpdyWindowUpdateFrame(streamId, deltaWindowSize);
          
          ctx.writeAndFlush(spdyWindowUpdateFrame);
        }
      }
      

      if (spdyDataFrame.isLast()) {
        halfCloseStream(streamId, true, ctx.newSucceededFuture());
      }
    }
    else if ((msg instanceof SpdySynStreamFrame))
    {














      SpdySynStreamFrame spdySynStreamFrame = (SpdySynStreamFrame)msg;
      int streamId = spdySynStreamFrame.getStreamId();
      

      if ((spdySynStreamFrame.isInvalid()) || (!isRemoteInitiatedID(streamId)) || (spdySession.isActiveStream(streamId)))
      {

        issueStreamError(ctx, streamId, SpdyStreamStatus.PROTOCOL_ERROR);
        return;
      }
      

      if (streamId <= lastGoodStreamId) {
        issueSessionError(ctx, SpdySessionStatus.PROTOCOL_ERROR);
        return;
      }
      

      byte priority = spdySynStreamFrame.getPriority();
      boolean remoteSideClosed = spdySynStreamFrame.isLast();
      boolean localSideClosed = spdySynStreamFrame.isUnidirectional();
      if (!acceptStream(streamId, priority, remoteSideClosed, localSideClosed)) {
        issueStreamError(ctx, streamId, SpdyStreamStatus.REFUSED_STREAM);
        return;
      }
    }
    else if ((msg instanceof SpdySynReplyFrame))
    {







      SpdySynReplyFrame spdySynReplyFrame = (SpdySynReplyFrame)msg;
      int streamId = spdySynReplyFrame.getStreamId();
      

      if ((spdySynReplyFrame.isInvalid()) || (isRemoteInitiatedID(streamId)) || (spdySession.isRemoteSideClosed(streamId)))
      {

        issueStreamError(ctx, streamId, SpdyStreamStatus.INVALID_STREAM);
        return;
      }
      

      if (spdySession.hasReceivedReply(streamId)) {
        issueStreamError(ctx, streamId, SpdyStreamStatus.STREAM_IN_USE);
        return;
      }
      
      spdySession.receivedReply(streamId);
      

      if (spdySynReplyFrame.isLast()) {
        halfCloseStream(streamId, true, ctx.newSucceededFuture());
      }
    }
    else if ((msg instanceof SpdyRstStreamFrame))
    {









      SpdyRstStreamFrame spdyRstStreamFrame = (SpdyRstStreamFrame)msg;
      removeStream(spdyRstStreamFrame.getStreamId(), ctx.newSucceededFuture());
    }
    else if ((msg instanceof SpdySettingsFrame))
    {
      SpdySettingsFrame spdySettingsFrame = (SpdySettingsFrame)msg;
      
      int newConcurrentStreams = spdySettingsFrame.getValue(4);
      
      if (newConcurrentStreams >= 0) {
        updateConcurrentStreams(newConcurrentStreams, true);
      }
      



      if (spdySettingsFrame.isPersisted(7)) {
        spdySettingsFrame.removeValue(7);
      }
      spdySettingsFrame.setPersistValue(7, false);
      
      if (flowControl) {
        int newInitialWindowSize = spdySettingsFrame.getValue(7);
        
        if (newInitialWindowSize >= 0) {
          updateInitialSendWindowSize(newInitialWindowSize);
        }
      }
    }
    else if ((msg instanceof SpdyPingFrame))
    {









      SpdyPingFrame spdyPingFrame = (SpdyPingFrame)msg;
      
      if (isRemoteInitiatedID(spdyPingFrame.getId())) {
        ctx.writeAndFlush(spdyPingFrame);
        return;
      }
      

      if (pings.get() == 0) {
        return;
      }
      pings.getAndDecrement();
    }
    else if ((msg instanceof SpdyGoAwayFrame))
    {
      receivedGoAwayFrame = true;
    }
    else if ((msg instanceof SpdyHeadersFrame))
    {
      SpdyHeadersFrame spdyHeadersFrame = (SpdyHeadersFrame)msg;
      int streamId = spdyHeadersFrame.getStreamId();
      

      if (spdyHeadersFrame.isInvalid()) {
        issueStreamError(ctx, streamId, SpdyStreamStatus.PROTOCOL_ERROR);
        return;
      }
      
      if (spdySession.isRemoteSideClosed(streamId)) {
        issueStreamError(ctx, streamId, SpdyStreamStatus.INVALID_STREAM);
        return;
      }
      

      if (spdyHeadersFrame.isLast()) {
        halfCloseStream(streamId, true, ctx.newSucceededFuture());
      }
    }
    else if ((msg instanceof SpdyWindowUpdateFrame))
    {










      if (flowControl) {
        SpdyWindowUpdateFrame spdyWindowUpdateFrame = (SpdyWindowUpdateFrame)msg;
        int streamId = spdyWindowUpdateFrame.getStreamId();
        int deltaWindowSize = spdyWindowUpdateFrame.getDeltaWindowSize();
        

        if (spdySession.isLocalSideClosed(streamId)) {
          return;
        }
        

        if (spdySession.getSendWindowSize(streamId) > 2147483647 - deltaWindowSize) {
          issueStreamError(ctx, streamId, SpdyStreamStatus.FLOW_CONTROL_ERROR);
          return;
        }
        
        updateSendWindowSize(ctx, streamId, deltaWindowSize);
      }
    }
    
    ctx.fireChannelRead(msg);
  }
  
  public void channelInactive(ChannelHandlerContext ctx) throws Exception
  {
    for (Integer streamId : spdySession.getActiveStreams()) {
      removeStream(streamId.intValue(), ctx.newSucceededFuture());
    }
    ctx.fireChannelInactive();
  }
  
  public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception
  {
    if ((cause instanceof SpdyProtocolException)) {
      issueSessionError(ctx, SpdySessionStatus.PROTOCOL_ERROR);
    }
    
    ctx.fireExceptionCaught(cause);
  }
  
  public void close(ChannelHandlerContext ctx, ChannelPromise promise) throws Exception
  {
    sendGoAwayFrame(ctx, promise);
  }
  
  public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise) throws Exception
  {
    if (((msg instanceof SpdyDataFrame)) || ((msg instanceof SpdySynStreamFrame)) || ((msg instanceof SpdySynReplyFrame)) || ((msg instanceof SpdyRstStreamFrame)) || ((msg instanceof SpdySettingsFrame)) || ((msg instanceof SpdyPingFrame)) || ((msg instanceof SpdyGoAwayFrame)) || ((msg instanceof SpdyHeadersFrame)) || ((msg instanceof SpdyWindowUpdateFrame)))
    {








      handleOutboundMessage(ctx, msg, promise);
    } else {
      ctx.write(msg, promise);
    }
  }
  
  private void handleOutboundMessage(ChannelHandlerContext ctx, Object msg, ChannelPromise promise) throws Exception {
    if ((msg instanceof SpdyDataFrame))
    {
      SpdyDataFrame spdyDataFrame = (SpdyDataFrame)msg;
      final int streamId = spdyDataFrame.getStreamId();
      

      if (spdySession.isLocalSideClosed(streamId)) {
        spdyDataFrame.release();
        promise.setFailure(PROTOCOL_EXCEPTION);
        return;
      }
      













      if (flowControl) {
        synchronized (flowControlLock) {
          int dataLength = spdyDataFrame.content().readableBytes();
          int sendWindowSize = spdySession.getSendWindowSize(streamId);
          
          if (sendWindowSize <= 0)
          {
            spdySession.putPendingWrite(streamId, new SpdySession.PendingWrite(spdyDataFrame, promise));
            return; }
          if (sendWindowSize < dataLength)
          {
            spdySession.updateSendWindowSize(streamId, -1 * sendWindowSize);
            

            SpdyDataFrame partialDataFrame = new DefaultSpdyDataFrame(streamId, spdyDataFrame.content().readSlice(sendWindowSize).retain());
            


            spdySession.putPendingWrite(streamId, new SpdySession.PendingWrite(spdyDataFrame, promise));
            


            final ChannelHandlerContext context = ctx;
            ctx.write(partialDataFrame).addListener(new ChannelFutureListener()
            {
              public void operationComplete(ChannelFuture future) throws Exception {
                if (!future.isSuccess()) {
                  SpdySessionHandler.this.issueStreamError(context, streamId, SpdyStreamStatus.INTERNAL_ERROR);
                }
              }
            });
            return;
          }
          
          spdySession.updateSendWindowSize(streamId, -1 * dataLength);
          


          final ChannelHandlerContext context = ctx;
          promise.addListener(new ChannelFutureListener()
          {
            public void operationComplete(ChannelFuture future) throws Exception {
              if (!future.isSuccess()) {
                SpdySessionHandler.this.issueStreamError(context, streamId, SpdyStreamStatus.INTERNAL_ERROR);
              }
            }
          });
        }
      }
      


      if (spdyDataFrame.isLast()) {
        halfCloseStream(streamId, false, promise);
      }
    }
    else if ((msg instanceof SpdySynStreamFrame))
    {
      SpdySynStreamFrame spdySynStreamFrame = (SpdySynStreamFrame)msg;
      int streamId = spdySynStreamFrame.getStreamId();
      
      if (isRemoteInitiatedID(streamId)) {
        promise.setFailure(PROTOCOL_EXCEPTION);
        return;
      }
      
      byte priority = spdySynStreamFrame.getPriority();
      boolean remoteSideClosed = spdySynStreamFrame.isUnidirectional();
      boolean localSideClosed = spdySynStreamFrame.isLast();
      if (!acceptStream(streamId, priority, remoteSideClosed, localSideClosed)) {
        promise.setFailure(PROTOCOL_EXCEPTION);
        return;
      }
    }
    else if ((msg instanceof SpdySynReplyFrame))
    {
      SpdySynReplyFrame spdySynReplyFrame = (SpdySynReplyFrame)msg;
      int streamId = spdySynReplyFrame.getStreamId();
      

      if ((!isRemoteInitiatedID(streamId)) || (spdySession.isLocalSideClosed(streamId))) {
        promise.setFailure(PROTOCOL_EXCEPTION);
        return;
      }
      

      if (spdySynReplyFrame.isLast()) {
        halfCloseStream(streamId, false, promise);
      }
    }
    else if ((msg instanceof SpdyRstStreamFrame))
    {
      SpdyRstStreamFrame spdyRstStreamFrame = (SpdyRstStreamFrame)msg;
      removeStream(spdyRstStreamFrame.getStreamId(), promise);
    }
    else if ((msg instanceof SpdySettingsFrame))
    {
      SpdySettingsFrame spdySettingsFrame = (SpdySettingsFrame)msg;
      
      int newConcurrentStreams = spdySettingsFrame.getValue(4);
      
      if (newConcurrentStreams >= 0) {
        updateConcurrentStreams(newConcurrentStreams, false);
      }
      



      if (spdySettingsFrame.isPersisted(7)) {
        spdySettingsFrame.removeValue(7);
      }
      spdySettingsFrame.setPersistValue(7, false);
      
      if (flowControl) {
        int newInitialWindowSize = spdySettingsFrame.getValue(7);
        
        if (newInitialWindowSize >= 0) {
          updateInitialReceiveWindowSize(newInitialWindowSize);
        }
      }
    }
    else if ((msg instanceof SpdyPingFrame))
    {
      SpdyPingFrame spdyPingFrame = (SpdyPingFrame)msg;
      if (isRemoteInitiatedID(spdyPingFrame.getId())) {
        ctx.fireExceptionCaught(new IllegalArgumentException("invalid PING ID: " + spdyPingFrame.getId()));
        
        return;
      }
      pings.getAndIncrement();
    } else {
      if ((msg instanceof SpdyGoAwayFrame))
      {


        promise.setFailure(PROTOCOL_EXCEPTION);
        return;
      }
      if ((msg instanceof SpdyHeadersFrame))
      {
        SpdyHeadersFrame spdyHeadersFrame = (SpdyHeadersFrame)msg;
        int streamId = spdyHeadersFrame.getStreamId();
        

        if (spdySession.isLocalSideClosed(streamId)) {
          promise.setFailure(PROTOCOL_EXCEPTION);
          return;
        }
        

        if (spdyHeadersFrame.isLast()) {
          halfCloseStream(streamId, false, promise);
        }
      }
      else if ((msg instanceof SpdyWindowUpdateFrame))
      {

        promise.setFailure(PROTOCOL_EXCEPTION);
        return;
      }
    }
    ctx.write(msg, promise);
  }
  










  private void issueSessionError(ChannelHandlerContext ctx, SpdySessionStatus status)
  {
    sendGoAwayFrame(ctx, status).addListener(new ClosingChannelFutureListener(ctx, ctx.newPromise()));
  }
  










  private void issueStreamError(ChannelHandlerContext ctx, int streamId, SpdyStreamStatus status)
  {
    boolean fireChannelRead = !spdySession.isRemoteSideClosed(streamId);
    ChannelPromise promise = ctx.newPromise();
    removeStream(streamId, promise);
    
    SpdyRstStreamFrame spdyRstStreamFrame = new DefaultSpdyRstStreamFrame(streamId, status);
    ctx.writeAndFlush(spdyRstStreamFrame, promise);
    if (fireChannelRead) {
      ctx.fireChannelRead(spdyRstStreamFrame);
    }
  }
  



  private boolean isRemoteInitiatedID(int id)
  {
    boolean serverId = SpdyCodecUtil.isServerId(id);
    return ((server) && (!serverId)) || ((!server) && (serverId));
  }
  
  private void updateConcurrentStreams(int newConcurrentStreams, boolean remote) {
    if (remote) {
      remoteConcurrentStreams = newConcurrentStreams;
    } else {
      localConcurrentStreams = newConcurrentStreams;
    }
    maxConcurrentStreams = Math.min(localConcurrentStreams, remoteConcurrentStreams);
  }
  
  private synchronized void updateInitialSendWindowSize(int newInitialWindowSize)
  {
    int deltaWindowSize = newInitialWindowSize - initialSendWindowSize;
    initialSendWindowSize = newInitialWindowSize;
    spdySession.updateAllSendWindowSizes(deltaWindowSize);
  }
  
  private synchronized void updateInitialReceiveWindowSize(int newInitialWindowSize)
  {
    int deltaWindowSize = newInitialWindowSize - initialReceiveWindowSize;
    initialReceiveWindowSize = newInitialWindowSize;
    spdySession.updateAllReceiveWindowSizes(deltaWindowSize);
  }
  


  private synchronized boolean acceptStream(int streamId, byte priority, boolean remoteSideClosed, boolean localSideClosed)
  {
    if ((receivedGoAwayFrame) || (sentGoAwayFrame)) {
      return false;
    }
    
    if (spdySession.numActiveStreams() >= maxConcurrentStreams) {
      return false;
    }
    spdySession.acceptStream(streamId, priority, remoteSideClosed, localSideClosed, initialSendWindowSize, initialReceiveWindowSize);
    

    if (isRemoteInitiatedID(streamId)) {
      lastGoodStreamId = streamId;
    }
    return true;
  }
  
  private void halfCloseStream(int streamId, boolean remote, ChannelFuture future) {
    if (remote) {
      spdySession.closeRemoteSide(streamId);
    } else {
      spdySession.closeLocalSide(streamId);
    }
    if ((closeSessionFutureListener != null) && (spdySession.noActiveStreams())) {
      future.addListener(closeSessionFutureListener);
    }
  }
  
  private void removeStream(int streamId, ChannelFuture future) {
    spdySession.removeStream(streamId, STREAM_CLOSED);
    
    if ((closeSessionFutureListener != null) && (spdySession.noActiveStreams())) {
      future.addListener(closeSessionFutureListener);
    }
  }
  
  private void updateSendWindowSize(final ChannelHandlerContext ctx, final int streamId, int deltaWindowSize) {
    synchronized (flowControlLock) {
      int newWindowSize = spdySession.updateSendWindowSize(streamId, deltaWindowSize);
      
      while (newWindowSize > 0)
      {
        SpdySession.PendingWrite pendingWrite = spdySession.getPendingWrite(streamId);
        if (pendingWrite == null) {
          break;
        }
        
        SpdyDataFrame spdyDataFrame = spdyDataFrame;
        int dataFrameSize = spdyDataFrame.content().readableBytes();
        
        if (newWindowSize >= dataFrameSize)
        {
          spdySession.removePendingWrite(streamId);
          newWindowSize = spdySession.updateSendWindowSize(streamId, -1 * dataFrameSize);
          

          if (spdyDataFrame.isLast()) {
            halfCloseStream(streamId, false, promise);
          }
          


          ctx.writeAndFlush(spdyDataFrame, promise).addListener(new ChannelFutureListener()
          {
            public void operationComplete(ChannelFuture future) throws Exception {
              if (!future.isSuccess()) {
                SpdySessionHandler.this.issueStreamError(ctx, streamId, SpdyStreamStatus.INTERNAL_ERROR);
              }
            }
          });
        }
        else {
          spdySession.updateSendWindowSize(streamId, -1 * newWindowSize);
          

          SpdyDataFrame partialDataFrame = new DefaultSpdyDataFrame(streamId, spdyDataFrame.content().readSlice(newWindowSize).retain());
          



          ctx.writeAndFlush(partialDataFrame).addListener(new ChannelFutureListener()
          {
            public void operationComplete(ChannelFuture future) throws Exception {
              if (!future.isSuccess()) {
                SpdySessionHandler.this.issueStreamError(ctx, streamId, SpdyStreamStatus.INTERNAL_ERROR);
              }
              
            }
          });
          newWindowSize = 0;
        }
      }
    }
  }
  
  private void sendGoAwayFrame(ChannelHandlerContext ctx, ChannelPromise future)
  {
    if (!ctx.channel().isActive()) {
      ctx.close(future);
      return;
    }
    
    ChannelFuture f = sendGoAwayFrame(ctx, SpdySessionStatus.OK);
    if (spdySession.noActiveStreams()) {
      f.addListener(new ClosingChannelFutureListener(ctx, future));
    } else {
      closeSessionFutureListener = new ClosingChannelFutureListener(ctx, future);
    }
  }
  

  private synchronized ChannelFuture sendGoAwayFrame(ChannelHandlerContext ctx, SpdySessionStatus status)
  {
    if (!sentGoAwayFrame) {
      sentGoAwayFrame = true;
      SpdyGoAwayFrame spdyGoAwayFrame = new DefaultSpdyGoAwayFrame(lastGoodStreamId, status);
      return ctx.writeAndFlush(spdyGoAwayFrame);
    }
    return ctx.newSucceededFuture();
  }
  
  private static final class ClosingChannelFutureListener implements ChannelFutureListener
  {
    private final ChannelHandlerContext ctx;
    private final ChannelPromise promise;
    
    ClosingChannelFutureListener(ChannelHandlerContext ctx, ChannelPromise promise) {
      this.ctx = ctx;
      this.promise = promise;
    }
    
    public void operationComplete(ChannelFuture sentGoAwayFuture) throws Exception
    {
      ctx.close(promise);
    }
  }
}
