package net.minecraft.util.io.netty.handler.ssl;

import java.io.IOException;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.net.ssl.SSLEngine;
import javax.net.ssl.SSLEngineResult;
import javax.net.ssl.SSLEngineResult.HandshakeStatus;
import javax.net.ssl.SSLEngineResult.Status;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLSession;
import net.minecraft.util.io.netty.buffer.ByteBuf;
import net.minecraft.util.io.netty.buffer.ByteBufAllocator;
import net.minecraft.util.io.netty.buffer.ByteBufUtil;
import net.minecraft.util.io.netty.buffer.Unpooled;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.ChannelException;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.channel.ChannelOutboundHandler;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.handler.codec.ByteToMessageDecoder;
import net.minecraft.util.io.netty.util.concurrent.DefaultPromise;
import net.minecraft.util.io.netty.util.concurrent.EventExecutor;
import net.minecraft.util.io.netty.util.concurrent.Future;
import net.minecraft.util.io.netty.util.concurrent.GenericFutureListener;
import net.minecraft.util.io.netty.util.concurrent.ImmediateExecutor;
import net.minecraft.util.io.netty.util.internal.EmptyArrays;
import net.minecraft.util.io.netty.util.internal.PendingWrite;
import net.minecraft.util.io.netty.util.internal.PlatformDependent;
import net.minecraft.util.io.netty.util.internal.RecyclableArrayList;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;






























































































public class SslHandler
  extends ByteToMessageDecoder
  implements ChannelOutboundHandler
{
  private static final InternalLogger logger;
  private static final Pattern IGNORABLE_CLASS_IN_STACK;
  private static final Pattern IGNORABLE_ERROR_MESSAGE;
  private static final SSLException SSLENGINE_CLOSED;
  private static final SSLException HANDSHAKE_TIMED_OUT;
  private static final ClosedChannelException CHANNEL_CLOSED;
  private volatile ChannelHandlerContext ctx;
  private final SSLEngine engine;
  private final Executor delegatedTaskExecutor;
  private final boolean startTls;
  private boolean sentFirstMessage;
  
  static
  {
    logger = InternalLoggerFactory.getInstance(SslHandler.class);
    

    IGNORABLE_CLASS_IN_STACK = Pattern.compile("^.*(?:Socket|Datagram|Sctp|Udt)Channel.*$");
    
    IGNORABLE_ERROR_MESSAGE = Pattern.compile("^.*(?:connection.*(?:reset|closed|abort|broken)|broken.*pipe).*$", 2);
    

    SSLENGINE_CLOSED = new SSLException("SSLEngine closed already");
    HANDSHAKE_TIMED_OUT = new SSLException("handshake timed out");
    CHANNEL_CLOSED = new ClosedChannelException();
    

    SSLENGINE_CLOSED.setStackTrace(EmptyArrays.EMPTY_STACK_TRACE);
    HANDSHAKE_TIMED_OUT.setStackTrace(EmptyArrays.EMPTY_STACK_TRACE);
    CHANNEL_CLOSED.setStackTrace(EmptyArrays.EMPTY_STACK_TRACE);
  }
  







  private final LazyChannelPromise handshakePromise = new LazyChannelPromise(null);
  private final LazyChannelPromise sslCloseFuture = new LazyChannelPromise(null);
  private final Deque<PendingWrite> pendingUnencryptedWrites = new ArrayDeque();
  
  private int packetLength;
  
  private ByteBuf decodeOut;
  private volatile long handshakeTimeoutMillis = 10000L;
  private volatile long closeNotifyTimeoutMillis = 3000L;
  




  public SslHandler(SSLEngine engine)
  {
    this(engine, ImmediateExecutor.INSTANCE);
  }
  






  public SslHandler(SSLEngine engine, boolean startTls)
  {
    this(engine, startTls, ImmediateExecutor.INSTANCE);
  }
  








  public SslHandler(SSLEngine engine, Executor delegatedTaskExecutor)
  {
    this(engine, false, delegatedTaskExecutor);
  }
  











  public SslHandler(SSLEngine engine, boolean startTls, Executor delegatedTaskExecutor)
  {
    if (engine == null) {
      throw new NullPointerException("engine");
    }
    if (delegatedTaskExecutor == null) {
      throw new NullPointerException("delegatedTaskExecutor");
    }
    this.engine = engine;
    this.delegatedTaskExecutor = delegatedTaskExecutor;
    this.startTls = startTls;
  }
  
  public long getHandshakeTimeoutMillis() {
    return handshakeTimeoutMillis;
  }
  
  public void setHandshakeTimeout(long handshakeTimeout, TimeUnit unit) {
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    
    setHandshakeTimeoutMillis(unit.toMillis(handshakeTimeout));
  }
  
  public void setHandshakeTimeoutMillis(long handshakeTimeoutMillis) {
    if (handshakeTimeoutMillis < 0L) {
      throw new IllegalArgumentException("handshakeTimeoutMillis: " + handshakeTimeoutMillis + " (expected: >= 0)");
    }
    
    this.handshakeTimeoutMillis = handshakeTimeoutMillis;
  }
  
  public long getCloseNotifyTimeoutMillis() {
    return closeNotifyTimeoutMillis;
  }
  
  public void setCloseNotifyTimeout(long closeNotifyTimeout, TimeUnit unit) {
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    
    setCloseNotifyTimeoutMillis(unit.toMillis(closeNotifyTimeout));
  }
  
  public void setCloseNotifyTimeoutMillis(long closeNotifyTimeoutMillis) {
    if (closeNotifyTimeoutMillis < 0L) {
      throw new IllegalArgumentException("closeNotifyTimeoutMillis: " + closeNotifyTimeoutMillis + " (expected: >= 0)");
    }
    
    this.closeNotifyTimeoutMillis = closeNotifyTimeoutMillis;
  }
  


  public SSLEngine engine()
  {
    return engine;
  }
  


  public Future<Channel> handshakeFuture()
  {
    return handshakePromise;
  }
  



  public ChannelFuture close()
  {
    return close(ctx.newPromise());
  }
  


  public ChannelFuture close(final ChannelPromise future)
  {
    final ChannelHandlerContext ctx = this.ctx;
    ctx.executor().execute(new Runnable()
    {
      public void run() {
        engine.closeOutbound();
        try {
          write(ctx, Unpooled.EMPTY_BUFFER, future);
          flush(ctx);
        } catch (Exception e) {
          if (!future.tryFailure(e)) {
            SslHandler.logger.warn("flush() raised a masked exception.", e);
          }
          
        }
      }
    });
    return future;
  }
  







  public Future<Channel> sslCloseFuture()
  {
    return sslCloseFuture;
  }
  
  public void handlerRemoved0(ChannelHandlerContext ctx) throws Exception
  {
    if (decodeOut != null) {
      decodeOut.release();
      decodeOut = null;
    }
    for (;;) {
      PendingWrite write = (PendingWrite)pendingUnencryptedWrites.poll();
      if (write == null) {
        break;
      }
      write.failAndRecycle(new ChannelException("Pending write on removal of SslHandler"));
    }
  }
  
  public void bind(ChannelHandlerContext ctx, SocketAddress localAddress, ChannelPromise promise) throws Exception
  {
    ctx.bind(localAddress, promise);
  }
  
  public void connect(ChannelHandlerContext ctx, SocketAddress remoteAddress, SocketAddress localAddress, ChannelPromise promise)
    throws Exception
  {
    ctx.connect(remoteAddress, localAddress, promise);
  }
  
  public void deregister(ChannelHandlerContext ctx, ChannelPromise promise) throws Exception
  {
    ctx.deregister(promise);
  }
  
  public void disconnect(ChannelHandlerContext ctx, ChannelPromise promise)
    throws Exception
  {
    closeOutboundAndChannel(ctx, promise, true);
  }
  
  public void close(ChannelHandlerContext ctx, ChannelPromise promise)
    throws Exception
  {
    closeOutboundAndChannel(ctx, promise, false);
  }
  
  public void read(ChannelHandlerContext ctx)
  {
    ctx.read();
  }
  
  public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise) throws Exception
  {
    pendingUnencryptedWrites.add(PendingWrite.newInstance((ByteBuf)msg, promise));
  }
  

  public void flush(ChannelHandlerContext ctx)
    throws Exception
  {
    if ((startTls) && (!sentFirstMessage)) {
      sentFirstMessage = true;
      for (;;) {
        PendingWrite pendingWrite = (PendingWrite)pendingUnencryptedWrites.poll();
        if (pendingWrite == null) {
          break;
        }
        ctx.write(pendingWrite.msg(), (ChannelPromise)pendingWrite.recycleAndGet());
      }
      ctx.flush();
      return;
    }
    if (pendingUnencryptedWrites.isEmpty()) {
      pendingUnencryptedWrites.add(PendingWrite.newInstance(Unpooled.EMPTY_BUFFER, null));
    }
    flush0(ctx);
  }
  
  private void flush0(ChannelHandlerContext ctx) throws SSLException {
    ByteBuf out = null;
    ChannelPromise promise = null;
    try {
      for (;;) {
        PendingWrite pending = (PendingWrite)pendingUnencryptedWrites.peek();
        if (pending == null) {
          break;
        }
        if (out == null) {
          out = ctx.alloc().buffer();
        }
        ByteBuf buf = (ByteBuf)pending.msg();
        SSLEngineResult result = wrap(engine, buf, out);
        
        if (!buf.isReadable()) {
          buf.release();
          promise = (ChannelPromise)pending.recycleAndGet();
          pendingUnencryptedWrites.remove();
        } else {
          promise = null;
        }
        
        if (result.getStatus() == SSLEngineResult.Status.CLOSED)
        {
          for (;;)
          {
            PendingWrite w = (PendingWrite)pendingUnencryptedWrites.poll();
            if (w == null) {
              break;
            }
            w.failAndRecycle(SSLENGINE_CLOSED);
          }
          return;
        }
        switch (7.$SwitchMap$javax$net$ssl$SSLEngineResult$HandshakeStatus[result.getHandshakeStatus().ordinal()]) {
        case 1: 
          runDelegatedTasks();
          break;
        case 2: 
          setHandshakeSuccess();
        
        case 3: 
        case 4: 
          if (promise != null) {
            ctx.writeAndFlush(out, promise);
            promise = null;
          } else {
            ctx.writeAndFlush(out);
          }
          out = ctx.alloc().buffer();
          break;
        case 5: 
          return;
        default: 
          throw new IllegalStateException("Unknown handshake status: " + result.getHandshakeStatus());
        }
      }
    }
    catch (SSLException e) {
      setHandshakeFailure(e);
      throw e;
    } finally {
      if ((out != null) && (out.isReadable())) {
        if (promise != null) {
          ctx.writeAndFlush(out, promise);
        } else {
          ctx.writeAndFlush(out);
        }
        out = null;
      } else if (promise != null) {
        promise.trySuccess();
      }
      if (out != null) {
        out.release();
      }
    }
  }
  
  private void flushNonAppData0(ChannelHandlerContext ctx, boolean inUnwrap) throws SSLException {
    ByteBuf out = null;
    try {
      for (;;) {
        if (out == null) {
          out = ctx.alloc().buffer();
        }
        SSLEngineResult result = wrap(engine, Unpooled.EMPTY_BUFFER, out);
        
        if (result.bytesProduced() > 0) {
          ctx.writeAndFlush(out);
          out = null;
        }
        
        switch (7.$SwitchMap$javax$net$ssl$SSLEngineResult$HandshakeStatus[result.getHandshakeStatus().ordinal()]) {
        case 2: 
          setHandshakeSuccess();
          break;
        case 1: 
          runDelegatedTasks();
          break;
        case 5: 
          if (!inUnwrap) {
            unwrap(ctx);
          }
          

          break;
        case 4: 
          break;
        case 3: 
          if (!inUnwrap) {
            unwrap(ctx);
          }
          break;
        default: 
          throw new IllegalStateException("Unknown handshake status: " + result.getHandshakeStatus());
        }
        
        if (result.bytesProduced() == 0) {
          break;
        }
      }
    } catch (SSLException e) {
      setHandshakeFailure(e);
      throw e;
    } finally {
      if (out != null) {
        out.release();
      }
    }
  }
  
  private static SSLEngineResult wrap(SSLEngine engine, ByteBuf in, ByteBuf out) throws SSLException {
    ByteBuffer in0 = in.nioBuffer();
    for (;;) {
      ByteBuffer out0 = out.nioBuffer(out.writerIndex(), out.writableBytes());
      SSLEngineResult result = engine.wrap(in0, out0);
      in.skipBytes(result.bytesConsumed());
      out.writerIndex(out.writerIndex() + result.bytesProduced());
      if (result.getStatus() == SSLEngineResult.Status.BUFFER_OVERFLOW) {
        out.ensureWritable(engine.getSession().getPacketBufferSize());
      } else {
        return result;
      }
    }
  }
  

  public void channelInactive(ChannelHandlerContext ctx)
    throws Exception
  {
    setHandshakeFailure(CHANNEL_CLOSED);
    super.channelInactive(ctx);
  }
  
  public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception
  {
    if (ignoreException(cause))
    {

      if (logger.isDebugEnabled()) {
        logger.debug("Swallowing a harmless 'connection reset by peer / broken pipe' error that occurred while writing close_notify in response to the peer's close_notify", cause);
      }
      




      if (ctx.channel().isActive()) {
        ctx.close();
      }
    } else {
      ctx.fireExceptionCaught(cause);
    }
  }
  








  private boolean ignoreException(Throwable t)
  {
    if ((!(t instanceof SSLException)) && ((t instanceof IOException)) && (sslCloseFuture.isDone())) {
      String message = String.valueOf(t.getMessage()).toLowerCase();
      


      if (IGNORABLE_ERROR_MESSAGE.matcher(message).matches()) {
        return true;
      }
      

      StackTraceElement[] elements = t.getStackTrace();
      for (StackTraceElement element : elements) {
        String classname = element.getClassName();
        String methodname = element.getMethodName();
        

        if (!classname.startsWith("io.netty."))
        {



          if ("read".equals(methodname))
          {




            if (IGNORABLE_CLASS_IN_STACK.matcher(classname).matches()) {
              return true;
            }
            


            try
            {
              Class<?> clazz = getClass().getClassLoader().loadClass(classname);
              
              if ((SocketChannel.class.isAssignableFrom(clazz)) || (DatagramChannel.class.isAssignableFrom(clazz)))
              {
                return true;
              }
              

              if ((PlatformDependent.javaVersion() >= 7) && ("com.sun.nio.sctp.SctpChannel".equals(clazz.getSuperclass().getName())))
              {
                return true;
              }
            }
            catch (ClassNotFoundException e) {}
          }
        }
      }
    }
    return false;
  }
  











  public static boolean isEncrypted(ByteBuf buffer)
  {
    if (buffer.readableBytes() < 5) {
      throw new IllegalArgumentException("buffer must have at least 5 readable bytes");
    }
    return getEncryptedPacketLength(buffer) != -1;
  }
  












  private static int getEncryptedPacketLength(ByteBuf buffer)
  {
    int first = buffer.readerIndex();
    int packetLength = 0;
    
    boolean tls;
    
    switch (buffer.getUnsignedByte(first)) {
    case 20: 
    case 21: 
    case 22: 
    case 23: 
      tls = true;
      break;
    
    default: 
      tls = false;
    }
    
    if (tls)
    {
      int majorVersion = buffer.getUnsignedByte(first + 1);
      if (majorVersion == 3)
      {
        packetLength = buffer.getUnsignedShort(first + 3) + 5;
        if (packetLength <= 5)
        {
          tls = false;
        }
      }
      else {
        tls = false;
      }
    }
    
    if (!tls)
    {
      boolean sslv2 = true;
      int headerLength = (buffer.getUnsignedByte(first) & 0x80) != 0 ? 2 : 3;
      int majorVersion = buffer.getUnsignedByte(first + headerLength + 1);
      if ((majorVersion == 2) || (majorVersion == 3))
      {
        if (headerLength == 2) {
          packetLength = (buffer.getShort(first) & 0x7FFF) + 2;
        } else {
          packetLength = (buffer.getShort(first) & 0x3FFF) + 3;
        }
        if (packetLength <= headerLength) {
          sslv2 = false;
        }
      } else {
        sslv2 = false;
      }
      
      if (!sslv2) {
        return -1;
      }
    }
    return packetLength;
  }
  
  protected void decode(ChannelHandlerContext ctx, ByteBuf in, List<Object> out) throws SSLException
  {
    int packetLength = this.packetLength;
    int readableBytes = in.readableBytes();
    
    if (packetLength == 0)
    {
      if (readableBytes < 5)
      {
        return;
      }
      
      packetLength = getEncryptedPacketLength(in);
      if (packetLength == -1)
      {
        NotSslRecordException e = new NotSslRecordException("not an SSL/TLS record: " + ByteBufUtil.hexDump(in));
        
        in.skipBytes(readableBytes);
        ctx.fireExceptionCaught(e);
        setHandshakeFailure(e);
        return;
      }
      
      assert (packetLength > 0);
      this.packetLength = packetLength;
    }
    
    if (readableBytes < packetLength)
    {
      return;
    }
    









    int readerIndex = in.readerIndex();
    in.skipBytes(packetLength);
    ByteBuffer buffer = in.nioBuffer(readerIndex, packetLength);
    this.packetLength = 0;
    
    unwrap(ctx, buffer, out);
  }
  
  private void unwrap(ChannelHandlerContext ctx) throws SSLException {
    RecyclableArrayList out = RecyclableArrayList.newInstance();
    try {
      unwrap(ctx, Unpooled.EMPTY_BUFFER.nioBuffer(), out);
      int size = out.size();
      for (int i = 0; i < size; i++) {
        ctx.fireChannelRead(out.get(i));
      }
    } finally {
      out.recycle();
    }
  }
  
  private void unwrap(ChannelHandlerContext ctx, ByteBuffer packet, List<Object> out) throws SSLException {
    boolean wrapLater = false;
    int bytesProduced = 0;
    try
    {
      for (;;) {
        if (this.decodeOut == null) {
          this.decodeOut = ctx.alloc().buffer();
        }
        SSLEngineResult result = unwrap(engine, packet, this.decodeOut);
        bytesProduced += result.bytesProduced();
        switch (result.getStatus())
        {
        case CLOSED: 
          sslCloseFuture.trySuccess(ctx.channel());
          break;
        case BUFFER_UNDERFLOW: 
          break;
        }
        
        switch (7.$SwitchMap$javax$net$ssl$SSLEngineResult$HandshakeStatus[result.getHandshakeStatus().ordinal()]) {
        case 5: 
          break;
        case 4: 
          flushNonAppData0(ctx, true);
          break;
        case 1: 
          runDelegatedTasks();
          break;
        case 2: 
          setHandshakeSuccess();
          wrapLater = true;
          break;
        case 3: 
          break;
        default: 
          throw new IllegalStateException("Unknown handshake status: " + result.getHandshakeStatus());
          


          if ((result.bytesConsumed() == 0) && (result.bytesProduced() == 0))
            break label246;
        }
      }
      label246:
      if (wrapLater)
        flush0(ctx);
    } catch (SSLException e) {
      ByteBuf decodeOut;
      setHandshakeFailure(e);
      throw e;
    } finally {
      if (bytesProduced > 0) {
        ByteBuf decodeOut = this.decodeOut;
        this.decodeOut = null;
        out.add(decodeOut);
      }
    }
  }
  
  private static SSLEngineResult unwrap(SSLEngine engine, ByteBuffer in, ByteBuf out) throws SSLException {
    for (;;) {
      ByteBuffer out0 = out.nioBuffer(out.writerIndex(), out.writableBytes());
      SSLEngineResult result = engine.unwrap(in, out0);
      out.writerIndex(out.writerIndex() + result.bytesProduced());
      switch (result.getStatus()) {
      case BUFFER_OVERFLOW: 
        out.ensureWritable(engine.getSession().getApplicationBufferSize());
        break;
      default: 
        return result;
      }
    }
  }
  
  private void runDelegatedTasks() {
    for (;;) {
      Runnable task = engine.getDelegatedTask();
      if (task == null) {
        break;
      }
      
      delegatedTaskExecutor.execute(task);
    }
  }
  


  private void setHandshakeSuccess()
  {
    if (handshakePromise.trySuccess(ctx.channel())) {
      ctx.fireUserEventTriggered(SslHandshakeCompletionEvent.SUCCESS);
    }
  }
  




  private void setHandshakeFailure(Throwable cause)
  {
    engine.closeOutbound();
    try
    {
      engine.closeInbound();

    }
    catch (SSLException e)
    {

      String msg = e.getMessage();
      if ((msg == null) || (!msg.contains("possible truncation attack"))) {
        logger.debug("SSLEngine.closeInbound() raised an exception.", e);
      }
    }
    notifyHandshakeFailure(cause);
    for (;;) {
      PendingWrite write = (PendingWrite)pendingUnencryptedWrites.poll();
      if (write == null) {
        break;
      }
      write.failAndRecycle(cause);
    }
  }
  
  private void notifyHandshakeFailure(Throwable cause) {
    if (handshakePromise.tryFailure(cause)) {
      ctx.fireUserEventTriggered(new SslHandshakeCompletionEvent(cause));
      ctx.close();
    }
  }
  
  private void closeOutboundAndChannel(ChannelHandlerContext ctx, ChannelPromise promise, boolean disconnect) throws Exception
  {
    if (!ctx.channel().isActive()) {
      if (disconnect) {
        ctx.disconnect(promise);
      } else {
        ctx.close(promise);
      }
      return;
    }
    
    engine.closeOutbound();
    
    ChannelPromise closeNotifyFuture = ctx.newPromise();
    write(ctx, Unpooled.EMPTY_BUFFER, closeNotifyFuture);
    flush(ctx);
    safeClose(ctx, closeNotifyFuture, promise);
  }
  
  public void handlerAdded(ChannelHandlerContext ctx) throws Exception
  {
    this.ctx = ctx;
    
    if (ctx.channel().isActive())
    {

      handshake0();
    }
  }
  

  private Future<Channel> handshake0()
  {
    ScheduledFuture<?> timeoutFuture;
    final ScheduledFuture<?> timeoutFuture;
    if (handshakeTimeoutMillis > 0L) {
      timeoutFuture = ctx.executor().schedule(new Runnable()
      {
        public void run() {
          if (handshakePromise.isDone()) {
            return;
          }
          SslHandler.this.notifyHandshakeFailure(SslHandler.HANDSHAKE_TIMED_OUT); } }, handshakeTimeoutMillis, TimeUnit.MILLISECONDS);
    }
    else
    {
      timeoutFuture = null;
    }
    
    handshakePromise.addListener(new GenericFutureListener()
    {
      public void operationComplete(Future<Channel> f) throws Exception {
        if (timeoutFuture != null) {
          timeoutFuture.cancel(false);
        }
      }
    });
    try {
      engine.beginHandshake();
      flushNonAppData0(ctx, false);
    } catch (Exception e) {
      notifyHandshakeFailure(e);
    }
    return handshakePromise;
  }
  


  public void channelActive(final ChannelHandlerContext ctx)
    throws Exception
  {
    if ((!startTls) && (engine.getUseClientMode()))
    {

      handshake0().addListener(new GenericFutureListener()
      {
        public void operationComplete(Future<Channel> future) throws Exception {
          if (!future.isSuccess()) {
            future.cause().printStackTrace();
            ctx.close();
          }
        }
      });
    }
    ctx.fireChannelActive();
  }
  
  private void safeClose(final ChannelHandlerContext ctx, ChannelFuture flushFuture, final ChannelPromise promise)
  {
    if (!ctx.channel().isActive()) {
      ctx.close(promise); return;
    }
    
    ScheduledFuture<?> timeoutFuture;
    final ScheduledFuture<?> timeoutFuture;
    if (closeNotifyTimeoutMillis > 0L)
    {
      timeoutFuture = ctx.executor().schedule(new Runnable()
      {
        public void run() {
          SslHandler.logger.warn(ctx.channel() + " last write attempt timed out." + " Force-closing the connection.");
          

          ctx.close(promise); } }, closeNotifyTimeoutMillis, TimeUnit.MILLISECONDS);
    }
    else
    {
      timeoutFuture = null;
    }
    

    flushFuture.addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture f) throws Exception
      {
        if (timeoutFuture != null) {
          timeoutFuture.cancel(false);
        }
        if (ctx.channel().isActive()) {
          ctx.close(promise);
        }
      }
    });
  }
  
  private final class LazyChannelPromise extends DefaultPromise<Channel> {
    private LazyChannelPromise() {}
    
    protected EventExecutor executor() {
      if (ctx == null) {
        throw new IllegalStateException();
      }
      return ctx.executor();
    }
  }
}
