package net.minecraft.util.io.netty.handler.timeout;

import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import net.minecraft.util.io.netty.channel.Channel;
import net.minecraft.util.io.netty.channel.ChannelDuplexHandler;
import net.minecraft.util.io.netty.channel.ChannelFuture;
import net.minecraft.util.io.netty.channel.ChannelFutureListener;
import net.minecraft.util.io.netty.channel.ChannelHandlerContext;
import net.minecraft.util.io.netty.channel.ChannelPromise;
import net.minecraft.util.io.netty.util.concurrent.EventExecutor;





















































































public class IdleStateHandler
  extends ChannelDuplexHandler
{
  private final long readerIdleTimeMillis;
  private final long writerIdleTimeMillis;
  private final long allIdleTimeMillis;
  volatile ScheduledFuture<?> readerIdleTimeout;
  volatile long lastReadTime;
  private boolean firstReaderIdleEvent = true;
  
  volatile ScheduledFuture<?> writerIdleTimeout;
  volatile long lastWriteTime;
  private boolean firstWriterIdleEvent = true;
  
  volatile ScheduledFuture<?> allIdleTimeout;
  private boolean firstAllIdleEvent = true;
  









  private volatile int state;
  










  public IdleStateHandler(int readerIdleTimeSeconds, int writerIdleTimeSeconds, int allIdleTimeSeconds)
  {
    this(readerIdleTimeSeconds, writerIdleTimeSeconds, allIdleTimeSeconds, TimeUnit.SECONDS);
  }
  





















  public IdleStateHandler(long readerIdleTime, long writerIdleTime, long allIdleTime, TimeUnit unit)
  {
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    
    if (readerIdleTime <= 0L) {
      readerIdleTimeMillis = 0L;
    } else {
      readerIdleTimeMillis = Math.max(unit.toMillis(readerIdleTime), 1L);
    }
    if (writerIdleTime <= 0L) {
      writerIdleTimeMillis = 0L;
    } else {
      writerIdleTimeMillis = Math.max(unit.toMillis(writerIdleTime), 1L);
    }
    if (allIdleTime <= 0L) {
      allIdleTimeMillis = 0L;
    } else {
      allIdleTimeMillis = Math.max(unit.toMillis(allIdleTime), 1L);
    }
  }
  



  public long getReaderIdleTimeInMillis()
  {
    return readerIdleTimeMillis;
  }
  



  public long getWriterIdleTimeInMillis()
  {
    return writerIdleTimeMillis;
  }
  



  public long getAllIdleTimeInMillis()
  {
    return allIdleTimeMillis;
  }
  
  public void handlerAdded(ChannelHandlerContext ctx) throws Exception
  {
    if ((ctx.channel().isActive() & ctx.channel().isRegistered()))
    {

      initialize(ctx);
    }
  }
  


  public void handlerRemoved(ChannelHandlerContext ctx)
    throws Exception
  {
    destroy();
  }
  
  public void channelRegistered(ChannelHandlerContext ctx)
    throws Exception
  {
    if (ctx.channel().isActive()) {
      initialize(ctx);
    }
    super.channelRegistered(ctx);
  }
  


  public void channelActive(ChannelHandlerContext ctx)
    throws Exception
  {
    initialize(ctx);
    super.channelActive(ctx);
  }
  
  public void channelInactive(ChannelHandlerContext ctx) throws Exception
  {
    destroy();
    super.channelInactive(ctx);
  }
  
  public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception
  {
    lastReadTime = System.currentTimeMillis();
    firstReaderIdleEvent = (this.firstAllIdleEvent = 1);
    ctx.fireChannelRead(msg);
  }
  
  public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise) throws Exception
  {
    promise.addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture future) throws Exception {
        lastWriteTime = System.currentTimeMillis();
        firstWriterIdleEvent = IdleStateHandler.access$102(IdleStateHandler.this, true);
      }
    });
    ctx.write(msg, promise);
  }
  

  private void initialize(ChannelHandlerContext ctx)
  {
    switch (state) {
    case 1: 
    case 2: 
      return;
    }
    
    state = 1;
    
    EventExecutor loop = ctx.executor();
    
    lastReadTime = (this.lastWriteTime = System.currentTimeMillis());
    if (readerIdleTimeMillis > 0L) {
      readerIdleTimeout = loop.schedule(new ReaderIdleTimeoutTask(ctx), readerIdleTimeMillis, TimeUnit.MILLISECONDS);
    }
    

    if (writerIdleTimeMillis > 0L) {
      writerIdleTimeout = loop.schedule(new WriterIdleTimeoutTask(ctx), writerIdleTimeMillis, TimeUnit.MILLISECONDS);
    }
    

    if (allIdleTimeMillis > 0L) {
      allIdleTimeout = loop.schedule(new AllIdleTimeoutTask(ctx), allIdleTimeMillis, TimeUnit.MILLISECONDS);
    }
  }
  

  private void destroy()
  {
    state = 2;
    
    if (readerIdleTimeout != null) {
      readerIdleTimeout.cancel(false);
      readerIdleTimeout = null;
    }
    if (writerIdleTimeout != null) {
      writerIdleTimeout.cancel(false);
      writerIdleTimeout = null;
    }
    if (allIdleTimeout != null) {
      allIdleTimeout.cancel(false);
      allIdleTimeout = null;
    }
  }
  


  protected void channelIdle(ChannelHandlerContext ctx, IdleStateEvent evt)
    throws Exception
  {
    ctx.fireUserEventTriggered(evt);
  }
  
  private final class ReaderIdleTimeoutTask implements Runnable
  {
    private final ChannelHandlerContext ctx;
    
    ReaderIdleTimeoutTask(ChannelHandlerContext ctx) {
      this.ctx = ctx;
    }
    
    public void run()
    {
      if (!ctx.channel().isOpen()) {
        return;
      }
      
      long currentTime = System.currentTimeMillis();
      long lastReadTime = IdleStateHandler.this.lastReadTime;
      long nextDelay = readerIdleTimeMillis - (currentTime - lastReadTime);
      if (nextDelay <= 0L)
      {
        readerIdleTimeout = ctx.executor().schedule(this, readerIdleTimeMillis, TimeUnit.MILLISECONDS);
        try {
          IdleStateEvent event;
          IdleStateEvent event;
          if (firstReaderIdleEvent) {
            firstReaderIdleEvent = false;
            event = IdleStateEvent.FIRST_READER_IDLE_STATE_EVENT;
          } else {
            event = IdleStateEvent.READER_IDLE_STATE_EVENT;
          }
          channelIdle(ctx, event);
        } catch (Throwable t) {
          ctx.fireExceptionCaught(t);
        }
      }
      else {
        readerIdleTimeout = ctx.executor().schedule(this, nextDelay, TimeUnit.MILLISECONDS);
      }
    }
  }
  
  private final class WriterIdleTimeoutTask implements Runnable
  {
    private final ChannelHandlerContext ctx;
    
    WriterIdleTimeoutTask(ChannelHandlerContext ctx) {
      this.ctx = ctx;
    }
    
    public void run()
    {
      if (!ctx.channel().isOpen()) {
        return;
      }
      
      long currentTime = System.currentTimeMillis();
      long lastWriteTime = IdleStateHandler.this.lastWriteTime;
      long nextDelay = writerIdleTimeMillis - (currentTime - lastWriteTime);
      if (nextDelay <= 0L)
      {
        writerIdleTimeout = ctx.executor().schedule(this, writerIdleTimeMillis, TimeUnit.MILLISECONDS);
        try {
          IdleStateEvent event;
          IdleStateEvent event;
          if (firstWriterIdleEvent) {
            firstWriterIdleEvent = false;
            event = IdleStateEvent.FIRST_WRITER_IDLE_STATE_EVENT;
          } else {
            event = IdleStateEvent.WRITER_IDLE_STATE_EVENT;
          }
          channelIdle(ctx, event);
        } catch (Throwable t) {
          ctx.fireExceptionCaught(t);
        }
      }
      else {
        writerIdleTimeout = ctx.executor().schedule(this, nextDelay, TimeUnit.MILLISECONDS);
      }
    }
  }
  
  private final class AllIdleTimeoutTask implements Runnable
  {
    private final ChannelHandlerContext ctx;
    
    AllIdleTimeoutTask(ChannelHandlerContext ctx) {
      this.ctx = ctx;
    }
    
    public void run()
    {
      if (!ctx.channel().isOpen()) {
        return;
      }
      
      long currentTime = System.currentTimeMillis();
      long lastIoTime = Math.max(lastReadTime, lastWriteTime);
      long nextDelay = allIdleTimeMillis - (currentTime - lastIoTime);
      if (nextDelay <= 0L)
      {

        allIdleTimeout = ctx.executor().schedule(this, allIdleTimeMillis, TimeUnit.MILLISECONDS);
        try {
          IdleStateEvent event;
          IdleStateEvent event;
          if (firstAllIdleEvent) {
            firstAllIdleEvent = false;
            event = IdleStateEvent.FIRST_ALL_IDLE_STATE_EVENT;
          } else {
            event = IdleStateEvent.ALL_IDLE_STATE_EVENT;
          }
          channelIdle(ctx, event);
        } catch (Throwable t) {
          ctx.fireExceptionCaught(t);
        }
      }
      else
      {
        allIdleTimeout = ctx.executor().schedule(this, nextDelay, TimeUnit.MILLISECONDS);
      }
    }
  }
}
