package net.minecraft.util.io.netty.handler.traffic;

import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;





























public class TrafficCounter
{
  private final AtomicLong currentWrittenBytes = new AtomicLong();
  



  private final AtomicLong currentReadBytes = new AtomicLong();
  



  private final AtomicLong cumulativeWrittenBytes = new AtomicLong();
  



  private final AtomicLong cumulativeReadBytes = new AtomicLong();
  



  private long lastCumulativeTime;
  



  private long lastWriteThroughput;
  



  private long lastReadThroughput;
  



  private final AtomicLong lastTime = new AtomicLong();
  



  private long lastWrittenBytes;
  



  private long lastReadBytes;
  



  final AtomicLong checkInterval = new AtomicLong(1000L);
  




  final String name;
  



  private final AbstractTrafficShapingHandler trafficShapingHandler;
  



  private final ScheduledExecutorService executor;
  



  private Runnable monitor;
  



  private volatile ScheduledFuture<?> scheduledFuture;
  



  final AtomicBoolean monitorActive = new AtomicBoolean();
  





  private static class TrafficMonitoringTask
    implements Runnable
  {
    private final AbstractTrafficShapingHandler trafficShapingHandler1;
    




    private final TrafficCounter counter;
    




    protected TrafficMonitoringTask(AbstractTrafficShapingHandler trafficShapingHandler, TrafficCounter counter)
    {
      trafficShapingHandler1 = trafficShapingHandler;
      this.counter = counter;
    }
    
    public void run()
    {
      if (!counter.monitorActive.get()) {
        return;
      }
      long endTime = System.currentTimeMillis();
      counter.resetAccounting(endTime);
      if (trafficShapingHandler1 != null) {
        trafficShapingHandler1.doAccounting(counter);
      }
      counter.scheduledFuture = counter.executor.schedule(this, counter.checkInterval.get(), TimeUnit.MILLISECONDS);
    }
  }
  



  public void start()
  {
    synchronized (lastTime) {
      if (monitorActive.get()) {
        return;
      }
      lastTime.set(System.currentTimeMillis());
      if (checkInterval.get() > 0L) {
        monitorActive.set(true);
        monitor = new TrafficMonitoringTask(trafficShapingHandler, this);
        scheduledFuture = executor.schedule(monitor, checkInterval.get(), TimeUnit.MILLISECONDS);
      }
    }
  }
  



  public void stop()
  {
    synchronized (lastTime) {
      if (!monitorActive.get()) {
        return;
      }
      monitorActive.set(false);
      resetAccounting(System.currentTimeMillis());
      if (trafficShapingHandler != null) {
        trafficShapingHandler.doAccounting(this);
      }
      if (scheduledFuture != null) {
        scheduledFuture.cancel(true);
      }
    }
  }
  




  void resetAccounting(long newLastTime)
  {
    synchronized (lastTime) {
      long interval = newLastTime - lastTime.getAndSet(newLastTime);
      if (interval == 0L)
      {
        return;
      }
      lastReadBytes = currentReadBytes.getAndSet(0L);
      lastWrittenBytes = currentWrittenBytes.getAndSet(0L);
      lastReadThroughput = (lastReadBytes / interval * 1000L);
      
      lastWriteThroughput = (lastWrittenBytes / interval * 1000L);
    }
  }
  









  public TrafficCounter(AbstractTrafficShapingHandler trafficShapingHandler, ScheduledExecutorService executor, String name, long checkInterval)
  {
    this.trafficShapingHandler = trafficShapingHandler;
    this.executor = executor;
    this.name = name;
    lastCumulativeTime = System.currentTimeMillis();
    configure(checkInterval);
  }
  




  public void configure(long newcheckInterval)
  {
    long newInterval = newcheckInterval / 10L * 10L;
    if (checkInterval.get() != newInterval) {
      checkInterval.set(newInterval);
      if (newInterval <= 0L) {
        stop();
        
        lastTime.set(System.currentTimeMillis());
      }
      else {
        start();
      }
    }
  }
  





  void bytesRecvFlowControl(long recv)
  {
    currentReadBytes.addAndGet(recv);
    cumulativeReadBytes.addAndGet(recv);
  }
  





  void bytesWriteFlowControl(long write)
  {
    currentWrittenBytes.addAndGet(write);
    cumulativeWrittenBytes.addAndGet(write);
  }
  




  public long checkInterval()
  {
    return checkInterval.get();
  }
  



  public long lastReadThroughput()
  {
    return lastReadThroughput;
  }
  



  public long lastWriteThroughput()
  {
    return lastWriteThroughput;
  }
  



  public long lastReadBytes()
  {
    return lastReadBytes;
  }
  



  public long lastWrittenBytes()
  {
    return lastWrittenBytes;
  }
  



  public long currentReadBytes()
  {
    return currentReadBytes.get();
  }
  



  public long currentWrittenBytes()
  {
    return currentWrittenBytes.get();
  }
  


  public long lastTime()
  {
    return lastTime.get();
  }
  


  public long cumulativeWrittenBytes()
  {
    return cumulativeWrittenBytes.get();
  }
  


  public long cumulativeReadBytes()
  {
    return cumulativeReadBytes.get();
  }
  



  public long lastCumulativeTime()
  {
    return lastCumulativeTime;
  }
  


  public void resetCumulativeTime()
  {
    lastCumulativeTime = System.currentTimeMillis();
    cumulativeReadBytes.set(0L);
    cumulativeWrittenBytes.set(0L);
  }
  


  public String name()
  {
    return name;
  }
  



  public String toString()
  {
    return "Monitor " + name + " Current Speed Read: " + (lastReadThroughput >> 10) + " KB/s, Write: " + (lastWriteThroughput >> 10) + " KB/s Current Read: " + (currentReadBytes.get() >> 10) + " KB Current Write: " + (currentWrittenBytes.get() >> 10) + " KB";
  }
}
