package net.minecraft.util.io.netty.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import net.minecraft.util.io.netty.util.internal.PlatformDependent;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;
























































public class HashedWheelTimer
  implements Timer
{
  static final InternalLogger logger = InternalLoggerFactory.getInstance(HashedWheelTimer.class);
  

  private static final ResourceLeakDetector<HashedWheelTimer> leakDetector = new ResourceLeakDetector(HashedWheelTimer.class, 1, Runtime.getRuntime().availableProcessors() * 4);
  

  private final ResourceLeak leak;
  
  private final Worker worker = new Worker();
  
  final Thread workerThread;
  public static final int WORKER_STATE_INIT = 0;
  public static final int WORKER_STATE_STARTED = 1;
  public static final int WORKER_STATE_SHUTDOWN = 2;
  final AtomicInteger workerState = new AtomicInteger();
  
  final long tickDuration;
  final Set<HashedWheelTimeout>[] wheel;
  final int mask;
  final ReadWriteLock lock = new ReentrantReadWriteLock();
  

  volatile int wheelCursor;
  


  public HashedWheelTimer()
  {
    this(Executors.defaultThreadFactory());
  }
  









  public HashedWheelTimer(long tickDuration, TimeUnit unit)
  {
    this(Executors.defaultThreadFactory(), tickDuration, unit);
  }
  









  public HashedWheelTimer(long tickDuration, TimeUnit unit, int ticksPerWheel)
  {
    this(Executors.defaultThreadFactory(), tickDuration, unit, ticksPerWheel);
  }
  








  public HashedWheelTimer(ThreadFactory threadFactory)
  {
    this(threadFactory, 100L, TimeUnit.MILLISECONDS);
  }
  











  public HashedWheelTimer(ThreadFactory threadFactory, long tickDuration, TimeUnit unit)
  {
    this(threadFactory, tickDuration, unit, 512);
  }
  














  public HashedWheelTimer(ThreadFactory threadFactory, long tickDuration, TimeUnit unit, int ticksPerWheel)
  {
    if (threadFactory == null) {
      throw new NullPointerException("threadFactory");
    }
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    if (tickDuration <= 0L) {
      throw new IllegalArgumentException("tickDuration must be greater than 0: " + tickDuration);
    }
    if (ticksPerWheel <= 0) {
      throw new IllegalArgumentException("ticksPerWheel must be greater than 0: " + ticksPerWheel);
    }
    

    wheel = createWheel(ticksPerWheel);
    mask = (wheel.length - 1);
    

    this.tickDuration = unit.toNanos(tickDuration);
    

    if (this.tickDuration >= 9223372036854775807L / wheel.length) {
      throw new IllegalArgumentException(String.format("tickDuration: %d (expected: 0 < tickDuration in nanos < %d", new Object[] { Long.valueOf(tickDuration), Long.valueOf(9223372036854775807L / wheel.length) }));
    }
    


    workerThread = threadFactory.newThread(worker);
    leak = leakDetector.open(this);
  }
  
  private static Set<HashedWheelTimeout>[] createWheel(int ticksPerWheel)
  {
    if (ticksPerWheel <= 0) {
      throw new IllegalArgumentException("ticksPerWheel must be greater than 0: " + ticksPerWheel);
    }
    
    if (ticksPerWheel > 1073741824) {
      throw new IllegalArgumentException("ticksPerWheel may not be greater than 2^30: " + ticksPerWheel);
    }
    

    ticksPerWheel = normalizeTicksPerWheel(ticksPerWheel);
    Set<HashedWheelTimeout>[] wheel = new Set[ticksPerWheel];
    for (int i = 0; i < wheel.length; i++) {
      wheel[i] = Collections.newSetFromMap(PlatformDependent.newConcurrentHashMap());
    }
    
    return wheel;
  }
  
  private static int normalizeTicksPerWheel(int ticksPerWheel) {
    int normalizedTicksPerWheel = 1;
    while (normalizedTicksPerWheel < ticksPerWheel) {
      normalizedTicksPerWheel <<= 1;
    }
    return normalizedTicksPerWheel;
  }
  






  public void start()
  {
    switch (workerState.get()) {
    case 0: 
      if (workerState.compareAndSet(0, 1)) {
        workerThread.start();
      }
      break;
    case 1: 
      break;
    case 2: 
      throw new IllegalStateException("cannot be started once stopped");
    default: 
      throw new Error("Invalid WorkerState");
    }
  }
  
  public Set<Timeout> stop()
  {
    if (Thread.currentThread() == workerThread) {
      throw new IllegalStateException(HashedWheelTimer.class.getSimpleName() + ".stop() cannot be called from " + TimerTask.class.getSimpleName());
    }
    



    if (!workerState.compareAndSet(1, 2))
    {
      workerState.set(2);
      
      if (leak != null) {
        leak.close();
      }
      
      return Collections.emptySet();
    }
    
    boolean interrupted = false;
    while (workerThread.isAlive()) {
      workerThread.interrupt();
      try {
        workerThread.join(100L);
      } catch (InterruptedException e) {
        interrupted = true;
      }
    }
    
    if (interrupted) {
      Thread.currentThread().interrupt();
    }
    
    if (leak != null) {
      leak.close();
    }
    
    Set<Timeout> unprocessedTimeouts = new HashSet();
    for (Set<HashedWheelTimeout> bucket : wheel) {
      unprocessedTimeouts.addAll(bucket);
      bucket.clear();
    }
    
    return Collections.unmodifiableSet(unprocessedTimeouts);
  }
  
  public Timeout newTimeout(TimerTask task, long delay, TimeUnit unit)
  {
    long currentTime = System.nanoTime();
    
    if (task == null) {
      throw new NullPointerException("task");
    }
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    
    start();
    
    long delayInNanos = unit.toNanos(delay);
    HashedWheelTimeout timeout = new HashedWheelTimeout(task, currentTime + delayInNanos);
    scheduleTimeout(timeout, delayInNanos);
    return timeout;
  }
  
  void scheduleTimeout(HashedWheelTimeout timeout, long delay)
  {
    long relativeIndex = (delay + tickDuration - 1L) / tickDuration;
    

    if (relativeIndex < 0L) {
      relativeIndex = delay / tickDuration;
    }
    if (relativeIndex == 0L) {
      relativeIndex = 1L;
    }
    if ((relativeIndex & mask) == 0L) {
      relativeIndex -= 1L;
    }
    long remainingRounds = relativeIndex / wheel.length;
    

    lock.readLock().lock();
    try {
      if (workerState.get() == 2) {
        throw new IllegalStateException("Cannot enqueue after shutdown");
      }
      int stopIndex = (int)(wheelCursor + relativeIndex & mask);
      stopIndex = stopIndex;
      remainingRounds = remainingRounds;
      wheel[stopIndex].add(timeout);
    } finally {
      lock.readLock().unlock();
    }
  }
  
  private final class Worker
    implements Runnable
  {
    private long startTime;
    private long tick;
    
    Worker() {}
    
    public void run()
    {
      List<HashedWheelTimer.HashedWheelTimeout> expiredTimeouts = new ArrayList();
      

      startTime = System.nanoTime();
      tick = 1L;
      
      while (workerState.get() == 1) {
        long deadline = waitForNextTick();
        if (deadline > 0L) {
          fetchExpiredTimeouts(expiredTimeouts, deadline);
          notifyExpiredTimeouts(expiredTimeouts);
        }
      }
    }
    





    private void fetchExpiredTimeouts(List<HashedWheelTimer.HashedWheelTimeout> expiredTimeouts, long deadline)
    {
      lock.writeLock().lock();
      try {
        int newWheelCursor = HashedWheelTimer.this.wheelCursor = wheelCursor + 1 & mask;
        fetchExpiredTimeouts(expiredTimeouts, wheel[newWheelCursor].iterator(), deadline);
      } finally {
        lock.writeLock().unlock();
      }
    }
    


    private void fetchExpiredTimeouts(List<HashedWheelTimer.HashedWheelTimeout> expiredTimeouts, Iterator<HashedWheelTimer.HashedWheelTimeout> i, long deadline)
    {
      List<HashedWheelTimer.HashedWheelTimeout> slipped = null;
      while (i.hasNext()) {
        HashedWheelTimer.HashedWheelTimeout timeout = (HashedWheelTimer.HashedWheelTimeout)i.next();
        if (remainingRounds <= 0L) {
          i.remove();
          if (deadline <= deadline) {
            expiredTimeouts.add(timeout);

          }
          else
          {

            if (slipped == null) {
              slipped = new ArrayList();
            }
            slipped.add(timeout);
          }
        } else {
          remainingRounds -= 1L;
        }
      }
      

      if (slipped != null) {
        for (HashedWheelTimer.HashedWheelTimeout timeout : slipped) {
          scheduleTimeout(timeout, deadline - deadline);
        }
      }
    }
    

    private void notifyExpiredTimeouts(List<HashedWheelTimer.HashedWheelTimeout> expiredTimeouts)
    {
      for (int i = expiredTimeouts.size() - 1; i >= 0; i--) {
        ((HashedWheelTimer.HashedWheelTimeout)expiredTimeouts.get(i)).expire();
      }
      

      expiredTimeouts.clear();
    }
    





    private long waitForNextTick()
    {
      long deadline = startTime + tickDuration * tick;
      for (;;)
      {
        long currentTime = System.nanoTime();
        long sleepTimeMs = (deadline - currentTime + 999999L) / 1000000L;
        
        if (sleepTimeMs <= 0L) {
          tick += 1L;
          if (currentTime == -9223372036854775808L) {
            return -9223372036854775807L;
          }
          return currentTime;
        }
        






        if (PlatformDependent.isWindows()) {
          sleepTimeMs = sleepTimeMs / 10L * 10L;
        }
        try
        {
          Thread.sleep(sleepTimeMs);
        } catch (InterruptedException e) {
          if (workerState.get() == 2) {
            return -9223372036854775808L;
          }
        }
      }
    }
  }
  
  private final class HashedWheelTimeout
    implements Timeout
  {
    private static final int ST_INIT = 0;
    private static final int ST_CANCELLED = 1;
    private static final int ST_EXPIRED = 2;
    private final TimerTask task;
    final long deadline;
    volatile int stopIndex;
    volatile long remainingRounds;
    private final AtomicInteger state = new AtomicInteger(0);
    
    HashedWheelTimeout(TimerTask task, long deadline) {
      this.task = task;
      this.deadline = deadline;
    }
    
    public Timer timer()
    {
      return HashedWheelTimer.this;
    }
    
    public TimerTask task()
    {
      return task;
    }
    
    public boolean cancel()
    {
      if (!state.compareAndSet(0, 1)) {
        return false;
      }
      
      wheel[stopIndex].remove(this);
      return true;
    }
    
    public boolean isCancelled()
    {
      return state.get() == 1;
    }
    
    public boolean isExpired()
    {
      return state.get() != 0;
    }
    
    public void expire() {
      if (!state.compareAndSet(0, 2)) {
        return;
      }
      try
      {
        task.run(this);
      } catch (Throwable t) {
        if (HashedWheelTimer.logger.isWarnEnabled()) {
          HashedWheelTimer.logger.warn("An exception was thrown by " + TimerTask.class.getSimpleName() + '.', t);
        }
      }
    }
    
    public String toString()
    {
      long currentTime = System.nanoTime();
      long remaining = deadline - currentTime;
      
      StringBuilder buf = new StringBuilder(192);
      buf.append(getClass().getSimpleName());
      buf.append('(');
      
      buf.append("deadline: ");
      if (remaining > 0L) {
        buf.append(remaining);
        buf.append(" ms later, ");
      } else if (remaining < 0L) {
        buf.append(-remaining);
        buf.append(" ms ago, ");
      } else {
        buf.append("now, ");
      }
      
      if (isCancelled()) {
        buf.append(", cancelled");
      }
      
      return ')';
    }
  }
}
