package net.minecraft.util.io.netty.util;

import java.lang.ref.PhantomReference;
import java.lang.ref.ReferenceQueue;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicBoolean;
import net.minecraft.util.io.netty.util.internal.PlatformDependent;
import net.minecraft.util.io.netty.util.internal.SystemPropertyUtil;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;



















public final class ResourceLeakDetector<T>
{
  private static boolean disabled;
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(ResourceLeakDetector.class);
  private static final int DEFAULT_SAMPLING_INTERVAL = 113;
  
  static { boolean DISABLED = SystemPropertyUtil.getBoolean("io.netty.noResourceLeakDetection", false);
    logger.debug("-Dio.netty.noResourceLeakDetection: {}", Boolean.valueOf(DISABLED));
    disabled = DISABLED;
  }
  




  public static void setEnabled(boolean enabled)
  {
    disabled = !enabled;
  }
  


  public static boolean isEnabled()
  {
    return !disabled;
  }
  

  private final ResourceLeakDetector<T>.DefaultResourceLeak head = new DefaultResourceLeak(null);
  private final ResourceLeakDetector<T>.DefaultResourceLeak tail = new DefaultResourceLeak(null);
  
  private final ReferenceQueue<Object> refQueue = new ReferenceQueue();
  private final ConcurrentMap<Exception, Boolean> reportedLeaks = PlatformDependent.newConcurrentHashMap();
  
  private final String resourceType;
  private final int samplingInterval;
  private final long maxActive;
  private long active;
  private final AtomicBoolean loggedTooManyActive = new AtomicBoolean();
  private long leakCheckCnt;
  
  public ResourceLeakDetector(Class<?> resourceType)
  {
    this(resourceType.getSimpleName());
  }
  
  public ResourceLeakDetector(String resourceType) {
    this(resourceType, 113, 9223372036854775807L);
  }
  
  public ResourceLeakDetector(Class<?> resourceType, int samplingInterval, long maxActive) {
    this(resourceType.getSimpleName(), samplingInterval, maxActive);
  }
  
  public ResourceLeakDetector(String resourceType, int samplingInterval, long maxActive) {
    if (resourceType == null) {
      throw new NullPointerException("resourceType");
    }
    if (samplingInterval <= 0) {
      throw new IllegalArgumentException("samplingInterval: " + samplingInterval + " (expected: 1+)");
    }
    if (maxActive <= 0L) {
      throw new IllegalArgumentException("maxActive: " + maxActive + " (expected: 1+)");
    }
    
    this.resourceType = resourceType;
    this.samplingInterval = samplingInterval;
    this.maxActive = maxActive;
    
    head.next = tail;
    tail.prev = head;
  }
  





  public ResourceLeak open(T obj)
  {
    if ((disabled) || (leakCheckCnt++ % samplingInterval != 0L)) {
      return null;
    }
    
    reportLeak();
    
    return new DefaultResourceLeak(obj);
  }
  
  private void reportLeak() {
    if (!logger.isWarnEnabled())
    {
      for (;;) {
        ResourceLeakDetector<T>.DefaultResourceLeak ref = (DefaultResourceLeak)refQueue.poll();
        if (ref == null) {
          break;
        }
        ref.close();
      }
      return;
    }
    

    if ((active * samplingInterval > maxActive) && (loggedTooManyActive.compareAndSet(false, true))) {
      logger.warn("LEAK: You are creating too many " + resourceType + " instances.  " + resourceType + " is a shared resource that must be reused across the JVM," + "so that only a few instances are created.");
    }
    




    for (;;)
    {
      ResourceLeakDetector<T>.DefaultResourceLeak ref = (DefaultResourceLeak)refQueue.poll();
      if (ref == null) {
        break;
      }
      
      ref.clear();
      
      if (ref.close())
      {


        if (reportedLeaks.putIfAbsent(exception, Boolean.TRUE) == null) {
          logger.warn("LEAK: " + resourceType + " was GC'd before being released correctly.  " + "The following stack trace shows where the leaked object was created, " + "rather than where you failed to release it.", exception);
        }
      }
    }
  }
  
  private final class DefaultResourceLeak
    extends PhantomReference<Object> implements ResourceLeak
  {
    private final ResourceLeakException exception;
    private final AtomicBoolean freed;
    private ResourceLeakDetector<T>.DefaultResourceLeak prev;
    private ResourceLeakDetector<T>.DefaultResourceLeak next;
    
    public DefaultResourceLeak(Object referent)
    {
      super(referent != null ? refQueue : null);
      
      if (referent != null) {
        exception = new ResourceLeakException(referent.getClass().getName() + '@' + Integer.toHexString(System.identityHashCode(referent)));
        


        synchronized (head) {
          prev = head;
          next = head.next;
          head.next.prev = this;
          head.next = this;
          ResourceLeakDetector.access$508(ResourceLeakDetector.this);
        }
        freed = new AtomicBoolean();
      } else {
        exception = null;
        freed = new AtomicBoolean(true);
      }
    }
    
    public boolean close()
    {
      if (freed.compareAndSet(false, true)) {
        synchronized (head) {
          ResourceLeakDetector.access$510(ResourceLeakDetector.this);
          prev.next = next;
          next.prev = prev;
          prev = null;
          next = null;
        }
        return true;
      }
      return false;
    }
  }
}
