package net.minecraft.util.io.netty.util.concurrent;

import java.util.concurrent.CancellationException;
import java.util.concurrent.TimeUnit;
import net.minecraft.util.io.netty.util.Signal;
import net.minecraft.util.io.netty.util.internal.PlatformDependent;
import net.minecraft.util.io.netty.util.internal.StringUtil;
import net.minecraft.util.io.netty.util.internal.logging.InternalLogger;
import net.minecraft.util.io.netty.util.internal.logging.InternalLoggerFactory;


















public class DefaultPromise<V>
  extends AbstractFuture<V>
  implements Promise<V>
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(DefaultPromise.class);
  
  private static final int MAX_LISTENER_STACK_DEPTH = 8;
  
  private static final ThreadLocal<Integer> LISTENER_STACK_DEPTH = new ThreadLocal()
  {
    protected Integer initialValue() {
      return Integer.valueOf(0);
    }
  };
  private static final Signal SUCCESS = new Signal(DefaultPromise.class.getName() + ".SUCCESS");
  private static final Signal UNCANCELLABLE = new Signal(DefaultPromise.class.getName() + ".UNCANCELLABLE");
  

  private final EventExecutor executor;
  

  private volatile Object result;
  

  private Object listeners;
  

  private short waiters;
  

  public DefaultPromise(EventExecutor executor)
  {
    if (executor == null) {
      throw new NullPointerException("executor");
    }
    this.executor = executor;
  }
  
  protected DefaultPromise()
  {
    executor = null;
  }
  
  protected EventExecutor executor() {
    return executor;
  }
  
  public boolean isCancelled()
  {
    return isCancelled0(result);
  }
  
  private static boolean isCancelled0(Object result) {
    return ((result instanceof CauseHolder)) && ((cause instanceof CancellationException));
  }
  
  public boolean isCancellable()
  {
    return result == null;
  }
  
  public boolean isDone()
  {
    return isDone0(result);
  }
  
  private static boolean isDone0(Object result) {
    return (result != null) && (result != UNCANCELLABLE);
  }
  
  public boolean isSuccess()
  {
    Object result = this.result;
    if ((result == null) || (result == UNCANCELLABLE)) {
      return false;
    }
    return !(result instanceof CauseHolder);
  }
  
  public Throwable cause()
  {
    Object result = this.result;
    if ((result instanceof CauseHolder)) {
      return cause;
    }
    return null;
  }
  
  public Promise<V> addListener(GenericFutureListener<? extends Future<? super V>> listener)
  {
    if (listener == null) {
      throw new NullPointerException("listener");
    }
    
    if (isDone()) {
      notifyListener(executor(), this, listener);
      return this;
    }
    
    synchronized (this) {
      if (!isDone()) {
        if (listeners == null) {
          listeners = listener;
        }
        else if ((listeners instanceof DefaultFutureListeners)) {
          ((DefaultFutureListeners)listeners).add(listener);
        }
        else {
          GenericFutureListener<? extends Future<V>> firstListener = (GenericFutureListener)listeners;
          
          listeners = new DefaultFutureListeners(firstListener, listener);
        }
        
        return this;
      }
    }
    
    notifyListener(executor(), this, listener);
    return this;
  }
  
  public Promise<V> addListeners(GenericFutureListener<? extends Future<? super V>>... listeners)
  {
    if (listeners == null) {
      throw new NullPointerException("listeners");
    }
    
    for (GenericFutureListener<? extends Future<? super V>> l : listeners) {
      if (l == null) {
        break;
      }
      addListener(l);
    }
    return this;
  }
  
  public Promise<V> removeListener(GenericFutureListener<? extends Future<? super V>> listener)
  {
    if (listener == null) {
      throw new NullPointerException("listener");
    }
    
    if (isDone()) {
      return this;
    }
    
    synchronized (this) {
      if (!isDone()) {
        if ((listeners instanceof DefaultFutureListeners)) {
          ((DefaultFutureListeners)listeners).remove(listener);
        } else if (listeners == listener) {
          listeners = null;
        }
      }
    }
    
    return this;
  }
  
  public Promise<V> removeListeners(GenericFutureListener<? extends Future<? super V>>... listeners)
  {
    if (listeners == null) {
      throw new NullPointerException("listeners");
    }
    
    for (GenericFutureListener<? extends Future<? super V>> l : listeners) {
      if (l == null) {
        break;
      }
      removeListener(l);
    }
    return this;
  }
  
  public Promise<V> sync() throws InterruptedException
  {
    await();
    rethrowIfFailed();
    return this;
  }
  
  public Promise<V> syncUninterruptibly()
  {
    awaitUninterruptibly();
    rethrowIfFailed();
    return this;
  }
  
  private void rethrowIfFailed() {
    Throwable cause = cause();
    if (cause == null) {
      return;
    }
    
    PlatformDependent.throwException(cause);
  }
  
  public Promise<V> await() throws InterruptedException
  {
    if (isDone()) {
      return this;
    }
    
    if (Thread.interrupted()) {
      throw new InterruptedException(toString());
    }
    
    synchronized (this) {
      while (!isDone()) {
        checkDeadLock();
        incWaiters();
        try {
          wait();
        } finally {
          decWaiters();
        }
      }
    }
    return this;
  }
  
  public boolean await(long timeout, TimeUnit unit)
    throws InterruptedException
  {
    return await0(unit.toNanos(timeout), true);
  }
  
  public boolean await(long timeoutMillis) throws InterruptedException
  {
    return await0(TimeUnit.MILLISECONDS.toNanos(timeoutMillis), true);
  }
  
  public Promise<V> awaitUninterruptibly()
  {
    if (isDone()) {
      return this;
    }
    
    boolean interrupted = false;
    synchronized (this) {
      while (!isDone()) {
        checkDeadLock();
        incWaiters();
        try {
          wait();
        } catch (InterruptedException e) {
          interrupted = true;
        } finally {
          decWaiters();
        }
      }
    }
    
    if (interrupted) {
      Thread.currentThread().interrupt();
    }
    
    return this;
  }
  
  public boolean awaitUninterruptibly(long timeout, TimeUnit unit)
  {
    try {
      return await0(unit.toNanos(timeout), false);
    } catch (InterruptedException e) {
      throw new InternalError();
    }
  }
  
  public boolean awaitUninterruptibly(long timeoutMillis)
  {
    try {
      return await0(TimeUnit.MILLISECONDS.toNanos(timeoutMillis), false);
    } catch (InterruptedException e) {
      throw new InternalError();
    }
  }
  
  /* Error */
  private boolean await0(long timeoutNanos, boolean interruptable)
    throws InterruptedException
  {
    // Byte code:
    //   0: aload_0
    //   1: invokevirtual 91	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:isDone	()Z
    //   4: ifeq +5 -> 9
    //   7: iconst_1
    //   8: ireturn
    //   9: lload_1
    //   10: lconst_0
    //   11: lcmp
    //   12: ifgt +8 -> 20
    //   15: aload_0
    //   16: invokevirtual 91	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:isDone	()Z
    //   19: ireturn
    //   20: iload_3
    //   21: ifeq +21 -> 42
    //   24: invokestatic 162	java/lang/Thread:interrupted	()Z
    //   27: ifeq +15 -> 42
    //   30: new 139	java/lang/InterruptedException
    //   33: dup
    //   34: aload_0
    //   35: invokevirtual 166	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:toString	()Ljava/lang/String;
    //   38: invokespecial 167	java/lang/InterruptedException:<init>	(Ljava/lang/String;)V
    //   41: athrow
    //   42: lload_1
    //   43: lconst_0
    //   44: lcmp
    //   45: ifgt +7 -> 52
    //   48: lconst_0
    //   49: goto +6 -> 55
    //   52: invokestatic 218	java/lang/System:nanoTime	()J
    //   55: lstore 4
    //   57: lload_1
    //   58: lstore 6
    //   60: iconst_0
    //   61: istore 8
    //   63: aload_0
    //   64: dup
    //   65: astore 9
    //   67: monitorenter
    //   68: aload_0
    //   69: invokevirtual 91	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:isDone	()Z
    //   72: ifeq +23 -> 95
    //   75: iconst_1
    //   76: istore 10
    //   78: aload 9
    //   80: monitorexit
    //   81: iload 8
    //   83: ifeq +9 -> 92
    //   86: invokestatic 203	java/lang/Thread:currentThread	()Ljava/lang/Thread;
    //   89: invokevirtual 206	java/lang/Thread:interrupt	()V
    //   92: iload 10
    //   94: ireturn
    //   95: lload 6
    //   97: lconst_0
    //   98: lcmp
    //   99: ifgt +26 -> 125
    //   102: aload_0
    //   103: invokevirtual 91	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:isDone	()Z
    //   106: istore 10
    //   108: aload 9
    //   110: monitorexit
    //   111: iload 8
    //   113: ifeq +9 -> 122
    //   116: invokestatic 203	java/lang/Thread:currentThread	()Ljava/lang/Thread;
    //   119: invokevirtual 206	java/lang/Thread:interrupt	()V
    //   122: iload 10
    //   124: ireturn
    //   125: aload_0
    //   126: invokevirtual 170	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:checkDeadLock	()V
    //   129: aload_0
    //   130: invokespecial 173	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:incWaiters	()V
    //   133: aload_0
    //   134: lload 6
    //   136: ldc2_w 219
    //   139: ldiv
    //   140: lload 6
    //   142: ldc2_w 219
    //   145: lrem
    //   146: l2i
    //   147: invokevirtual 223	java/lang/Object:wait	(JI)V
    //   150: goto +15 -> 165
    //   153: astore 10
    //   155: iload_3
    //   156: ifeq +6 -> 162
    //   159: aload 10
    //   161: athrow
    //   162: iconst_1
    //   163: istore 8
    //   165: aload_0
    //   166: invokevirtual 91	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:isDone	()Z
    //   169: ifeq +27 -> 196
    //   172: iconst_1
    //   173: istore 10
    //   175: aload_0
    //   176: invokespecial 179	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:decWaiters	()V
    //   179: aload 9
    //   181: monitorexit
    //   182: iload 8
    //   184: ifeq +9 -> 193
    //   187: invokestatic 203	java/lang/Thread:currentThread	()Ljava/lang/Thread;
    //   190: invokevirtual 206	java/lang/Thread:interrupt	()V
    //   193: iload 10
    //   195: ireturn
    //   196: lload_1
    //   197: invokestatic 218	java/lang/System:nanoTime	()J
    //   200: lload 4
    //   202: lsub
    //   203: lsub
    //   204: lstore 6
    //   206: lload 6
    //   208: lconst_0
    //   209: lcmp
    //   210: ifgt -77 -> 133
    //   213: aload_0
    //   214: invokevirtual 91	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:isDone	()Z
    //   217: istore 10
    //   219: aload_0
    //   220: invokespecial 179	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:decWaiters	()V
    //   223: aload 9
    //   225: monitorexit
    //   226: iload 8
    //   228: ifeq +9 -> 237
    //   231: invokestatic 203	java/lang/Thread:currentThread	()Ljava/lang/Thread;
    //   234: invokevirtual 206	java/lang/Thread:interrupt	()V
    //   237: iload 10
    //   239: ireturn
    //   240: astore 11
    //   242: aload_0
    //   243: invokespecial 179	net/minecraft/util/io/netty/util/concurrent/DefaultPromise:decWaiters	()V
    //   246: aload 11
    //   248: athrow
    //   249: astore 12
    //   251: aload 9
    //   253: monitorexit
    //   254: aload 12
    //   256: athrow
    //   257: astore 13
    //   259: iload 8
    //   261: ifeq +9 -> 270
    //   264: invokestatic 203	java/lang/Thread:currentThread	()Ljava/lang/Thread;
    //   267: invokevirtual 206	java/lang/Thread:interrupt	()V
    //   270: aload 13
    //   272: athrow
    // Line number table:
    //   Java source line #307	-> byte code offset #0
    //   Java source line #308	-> byte code offset #7
    //   Java source line #311	-> byte code offset #9
    //   Java source line #312	-> byte code offset #15
    //   Java source line #315	-> byte code offset #20
    //   Java source line #316	-> byte code offset #30
    //   Java source line #319	-> byte code offset #42
    //   Java source line #320	-> byte code offset #57
    //   Java source line #321	-> byte code offset #60
    //   Java source line #324	-> byte code offset #63
    //   Java source line #325	-> byte code offset #68
    //   Java source line #326	-> byte code offset #75
    //   Java source line #361	-> byte code offset #81
    //   Java source line #362	-> byte code offset #86
    //   Java source line #329	-> byte code offset #95
    //   Java source line #330	-> byte code offset #102
    //   Java source line #361	-> byte code offset #111
    //   Java source line #362	-> byte code offset #116
    //   Java source line #333	-> byte code offset #125
    //   Java source line #334	-> byte code offset #129
    //   Java source line #338	-> byte code offset #133
    //   Java source line #345	-> byte code offset #150
    //   Java source line #339	-> byte code offset #153
    //   Java source line #340	-> byte code offset #155
    //   Java source line #341	-> byte code offset #159
    //   Java source line #343	-> byte code offset #162
    //   Java source line #347	-> byte code offset #165
    //   Java source line #348	-> byte code offset #172
    //   Java source line #357	-> byte code offset #175
    //   Java source line #361	-> byte code offset #182
    //   Java source line #362	-> byte code offset #187
    //   Java source line #350	-> byte code offset #196
    //   Java source line #351	-> byte code offset #206
    //   Java source line #352	-> byte code offset #213
    //   Java source line #357	-> byte code offset #219
    //   Java source line #361	-> byte code offset #226
    //   Java source line #362	-> byte code offset #231
    //   Java source line #357	-> byte code offset #240
    //   Java source line #359	-> byte code offset #249
    //   Java source line #361	-> byte code offset #257
    //   Java source line #362	-> byte code offset #264
    // Local variable table:
    //   start	length	slot	name	signature
    //   0	273	0	this	DefaultPromise<V>
    //   0	273	1	timeoutNanos	long
    //   0	273	3	interruptable	boolean
    //   55	146	4	startTime	long
    //   58	149	6	waitTime	long
    //   61	199	8	interrupted	boolean
    //   76	47	10	bool1	boolean
    //   153	85	10	e	InterruptedException
    //   173	65	10	bool2	boolean
    //   240	7	11	localObject1	Object
    //   249	6	12	localObject2	Object
    //   257	14	13	localObject3	Object
    // Exception table:
    //   from	to	target	type
    //   133	150	153	java/lang/InterruptedException
    //   133	175	240	finally
    //   196	219	240	finally
    //   240	242	240	finally
    //   68	81	249	finally
    //   95	111	249	finally
    //   125	182	249	finally
    //   196	226	249	finally
    //   240	254	249	finally
    //   63	81	257	finally
    //   95	111	257	finally
    //   125	182	257	finally
    //   196	226	257	finally
    //   240	259	257	finally
  }
  
  protected void checkDeadLock()
  {
    EventExecutor e = executor();
    if ((e != null) && (e.inEventLoop())) {
      throw new BlockingOperationException(toString());
    }
  }
  
  public Promise<V> setSuccess(V result)
  {
    if (setSuccess0(result)) {
      notifyListeners();
      return this;
    }
    throw new IllegalStateException("complete already: " + this);
  }
  
  public boolean trySuccess(V result)
  {
    if (setSuccess0(result)) {
      notifyListeners();
      return true;
    }
    return false;
  }
  
  public Promise<V> setFailure(Throwable cause)
  {
    if (setFailure0(cause)) {
      notifyListeners();
      return this;
    }
    throw new IllegalStateException("complete already: " + this, cause);
  }
  
  public boolean tryFailure(Throwable cause)
  {
    if (setFailure0(cause)) {
      notifyListeners();
      return true;
    }
    return false;
  }
  
  public boolean cancel(boolean mayInterruptIfRunning)
  {
    Object result = this.result;
    if ((isDone0(result)) || (result == UNCANCELLABLE)) {
      return false;
    }
    
    synchronized (this)
    {
      result = this.result;
      if ((isDone0(result)) || (result == UNCANCELLABLE)) {
        return false;
      }
      
      this.result = new CauseHolder(new CancellationException(), null);
      if (hasWaiters()) {
        notifyAll();
      }
    }
    
    notifyListeners();
    return true;
  }
  
  public boolean setUncancellable()
  {
    Object result = this.result;
    if (isDone0(result)) {
      return false;
    }
    
    synchronized (this)
    {
      result = this.result;
      if (isDone0(result)) {
        return false;
      }
      
      this.result = UNCANCELLABLE;
    }
    return true;
  }
  
  private boolean setFailure0(Throwable cause) {
    if (isDone()) {
      return false;
    }
    
    synchronized (this)
    {
      if (isDone()) {
        return false;
      }
      
      result = new CauseHolder(cause, null);
      if (hasWaiters()) {
        notifyAll();
      }
    }
    return true;
  }
  
  private boolean setSuccess0(V result) {
    if (isDone()) {
      return false;
    }
    
    synchronized (this)
    {
      if (isDone()) {
        return false;
      }
      if (result == null) {
        this.result = SUCCESS;
      } else {
        this.result = result;
      }
      if (hasWaiters()) {
        notifyAll();
      }
    }
    return true;
  }
  

  public V getNow()
  {
    Object result = this.result;
    if (((result instanceof CauseHolder)) || (result == SUCCESS)) {
      return null;
    }
    return result;
  }
  
  private boolean hasWaiters() {
    return waiters > 0;
  }
  
  private void incWaiters() {
    if (waiters == 32767) {
      throw new IllegalStateException("too many waiters: " + this);
    }
    waiters = ((short)(waiters + 1));
  }
  
  private void decWaiters() {
    waiters = ((short)(waiters - 1));
  }
  





  private void notifyListeners()
  {
    Object listeners = this.listeners;
    if (listeners == null) {
      return;
    }
    
    this.listeners = null;
    
    EventExecutor executor = executor();
    if (executor.inEventLoop()) {
      Integer stackDepth = (Integer)LISTENER_STACK_DEPTH.get();
      if (stackDepth.intValue() < 8) {
        LISTENER_STACK_DEPTH.set(Integer.valueOf(stackDepth.intValue() + 1));
        try {
          if ((listeners instanceof DefaultFutureListeners)) {
            notifyListeners0(this, (DefaultFutureListeners)listeners);
          }
          else {
            GenericFutureListener<? extends Future<V>> l = (GenericFutureListener)listeners;
            
            notifyListener0(this, l);
          }
        } finally {
          LISTENER_STACK_DEPTH.set(stackDepth);
        }
        return;
      }
    }
    try
    {
      if ((listeners instanceof DefaultFutureListeners)) {
        final DefaultFutureListeners dfl = (DefaultFutureListeners)listeners;
        executor.execute(new Runnable()
        {
          public void run() {
            DefaultPromise.notifyListeners0(DefaultPromise.this, dfl);
          }
        });
      }
      else {
        final GenericFutureListener<? extends Future<V>> l = (GenericFutureListener)listeners;
        
        executor.execute(new Runnable()
        {
          public void run() {
            DefaultPromise.notifyListener0(DefaultPromise.this, l);
          }
        });
      }
    } catch (Throwable t) {
      logger.error("Failed to notify listener(s). Event loop shut down?", t);
    }
  }
  
  private static void notifyListeners0(Future<?> future, DefaultFutureListeners listeners) {
    GenericFutureListener<?>[] a = listeners.listeners();
    int size = listeners.size();
    for (int i = 0; i < size; i++) {
      notifyListener0(future, a[i]);
    }
  }
  

  protected static void notifyListener(EventExecutor eventExecutor, Future<?> future, final GenericFutureListener<?> l)
  {
    if (eventExecutor.inEventLoop()) {
      Integer stackDepth = (Integer)LISTENER_STACK_DEPTH.get();
      if (stackDepth.intValue() < 8) {
        LISTENER_STACK_DEPTH.set(Integer.valueOf(stackDepth.intValue() + 1));
        try {
          notifyListener0(future, l);
        } finally {
          LISTENER_STACK_DEPTH.set(stackDepth);
        }
        return;
      }
    }
    try
    {
      eventExecutor.execute(new Runnable()
      {
        public void run() {
          DefaultPromise.notifyListener0(val$future, l);
        }
      });
    } catch (Throwable t) {
      logger.error("Failed to notify a listener. Event loop shut down?", t);
    }
  }
  
  static void notifyListener0(Future future, GenericFutureListener l)
  {
    try {
      l.operationComplete(future);
    } catch (Throwable t) {
      if (logger.isWarnEnabled()) {
        logger.warn("An exception was thrown by " + l.getClass().getName() + ".operationComplete()", t);
      }
    }
  }
  



  private synchronized Object progressiveListeners()
  {
    Object listeners = this.listeners;
    if (listeners == null)
    {
      return null;
    }
    
    if ((listeners instanceof DefaultFutureListeners))
    {
      DefaultFutureListeners dfl = (DefaultFutureListeners)listeners;
      int progressiveSize = dfl.progressiveSize();
      switch (progressiveSize) {
      case 0: 
        return null;
      case 1: 
        for (GenericFutureListener<?> l : dfl.listeners()) {
          if ((l instanceof GenericProgressiveFutureListener)) {
            return l;
          }
        }
        return null;
      }
      
      GenericFutureListener<?>[] array = dfl.listeners();
      GenericProgressiveFutureListener<?>[] copy = new GenericProgressiveFutureListener[progressiveSize];
      int i = 0; for (int j = 0; j < progressiveSize; i++) {
        GenericFutureListener<?> l = array[i];
        if ((l instanceof GenericProgressiveFutureListener)) {
          copy[(j++)] = ((GenericProgressiveFutureListener)l);
        }
      }
      
      return copy; }
    if ((listeners instanceof GenericProgressiveFutureListener)) {
      return listeners;
    }
    
    return null;
  }
  

  void notifyProgressiveListeners(final long progress, long total)
  {
    Object listeners = progressiveListeners();
    if (listeners == null) {
      return;
    }
    
    final ProgressiveFuture<V> self = (ProgressiveFuture)this;
    
    EventExecutor executor = executor();
    if (executor.inEventLoop()) {
      if ((listeners instanceof GenericProgressiveFutureListener[])) {
        notifyProgressiveListeners0(self, (GenericProgressiveFutureListener[])listeners, progress, total);
      }
      else {
        notifyProgressiveListener0(self, (GenericProgressiveFutureListener)listeners, progress, total);
      }
    }
    else {
      try {
        if ((listeners instanceof GenericProgressiveFutureListener[])) {
          final GenericProgressiveFutureListener<?>[] array = (GenericProgressiveFutureListener[])listeners;
          
          executor.execute(new Runnable()
          {
            public void run() {
              DefaultPromise.notifyProgressiveListeners0(self, array, progress, val$total);
            }
          });
        } else {
          final GenericProgressiveFutureListener<ProgressiveFuture<V>> l = (GenericProgressiveFutureListener)listeners;
          
          executor.execute(new Runnable()
          {
            public void run() {
              DefaultPromise.notifyProgressiveListener0(self, l, progress, val$total);
            }
          });
        }
      } catch (Throwable t) {
        logger.error("Failed to notify listener(s). Event loop shut down?", t);
      }
    }
  }
  
  private static void notifyProgressiveListeners0(ProgressiveFuture<?> future, GenericProgressiveFutureListener<?>[] listeners, long progress, long total)
  {
    for (GenericProgressiveFutureListener<?> l : listeners) {
      if (l == null) {
        break;
      }
      notifyProgressiveListener0(future, l, progress, total);
    }
  }
  
  private static void notifyProgressiveListener0(ProgressiveFuture future, GenericProgressiveFutureListener l, long progress, long total)
  {
    try
    {
      l.operationProgressed(future, progress, total);
    } catch (Throwable t) {
      if (logger.isWarnEnabled())
        logger.warn("An exception was thrown by " + l.getClass().getName() + ".operationProgressed()", t);
    }
  }
  
  private static final class CauseHolder {
    final Throwable cause;
    
    private CauseHolder(Throwable cause) {
      this.cause = cause;
    }
  }
  
  public String toString()
  {
    return toStringBuilder().toString();
  }
  
  protected StringBuilder toStringBuilder() {
    StringBuilder buf = new StringBuilder(64);
    buf.append(StringUtil.simpleClassName(this));
    buf.append('@');
    buf.append(Integer.toHexString(hashCode()));
    
    Object result = this.result;
    if (result == SUCCESS) {
      buf.append("(success)");
    } else if (result == UNCANCELLABLE) {
      buf.append("(uncancellable)");
    } else if ((result instanceof CauseHolder)) {
      buf.append("(failure(");
      buf.append(cause);
      buf.append(')');
    } else {
      buf.append("(incomplete)");
    }
    return buf;
  }
}
