package org.apache.logging.log4j.core;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.core.config.Configuration;
import org.apache.logging.log4j.core.config.ConfigurationFactory;
import org.apache.logging.log4j.core.config.ConfigurationListener;
import org.apache.logging.log4j.core.config.DefaultConfiguration;
import org.apache.logging.log4j.core.config.NullConfiguration;
import org.apache.logging.log4j.core.config.Reconfigurable;
import org.apache.logging.log4j.core.helpers.Assert;
import org.apache.logging.log4j.core.helpers.NetUtils;
import org.apache.logging.log4j.message.MessageFactory;
import org.apache.logging.log4j.spi.AbstractLogger;
import org.apache.logging.log4j.spi.LoggerContextFactory;
import org.apache.logging.log4j.status.StatusLogger;























public class LoggerContext
  implements org.apache.logging.log4j.spi.LoggerContext, ConfigurationListener, LifeCycle
{
  public static final String PROPERTY_CONFIG = "config";
  private static final StatusLogger LOGGER = ;
  
  private final ConcurrentMap<String, Logger> loggers = new ConcurrentHashMap();
  private final CopyOnWriteArrayList<PropertyChangeListener> propertyChangeListeners = new CopyOnWriteArrayList();
  




  private volatile Configuration config = new DefaultConfiguration();
  
  private Object externalContext;
  private final String name;
  private URI configLocation;
  private ShutdownThread shutdownThread = null;
  



  public static enum Status
  {
    INITIALIZED, 
    
    STARTING, 
    
    STARTED, 
    
    STOPPING, 
    
    STOPPED;
    
    private Status() {} }
  private volatile Status status = Status.INITIALIZED;
  
  private final Lock configLock = new ReentrantLock();
  



  public LoggerContext(String name)
  {
    this(name, null, (URI)null);
  }
  




  public LoggerContext(String name, Object externalContext)
  {
    this(name, externalContext, (URI)null);
  }
  





  public LoggerContext(String name, Object externalContext, URI configLocn)
  {
    this.name = name;
    this.externalContext = externalContext;
    configLocation = configLocn;
  }
  







  public LoggerContext(String name, Object externalContext, String configLocn)
  {
    this.name = name;
    this.externalContext = externalContext;
    if (configLocn != null) {
      URI uri;
      try {
        uri = new File(configLocn).toURI();
      } catch (Exception ex) {
        uri = null;
      }
      configLocation = uri;
    } else {
      configLocation = null;
    }
  }
  
  public void start()
  {
    if (configLock.tryLock()) {
      try {
        if ((status == Status.INITIALIZED) || (status == Status.STOPPED)) {
          status = Status.STARTING;
          reconfigure();
          if (config.isShutdownHookEnabled()) {
            shutdownThread = new ShutdownThread(this);
            try {
              Runtime.getRuntime().addShutdownHook(shutdownThread);
            } catch (IllegalStateException ise) {
              LOGGER.warn("Unable to register shutdown hook due to JVM state");
              shutdownThread = null;
            } catch (SecurityException se) {
              LOGGER.warn("Unable to register shutdown hook due to security restrictions");
              shutdownThread = null;
            }
          }
          status = Status.STARTED;
        }
      } finally {
        configLock.unlock();
      }
    }
  }
  



  public void start(Configuration config)
  {
    if (configLock.tryLock()) {
      try {
        if (((status == Status.INITIALIZED) || (status == Status.STOPPED)) && (config.isShutdownHookEnabled())) {
          shutdownThread = new ShutdownThread(this);
          try {
            Runtime.getRuntime().addShutdownHook(shutdownThread);
          } catch (IllegalStateException ise) {
            LOGGER.warn("Unable to register shutdown hook due to JVM state");
            shutdownThread = null;
          } catch (SecurityException se) {
            LOGGER.warn("Unable to register shutdown hook due to security restrictions");
            shutdownThread = null;
          }
          status = Status.STARTED;
        }
      } finally {
        configLock.unlock();
      }
    }
    setConfiguration(config);
  }
  
  public void stop()
  {
    configLock.lock();
    try {
      if (status == Status.STOPPED) {
        return;
      }
      status = Status.STOPPING;
      if (shutdownThread != null) {
        Runtime.getRuntime().removeShutdownHook(shutdownThread);
        shutdownThread = null;
      }
      Configuration prev = config;
      config = new NullConfiguration();
      updateLoggers();
      prev.stop();
      externalContext = null;
      LogManager.getFactory().removeContext(this);
      status = Status.STOPPED;
    } finally {
      configLock.unlock();
    }
  }
  




  public String getName()
  {
    return name;
  }
  
  public Status getStatus() {
    return status;
  }
  
  public boolean isStarted()
  {
    return status == Status.STARTED;
  }
  



  public void setExternalContext(Object context)
  {
    externalContext = context;
  }
  




  public Object getExternalContext()
  {
    return externalContext;
  }
  





  public Logger getLogger(String name)
  {
    return getLogger(name, null);
  }
  








  public Logger getLogger(String name, MessageFactory messageFactory)
  {
    Logger logger = (Logger)loggers.get(name);
    if (logger != null) {
      AbstractLogger.checkMessageFactory(logger, messageFactory);
      return logger;
    }
    
    logger = newInstance(this, name, messageFactory);
    Logger prev = (Logger)loggers.putIfAbsent(name, logger);
    return prev == null ? logger : prev;
  }
  





  public boolean hasLogger(String name)
  {
    return loggers.containsKey(name);
  }
  





  public Configuration getConfiguration()
  {
    return config;
  }
  




  public void addFilter(Filter filter)
  {
    config.addFilter(filter);
  }
  



  public void removeFilter(Filter filter)
  {
    config.removeFilter(filter);
  }
  




  private synchronized Configuration setConfiguration(Configuration config)
  {
    if (config == null) {
      throw new NullPointerException("No Configuration was provided");
    }
    Configuration prev = this.config;
    config.addListener(this);
    Map<String, String> map = new HashMap();
    map.put("hostName", NetUtils.getLocalHostname());
    map.put("contextName", name);
    config.addComponent("ContextProperties", map);
    config.start();
    this.config = config;
    updateLoggers();
    if (prev != null) {
      prev.removeListener(this);
      prev.stop();
    }
    

    PropertyChangeEvent evt = new PropertyChangeEvent(this, "config", prev, config);
    for (PropertyChangeListener listener : propertyChangeListeners) {
      listener.propertyChange(evt);
    }
    return prev;
  }
  
  public void addPropertyChangeListener(PropertyChangeListener listener) {
    propertyChangeListeners.add(Assert.isNotNull(listener, "listener"));
  }
  
  public void removePropertyChangeListener(PropertyChangeListener listener) {
    propertyChangeListeners.remove(listener);
  }
  
  public synchronized URI getConfigLocation() {
    return configLocation;
  }
  
  public synchronized void setConfigLocation(URI configLocation) {
    this.configLocation = configLocation;
    reconfigure();
  }
  


  public synchronized void reconfigure()
  {
    LOGGER.debug("Reconfiguration started for context " + name);
    Configuration instance = ConfigurationFactory.getInstance().getConfiguration(name, configLocation);
    setConfiguration(instance);
    



    LOGGER.debug("Reconfiguration completed");
  }
  


  public void updateLoggers()
  {
    updateLoggers(config);
  }
  



  public void updateLoggers(Configuration config)
  {
    for (Logger logger : loggers.values()) {
      logger.updateConfiguration(config);
    }
  }
  






  public synchronized void onChange(Reconfigurable reconfigurable)
  {
    LOGGER.debug("Reconfiguration started for context " + name);
    Configuration config = reconfigurable.reconfigure();
    if (config != null) {
      setConfiguration(config);
      LOGGER.debug("Reconfiguration completed");
    } else {
      LOGGER.debug("Reconfiguration failed");
    }
  }
  
  protected Logger newInstance(LoggerContext ctx, String name, MessageFactory messageFactory)
  {
    return new Logger(ctx, name, messageFactory);
  }
  
  private class ShutdownThread extends Thread
  {
    private final LoggerContext context;
    
    public ShutdownThread(LoggerContext context) {
      this.context = context;
    }
    
    public void run()
    {
      context.shutdownThread = null;
      context.stop();
    }
  }
}
