package org.apache.logging.log4j.core.filter;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.Marker;
import org.apache.logging.log4j.ThreadContext;
import org.apache.logging.log4j.core.Filter.Result;
import org.apache.logging.log4j.core.LogEvent;
import org.apache.logging.log4j.core.Logger;
import org.apache.logging.log4j.core.config.plugins.Plugin;
import org.apache.logging.log4j.core.config.plugins.PluginAttribute;
import org.apache.logging.log4j.core.config.plugins.PluginElement;
import org.apache.logging.log4j.core.config.plugins.PluginFactory;
import org.apache.logging.log4j.core.helpers.KeyValuePair;
import org.apache.logging.log4j.message.Message;

















@Plugin(name="DynamicThresholdFilter", category="Core", elementType="filter", printObject=true)
public final class DynamicThresholdFilter
  extends AbstractFilter
{
  private Map<String, Level> levelMap = new HashMap();
  private Level defaultThreshold = Level.ERROR;
  private final String key;
  
  private DynamicThresholdFilter(String key, Map<String, Level> pairs, Level defaultLevel, Filter.Result onMatch, Filter.Result onMismatch)
  {
    super(onMatch, onMismatch);
    if (key == null) {
      throw new NullPointerException("key cannot be null");
    }
    this.key = key;
    levelMap = pairs;
    defaultThreshold = defaultLevel;
  }
  
  public String getKey() {
    return key;
  }
  

  public Filter.Result filter(Logger logger, Level level, Marker marker, String msg, Object... params)
  {
    return filter(level);
  }
  

  public Filter.Result filter(Logger logger, Level level, Marker marker, Object msg, Throwable t)
  {
    return filter(level);
  }
  

  public Filter.Result filter(Logger logger, Level level, Marker marker, Message msg, Throwable t)
  {
    return filter(level);
  }
  
  public Filter.Result filter(LogEvent event)
  {
    return filter(event.getLevel());
  }
  
  private Filter.Result filter(Level level) {
    Object value = ThreadContext.get(key);
    if (value != null) {
      Level ctxLevel = (Level)levelMap.get(value);
      if (ctxLevel == null) {
        ctxLevel = defaultThreshold;
      }
      return level.isAtLeastAsSpecificAs(ctxLevel) ? onMatch : onMismatch;
    }
    return Filter.Result.NEUTRAL;
  }
  
  public Map<String, Level> getLevelMap()
  {
    return levelMap;
  }
  
  public String toString()
  {
    StringBuilder sb = new StringBuilder();
    sb.append("key=").append(key);
    sb.append(", default=").append(defaultThreshold);
    if (levelMap.size() > 0) {
      sb.append("{");
      boolean first = true;
      for (Map.Entry<String, Level> entry : levelMap.entrySet()) {
        if (!first) {
          sb.append(", ");
          first = false;
        }
        sb.append((String)entry.getKey()).append("=").append(entry.getValue());
      }
      sb.append("}");
    }
    return sb.toString();
  }
  













  @PluginFactory
  public static DynamicThresholdFilter createFilter(@PluginAttribute("key") String key, @PluginElement("Pairs") KeyValuePair[] pairs, @PluginAttribute("defaultThreshold") String levelName, @PluginAttribute("onMatch") String match, @PluginAttribute("onMismatch") String mismatch)
  {
    Filter.Result onMatch = Filter.Result.toResult(match);
    Filter.Result onMismatch = Filter.Result.toResult(mismatch);
    Map<String, Level> map = new HashMap();
    for (KeyValuePair pair : pairs) {
      map.put(pair.getKey(), Level.toLevel(pair.getValue()));
    }
    Level level = Level.toLevel(levelName, Level.ERROR);
    return new DynamicThresholdFilter(key, map, level, onMatch, onMismatch);
  }
}
