package org.apache.logging.log4j.core.pattern;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.core.config.Configuration;
import org.apache.logging.log4j.core.config.plugins.PluginManager;
import org.apache.logging.log4j.core.config.plugins.PluginType;
import org.apache.logging.log4j.core.helpers.Strings;
import org.apache.logging.log4j.status.StatusLogger;































public final class PatternParser
{
  private static final char ESCAPE_CHAR = '%';
  
  private static enum ParserState
  {
    LITERAL_STATE, 
    



    CONVERTER_STATE, 
    



    DOT_STATE, 
    



    MIN_STATE, 
    



    MAX_STATE;
    
    private ParserState() {} }
  private static final Logger LOGGER = ;
  

  private static final int BUF_SIZE = 32;
  

  private static final int DECIMAL = 10;
  

  private final Configuration config;
  
  private final Map<String, Class<PatternConverter>> converterRules;
  

  public PatternParser(String converterKey)
  {
    this(null, converterKey, null, null);
  }
  





  public PatternParser(Configuration config, String converterKey, Class<?> expected)
  {
    this(config, converterKey, expected, null);
  }
  







  public PatternParser(Configuration config, String converterKey, Class<?> expectedClass, Class<?> filterClass)
  {
    this.config = config;
    PluginManager manager = new PluginManager(converterKey, expectedClass);
    manager.collectPlugins();
    Map<String, PluginType<?>> plugins = manager.getPlugins();
    Map<String, Class<PatternConverter>> converters = new HashMap();
    
    for (PluginType<?> type : plugins.values()) {
      try
      {
        Class<PatternConverter> clazz = type.getPluginClass();
        if ((filterClass == null) || (filterClass.isAssignableFrom(clazz)))
        {

          ConverterKeys keys = (ConverterKeys)clazz.getAnnotation(ConverterKeys.class);
          if (keys != null) {
            for (String key : keys.value())
              converters.put(key, clazz);
          }
        }
      } catch (Exception ex) {
        LOGGER.error("Error processing plugin " + type.getElementName(), ex);
      }
    }
    converterRules = converters;
  }
  
  public List<PatternFormatter> parse(String pattern) {
    return parse(pattern, false);
  }
  
  public List<PatternFormatter> parse(String pattern, boolean alwaysWriteExceptions) {
    List<PatternFormatter> list = new ArrayList();
    List<PatternConverter> converters = new ArrayList();
    List<FormattingInfo> fields = new ArrayList();
    
    parse(pattern, converters, fields);
    
    Iterator<FormattingInfo> fieldIter = fields.iterator();
    boolean handlesThrowable = false;
    
    for (PatternConverter converter : converters) {
      LogEventPatternConverter pc;
      if ((converter instanceof LogEventPatternConverter)) {
        LogEventPatternConverter pc = (LogEventPatternConverter)converter;
        handlesThrowable |= pc.handlesThrowable();
      } else {
        pc = new LiteralPatternConverter(config, "");
      }
      FormattingInfo field;
      FormattingInfo field;
      if (fieldIter.hasNext()) {
        field = (FormattingInfo)fieldIter.next();
      } else {
        field = FormattingInfo.getDefault();
      }
      list.add(new PatternFormatter(pc, field));
    }
    if ((alwaysWriteExceptions) && (!handlesThrowable)) {
      LogEventPatternConverter pc = ExtendedThrowablePatternConverter.newInstance(null);
      list.add(new PatternFormatter(pc, FormattingInfo.getDefault()));
    }
    return list;
  }
  

















  private static int extractConverter(char lastChar, String pattern, int i, StringBuilder convBuf, StringBuilder currentLiteral)
  {
    convBuf.setLength(0);
    





    if (!Character.isUnicodeIdentifierStart(lastChar)) {
      return i;
    }
    
    convBuf.append(lastChar);
    
    while ((i < pattern.length()) && (Character.isUnicodeIdentifierPart(pattern.charAt(i)))) {
      convBuf.append(pattern.charAt(i));
      currentLiteral.append(pattern.charAt(i));
      i++;
    }
    
    return i;
  }
  







  private static int extractOptions(String pattern, int i, List<String> options)
  {
    while ((i < pattern.length()) && (pattern.charAt(i) == '{')) {
      int begin = i++;
      
      int depth = 0;
      int end;
      do { end = pattern.indexOf('}', i);
        if (end != -1) {
          int next = pattern.indexOf("{", i);
          if ((next != -1) && (next < end)) {
            i = end + 1;
            depth++;
          } else if (depth > 0) {
            depth--;
          }
        }
      } while (depth > 0);
      
      if (end == -1) {
        break;
      }
      
      String r = pattern.substring(begin + 1, end);
      options.add(r);
      i = end + 1;
    }
    
    return i;
  }
  







  public void parse(String pattern, List<PatternConverter> patternConverters, List<FormattingInfo> formattingInfos)
  {
    if (pattern == null) {
      throw new NullPointerException("pattern");
    }
    
    StringBuilder currentLiteral = new StringBuilder(32);
    
    int patternLength = pattern.length();
    ParserState state = ParserState.LITERAL_STATE;
    
    int i = 0;
    FormattingInfo formattingInfo = FormattingInfo.getDefault();
    
    while (i < patternLength) {
      char c = pattern.charAt(i++);
      
      switch (1.$SwitchMap$org$apache$logging$log4j$core$pattern$PatternParser$ParserState[state.ordinal()])
      {

      case 1: 
        if (i == patternLength) {
          currentLiteral.append(c);



        }
        else if (c == '%')
        {
          switch (pattern.charAt(i)) {
          case '%': 
            currentLiteral.append(c);
            i++;
            
            break;
          

          default: 
            if (currentLiteral.length() != 0) {
              patternConverters.add(new LiteralPatternConverter(config, currentLiteral.toString()));
              
              formattingInfos.add(FormattingInfo.getDefault());
            }
            
            currentLiteral.setLength(0);
            currentLiteral.append(c);
            state = ParserState.CONVERTER_STATE;
            formattingInfo = FormattingInfo.getDefault(); break;
          }
        } else {
          currentLiteral.append(c);
        }
        
        break;
      
      case 2: 
        currentLiteral.append(c);
        
        switch (c) {
        case '-': 
          formattingInfo = new FormattingInfo(true, formattingInfo.getMinLength(), formattingInfo.getMaxLength());
          

          break;
        
        case '.': 
          state = ParserState.DOT_STATE;
          break;
        

        default: 
          if ((c >= '0') && (c <= '9')) {
            formattingInfo = new FormattingInfo(formattingInfo.isLeftAligned(), c - '0', formattingInfo.getMaxLength());
            
            state = ParserState.MIN_STATE;
          } else {
            i = finalizeConverter(c, pattern, i, currentLiteral, formattingInfo, converterRules, patternConverters, formattingInfos);
            


            state = ParserState.LITERAL_STATE;
            formattingInfo = FormattingInfo.getDefault();
            currentLiteral.setLength(0);
          }
          break;
        }
        break;
      
      case 3: 
        currentLiteral.append(c);
        
        if ((c >= '0') && (c <= '9'))
        {
          formattingInfo = new FormattingInfo(formattingInfo.isLeftAligned(), formattingInfo.getMinLength() * 10 + c - 48, formattingInfo.getMaxLength());

        }
        else if (c == '.') {
          state = ParserState.DOT_STATE;
        } else {
          i = finalizeConverter(c, pattern, i, currentLiteral, formattingInfo, converterRules, patternConverters, formattingInfos);
          
          state = ParserState.LITERAL_STATE;
          formattingInfo = FormattingInfo.getDefault();
          currentLiteral.setLength(0);
        }
        
        break;
      
      case 4: 
        currentLiteral.append(c);
        
        if ((c >= '0') && (c <= '9')) {
          formattingInfo = new FormattingInfo(formattingInfo.isLeftAligned(), formattingInfo.getMinLength(), c - '0');
          
          state = ParserState.MAX_STATE;
        } else {
          LOGGER.error("Error occurred in position " + i + ".\n Was expecting digit, instead got char \"" + c + "\".");
          

          state = ParserState.LITERAL_STATE;
        }
        
        break;
      
      case 5: 
        currentLiteral.append(c);
        
        if ((c >= '0') && (c <= '9'))
        {
          formattingInfo = new FormattingInfo(formattingInfo.isLeftAligned(), formattingInfo.getMinLength(), formattingInfo.getMaxLength() * 10 + c - 48);
        }
        else
        {
          i = finalizeConverter(c, pattern, i, currentLiteral, formattingInfo, converterRules, patternConverters, formattingInfos);
          
          state = ParserState.LITERAL_STATE;
          formattingInfo = FormattingInfo.getDefault();
          currentLiteral.setLength(0);
        }
        
        break;
      }
      
    }
    
    if (currentLiteral.length() != 0) {
      patternConverters.add(new LiteralPatternConverter(config, currentLiteral.toString()));
      formattingInfos.add(FormattingInfo.getDefault());
    }
  }
  











  private PatternConverter createConverter(String converterId, StringBuilder currentLiteral, Map<String, Class<PatternConverter>> rules, List<String> options)
  {
    String converterName = converterId;
    Class<PatternConverter> converterClass = null;
    
    for (int i = converterId.length(); (i > 0) && (converterClass == null); i--) {
      converterName = converterName.substring(0, i);
      
      if ((converterClass == null) && (rules != null)) {
        converterClass = (Class)rules.get(converterName);
      }
    }
    
    if (converterClass == null) {
      LOGGER.error("Unrecognized format specifier [" + converterId + "]");
      return null;
    }
    


    Method[] methods = converterClass.getDeclaredMethods();
    Method newInstanceMethod = null;
    for (Method method : methods) {
      if ((Modifier.isStatic(method.getModifiers())) && (method.getDeclaringClass().equals(converterClass)) && (method.getName().equals("newInstance")))
      {
        if (newInstanceMethod == null) {
          newInstanceMethod = method;
        } else if (method.getReturnType().equals(newInstanceMethod.getReturnType())) {
          LOGGER.error("Class " + converterClass + " cannot contain multiple static newInstance methods");
          return null;
        }
      }
    }
    if (newInstanceMethod == null) {
      LOGGER.error("Class " + converterClass + " does not contain a static newInstance method");
      return null;
    }
    
    Class<?>[] parmTypes = newInstanceMethod.getParameterTypes();
    Object[] parms = parmTypes.length > 0 ? new Object[parmTypes.length] : null;
    
    if (parms != null) {
      int i = 0;
      boolean errors = false;
      for (Class<?> clazz : parmTypes) {
        if ((clazz.isArray()) && (clazz.getName().equals("[Ljava.lang.String;"))) {
          String[] optionsArray = (String[])options.toArray(new String[options.size()]);
          parms[i] = optionsArray;
        } else if (clazz.isAssignableFrom(Configuration.class)) {
          parms[i] = config;
        } else {
          LOGGER.error("Unknown parameter type " + clazz.getName() + " for static newInstance method of " + converterClass.getName());
          
          errors = true;
        }
        i++;
      }
      if (errors) {
        return null;
      }
    }
    try
    {
      Object newObj = newInstanceMethod.invoke(null, parms);
      
      if ((newObj instanceof PatternConverter)) {
        currentLiteral.delete(0, currentLiteral.length() - (converterId.length() - converterName.length()));
        

        return (PatternConverter)newObj;
      }
      LOGGER.warn("Class " + converterClass.getName() + " does not extend PatternConverter.");
    }
    catch (Exception ex) {
      LOGGER.error("Error creating converter for " + converterId, ex);
    }
    
    return null;
  }
  















  private int finalizeConverter(char c, String pattern, int i, StringBuilder currentLiteral, FormattingInfo formattingInfo, Map<String, Class<PatternConverter>> rules, List<PatternConverter> patternConverters, List<FormattingInfo> formattingInfos)
  {
    StringBuilder convBuf = new StringBuilder();
    i = extractConverter(c, pattern, i, convBuf, currentLiteral);
    
    String converterId = convBuf.toString();
    
    List<String> options = new ArrayList();
    i = extractOptions(pattern, i, options);
    
    PatternConverter pc = createConverter(converterId, currentLiteral, rules, options);
    
    if (pc == null) {
      StringBuilder msg;
      StringBuilder msg;
      if (Strings.isEmpty(converterId)) {
        msg = new StringBuilder("Empty conversion specifier starting at position ");
      } else {
        msg = new StringBuilder("Unrecognized conversion specifier [");
        msg.append(converterId);
        msg.append("] starting at position ");
      }
      
      msg.append(Integer.toString(i));
      msg.append(" in conversion pattern.");
      
      LOGGER.error(msg.toString());
      
      patternConverters.add(new LiteralPatternConverter(config, currentLiteral.toString()));
      formattingInfos.add(FormattingInfo.getDefault());
    } else {
      patternConverters.add(pc);
      formattingInfos.add(formattingInfo);
      
      if (currentLiteral.length() > 0) {
        patternConverters.add(new LiteralPatternConverter(config, currentLiteral.toString()));
        formattingInfos.add(FormattingInfo.getDefault());
      }
    }
    
    currentLiteral.setLength(0);
    
    return i;
  }
}
