package org.apache.logging.log4j.status;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.Marker;
import org.apache.logging.log4j.message.Message;
import org.apache.logging.log4j.simple.SimpleLogger;
import org.apache.logging.log4j.spi.AbstractLogger;
import org.apache.logging.log4j.util.PropertiesUtil;

























public final class StatusLogger
  extends AbstractLogger
{
  public static final String MAX_STATUS_ENTRIES = "log4j2.status.entries";
  private static final String NOT_AVAIL = "?";
  private static final PropertiesUtil PROPS = new PropertiesUtil("log4j2.StatusLogger.properties");
  
  private static final int MAX_ENTRIES = PROPS.getIntegerProperty("log4j2.status.entries", 200);
  
  private static final String DEFAULT_STATUS_LEVEL = PROPS.getStringProperty("log4j2.StatusLogger.level");
  


  private static final StatusLogger STATUS_LOGGER = new StatusLogger();
  
  private final SimpleLogger logger;
  
  private final CopyOnWriteArrayList<StatusListener> listeners = new CopyOnWriteArrayList();
  private final ReentrantReadWriteLock listenersLock = new ReentrantReadWriteLock();
  
  private final Queue<StatusData> messages = new BoundedQueue(MAX_ENTRIES);
  private final ReentrantLock msgLock = new ReentrantLock();
  private int listenersLevel;
  
  private StatusLogger()
  {
    logger = new SimpleLogger("StatusLogger", Level.ERROR, false, true, false, false, "", null, PROPS, System.err);
    
    listenersLevel = Level.toLevel(DEFAULT_STATUS_LEVEL, Level.WARN).intLevel();
  }
  



  public static StatusLogger getLogger()
  {
    return STATUS_LOGGER;
  }
  
  public Level getLevel() {
    return logger.getLevel();
  }
  
  public void setLevel(Level level) {
    logger.setLevel(level);
  }
  



  public void registerListener(StatusListener listener)
  {
    listenersLock.writeLock().lock();
    try {
      listeners.add(listener);
      Level lvl = listener.getStatusLevel();
      if (listenersLevel < lvl.intLevel()) {
        listenersLevel = lvl.intLevel();
      }
    } finally {
      listenersLock.writeLock().unlock();
    }
  }
  



  public void removeListener(StatusListener listener)
  {
    listenersLock.writeLock().lock();
    try {
      listeners.remove(listener);
      int lowest = Level.toLevel(DEFAULT_STATUS_LEVEL, Level.WARN).intLevel();
      for (StatusListener l : listeners) {
        int level = l.getStatusLevel().intLevel();
        if (lowest < level) {
          lowest = level;
        }
      }
      listenersLevel = lowest;
    } finally {
      listenersLock.writeLock().unlock();
    }
  }
  



  public Iterator<StatusListener> getListeners()
  {
    return listeners.iterator();
  }
  


  public void reset()
  {
    listeners.clear();
    clear();
  }
  



  public List<StatusData> getStatusData()
  {
    msgLock.lock();
    try {
      return new ArrayList(messages);
    } finally {
      msgLock.unlock();
    }
  }
  


  public void clear()
  {
    msgLock.lock();
    try {
      messages.clear();
    } finally {
      msgLock.unlock();
    }
  }
  









  public void log(Marker marker, String fqcn, Level level, Message msg, Throwable t)
  {
    StackTraceElement element = null;
    if (fqcn != null) {
      element = getStackTraceElement(fqcn, Thread.currentThread().getStackTrace());
    }
    StatusData data = new StatusData(element, level, msg, t);
    msgLock.lock();
    try {
      messages.add(data);
    } finally {
      msgLock.unlock();
    }
    if (listeners.size() > 0) {
      for (StatusListener listener : listeners) {
        if (data.getLevel().isAtLeastAsSpecificAs(listener.getStatusLevel())) {
          listener.log(data);
        }
      }
    } else {
      logger.log(marker, fqcn, level, msg, t);
    }
  }
  
  private StackTraceElement getStackTraceElement(String fqcn, StackTraceElement[] stackTrace) {
    if (fqcn == null) {
      return null;
    }
    boolean next = false;
    for (StackTraceElement element : stackTrace) {
      if (next) {
        return element;
      }
      String className = element.getClassName();
      if (fqcn.equals(className))
        next = true; else {
        if ("?".equals(className))
          break;
      }
    }
    return null;
  }
  
  protected boolean isEnabled(Level level, Marker marker, String data)
  {
    return isEnabled(level, marker);
  }
  
  protected boolean isEnabled(Level level, Marker marker, String data, Throwable t)
  {
    return isEnabled(level, marker);
  }
  
  protected boolean isEnabled(Level level, Marker marker, String data, Object... p1)
  {
    return isEnabled(level, marker);
  }
  
  protected boolean isEnabled(Level level, Marker marker, Object data, Throwable t)
  {
    return isEnabled(level, marker);
  }
  
  protected boolean isEnabled(Level level, Marker marker, Message data, Throwable t)
  {
    return isEnabled(level, marker);
  }
  
  public boolean isEnabled(Level level, Marker marker)
  {
    if (listeners.size() > 0) {
      return listenersLevel >= level.intLevel();
    }
    switch (1.$SwitchMap$org$apache$logging$log4j$Level[level.ordinal()]) {
    case 1: 
      return logger.isFatalEnabled(marker);
    case 2: 
      return logger.isTraceEnabled(marker);
    case 3: 
      return logger.isDebugEnabled(marker);
    case 4: 
      return logger.isInfoEnabled(marker);
    case 5: 
      return logger.isWarnEnabled(marker);
    case 6: 
      return logger.isErrorEnabled(marker);
    }
    return false;
  }
  


  private class BoundedQueue<E>
    extends ConcurrentLinkedQueue<E>
  {
    private static final long serialVersionUID = -3945953719763255337L;
    
    private final int size;
    

    public BoundedQueue(int size)
    {
      this.size = size;
    }
    
    public boolean add(E object)
    {
      while (messages.size() > size) {
        messages.poll();
      }
      return super.add(object);
    }
  }
}
