package org.bukkit;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableList.Builder;
import java.util.List;
import java.util.Map;
import org.apache.commons.lang.Validate;
import org.bukkit.configuration.serialization.ConfigurationSerializable;

@org.bukkit.configuration.serialization.SerializableAs("Firework")
public final class FireworkEffect implements ConfigurationSerializable
{
  private static final String FLICKER = "flicker";
  private static final String TRAIL = "trail";
  private static final String COLORS = "colors";
  private static final String FADE_COLORS = "fade-colors";
  private static final String TYPE = "type";
  private final boolean flicker;
  private final boolean trail;
  private final ImmutableList<Color> colors;
  private final ImmutableList<Color> fadeColors;
  private final Type type;
  
  public static enum Type
  {
    BALL, 
    


    BALL_LARGE, 
    


    STAR, 
    


    BURST, 
    


    CREEPER;
    


    private Type() {}
  }
  

  public static Builder builder()
  {
    return new Builder();
  }
  




  public static final class Builder
  {
    boolean flicker = false;
    boolean trail = false;
    final ImmutableList.Builder<Color> colors = ImmutableList.builder();
    ImmutableList.Builder<Color> fadeColors = null;
    FireworkEffect.Type type = FireworkEffect.Type.BALL;
    







    public Builder with(FireworkEffect.Type type)
      throws IllegalArgumentException
    {
      Validate.notNull(type, "Cannot have null type");
      this.type = type;
      return this;
    }
    




    public Builder withFlicker()
    {
      flicker = true;
      return this;
    }
    





    public Builder flicker(boolean flicker)
    {
      this.flicker = flicker;
      return this;
    }
    




    public Builder withTrail()
    {
      trail = true;
      return this;
    }
    





    public Builder trail(boolean trail)
    {
      this.trail = trail;
      return this;
    }
    





    public Builder withColor(Color color)
      throws IllegalArgumentException
    {
      Validate.notNull(color, "Cannot have null color");
      
      colors.add(color);
      
      return this;
    }
    







    public Builder withColor(Color... colors)
      throws IllegalArgumentException
    {
      Validate.notNull(colors, "Cannot have null colors");
      if (colors.length == 0) {
        return this;
      }
      
      ImmutableList.Builder<Color> list = this.colors;
      for (Color color : colors) {
        Validate.notNull(color, "Color cannot be null");
        list.add(color);
      }
      
      return this;
    }
    








    public Builder withColor(Iterable<?> colors)
      throws IllegalArgumentException
    {
      Validate.notNull(colors, "Cannot have null colors");
      
      ImmutableList.Builder<Color> list = this.colors;
      for (Object color : colors) {
        if (!(color instanceof Color)) {
          throw new IllegalArgumentException(color + " is not a Color in " + colors);
        }
        list.add((Color)color);
      }
      
      return this;
    }
    







    public Builder withFade(Color color)
      throws IllegalArgumentException
    {
      Validate.notNull(color, "Cannot have null color");
      
      if (fadeColors == null) {
        fadeColors = ImmutableList.builder();
      }
      
      fadeColors.add(color);
      
      return this;
    }
    







    public Builder withFade(Color... colors)
      throws IllegalArgumentException
    {
      Validate.notNull(colors, "Cannot have null colors");
      if (colors.length == 0) {
        return this;
      }
      
      ImmutableList.Builder<Color> list = fadeColors;
      if (list == null) {
        list = this.fadeColors = ImmutableList.builder();
      }
      
      for (Color color : colors) {
        Validate.notNull(color, "Color cannot be null");
        list.add(color);
      }
      
      return this;
    }
    








    public Builder withFade(Iterable<?> colors)
      throws IllegalArgumentException
    {
      Validate.notNull(colors, "Cannot have null colors");
      
      ImmutableList.Builder<Color> list = fadeColors;
      if (list == null) {
        list = this.fadeColors = ImmutableList.builder();
      }
      
      for (Object color : colors) {
        if (!(color instanceof Color)) {
          throw new IllegalArgumentException(color + " is not a Color in " + colors);
        }
        list.add((Color)color);
      }
      
      return this;
    }
    







    public FireworkEffect build()
    {
      return new FireworkEffect(flicker, trail, colors.build(), fadeColors == null ? ImmutableList.of() : fadeColors.build(), type);
    }
  }
  

















  private String string = null;
  
  FireworkEffect(boolean flicker, boolean trail, ImmutableList<Color> colors, ImmutableList<Color> fadeColors, Type type) {
    if (colors.isEmpty()) {
      throw new IllegalStateException("Cannot make FireworkEffect without any color");
    }
    this.flicker = flicker;
    this.trail = trail;
    this.colors = colors;
    this.fadeColors = fadeColors;
    this.type = type;
  }
  




  public boolean hasFlicker()
  {
    return flicker;
  }
  




  public boolean hasTrail()
  {
    return trail;
  }
  




  public List<Color> getColors()
  {
    return colors;
  }
  




  public List<Color> getFadeColors()
  {
    return fadeColors;
  }
  




  public Type getType()
  {
    return type;
  }
  


  public static ConfigurationSerializable deserialize(Map<String, Object> map)
  {
    Type type = Type.valueOf((String)map.get("type"));
    if (type == null) {
      throw new IllegalArgumentException(map.get("type") + " is not a valid Type");
    }
    
    return builder().flicker(((Boolean)map.get("flicker")).booleanValue()).trail(((Boolean)map.get("trail")).booleanValue()).withColor((Iterable)map.get("colors")).withFade((Iterable)map.get("fade-colors")).with(type).build();
  }
  





  public Map<String, Object> serialize()
  {
    return com.google.common.collect.ImmutableMap.of("flicker", Boolean.valueOf(flicker), "trail", Boolean.valueOf(trail), "colors", colors, "fade-colors", fadeColors, "type", type.name());
  }
  






  public String toString()
  {
    String string = this.string;
    if (string == null) {
      return this.string = "FireworkEffect:" + serialize();
    }
    return string;
  }
  



  public int hashCode()
  {
    int PRIME = 31;int TRUE = 1231;int FALSE = 1237;
    int hash = 1;
    hash = hash * 31 + (flicker ? 1231 : 1237);
    hash = hash * 31 + (trail ? 1231 : 1237);
    hash = hash * 31 + type.hashCode();
    hash = hash * 31 + colors.hashCode();
    hash = hash * 31 + fadeColors.hashCode();
    return hash;
  }
  
  public boolean equals(Object obj)
  {
    if (this == obj) {
      return true;
    }
    
    if (!(obj instanceof FireworkEffect)) {
      return false;
    }
    
    FireworkEffect that = (FireworkEffect)obj;
    return (flicker == flicker) && (trail == trail) && (type == type) && (colors.equals(colors)) && (fadeColors.equals(fadeColors));
  }
}
