package org.bukkit.command;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Pattern;
import org.apache.commons.lang.Validate;
import org.bukkit.Server;
import org.bukkit.command.defaults.AchievementCommand;
import org.bukkit.command.defaults.SaveOffCommand;
import org.bukkit.command.defaults.SetIdleTimeoutCommand;
import org.bukkit.command.defaults.TimeCommand;
import org.bukkit.command.defaults.VanillaCommand;
import org.bukkit.util.StringUtil;

public class SimpleCommandMap implements CommandMap
{
  private static final Pattern PATTERN_ON_SPACE = Pattern.compile(" ", 16);
  protected final Map<String, Command> knownCommands = new java.util.HashMap();
  protected final Set<String> aliases = new HashSet();
  private final Server server;
  protected static final Set<VanillaCommand> fallbackCommands = new HashSet();
  
  static {
    fallbackCommands.add(new org.bukkit.command.defaults.ListCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.OpCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.DeopCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.BanIpCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.PardonIpCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.BanCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.PardonCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.KickCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.TeleportCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.GiveCommand());
    fallbackCommands.add(new TimeCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.SayCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.WhitelistCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.TellCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.MeCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.KillCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.GameModeCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.HelpCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.ExpCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.ToggleDownfallCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.BanListCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.DefaultGameModeCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.SeedCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.DifficultyCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.WeatherCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.SpawnpointCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.ClearCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.GameRuleCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.EnchantCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.TestForCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.EffectCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.ScoreboardCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.PlaySoundCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.SpreadPlayersCommand());
    fallbackCommands.add(new org.bukkit.command.defaults.SetWorldSpawnCommand());
    fallbackCommands.add(new SetIdleTimeoutCommand());
    fallbackCommands.add(new AchievementCommand());
  }
  
  public SimpleCommandMap(Server server) {
    this.server = server;
    setDefaultCommands(server);
  }
  
  private void setDefaultCommands(Server server) {
    register("bukkit", new org.bukkit.command.defaults.SaveCommand());
    register("bukkit", new org.bukkit.command.defaults.SaveOnCommand());
    register("bukkit", new SaveOffCommand());
    register("bukkit", new org.bukkit.command.defaults.StopCommand());
    register("bukkit", new org.bukkit.command.defaults.VersionCommand("version"));
    register("bukkit", new org.bukkit.command.defaults.ReloadCommand("reload"));
    register("bukkit", new org.bukkit.command.defaults.PluginsCommand("plugins"));
    register("bukkit", new org.bukkit.command.defaults.TimingsCommand("timings"));
  }
  


  public void registerAll(String fallbackPrefix, List<Command> commands)
  {
    if (commands != null) {
      for (Command c : commands) {
        register(fallbackPrefix, c);
      }
    }
  }
  


  public boolean register(String fallbackPrefix, Command command)
  {
    return register(command.getName(), fallbackPrefix, command);
  }
  


  public boolean register(String label, String fallbackPrefix, Command command)
  {
    boolean registeredPassedLabel = register(label, fallbackPrefix, command, false);
    
    Iterator<String> iterator = command.getAliases().iterator();
    while (iterator.hasNext()) {
      if (!register((String)iterator.next(), fallbackPrefix, command, true)) {
        iterator.remove();
      }
    }
    

    command.register(this);
    
    return registeredPassedLabel;
  }
  













  private synchronized boolean register(String label, String fallbackPrefix, Command command, boolean isAlias)
  {
    String lowerLabel = label.trim().toLowerCase();
    
    if ((isAlias) && (knownCommands.containsKey(lowerLabel)))
    {

      return false;
    }
    
    String lowerPrefix = fallbackPrefix.trim().toLowerCase();
    boolean registerdPassedLabel = true;
    

    while ((knownCommands.containsKey(lowerLabel)) && (!aliases.contains(lowerLabel))) {
      lowerLabel = lowerPrefix + ":" + lowerLabel;
      registerdPassedLabel = false;
    }
    
    if (isAlias) {
      aliases.add(lowerLabel);
    }
    else {
      aliases.remove(lowerLabel);
      command.setLabel(lowerLabel);
    }
    knownCommands.put(lowerLabel, command);
    
    return registerdPassedLabel;
  }
  
  protected Command getFallback(String label) {
    for (VanillaCommand cmd : fallbackCommands) {
      if (cmd.matches(label)) {
        return cmd;
      }
    }
    
    return null;
  }
  
  public Set<VanillaCommand> getFallbackCommands() {
    return Collections.unmodifiableSet(fallbackCommands);
  }
  

  public boolean dispatch(CommandSender sender, String commandLine)
    throws CommandException
  {
    String[] args = PATTERN_ON_SPACE.split(commandLine);
    
    if (args.length == 0) {
      return false;
    }
    
    String sentCommandLabel = args[0].toLowerCase();
    Command target = getCommand(sentCommandLabel);
    
    if (target == null) {
      return false;
    }
    
    try
    {
      target.execute(sender, sentCommandLabel, (String[])org.bukkit.util.Java15Compat.Arrays_copyOfRange(args, 1, args.length));
    } catch (CommandException ex) {
      throw ex;
    } catch (Throwable ex) {
      throw new CommandException("Unhandled exception executing '" + commandLine + "' in " + target, ex);
    }
    

    return true;
  }
  
  public synchronized void clearCommands() {
    for (Map.Entry<String, Command> entry : knownCommands.entrySet()) {
      ((Command)entry.getValue()).unregister(this);
    }
    knownCommands.clear();
    aliases.clear();
    setDefaultCommands(server);
  }
  
  public Command getCommand(String name) {
    Command target = (Command)knownCommands.get(name.toLowerCase());
    if (target == null) {
      target = getFallback(name);
    }
    return target;
  }
  
  public List<String> tabComplete(CommandSender sender, String cmdLine) {
    Validate.notNull(sender, "Sender cannot be null");
    Validate.notNull(cmdLine, "Command line cannot null");
    
    int spaceIndex = cmdLine.indexOf(' ');
    
    if (spaceIndex == -1) {
      ArrayList<String> completions = new ArrayList();
      Map<String, Command> knownCommands = this.knownCommands;
      
      String prefix = (sender instanceof org.bukkit.entity.Player) ? "/" : "";
      
      for (VanillaCommand command : fallbackCommands) {
        String name = command.getName();
        
        if ((command.testPermissionSilent(sender)) && 
        

          (!knownCommands.containsKey(name)) && 
          



          (StringUtil.startsWithIgnoreCase(name, cmdLine)))
        {


          completions.add(prefix + name);
        }
      }
      for (Map.Entry<String, Command> commandEntry : knownCommands.entrySet()) {
        Command command = (Command)commandEntry.getValue();
        
        if (command.testPermissionSilent(sender))
        {


          String name = (String)commandEntry.getKey();
          
          if (StringUtil.startsWithIgnoreCase(name, cmdLine)) {
            completions.add(prefix + name);
          }
        }
      }
      Collections.sort(completions, String.CASE_INSENSITIVE_ORDER);
      return completions;
    }
    
    String commandName = cmdLine.substring(0, spaceIndex);
    Command target = getCommand(commandName);
    
    if (target == null) {
      return null;
    }
    
    if (!target.testPermissionSilent(sender)) {
      return null;
    }
    
    String argLine = cmdLine.substring(spaceIndex + 1, cmdLine.length());
    String[] args = PATTERN_ON_SPACE.split(argLine, -1);
    try
    {
      return target.tabComplete(sender, commandName, args);
    } catch (CommandException ex) {
      throw ex;
    } catch (Throwable ex) {
      throw new CommandException("Unhandled exception executing tab-completer for '" + cmdLine + "' in " + target, ex);
    }
  }
  
  public java.util.Collection<Command> getCommands() {
    return knownCommands.values();
  }
  
  public void registerServerAliases() {
    Map<String, String[]> values = server.getCommandAliases();
    
    for (String alias : values.keySet()) {
      String[] targetNames = (String[])values.get(alias);
      List<Command> targets = new ArrayList();
      StringBuilder bad = new StringBuilder();
      
      for (String name : targetNames) {
        Command command = getCommand(name);
        
        if (command == null) {
          if (bad.length() > 0) {
            bad.append(", ");
          }
          bad.append(name);
        } else {
          targets.add(command);
        }
      }
      


      if (targets.size() > 0) {
        knownCommands.put(alias.toLowerCase(), new MultipleCommandAlias(alias.toLowerCase(), (Command[])targets.toArray(new Command[0])));
      } else {
        knownCommands.remove(alias.toLowerCase());
      }
      
      if (bad.length() > 0) {
        server.getLogger().warning("The following command(s) could not be aliased under '" + alias + "' because they do not exist: " + bad);
      }
    }
  }
}
