package org.bukkit.craftbukkit.v1_7_R1.block;

import java.util.List;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.block.Block;
import org.bukkit.block.BlockState;
import org.bukkit.craftbukkit.v1_7_R1.CraftChunk;
import org.bukkit.craftbukkit.v1_7_R1.CraftWorld;
import org.bukkit.craftbukkit.v1_7_R1.metadata.BlockMetadataStore;
import org.bukkit.material.MaterialData;
import org.bukkit.metadata.MetadataValue;
import org.bukkit.plugin.Plugin;

public class CraftBlockState implements BlockState
{
  private final CraftWorld world;
  private final CraftChunk chunk;
  private final int x;
  private final int y;
  private final int z;
  protected int type;
  protected MaterialData data;
  protected final byte light;
  
  public CraftBlockState(Block block)
  {
    world = ((CraftWorld)block.getWorld());
    x = block.getX();
    y = block.getY();
    z = block.getZ();
    type = block.getTypeId();
    light = block.getLightLevel();
    chunk = ((CraftChunk)block.getChunk());
    
    createData(block.getData());
  }
  
  public static CraftBlockState getBlockState(net.minecraft.server.v1_7_R1.World world, int x, int y, int z) {
    return new CraftBlockState(world.getWorld().getBlockAt(x, y, z));
  }
  
  public org.bukkit.World getWorld() {
    return world;
  }
  
  public int getX() {
    return x;
  }
  
  public int getY() {
    return y;
  }
  
  public int getZ() {
    return z;
  }
  
  public org.bukkit.Chunk getChunk() {
    return chunk;
  }
  
  public void setData(MaterialData data) {
    Material mat = getType();
    
    if ((mat == null) || (mat.getData() == null)) {
      this.data = data;
    }
    else if ((data.getClass() == mat.getData()) || (data.getClass() == MaterialData.class)) {
      this.data = data;
    } else {
      throw new IllegalArgumentException("Provided data is not of type " + mat.getData().getName() + ", found " + data.getClass().getName());
    }
  }
  

  public MaterialData getData()
  {
    return data;
  }
  
  public void setType(Material type) {
    setTypeId(type.getId());
  }
  
  public boolean setTypeId(int type) {
    if (this.type != type) {
      this.type = type;
      
      createData((byte)0);
    }
    return true;
  }
  
  public Material getType() {
    return Material.getMaterial(getTypeId());
  }
  
  public int getTypeId() {
    return type;
  }
  
  public byte getLightLevel() {
    return light;
  }
  
  public Block getBlock() {
    return world.getBlockAt(x, y, z);
  }
  
  public boolean update() {
    return update(false);
  }
  
  public boolean update(boolean force) {
    return update(force, true);
  }
  
  public boolean update(boolean force, boolean applyPhysics) {
    Block block = getBlock();
    
    if (block.getType() != getType()) {
      if (force) {
        block.setTypeId(getTypeId(), applyPhysics);
      } else {
        return false;
      }
    }
    
    block.setData(getRawData(), applyPhysics);
    world.getHandle().notify(x, y, z);
    
    return true;
  }
  
  private void createData(byte data) {
    Material mat = getType();
    if ((mat == null) || (mat.getData() == null)) {
      this.data = new MaterialData(type, data);
    } else {
      this.data = mat.getNewData(data);
    }
  }
  
  public byte getRawData() {
    return data.getData();
  }
  
  public Location getLocation() {
    return new Location(world, x, y, z);
  }
  
  public Location getLocation(Location loc) {
    if (loc != null) {
      loc.setWorld(world);
      loc.setX(x);
      loc.setY(y);
      loc.setZ(z);
      loc.setYaw(0.0F);
      loc.setPitch(0.0F);
    }
    
    return loc;
  }
  
  public void setRawData(byte data) {
    this.data.setData(data);
  }
  
  public boolean equals(Object obj)
  {
    if (obj == null) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }
    CraftBlockState other = (CraftBlockState)obj;
    if ((world != world) && ((world == null) || (!world.equals(world)))) {
      return false;
    }
    if (x != x) {
      return false;
    }
    if (y != y) {
      return false;
    }
    if (z != z) {
      return false;
    }
    if (type != type) {
      return false;
    }
    if ((data != data) && ((data == null) || (!data.equals(data)))) {
      return false;
    }
    return true;
  }
  
  public int hashCode()
  {
    int hash = 7;
    hash = 73 * hash + (world != null ? world.hashCode() : 0);
    hash = 73 * hash + x;
    hash = 73 * hash + y;
    hash = 73 * hash + z;
    hash = 73 * hash + type;
    hash = 73 * hash + (data != null ? data.hashCode() : 0);
    return hash;
  }
  
  public void setMetadata(String metadataKey, MetadataValue newMetadataValue) {
    chunk.getCraftWorld().getBlockMetadata().setMetadata(getBlock(), metadataKey, newMetadataValue);
  }
  
  public List<MetadataValue> getMetadata(String metadataKey) {
    return chunk.getCraftWorld().getBlockMetadata().getMetadata(getBlock(), metadataKey);
  }
  
  public boolean hasMetadata(String metadataKey) {
    return chunk.getCraftWorld().getBlockMetadata().hasMetadata(getBlock(), metadataKey);
  }
  
  public void removeMetadata(String metadataKey, Plugin owningPlugin) {
    chunk.getCraftWorld().getBlockMetadata().removeMetadata(getBlock(), metadataKey, owningPlugin);
  }
}
