package org.bukkit.craftbukkit.v1_7_R1.inventory;

import com.google.common.base.Strings;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap.Builder;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import net.minecraft.server.v1_7_R1.NBTTagCompound;
import net.minecraft.server.v1_7_R1.NBTTagList;
import org.apache.commons.lang.Validate;
import org.bukkit.Material;
import org.bukkit.configuration.serialization.DelegateDeserialization;
import org.bukkit.inventory.meta.BookMeta;


@DelegateDeserialization(CraftMetaItem.SerializableMeta.class)
class CraftMetaBook
  extends CraftMetaItem
  implements BookMeta
{
  static final CraftMetaItem.ItemMetaKey BOOK_TITLE = new CraftMetaItem.ItemMetaKey("title");
  static final CraftMetaItem.ItemMetaKey BOOK_AUTHOR = new CraftMetaItem.ItemMetaKey("author");
  static final CraftMetaItem.ItemMetaKey BOOK_PAGES = new CraftMetaItem.ItemMetaKey("pages");
  
  static final int MAX_PAGE_LENGTH = 256;
  static final int MAX_TITLE_LENGTH = 65535;
  private String title;
  private String author;
  private List<String> pages = new ArrayList();
  
  CraftMetaBook(CraftMetaItem meta) {
    super(meta);
    
    if (!(meta instanceof CraftMetaBook)) {
      return;
    }
    CraftMetaBook bookMeta = (CraftMetaBook)meta;
    title = title;
    author = author;
    pages.addAll(pages);
  }
  
  CraftMetaBook(NBTTagCompound tag) {
    super(tag);
    
    if (tag.hasKey(BOOK_TITLENBT)) {
      title = tag.getString(BOOK_TITLENBT);
    }
    
    if (tag.hasKey(BOOK_AUTHORNBT)) {
      author = tag.getString(BOOK_AUTHORNBT);
    }
    
    if (tag.hasKey(BOOK_PAGESNBT)) {
      NBTTagList pages = tag.getList(BOOK_PAGESNBT, 8);
      String[] pageArray = new String[pages.size()];
      
      for (int i = 0; i < pages.size(); i++) {
        String page = pages.f(i);
        pageArray[i] = page;
      }
      
      addPage(pageArray);
    }
  }
  
  CraftMetaBook(Map<String, Object> map) {
    super(map);
    
    setAuthor(CraftMetaItem.SerializableMeta.getString(map, BOOK_AUTHORBUKKIT, true));
    
    setTitle(CraftMetaItem.SerializableMeta.getString(map, BOOK_TITLEBUKKIT, true));
    
    Iterable<?> pages = (Iterable)CraftMetaItem.SerializableMeta.getObject(Iterable.class, map, BOOK_PAGESBUKKIT, true);
    CraftMetaItem.safelyAdd(pages, this.pages, 256);
  }
  
  void applyToItem(NBTTagCompound itemData)
  {
    super.applyToItem(itemData);
    
    if (hasTitle()) {
      itemData.setString(BOOK_TITLENBT, title);
    }
    
    if (hasAuthor()) {
      itemData.setString(BOOK_AUTHORNBT, author);
    }
    
    if (hasPages()) {
      itemData.set(BOOK_PAGESNBT, createStringList(pages));
    }
  }
  
  boolean isEmpty()
  {
    return (super.isEmpty()) && (isBookEmpty());
  }
  
  boolean isBookEmpty() {
    return (!hasPages()) && (!hasAuthor()) && (!hasTitle());
  }
  
  boolean applicableTo(Material type)
  {
    switch (1.$SwitchMap$org$bukkit$Material[type.ordinal()]) {
    case 1: 
    case 2: 
      return true;
    }
    return false;
  }
  
  public boolean hasAuthor()
  {
    return !Strings.isNullOrEmpty(author);
  }
  
  public boolean hasTitle() {
    return !Strings.isNullOrEmpty(title);
  }
  
  public boolean hasPages() {
    return !pages.isEmpty();
  }
  
  public String getTitle() {
    return title;
  }
  
  public boolean setTitle(String title) {
    if (title == null) {
      this.title = null;
      return true; }
    if (title.length() > 65535) {
      return false;
    }
    
    this.title = title;
    return true;
  }
  
  public String getAuthor() {
    return author;
  }
  
  public void setAuthor(String author) {
    this.author = author;
  }
  
  public String getPage(int page) {
    Validate.isTrue(isValidPage(page), "Invalid page number");
    return (String)pages.get(page - 1);
  }
  
  public void setPage(int page, String text) {
    if (!isValidPage(page)) {
      throw new IllegalArgumentException("Invalid page number " + page + "/" + pages.size());
    }
    
    pages.set(page - 1, text.length() > 256 ? text.substring(0, 256) : text == null ? "" : text);
  }
  
  public void setPages(String... pages) {
    this.pages.clear();
    
    addPage(pages);
  }
  
  public void addPage(String... pages) {
    for (String page : pages) {
      if (page == null) {
        page = "";
      } else if (page.length() > 256) {
        page = page.substring(0, 256);
      }
      
      this.pages.add(page);
    }
  }
  
  public int getPageCount() {
    return pages.size();
  }
  
  public List<String> getPages() {
    return ImmutableList.copyOf(pages);
  }
  
  public void setPages(List<String> pages) {
    this.pages.clear();
    CraftMetaItem.safelyAdd(pages, this.pages, 256);
  }
  
  private boolean isValidPage(int page) {
    return (page > 0) && (page <= pages.size());
  }
  
  public CraftMetaBook clone()
  {
    CraftMetaBook meta = (CraftMetaBook)super.clone();
    pages = new ArrayList(pages);
    return meta;
  }
  
  int applyHash()
  {
    int original;
    int hash = original = super.applyHash();
    if (hasTitle()) {
      hash = 61 * hash + title.hashCode();
    }
    if (hasAuthor()) {
      hash = 61 * hash + 13 * author.hashCode();
    }
    if (hasPages()) {
      hash = 61 * hash + 17 * pages.hashCode();
    }
    return original != hash ? CraftMetaBook.class.hashCode() ^ hash : hash;
  }
  
  boolean equalsCommon(CraftMetaItem meta)
  {
    if (!super.equalsCommon(meta)) {
      return false;
    }
    if ((meta instanceof CraftMetaBook)) {
      CraftMetaBook that = (CraftMetaBook)meta;
      
      return (hasTitle() ? (that.hasTitle()) && (title.equals(title)) : !that.hasTitle()) && (hasAuthor() ? (that.hasAuthor()) && (author.equals(author)) : !that.hasAuthor()) && (hasPages() ? (that.hasPages()) && (pages.equals(pages)) : !that.hasPages());
    }
    

    return true;
  }
  
  boolean notUncommon(CraftMetaItem meta)
  {
    return (super.notUncommon(meta)) && (((meta instanceof CraftMetaBook)) || (isBookEmpty()));
  }
  
  ImmutableMap.Builder<String, Object> serialize(ImmutableMap.Builder<String, Object> builder)
  {
    super.serialize(builder);
    
    if (hasTitle()) {
      builder.put(BOOK_TITLEBUKKIT, title);
    }
    
    if (hasAuthor()) {
      builder.put(BOOK_AUTHORBUKKIT, author);
    }
    
    if (hasPages()) {
      builder.put(BOOK_PAGESBUKKIT, pages);
    }
    
    return builder;
  }
}
