package org.bukkit.craftbukkit.v1_7_R1.inventory;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap.Builder;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import net.minecraft.server.v1_7_R1.NBTTagCompound;
import net.minecraft.server.v1_7_R1.NBTTagList;
import org.apache.commons.lang.Validate;
import org.bukkit.Material;
import org.bukkit.configuration.serialization.DelegateDeserialization;
import org.bukkit.inventory.meta.PotionMeta;
import org.bukkit.potion.PotionEffect;
import org.bukkit.potion.PotionEffectType;


@DelegateDeserialization(CraftMetaItem.SerializableMeta.class)
class CraftMetaPotion
  extends CraftMetaItem
  implements PotionMeta
{
  static final CraftMetaItem.ItemMetaKey AMPLIFIER = new CraftMetaItem.ItemMetaKey("Amplifier", "amplifier");
  static final CraftMetaItem.ItemMetaKey AMBIENT = new CraftMetaItem.ItemMetaKey("Ambient", "ambient");
  static final CraftMetaItem.ItemMetaKey DURATION = new CraftMetaItem.ItemMetaKey("Duration", "duration");
  static final CraftMetaItem.ItemMetaKey POTION_EFFECTS = new CraftMetaItem.ItemMetaKey("CustomPotionEffects", "custom-effects");
  static final CraftMetaItem.ItemMetaKey ID = new CraftMetaItem.ItemMetaKey("Id", "potion-id");
  private List<PotionEffect> customEffects;
  
  CraftMetaPotion(CraftMetaItem meta)
  {
    super(meta);
    if (!(meta instanceof CraftMetaPotion)) {
      return;
    }
    CraftMetaPotion potionMeta = (CraftMetaPotion)meta;
    if (potionMeta.hasCustomEffects()) {
      customEffects = new ArrayList(customEffects);
    }
  }
  
  CraftMetaPotion(NBTTagCompound tag) {
    super(tag);
    
    if (tag.hasKey(POTION_EFFECTSNBT)) {
      NBTTagList list = tag.getList(POTION_EFFECTSNBT, 10);
      int length = list.size();
      if (length > 0) {
        customEffects = new ArrayList(length);
        
        for (int i = 0; i < length; i++) {
          NBTTagCompound effect = list.get(i);
          PotionEffectType type = PotionEffectType.getById(effect.getByte(IDNBT));
          int amp = effect.getByte(AMPLIFIERNBT);
          int duration = effect.getInt(DURATIONNBT);
          boolean ambient = effect.getBoolean(AMBIENTNBT);
          customEffects.add(new PotionEffect(type, duration, amp, ambient));
        }
      }
    }
  }
  
  CraftMetaPotion(Map<String, Object> map) {
    super(map);
    
    Iterable<?> rawEffectList = (Iterable)CraftMetaItem.SerializableMeta.getObject(Iterable.class, map, POTION_EFFECTSBUKKIT, true);
    if (rawEffectList == null) {
      return;
    }
    
    for (Object obj : rawEffectList) {
      if (!(obj instanceof PotionEffect)) {
        throw new IllegalArgumentException("Object in effect list is not valid. " + obj.getClass());
      }
      addCustomEffect((PotionEffect)obj, true);
    }
  }
  
  void applyToItem(NBTTagCompound tag)
  {
    super.applyToItem(tag);
    NBTTagList effectList; if (hasCustomEffects()) {
      effectList = new NBTTagList();
      tag.set(POTION_EFFECTSNBT, effectList);
      
      for (PotionEffect effect : customEffects) {
        NBTTagCompound effectData = new NBTTagCompound();
        effectData.setByte(IDNBT, (byte)effect.getType().getId());
        effectData.setByte(AMPLIFIERNBT, (byte)effect.getAmplifier());
        effectData.setInt(DURATIONNBT, effect.getDuration());
        effectData.setBoolean(AMBIENTNBT, effect.isAmbient());
        effectList.add(effectData);
      }
    }
  }
  
  boolean isEmpty()
  {
    return (super.isEmpty()) && (isPotionEmpty());
  }
  
  boolean isPotionEmpty() {
    return !hasCustomEffects();
  }
  
  boolean applicableTo(Material type)
  {
    switch (1.$SwitchMap$org$bukkit$Material[type.ordinal()]) {
    case 1: 
      return true;
    }
    return false;
  }
  

  public CraftMetaPotion clone()
  {
    CraftMetaPotion clone = (CraftMetaPotion)super.clone();
    if (customEffects != null) {
      customEffects = new ArrayList(customEffects);
    }
    return clone;
  }
  
  public boolean hasCustomEffects() {
    return (customEffects != null) && (!customEffects.isEmpty());
  }
  
  public List<PotionEffect> getCustomEffects() {
    if (hasCustomEffects()) {
      return ImmutableList.copyOf(customEffects);
    }
    return ImmutableList.of();
  }
  
  public boolean addCustomEffect(PotionEffect effect, boolean overwrite) {
    Validate.notNull(effect, "Potion effect must not be null");
    
    int index = indexOfEffect(effect.getType());
    if (index != -1) {
      if (overwrite) {
        PotionEffect old = (PotionEffect)customEffects.get(index);
        if ((old.getAmplifier() == effect.getAmplifier()) && (old.getDuration() == effect.getDuration()) && (old.isAmbient() == effect.isAmbient())) {
          return false;
        }
        customEffects.set(index, effect);
        return true;
      }
      return false;
    }
    
    if (customEffects == null) {
      customEffects = new ArrayList();
    }
    customEffects.add(effect);
    return true;
  }
  
  public boolean removeCustomEffect(PotionEffectType type)
  {
    Validate.notNull(type, "Potion effect type must not be null");
    
    if (!hasCustomEffects()) {
      return false;
    }
    
    boolean changed = false;
    Iterator<PotionEffect> iterator = customEffects.iterator();
    while (iterator.hasNext()) {
      PotionEffect effect = (PotionEffect)iterator.next();
      if (effect.getType() == type) {
        iterator.remove();
        changed = true;
      }
    }
    return changed;
  }
  
  public boolean hasCustomEffect(PotionEffectType type) {
    Validate.notNull(type, "Potion effect type must not be null");
    return indexOfEffect(type) != -1;
  }
  
  public boolean setMainEffect(PotionEffectType type) {
    Validate.notNull(type, "Potion effect type must not be null");
    int index = indexOfEffect(type);
    if ((index == -1) || (index == 0)) {
      return false;
    }
    
    PotionEffect old = (PotionEffect)customEffects.get(0);
    customEffects.set(0, customEffects.get(index));
    customEffects.set(index, old);
    return true;
  }
  
  private int indexOfEffect(PotionEffectType type) {
    if (!hasCustomEffects()) {
      return -1;
    }
    
    for (int i = 0; i < customEffects.size(); i++) {
      if (((PotionEffect)customEffects.get(i)).getType().equals(type)) {
        return i;
      }
    }
    return -1;
  }
  
  public boolean clearCustomEffects() {
    boolean changed = hasCustomEffects();
    customEffects = null;
    return changed;
  }
  
  int applyHash()
  {
    int original;
    int hash = original = super.applyHash();
    if (hasCustomEffects()) {
      hash = 73 * hash + customEffects.hashCode();
    }
    return original != hash ? CraftMetaPotion.class.hashCode() ^ hash : hash;
  }
  
  public boolean equalsCommon(CraftMetaItem meta)
  {
    if (!super.equalsCommon(meta)) {
      return false;
    }
    if ((meta instanceof CraftMetaPotion)) {
      CraftMetaPotion that = (CraftMetaPotion)meta;
      
      return (that.hasCustomEffects()) && (customEffects.equals(customEffects));
    }
    return true;
  }
  
  boolean notUncommon(CraftMetaItem meta)
  {
    return (super.notUncommon(meta)) && (((meta instanceof CraftMetaPotion)) || (isPotionEmpty()));
  }
  
  ImmutableMap.Builder<String, Object> serialize(ImmutableMap.Builder<String, Object> builder)
  {
    super.serialize(builder);
    
    if (hasCustomEffects()) {
      builder.put(POTION_EFFECTSBUKKIT, ImmutableList.copyOf(customEffects));
    }
    
    return builder;
  }
}
