package org.bukkit.craftbukkit.v1_7_R1.scoreboard;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSet.Builder;
import java.util.HashMap;
import java.util.Map;
import net.minecraft.server.v1_7_R1.ScoreboardObjective;
import net.minecraft.server.v1_7_R1.ScoreboardTeam;
import org.apache.commons.lang.Validate;
import org.bukkit.Bukkit;
import org.bukkit.OfflinePlayer;
import org.bukkit.scoreboard.DisplaySlot;
import org.bukkit.scoreboard.Objective;
import org.bukkit.scoreboard.Score;
import org.bukkit.scoreboard.Team;


public final class CraftScoreboard
  implements org.bukkit.scoreboard.Scoreboard
{
  final net.minecraft.server.v1_7_R1.Scoreboard board;
  final Map<String, CraftObjective> objectives = new HashMap();
  final Map<String, CraftTeam> teams = new HashMap();
  
  CraftScoreboard(net.minecraft.server.v1_7_R1.Scoreboard board) {
    this.board = board;
    
    for (ScoreboardObjective objective : board.getObjectives()) {
      new CraftObjective(this, objective);
    }
    for (ScoreboardTeam team : board.getTeams()) {
      new CraftTeam(this, team);
    }
  }
  
  public CraftObjective registerNewObjective(String name, String criteria) throws IllegalArgumentException {
    Validate.notNull(name, "Objective name cannot be null");
    Validate.notNull(criteria, "Criteria cannot be null");
    Validate.isTrue(name.length() <= 16, "The name '" + name + "' is longer than the limit of 16 characters");
    Validate.isTrue(board.getObjective(name) == null, "An objective of name '" + name + "' already exists");
    
    CraftCriteria craftCriteria = CraftCriteria.getFromBukkit(criteria);
    ScoreboardObjective objective = board.registerObjective(name, criteria);
    return new CraftObjective(this, objective);
  }
  
  public Objective getObjective(String name) throws IllegalArgumentException {
    Validate.notNull(name, "Name cannot be null");
    return (Objective)objectives.get(name);
  }
  
  public ImmutableSet<Objective> getObjectivesByCriteria(String criteria) throws IllegalArgumentException {
    Validate.notNull(criteria, "Criteria cannot be null");
    
    ImmutableSet.Builder<Objective> objectives = ImmutableSet.builder();
    for (CraftObjective objective : this.objectives.values()) {
      if (objective.getCriteria().equals(criteria)) {
        objectives.add(objective);
      }
    }
    return objectives.build();
  }
  
  public ImmutableSet<Objective> getObjectives() {
    return ImmutableSet.copyOf(objectives.values());
  }
  
  public Objective getObjective(DisplaySlot slot) throws IllegalArgumentException {
    Validate.notNull(slot, "Display slot cannot be null");
    ScoreboardObjective objective = board.getObjectiveForSlot(CraftScoreboardTranslations.fromBukkitSlot(slot));
    if (objective == null) {
      return null;
    }
    return (Objective)objectives.get(objective.getName());
  }
  
  public ImmutableSet<Score> getScores(OfflinePlayer player) throws IllegalArgumentException {
    Validate.notNull(player, "OfflinePlayer cannot be null");
    
    ImmutableSet.Builder<Score> scores = ImmutableSet.builder();
    for (CraftObjective objective : objectives.values()) {
      scores.add(objective.getScore(player));
    }
    return scores.build();
  }
  
  public void resetScores(OfflinePlayer player) throws IllegalArgumentException {
    Validate.notNull(player, "OfflinePlayer cannot be null");
    
    board.resetPlayerScores(player.getName());
  }
  
  public Team getPlayerTeam(OfflinePlayer player) throws IllegalArgumentException {
    Validate.notNull(player, "OfflinePlayer cannot be null");
    
    ScoreboardTeam team = board.getPlayerTeam(player.getName());
    return team == null ? null : (CraftTeam)teams.get(team.getName());
  }
  
  public Team getTeam(String teamName) throws IllegalArgumentException {
    Validate.notNull(teamName, "Team name cannot be null");
    
    return (Team)teams.get(teamName);
  }
  
  public ImmutableSet<Team> getTeams() {
    return ImmutableSet.copyOf(teams.values());
  }
  
  public Team registerNewTeam(String name) throws IllegalArgumentException {
    Validate.notNull(name, "Team name cannot be null");
    Validate.isTrue(name.length() <= 16, "Team name '" + name + "' is longer than the limit of 16 characters");
    Validate.isTrue(board.getTeam(name) == null, "Team name '" + name + "' is already in use");
    
    return new CraftTeam(this, board.createTeam(name));
  }
  
  public ImmutableSet<OfflinePlayer> getPlayers() {
    ImmutableSet.Builder<OfflinePlayer> players = ImmutableSet.builder();
    for (Object playerName : board.getPlayers()) {
      players.add(Bukkit.getOfflinePlayer(playerName.toString()));
    }
    return players.build();
  }
  
  public void clearSlot(DisplaySlot slot) throws IllegalArgumentException {
    Validate.notNull(slot, "Slot cannot be null");
    board.setDisplaySlot(CraftScoreboardTranslations.fromBukkitSlot(slot), null);
  }
  
  public net.minecraft.server.v1_7_R1.Scoreboard getHandle()
  {
    return board;
  }
}
