package org.bukkit.craftbukkit.v1_7_R1.util;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;
import org.apache.commons.lang.Validate;











































public final class AsynchronousExecutor<P, T, C, E extends Throwable>
{
  static final AtomicIntegerFieldUpdater STATE_FIELD = AtomicIntegerFieldUpdater.newUpdater(Task.class, "state");
  final CallBackProvider<P, T, C, E> provider;
  
  private static boolean set(Task $this, int expected, int value) {
    return STATE_FIELD.compareAndSet($this, expected, value);
  }
  
  class Task implements Runnable
  {
    static final int PENDING = 0;
    static final int STAGE_1_ASYNC = 1;
    static final int STAGE_1_SYNC = 2;
    static final int STAGE_1_COMPLETE = 3;
    static final int FINISHED = 4;
    volatile int state = 0;
    final P parameter;
    T object;
    final List<C> callbacks = new LinkedList();
    E t = null;
    
    Task() {
      this.parameter = parameter;
    }
    
    public void run() {
      if (initAsync()) {
        finished.add(this);
      }
    }
    
    boolean initAsync() {
      if (AsynchronousExecutor.set(this, 0, 1)) {
        boolean ret = true;
        try
        {
          init();
        } finally {
          if (!AsynchronousExecutor.set(this, 1, 3))
          {


            synchronized (this) {
              if (state != 2)
              {
                notifyAll();
              }
              

              state = 3;
            }
            
            ret = false;
          }
        }
        
        return ret;
      }
      return false;
    }
    
    void initSync()
    {
      if (AsynchronousExecutor.set(this, 0, 3))
      {
        init();
      } else if (AsynchronousExecutor.set(this, 1, 2))
      {
        synchronized (this) {
          if (AsynchronousExecutor.set(this, 2, 0))
          {
            while (state != 3) {
              try {
                wait();
              } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                throw new RuntimeException("Unable to handle interruption on " + parameter, e);
              }
            }
          }
        }
      }
    }
    



    void init()
    {
      try
      {
        object = provider.callStage1(parameter);
      } catch (Throwable t) {
        this.t = t;
      }
    }
    
    T get() throws Throwable {
      initSync();
      finish();
      return object;
    }
    
    void finish() throws Throwable {
      switch (state) {
      case 0: 
      case 1: 
      case 2: 
      default: 
        throw new IllegalStateException("Attempting to finish unprepared(" + state + ") task(" + this.parameter + ")");
      case 3: 
        try {
          if (t != null) {
            throw t;
          }
          
          provider = AsynchronousExecutor.this.provider;
          parameter = this.parameter;
          object = this.object;
          
          provider.callStage2(parameter, object);
          for (C callback : callbacks)
            provider.callStage3(parameter, object, callback); } finally { AsynchronousExecutor.CallBackProvider<P, T, C, E> provider;
          P parameter;
          T object;
          tasks.remove(this.parameter);
          state = 4;
        }
      }
      
    }
  }
  

  final Queue<AsynchronousExecutor<P, T, C, E>.Task> finished = new ConcurrentLinkedQueue();
  final Map<P, AsynchronousExecutor<P, T, C, E>.Task> tasks = new HashMap();
  

  final ThreadPoolExecutor pool;
  

  public AsynchronousExecutor(CallBackProvider<P, T, C, E> provider, int coreSize)
  {
    Validate.notNull(provider, "Provider cannot be null");
    this.provider = provider;
    

    pool = new ThreadPoolExecutor(coreSize, 2147483647, 60L, TimeUnit.SECONDS, new LinkedBlockingQueue(), provider);
  }
  




  public void add(P parameter, C callback)
  {
    AsynchronousExecutor<P, T, C, E>.Task task = (Task)tasks.get(parameter);
    if (task == null) {
      tasks.put(parameter, task = new Task(parameter));
      pool.execute(task);
    }
    callbacks.add(callback);
  }
  




  public T get(P parameter)
    throws Throwable, IllegalStateException
  {
    AsynchronousExecutor<P, T, C, E>.Task task = (Task)tasks.get(parameter);
    if (task == null) {
      throw new IllegalStateException("Unknown " + parameter);
    }
    return task.get();
  }
  

  public T getSkipQueue(P parameter)
    throws Throwable
  {
    return skipQueue(parameter);
  }
  

  public T getSkipQueue(P parameter, C callback)
    throws Throwable
  {
    T object = skipQueue(parameter);
    provider.callStage3(parameter, object, callback);
    return object;
  }
  

  public T getSkipQueue(P parameter, C... callbacks)
    throws Throwable
  {
    CallBackProvider<P, T, C, E> provider = this.provider;
    T object = skipQueue(parameter);
    for (C callback : callbacks) {
      provider.callStage3(parameter, object, callback);
    }
    return object;
  }
  

  public T getSkipQueue(P parameter, Iterable<C> callbacks)
    throws Throwable
  {
    CallBackProvider<P, T, C, E> provider = this.provider;
    T object = skipQueue(parameter);
    for (C callback : callbacks) {
      provider.callStage3(parameter, object, callback);
    }
    return object;
  }
  
  private T skipQueue(P parameter) throws Throwable {
    AsynchronousExecutor<P, T, C, E>.Task task = (Task)tasks.get(parameter);
    if (task != null) {
      return task.get();
    }
    T object = provider.callStage1(parameter);
    provider.callStage2(parameter, object);
    return object;
  }
  

  public void finishActive()
    throws Throwable
  {
    Queue<AsynchronousExecutor<P, T, C, E>.Task> finished = this.finished;
    while (!finished.isEmpty()) {
      ((Task)finished.poll()).finish();
    }
  }
  
  public void setActiveThreads(int coreSize) {
    pool.setCorePoolSize(coreSize);
  }
  
  public static abstract interface CallBackProvider<P, T, C, E extends Throwable>
    extends ThreadFactory
  {
    public abstract T callStage1(P paramP)
      throws Throwable;
    
    public abstract void callStage2(P paramP, T paramT)
      throws Throwable;
    
    public abstract void callStage3(P paramP, T paramT, C paramC)
      throws Throwable;
  }
}
