package org.bukkit.plugin.java;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang.Validate;
import org.bukkit.plugin.InvalidPluginException;
import org.bukkit.plugin.PluginDescriptionFile;


final class PluginClassLoader
  extends URLClassLoader
{
  private final JavaPluginLoader loader;
  private final Map<String, Class<?>> classes = new HashMap();
  private final PluginDescriptionFile description;
  private final File dataFolder;
  private final File file;
  final JavaPlugin plugin;
  private JavaPlugin pluginInit;
  private IllegalStateException pluginState;
  
  PluginClassLoader(JavaPluginLoader loader, ClassLoader parent, PluginDescriptionFile description, File dataFolder, File file) throws InvalidPluginException, MalformedURLException {
    super(new URL[] { file.toURI().toURL() }, parent);
    Validate.notNull(loader, "Loader cannot be null");
    
    this.loader = loader;
    this.description = description;
    this.dataFolder = dataFolder;
    this.file = file;
    try
    {
      Class<?> jarClass;
      try {
        jarClass = Class.forName(description.getMain(), true, this);
      } catch (ClassNotFoundException ex) {
        throw new InvalidPluginException("Cannot find main class `" + description.getMain() + "'", ex);
      }
      Class<? extends JavaPlugin> pluginClass;
      try
      {
        pluginClass = jarClass.asSubclass(JavaPlugin.class);
      } catch (ClassCastException ex) {
        throw new InvalidPluginException("main class `" + description.getMain() + "' does not extend JavaPlugin", ex);
      }
      
      plugin = ((JavaPlugin)pluginClass.newInstance());
    } catch (IllegalAccessException ex) {
      throw new InvalidPluginException("No public constructor", ex);
    } catch (InstantiationException ex) {
      throw new InvalidPluginException("Abnormal plugin type", ex);
    }
  }
  
  protected Class<?> findClass(String name) throws ClassNotFoundException
  {
    return findClass(name, true);
  }
  
  Class<?> findClass(String name, boolean checkGlobal) throws ClassNotFoundException {
    if ((name.startsWith("org.bukkit.")) || (name.startsWith("net.minecraft."))) {
      throw new ClassNotFoundException(name);
    }
    Class<?> result = (Class)classes.get(name);
    
    if (result == null) {
      if (checkGlobal) {
        result = loader.getClassByName(name);
      }
      
      if (result == null) {
        result = super.findClass(name);
        
        if (result != null) {
          loader.setClass(name, result);
        }
      }
      
      classes.put(name, result);
    }
    
    return result;
  }
  
  Set<String> getClasses() {
    return classes.keySet();
  }
  
  synchronized void initialize(JavaPlugin javaPlugin) {
    Validate.notNull(javaPlugin, "Initializing plugin cannot be null");
    Validate.isTrue(javaPlugin.getClass().getClassLoader() == this, "Cannot initialize plugin outside of this class loader");
    if ((plugin != null) || (pluginInit != null)) {
      throw new IllegalArgumentException("Plugin already initialized!", pluginState);
    }
    
    pluginState = new IllegalStateException("Initial initialization");
    pluginInit = javaPlugin;
    
    javaPlugin.init(loader, loader.server, description, dataFolder, file, this);
  }
}
