package net.minecraft.util.com.google.common.util.concurrent;

import java.math.RoundingMode;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import net.minecraft.util.com.google.common.annotations.Beta;
import net.minecraft.util.com.google.common.base.Objects;
import net.minecraft.util.com.google.common.base.Preconditions;
import net.minecraft.util.com.google.common.base.Supplier;
import net.minecraft.util.com.google.common.collect.Iterables;
import net.minecraft.util.com.google.common.collect.MapMaker;
import net.minecraft.util.com.google.common.math.IntMath;





































































































@Beta
public abstract class Striped<L>
{
  public abstract L get(Object paramObject);
  
  public abstract L getAt(int paramInt);
  
  abstract int indexFor(Object paramObject);
  
  public abstract int size();
  
  public Iterable<L> bulkGet(Iterable<?> keys)
  {
    Object[] array = Iterables.toArray(keys, Object.class);
    int[] stripes = new int[array.length];
    for (int i = 0; i < array.length; i++) {
      stripes[i] = indexFor(array[i]);
    }
    Arrays.sort(stripes);
    for (int i = 0; i < array.length; i++) {
      array[i] = getAt(stripes[i]);
    }
    

















    List<L> asList = Arrays.asList(array);
    return Collections.unmodifiableList(asList);
  }
  








  public static Striped<Lock> lock(int stripes)
  {
    new CompactStriped(stripes, new Supplier()
    {
      public Lock get() { return new Striped.PaddedLock(); } }, null);
  }
  








  public static Striped<Lock> lazyWeakLock(int stripes)
  {
    new LazyStriped(stripes, new Supplier() {
      public Lock get() {
        return new ReentrantLock(false);
      }
    });
  }
  







  public static Striped<Semaphore> semaphore(int stripes, int permits)
  {
    new CompactStriped(stripes, new Supplier()
    {
      public Semaphore get() { return new Striped.PaddedSemaphore(val$permits); } }, null);
  }
  









  public static Striped<Semaphore> lazyWeakSemaphore(int stripes, int permits)
  {
    new LazyStriped(stripes, new Supplier() {
      public Semaphore get() {
        return new Semaphore(val$permits, false);
      }
    });
  }
  






  public static Striped<ReadWriteLock> readWriteLock(int stripes)
  {
    return new CompactStriped(stripes, READ_WRITE_LOCK_SUPPLIER, null);
  }
  






  public static Striped<ReadWriteLock> lazyWeakReadWriteLock(int stripes)
  {
    return new LazyStriped(stripes, READ_WRITE_LOCK_SUPPLIER);
  }
  

  private static final Supplier<ReadWriteLock> READ_WRITE_LOCK_SUPPLIER = new Supplier()
  {

    public ReadWriteLock get() { return new ReentrantReadWriteLock(); }
  };
  private static final int ALL_SET = -1;
  
  private static abstract class PowerOfTwoStriped<L> extends Striped<L> {
    final int mask;
    
    PowerOfTwoStriped(int stripes) {
      super();
      Preconditions.checkArgument(stripes > 0, "Stripes must be positive");
      mask = (stripes > 1073741824 ? -1 : Striped.ceilToPowerOfTwo(stripes) - 1);
    }
    
    final int indexFor(Object key) {
      int hash = Striped.smear(key.hashCode());
      return hash & mask;
    }
    
    public final L get(Object key) {
      return getAt(indexFor(key));
    }
  }
  

  private static class CompactStriped<L>
    extends Striped.PowerOfTwoStriped<L>
  {
    private final Object[] array;
    

    private CompactStriped(int stripes, Supplier<L> supplier)
    {
      super();
      Preconditions.checkArgument(stripes <= 1073741824, "Stripes must be <= 2^30)");
      
      array = new Object[mask + 1];
      for (int i = 0; i < array.length; i++) {
        array[i] = supplier.get();
      }
    }
    
    public L getAt(int index)
    {
      return array[index];
    }
    
    public int size() {
      return array.length;
    }
  }
  

  private static class LazyStriped<L>
    extends Striped.PowerOfTwoStriped<L>
  {
    final ConcurrentMap<Integer, L> locks;
    
    final Supplier<L> supplier;
    final int size;
    
    LazyStriped(int stripes, Supplier<L> supplier)
    {
      super();
      size = (mask == -1 ? 2147483647 : mask + 1);
      this.supplier = supplier;
      locks = new MapMaker().weakValues().makeMap();
    }
    
    public L getAt(int index) {
      if (size != 2147483647) {
        Preconditions.checkElementIndex(index, size());
      }
      L existing = locks.get(Integer.valueOf(index));
      if (existing != null) {
        return existing;
      }
      L created = supplier.get();
      existing = locks.putIfAbsent(Integer.valueOf(index), created);
      return Objects.firstNonNull(existing, created);
    }
    
    public int size() {
      return size;
    }
  }
  




  private static int ceilToPowerOfTwo(int x)
  {
    return 1 << IntMath.log2(x, RoundingMode.CEILING);
  }
  








  private static int smear(int hashCode)
  {
    hashCode ^= hashCode >>> 20 ^ hashCode >>> 12;
    return hashCode ^ hashCode >>> 7 ^ hashCode >>> 4;
  }
  

  private static class PaddedLock
    extends ReentrantLock
  {
    long q1;
    long q2;
    long q3;
    
    PaddedLock()
    {
      super();
    }
  }
  
  private static class PaddedSemaphore extends Semaphore {
    long q1;
    long q2;
    long q3;
    
    PaddedSemaphore(int permits) {
      super(false);
    }
  }
}
